from munch import Munch
from functools import partial

from infinishell.infishell import runtime
from infinishell.infishell.commands import commandgroup, command, deprecated
from infinishell.infishell.exceptions import BadValue, BadArguments, CommandAborted
from infinishell.infishell.types import YesNo, Timestamp, IPAddress
from infinishell.infishell.utils import strlist
from ...utils import yesno
from infinishell.infishell.exceptions import CommandAborted

from ..exceptions import ManagementException
from ..renderer import render_capacity
from .hierarchical_base import HierarchicalEntityCommandsBase
from .types import (
    NONE_CONSTANT,
    QOSPolicy,
    Pool,
    Filesystem,
    FilesystemSnapshot,
    FilesystemOrSnapshot,
    MasterFilesystem,
    Name,
    Size,
    SizeDelta,
    SizePositiveDelta,
    NasCapacityUnit,
    FSSeries,
    ReplicaRoleOrNone,
    DateDuration,
    LockType,
    SuspendState,
    TreeQ,
    TreeQPath,
    TreeQMode,
    TreeQState,
    TreeQName,
    TreeQInodeQuota,
    TreeQInodeQuotaOrNone,
    TreeQSoftInodeQuota,
    TreeQSoftInodeQuotaOrNone,
    TreeQCapacityQuota,
    TreeQCapacityQuotaOrNone,
    TreeQSoftCapacityQuota,
    TreeQSoftCapacityQuotaOrNone,
    FilePath,
    NlmLock,
    NlmLockState,
    NlmLockId,
    )
from .options import QueryOptions
from .utils import get_target_pool, format_savings
from . import render_collection, roles, set_rendered_unit, set_renderer_query_fields, EntityCommands, Null, LazyIdNameMapping

from ..system import system
from ..features import feature_command

FS_QUERY_FIELDS                     = strlist("name                        thin size used allocated tree_allocated pool_name write_protected mapped compression_enabled data_reduction replication_role")
DETAILED_FS_QUERY_FIELDS            = strlist("name nfs_filesystem_id      thin size used allocated tree_allocated pool_name write_protected mapped compression_enabled data_reduction ssd_enabled root_mode created_at updated_at snapdir_name snapdir_accessible replication_role parent")

INFINIDAT_FS_QUERY_FIELDS           = strlist("name                        thin size used allocated tree_allocated pool_name write_protected suspend_state mapped compression_enabled data_reduction  replication_role")
DETAILED_INFINIDAT_FS_QUERY_FIELDS  = strlist("name nfs_filesystem_id      thin size used allocated tree_allocated pool_name policy write_protected suspend_state mapped compression_enabled data_reduction ssd_enabled root_mode created_at updated_at snapdir_name snapdir_accessible replication_role parent")

SNAP_QUERY_FIELDS                   = strlist("name                        thin size used allocated                pool_name write_protected mapped                                                                                  lock_state created_at")
DETAILED_SNAP_QUERY_FIELDS          = strlist("name nfs_filesystem_id      thin size used allocated                pool_name write_protected mapped                                    ssd_enabled root_mode snapdir_name snapdir_accessible replication_role lock_state lock_expires_at created_at updated_at parent")

INFINIDAT_SNAP_QUERY_FIELDS         = strlist("name                        thin size used allocated                pool_name write_protected suspend_state mapped                                                                    lock_state created_at")
DETAILED_INFINIDAT_SNAP_QUERY_FIELDS= strlist("name nfs_filesystem_id      thin size used allocated                pool_name write_protected suspend_state mapped policy                      ssd_enabled root_mode snapdir_name snapdir_accessible replication_role lock_state lock_expires_at created_at updated_at parent")

CLONE_QUERY_FIELDS            = strlist("name             thin size used allocated tree_allocated pool_name write_protected mapped                              created_at")
DETAILED_CLONE_QUERY_FIELDS   = DETAILED_FS_QUERY_FIELDS

FS_TREE_FIELDS                      = strlist("name                   type      size used allocated                pool_name write_protected                                                                                  created_at")
DETAILED_FS_TREE_FIELDS             = strlist("name nfs_filesystem_id type thin size used allocated tree_allocated pool_name write_protected mapped compression_enabled data_reduction ssd_enabled root_mode replication_role created_at updated_at")
FS_TREE_SORTABLE_FIELDS             = strlist("name                   type      size                               pool_name write_protected                                                                 created_at")

INFINIDAT_FS_TREE_FIELDS            = strlist("name                   type      size used allocated                pool_name write_protected suspend_state                                                                                  created_at")
DETAILED_INFINIDAT_FS_TREE_FIELDS   = strlist("name nfs_filesystem_id type thin size used allocated tree_allocated pool_name write_protected suspend_state mapped compression_enabled data_reduction ssd_enabled root_mode replication_role created_at updated_at")

SORTABLE_FIELDS = strlist("name serial size write_protected mapped pool_name ssd_enabled created_at updated_at compression_enabled lock_expires_at")

QUERY_SUBSTITUTES = dict(
    write_protected="wp",
    ssd_enabled="ssd_cache",
    root_mode="default_filesystem_permissions",
    nfs_filesystem_id="nfs_fs_id",
    mapped="exported",
    compression_enabled="compressed",
    tree_allocated="snapshots",
    pool_name="pool",
    lock_expires_at="expiry_date",
    )

DETAILED_QUERY_SUBSTITUTES = dict(
    write_protected="write_protected",
    snapdir_name="snapshot_directory_name",
    snapdir_accessible="snapshot_directory_accessible",
    lock_expires_at="expiry_date",
    )


class FilesystemCommandsMixin(object):

    FS_TYPE_NAMES = Munch(
        MASTER="filesystem",
        SNAPSHOT="snapshot",
        )

    def _get_fs_type(self, fs):
        if fs.type == "MASTER":
            return "FILESYSTEM"
        return fs.type

    def _get_fs_type_name(self, volume):
        return self.FS_TYPE_NAMES.get(volume.type) or "filesystem"


class FilesystemCommandsBase(FilesystemCommandsMixin, HierarchicalEntityCommandsBase):

    HIERARCHICAL_TYPE = Filesystem

    def _query_hook(self, filesystems, _api):
        has_rmr_fields = None
        for fs in filesystems:
            fs.type = self._get_fs_type(fs)
            fs.thin = fs.provtype == self.THIN
            if 'capacity_savings' in fs:
                fs.data_reduction = format_savings(fs)
            if has_rmr_fields is None:
                has_rmr_fields = "rmr_source" in fs and "rmr_target" in fs
            if has_rmr_fields:
                fs.replication_role = ReplicaRoleOrNone.SOURCE if fs.rmr_source \
                                     else ReplicaRoleOrNone.TARGET if fs.rmr_target \
                                     else None
        return filesystems


def create_common_filesystem_commands_mixin(FilesystemType, fs_arg):

    class CommonFilesystemCommandsMixin(object):

        @command(
            description="Rename {}".format(FilesystemType.DESCRIPTION_SUBJECT),
            types=(FilesystemType, Name),
            hidden_substitue=dict(name="fs"),
            substitute=dict() if FilesystemType == MasterFilesystem else dict(fs="snap"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def rename(self, _api, fs, new_name):
            return self._rename(_api, fs, new_name)

        @command(
            description="Delete {}".format(FilesystemType.DESCRIPTION_SUBJECT),
            types=(FilesystemType.List,),
            hidden_substitue=dict(name="fs"),
            substitute=dict() if FilesystemType == MasterFilesystem else dict(fs="snap"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def delete(self, _api, fs):
            return self._delete(_api, fs)

        @command(
            description="Enable or disable SSD cache",
            types=(FilesystemType, YesNo, YesNo),
            hidden_substitue=dict(name="fs"),
            substitute=dict() if FilesystemType == MasterFilesystem else dict(fs="snap"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def cache(self, _api, fs, ssd_cache, recursive=False):
            return self._update_recursive(_api,
                entity=fs,
                recurse=recursive,
                success_msg="{} cached on SSD".format("now" if ssd_cache else "no longer"),
                success_msg_with_is=True,
                noop_msg="already{} cached on SSD".format("" if ssd_cache else " not"),
                noop_msg_with_is=True,
                props=dict(
                    ssd_enabled=ssd_cache,
                    ),
                predicate=lambda dataset: dataset.ssd_enabled != ssd_cache,
                )

        @feature_command(
            feature='qos_fs',
            status=True,
            description="Assign filesystem to policy",
            types=(FilesystemType.List, QOSPolicy),
            hidden_substitue=dict(name="fs"),
            substitute=dict() if FilesystemType == MasterFilesystem else dict(fs="snap"),
            )
        @roles.INFINIDAT.ADMIN
        def assign_qos(self, _api, fs, policy):
            with _api.aggregation() as aggregator:
                for f in fs:
                    aggregator.next_task(f.name, "Assigned")
                    data = dict(entity_id=f.id)
                    result = _api.post("{}/{}/assigned_entities".format(QOSPolicy.get_uri(), policy.id), data)
            if len(fs) == 1:
                return "Filesystem '{}' was assigned to policy '{}'".format(fs[0].name, policy.name)
            else:
                return "{} filesystems were assigned to policy '{}'".format(len(fs), policy.name)

        @feature_command(
            feature='qos_fs',
            status=True,
            description="Unassign fs from policy",
            types=(FilesystemType.List,),
            hidden_substitue=dict(name="fs"),
            substitute=dict() if FilesystemType == MasterFilesystem else dict(fs="snap"),
            )
        @roles.INFINIDAT.ADMIN
        def unassign_qos(self, _api, fs):
            with _api.aggregation() as aggregator:
                for f in fs:
                    if not f.qos_policy_id:
                        raise BadArguments("Filesystem '{}' does not belong to QoS policy".format(f.name))
                    data = dict(entity_id=f.id)
                    aggregator.next_task(f.name, "Unassigned")
                    result = _api.post("{}/{}/unassign".format(QOSPolicy.get_uri(), f.qos_policy_id), data)
            if len(fs) == 1:
                return "Filesystem '{}' was unassigned from policy '{}'".format(f.name, f.qos_policy_name)
            else:
                return "{} filesystems were unassigned from their policies".format(len(fs))
            
        @command(
            description="Suspend {}".format(FilesystemType.DESCRIPTION_SUBJECT),
            types=(FilesystemType, SuspendState),
            hidden_substitue=dict(name="fs"),
            substitute=dict() if FilesystemType == MasterFilesystem else dict(fs="snap"),
            )
        @roles.INFINIDAT
        def suspend(self, _api, fs, suspend_state):
            self._update(_api, fs, suspend_state=suspend_state)
            return self._result_message(fs, "suspended")
        
        @command(
            description="Resume {}".format(FilesystemType.DESCRIPTION_SUBJECT),
            types=(FilesystemType,),
            hidden_substitue=dict(name="fs"),
            substitute=dict() if FilesystemType == MasterFilesystem else dict(fs="snap"),
            )
        @roles.INFINIDAT
        def resume(self, _api, fs):
            self._update(_api, fs, suspend_state="NONE")
            return self._result_message(fs, "resumed")

    return CommonFilesystemCommandsMixin


def create_writable_filesystem_commands_mixin(FilesystemType, fs_arg):

    class WritableFilesystemCommandsMixin(object):

        @command(
            description="Resize {}".format(FilesystemType.DESCRIPTION_SUBJECT),
            types=(FilesystemType, SizePositiveDelta),
            hidden_substitue=dict(name="fs"),
            substitute=dict() if FilesystemType == MasterFilesystem else dict(fs="snap"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def resize(self, _api, fs, size):
            size = size.compute_size(fs.size)
            if fs.size == size:
                return self._result_message(fs, "size is already {}".format(render_capacity(fs.size)))
            self._update(_api, fs, size=size)
            return self._result_message(fs, "resized")

        @command(
            description="Restore data state of {} from previous snapshot".format(FilesystemType.DESCRIPTION_SUBJECT),
            types=(FilesystemType, FilesystemSnapshot),
            hidden_substitue=dict(target="fs"),
            substitute=dict() if FilesystemType == MasterFilesystem else dict(fs="snap"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def restore(self, _api, fs, source):
            if fs.id == source.id:
                raise BadValue('"{}" was specified as both arguments'.format(fs.name))
            # _api.post(FilesystemType.get_uri(fs.id, "restore"), source.id)
            self._post(_api, fs, "restore", data=dict(
                source_id=source.id
                ))
            return self._result_message(fs, 'restored from {} "{}"'.format(self._get_fs_type_name(source), source.name))

        @command(
            description="Disable write access on {}".format(FilesystemType.DESCRIPTION_SUBJECT),
            types=(FilesystemType, YesNo),
            hidden_substitue=dict(name="fs"),
            substitute=dict() if FilesystemType == MasterFilesystem else dict(fs="snap"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def write_protect(self, _api, fs, recursive=False):
            return self._update_recursive(_api,
                entity=fs,
                recurse=recursive,
                success_msg="now write-protected",
                success_msg_with_is=True,
                noop_msg="already write-protected",
                noop_msg_with_is=True,
                props=dict(
                    write_protected=True,
                    ),
                predicate=lambda dataset: dataset.type != "SNAP" and not dataset.write_protected,
                )

        @command(
            description="Enable write access on {}".format(FilesystemType.DESCRIPTION_SUBJECT),
            types=(FilesystemType, YesNo),
            hidden_substitue=dict(name="fs"),
            substitute=dict() if FilesystemType == MasterFilesystem else dict(fs="snap"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def write_enable(self, _api, fs, recursive=False):
            return self._update_recursive(_api,
                entity=fs,
                recurse=recursive,
                success_msg="now write-enabled",
                success_msg_with_is=True,
                noop_msg="already write-enabled",
                noop_msg_with_is=True,
                props=dict(
                    write_protected=False,
                    ),
                predicate=lambda dataset: dataset.type != "SNAP" and dataset.write_protected,
                )

    return WritableFilesystemCommandsMixin

def create_child_filesystem_commands_mixin(FilesystemType, ParentType, fs_arg, parent_arg, create_parent_arg):

    class ChildFilesystemCommandsMixin(object):

        @command(
            description="Create a new {} from {}".format(FilesystemType.DESCRIPTION_SUBJECT, ParentType.DESCRIPTION_SUBJECT),
            types=(ParentType, Name, YesNo, DateDuration, YesNo),
            hidden_substitue=dict(source="fs"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def create(self, _api, fs, name, ssd_cache=None, lock_duration=None, snapshot_directory_accessible=None):
            parent = self._create_and_return_entity(_api, parent_id=fs.id, name=name, ssd_enabled=ssd_cache, lock_expires_at=lock_duration['utc_timestamp'] if lock_duration else None, snapdir_accessible=snapshot_directory_accessible)
            if lock_duration:
                return self._result_message(parent, "created, locked until '{}'".format(lock_duration['local_date']), subject=FilesystemType.get_entity_name())
            return self._result_message(parent, "created", subject=FilesystemType.get_entity_name())

        @command(
            name="tree",
            description="Display hierarchy of snapshots",
            types=(FilesystemType, NasCapacityUnit),
            hidden_substitue=dict(name="fs"),
            substitute=dict() if FilesystemType == MasterFilesystem else dict(fs="snap"),
            options=QueryOptions(fields=DETAILED_FS_TREE_FIELDS, sortable=FS_TREE_SORTABLE_FIELDS, substitute=DETAILED_QUERY_SUBSTITUTES),
            )
        @render_collection(
            entity=Filesystem,
            fields=INFINIDAT_FS_TREE_FIELDS,
            substitute=QUERY_SUBSTITUTES,
        )
        @roles.INFINIDAT.TECHNICIAN
        def infinidat_tree(self, _api, fs, unit=None):
            set_rendered_unit(unit)
            set_renderer_query_fields()
            return self._get_tree(_api, fs)
        
        @command(
            description="Display hierarchy of snapshots",
            types=(FilesystemType, NasCapacityUnit),
            hidden_substitue=dict(name="fs"),
            substitute=dict() if FilesystemType == MasterFilesystem else dict(fs="snap"),
            options=QueryOptions(fields=DETAILED_FS_TREE_FIELDS, sortable=FS_TREE_SORTABLE_FIELDS, substitute=DETAILED_QUERY_SUBSTITUTES),
            )
        @render_collection(
            entity=Filesystem,
            fields=FS_TREE_FIELDS,
            substitute=QUERY_SUBSTITUTES,
        )
        @roles.ADMIN.POOLADMIN.READONLY
        def tree(self, _api, fs, unit=None):
            set_rendered_unit(unit)
            set_renderer_query_fields()
            return self._get_tree(_api, fs)

    return ChildFilesystemCommandsMixin

@commandgroup(prefix="fs")
class FilesystemCommands(FilesystemCommandsBase,
                         create_common_filesystem_commands_mixin(MasterFilesystem, "fs"),
                         create_writable_filesystem_commands_mixin(MasterFilesystem, "fs")):
    ENTITY_TYPE = MasterFilesystem
    @command(
        name='query',
        description="List existing filesystems",
        types=(MasterFilesystem.List, Pool, NasCapacityUnit, YesNo, YesNo, YesNo, YesNo, YesNo, YesNo),
        hidden_substitue=dict(name="fs"),
        options=QueryOptions(fields=DETAILED_FS_QUERY_FIELDS, sortable=SORTABLE_FIELDS, substitute=QUERY_SUBSTITUTES, detailed=True),
        )
    @render_collection(
        entity=Filesystem,
        fields=INFINIDAT_FS_QUERY_FIELDS,
        detailed_fields=DETAILED_INFINIDAT_FS_QUERY_FIELDS,
        detailed_substitute=DETAILED_QUERY_SUBSTITUTES,
        detailed_individually=True,
        substitute=QUERY_SUBSTITUTES,
        alignment=dict(
            data_reduction="right",
            ),
        )
    @roles.INFINIDAT.TECHNICIAN
    def infinidat_query(self, _api, fs=None, pool=None, unit=None, has_exports=None, write_protected=None, thin=None, compression_enabled=None, ssd_enabled=None, suspended=None):
        set_rendered_unit(unit)
        provtype = None
        if thin is not None:
            provtype = self.THIN if thin else self.THICK
        suspend_state = None
        if suspended is not None:
            suspend_state = ['drop', 'retry'] if suspended else 'none'
        return self._common_query(_api, fs, pool_id=pool, mapped=has_exports, write_protected=write_protected, provtype=provtype, compression_enabled=compression_enabled, ssd_enabled=ssd_enabled, suspend_state=suspend_state)

    @command(
        description="List existing filesystems",
        types=(MasterFilesystem.List, Pool, NasCapacityUnit, YesNo, YesNo, YesNo, YesNo, YesNo),
        hidden_substitue=dict(name="fs"),
        options=QueryOptions(fields=DETAILED_FS_QUERY_FIELDS, sortable=SORTABLE_FIELDS, substitute=QUERY_SUBSTITUTES, detailed=True),
        )
    @render_collection(
        entity=Filesystem,
        fields=FS_QUERY_FIELDS,
        detailed_fields=DETAILED_FS_QUERY_FIELDS,
        detailed_substitute=DETAILED_QUERY_SUBSTITUTES,
        detailed_individually=True,
        substitute=QUERY_SUBSTITUTES,
        alignment=dict(
            data_reduction="right",
            ),
        )
    @roles.ADMIN.POOLADMIN.READONLY
    def query(self, _api, fs=None, pool=None, unit=None, has_exports=None, write_protected=None, thin=None, compression_enabled=None, ssd_enabled=None):
        set_rendered_unit(unit)
        provtype = None
        if thin is not None:
            provtype = self.THIN if thin else self.THICK
        return self._common_query(_api, fs, pool_id=pool, mapped=has_exports, write_protected=write_protected, provtype=provtype, compression_enabled=compression_enabled, ssd_enabled=ssd_enabled)

    def _query_hook(self, entities, _api):
        for e in entities:
            if system.user.has_role('infinidat','technician'):
                if e.get('compression_enabled', False) and e.get('compression_suppressed', False):
                    e.compression_enabled = "yes (suppressed)"
            e.policy = e.qos_policy_name
        return super(FilesystemCommands, self)._query_hook(entities, _api)

    @command(
        description="Create a new filesystem",
        types=(Name.List, Size, Name, YesNo, Pool, YesNo, YesNo, FSSeries, YesNo),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def create(self, _api, name, size, snapshot_directory_name=None, snapshot_directory_accessible=None,
               pool=None, thin=None, ssd_cache=None, series=None, compression=None):
        thin = self.THIN == self._get_default_provtype() if thin == None else thin
        pool = get_target_pool(_api, pool)
        if series:
            # In case of multiple filesystems, since this is not an atomic
            # operation, ensure there's enough free space for all requested
            # filesystems.
            total_size = size * series.total
            if total_size > pool.free_virtual_space:
                raise ManagementException(409, Munch(code="POOL_VIRTUAL_CAPACITY", message="Pool virtual capacity reached."))
            if not thin and total_size > pool.free_physical_space:
                raise ManagementException(409, Munch(code="POOL_PHYSICAL_CAPACITY", message="Pool physical capacity reached."))
        return self._create_single_or_multiple_or_series(_api, name, series,
            size=size,
            pool_id=pool.id,
            provtype=self.THIN if thin else self.THICK,
            write_protected=False,
            ssd_enabled=ssd_cache,
            compression_enabled=compression,
            snapdir_name=snapshot_directory_name,
            snapdir_accessible=snapshot_directory_accessible,
            )

    @command(
        description="Move filesystem to a different pool",
        types=(MasterFilesystem, Pool, YesNo),
        hidden_substitue=dict(name="fs"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def move(self, _api, fs, pool, move_capacity=False):
        if fs.pool_id == pool.id:
            return self._result_message(fs, 'already in {} "{}"'.format(Pool.get_entity_name(), pool.name))
        _api.post(self.ENTITY_TYPE.get_uri(fs.id, "move"), dict(
            pool_id=pool.id,
            with_capacity=move_capacity,
            ))
        return self._result_message(fs, 'moved to pool "{}" {} its capacity'
                                        .format(pool.name, "along with" if move_capacity else "without"))

    @command(
        description="Set filesystem provisioning type",
        types=(MasterFilesystem, YesNo),
        hidden_substitue=dict(name="fs"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def set_thin(self, _api, fs, thin):
        prov_name = "thin" if thin else "thick"
        if fs.provtype.upper() == self._get_provtype(thin):
            raise CommandAborted(self._result_message(fs, "already set to {} provisioning".format(prov_name)))
        self._update(_api, fs, provtype=self._get_provtype(thin))
        return self._result_message(fs, "set to {} provisioning".format(prov_name))

    @command(
        description="Set filesystem compression",
        types=(MasterFilesystem, YesNo),
        hidden_substitue=dict(name="fs"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN.TECHNICIAN
    def set_compression(self, _api, fs, compression):
        if fs.compression_enabled  == compression:
            return self._result_message(fs, "compression is already {}".format("enabled" if compression else "disabled"))
        self._update(_api, fs, compression_enabled=compression)
        return 'Compression was {} for filesystem "{}"'.format("enabled" if compression else "disabled", fs.name)

    @command(
        description="Estimate reclaimed capacity (deletion simulation)",
        types=(Filesystem,),
        hidden_substitue=dict(name="fs"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def reclaimable(self, _api, fs):
        data = dict(entities=[fs.id])
        result = _api.post(Filesystem.get_uri("delete_simulation"), data)
        return 'Estimated reclaimable space for filesystem "{}" and its tree descendants: {}'.format(fs.name, render_capacity(result.space_reclaimable))

    @command(
        name="tree",
        description="Display hierarchy of filesystems and snapshots",
        types=(MasterFilesystem, Pool, NasCapacityUnit),
        hidden_substitue=dict(name="fs"),
        options=QueryOptions(fields=DETAILED_FS_TREE_FIELDS, sortable=FS_TREE_SORTABLE_FIELDS, substitute=DETAILED_QUERY_SUBSTITUTES),
        )
    @render_collection(
        entity=Filesystem,
        fields=INFINIDAT_FS_TREE_FIELDS,
        substitute=QUERY_SUBSTITUTES,
        )
    @roles.INFINIDAT.TECHNICIAN
    def infinidat_tree(self, _api, fs=None, pool=None, unit=None):
        set_rendered_unit(unit)
        set_renderer_query_fields()
        return self._get_tree(_api, fs, pool)
    
    @command(
        description="Display hierarchy of filesystems and snapshots",
        types=(MasterFilesystem, Pool, NasCapacityUnit),
        hidden_substitue=dict(name="fs"),
        options=QueryOptions(fields=DETAILED_FS_TREE_FIELDS, sortable=FS_TREE_SORTABLE_FIELDS, substitute=DETAILED_QUERY_SUBSTITUTES),
        )
    @render_collection(
        entity=Filesystem,
        fields=FS_TREE_FIELDS,
        substitute=QUERY_SUBSTITUTES,
        )
    @roles.ADMIN.POOLADMIN.READONLY
    def tree(self, _api, fs=None, pool=None, unit=None):
        set_rendered_unit(unit)
        set_renderer_query_fields()
        return self._get_tree(_api, fs, pool)
    
    @command(
        description="Determines whether the snapshot directory is accessible",
        types=(MasterFilesystem, YesNo),
        hidden_substitue=dict(name="fs"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def set_snapshot_directory_accessible(self, _api, fs, snapshot_directory_accessible):
        if fs.snapdir_accessible == snapshot_directory_accessible:
            return self._result_message(fs, "snapshot directory access is already {}".format("allowed" if snapshot_directory_accessible else "denied"))
        self._update(_api, fs, snapdir_accessible=snapshot_directory_accessible)
        return 'Filesystem "{}" set to {} access to snapshot directory'.format(fs.name, "allow" if snapshot_directory_accessible else "deny")

@commandgroup(prefix="fs.snap")
class FilesystemSnapshotCommands(FilesystemCommandsBase,
                                 create_common_filesystem_commands_mixin(
                                     FilesystemType=FilesystemSnapshot,
                                     fs_arg="snap",
                                     ),
                                 create_writable_filesystem_commands_mixin(
                                     FilesystemType=FilesystemSnapshot,
                                     fs_arg="snap",
                                     ),
                                 create_child_filesystem_commands_mixin(
                                     FilesystemType=FilesystemSnapshot,
                                     ParentType=FilesystemOrSnapshot,
                                     fs_arg="snap",
                                     parent_arg="fs",
                                     create_parent_arg="fs",
                                     )):
    ENTITY_TYPE = FilesystemSnapshot

    @command(
        name="query",
        description="List existing snapshots",
        types=(FilesystemSnapshot.List, FilesystemOrSnapshot, Timestamp, Timestamp, Pool, NasCapacityUnit, YesNo, YesNo, LockType, YesNo),
        hidden_substitue=dict(name="snap"),
        options=QueryOptions(fields=DETAILED_SNAP_QUERY_FIELDS+strlist("master"), sortable=SORTABLE_FIELDS + strlist('lock_expires_at'), substitute=QUERY_SUBSTITUTES, detailed=True),
        )
    @render_collection(
        entity=Filesystem,
        fields=INFINIDAT_SNAP_QUERY_FIELDS,
        detailed_fields=DETAILED_INFINIDAT_SNAP_QUERY_FIELDS,
        detailed_substitute=DETAILED_QUERY_SUBSTITUTES,
        detailed_individually=True,
        substitute=QUERY_SUBSTITUTES,
        )
    @roles.INFINIDAT.TECHNICIAN
    def infinidat_query(self, _api, snap=None, source=None, from_time=None, to_time=None, pool=None, unit=None, has_exports=None, write_protected=None, lock=None, suspended=None):
        set_rendered_unit(unit)
        filter = self._create_lock_filter(_api, lock, from_time, to_time)
        suspend_state = None
        if suspended is not None:
            suspend_state = ['drop', 'retry'] if suspended else 'none'
        return self._common_query(_api, snap, _filter=filter, parent_id=source, pool_id=pool, type=FilesystemSnapshot.ENTITY_MGMT_SUBTYPE, mapped=has_exports, write_protected=write_protected, suspend_state=suspend_state)

    @command(
        description="List existing snapshots",
        types=(FilesystemSnapshot.List, FilesystemOrSnapshot, Timestamp, Timestamp, Pool, NasCapacityUnit, YesNo, YesNo, LockType),
        hidden_substitue=dict(name="snap"),
        options=QueryOptions(fields=DETAILED_SNAP_QUERY_FIELDS+strlist("master"), sortable=SORTABLE_FIELDS + strlist('lock_expires_at'), substitute=QUERY_SUBSTITUTES, detailed=True),
        substitute=dict(from_time="expiry_date_from", to_time="expiry_date_to"),
        )
    @render_collection(
        entity=Filesystem,
        fields=SNAP_QUERY_FIELDS,
        detailed_fields=DETAILED_SNAP_QUERY_FIELDS,
        detailed_substitute=DETAILED_QUERY_SUBSTITUTES,
        detailed_individually=True,
        substitute=QUERY_SUBSTITUTES,
        )
    @roles.ADMIN.POOLADMIN.READONLY
    def query(self, _api, snap=None, source=None, from_time=None, to_time=None, pool=None, unit=None, has_exports=None, write_protected=None, lock=None):
        set_rendered_unit(unit)
        filter = self._create_lock_filter(_api, lock, from_time, to_time)
        return self._common_query(_api, snap, _filter=filter, parent_id=source, pool_id=pool, type=FilesystemSnapshot.ENTITY_MGMT_SUBTYPE, mapped=has_exports, write_protected=write_protected)

    def _query_hook(self, entities, _api):
        for e in entities:
            e.policy = e.qos_policy_name
            columns = runtime.options.get('columns', []) or []
            if runtime.options.get('detailed', False) or [i for i in columns if i in ['master', 'parent', '*']]:
                if 'parent' in columns or '*' in columns:
                    e.parent = FilesystemOrSnapshot.get_by_id(_api, e.parent_id).name
                if e.family_id and ('master' in columns or '*' in columns):
                    e.master = FilesystemOrSnapshot.get_by_id(_api, e.family_id).name
        return super(FilesystemSnapshotCommands, self)._query_hook(entities, _api)

    @command(description='Refresh a snapshot',
             types=(FilesystemSnapshot,),
             hidden_substitue=dict(name="snap"),
             )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def refresh(self, _api, snap):
        _api.post(self.ENTITY_TYPE.get_uri(snap.id, 'refresh'),
                  {'source_id': snap.parent_id})
        return self._result_message(snap, 'refreshed')

    @command(description='Locks a snapshot, temporarily preventing its deletion until the provided time, or extends an existing snapshot lock',
             types=(FilesystemSnapshot, DateDuration),
             hidden_substitue=dict(name="snap"),
             )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def secure_lock(self, _api, snap, lock_duration):
        self._update(_api, snap, lock_expires_at=lock_duration['utc_timestamp'])
        return self._result_message(snap, "is locked until '{}'".format(lock_duration['local_date']))
    
    @command(
        description="Determines whether the snapshot directory is accessible",
        types=(FilesystemSnapshot, YesNo),
        hidden_substitue=dict(name="snap"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def set_snapshot_directory_accessible(self, _api, snap, snapshot_directory_accessible):
        if snap.snapdir_accessible == snapshot_directory_accessible:
            return self._result_message(snap, "snapshot directory access is already {}".format("allowed" if snapshot_directory_accessible else "denied"))
        self._update(_api, snap, snapdir_accessible=snapshot_directory_accessible)
        return 'Filesystem snapshot "{}" set to {} access to snapshot directory'.format(snap.name, "allow" if snapshot_directory_accessible else "deny")

TREEQ_QUERY_FIELDS = strlist("fs name capacity_state used_capacity soft_capacity hard_capacity inodes_state used_inodes soft_inodes hard_inodes path")
TREEQ_DETAILED_QUERY_FIELDS = TREEQ_QUERY_FIELDS
TREEQ_QUERY_SUBSTITUTES = dict(
    name="treeq",
    hard_inodes="inodes_hard_quota",
    soft_inodes="inodes_soft_quota",
    hard_capacity="capacity_hard_quota",
    soft_capacity="capacity_soft_quota",
)
TREEQ_QUERY_SORTABLE = strlist("name capacity_state inodes_state")

@commandgroup(prefix="treeq", parent=FilesystemCommands)
class FilesystemTreeQCommands(EntityCommands):
    ENTITY_TYPE = TreeQ

    def _get_treeq(self, _api, fs, **kwargs):
        kwargs = {key:value for key,value in kwargs.items() if value}
        treeq = _api.get("filesystems/{}/treeqs".format(fs.id), params=kwargs)
        if not treeq:
            raise BadArguments, "TreeQ does not exist in filesystem '{}'".format(fs.name)
        return treeq

    @feature_command(
        feature='treeq',
        status=True,
        description="List existing treeqs",
        types=(FilesystemOrSnapshot, TreeQName, TreeQPath, TreeQState, TreeQState),
        hidden_substitue=dict(name="fs"),
        options=QueryOptions(fields=TREEQ_QUERY_FIELDS, sortable=TREEQ_QUERY_SORTABLE, substitute=TREEQ_QUERY_SUBSTITUTES, detailed=True),
        )
    @render_collection(
        entity=Filesystem,
        fields=TREEQ_QUERY_FIELDS,
        detailed_fields=TREEQ_DETAILED_QUERY_FIELDS,
        detailed_substitute=TREEQ_QUERY_SUBSTITUTES,
        detailed_individually=True,
        substitute=TREEQ_QUERY_SUBSTITUTES,
        alignment=dict(capacity_state="left",),
        )
    def query(self, _api, fs, treeq=None, path=None, inodes_state=None, capacity_state=None):
        params=Munch(
            name=treeq,
            path=path,
            capacity_state=capacity_state,
            inodes_state=inodes_state,
            )
        params = {k: v for k, v in params.items() if v is not None}
        self.ENTITY_TYPE.URI = self.ENTITY_TYPE.URI_TEMPLATE.format(fs.id)
        return self._query(_api, name=treeq, path=path, inodes_state=inodes_state, capacity_state=capacity_state, _hook=partial(self._query_hook, fs=fs))

    def _query_hook(self, data, fs):
        for treeq in data:
            treeq['fs'] = fs.name
        return data

    @feature_command(
        feature='treeq',
        status=True,
        description="Create new filesystem TreeQ",
        types=(FilesystemOrSnapshot, TreeQName, TreeQPath, TreeQMode, TreeQInodeQuota, TreeQSoftInodeQuota, TreeQCapacityQuota, TreeQSoftCapacityQuota),
    )
    @roles.INFINIDAT.ADMIN
    def create(self, _api, fs, name, path=None, mode=None, inode_quota=None, soft_inode_quota=None, capacity_quota=None, soft_capacity_quota=None):
        self.ENTITY_TYPE.URI = self.ENTITY_TYPE.URI_TEMPLATE.format(fs.id)
        path = path or "/{}".format(name)
        data = Munch(
            name=name,
            path=path,
            hard_inodes=inode_quota,
            soft_inodes=soft_inode_quota,
            hard_capacity=capacity_quota,
            soft_capacity=soft_capacity_quota,
            mode=mode,
        )
        data = {key:value for key,value in data.items() if value is not None}
        self._create(_api, **data)
        return "TreeQ '{}' created in filesystem '{}' with path '{}'".format(name, fs.name, path)

    @feature_command(
        feature='treeq',
        status=True,
        description="Modify filesystem TreeQ",
        types=(FilesystemOrSnapshot, TreeQName, TreeQPath, TreeQInodeQuotaOrNone, TreeQSoftInodeQuotaOrNone, TreeQCapacityQuotaOrNone, TreeQSoftCapacityQuotaOrNone),
    )
    @roles.INFINIDAT.ADMIN
    def modify(self, _api, fs, treeq=None, path=None, inode_quota=None, soft_inode_quota=None, capacity_quota=None, soft_capacity_quota=None):
        if  treeq is None and path is None:
            raise BadArguments, "treeq or path parameters must be used"
        if all(i is None for i in [inode_quota, soft_inode_quota, capacity_quota, soft_capacity_quota]):
            raise BadArguments, "At least one of inode_quota, soft_inode_quota, capacity_quota or soft_capacity_quota parameters must be set"
        self.ENTITY_TYPE.URI = self.ENTITY_TYPE.URI_TEMPLATE.format(fs.id)
        params=Munch(
            name=treeq,
            path=path,
            )
        results = self.ENTITY_TYPE.get_all(_api, params=params)
        wanted_treeq = results[0] if results else None
        if not wanted_treeq:
            msg = ""
            if treeq and path is None:
                msg = "TreeQ name '{}' was not found in filesystem '{}'".format(treeq, fs.name)
            elif path and treeq is None:
                msg = "TreeQ path '{}' was not found in filesystem '{}'".format(path, fs.name)
            else:
                msg = "TreeQ name '{}' or path '{}' were not found in filesystem '{}'".format(treeq, path, fs.name)
            raise BadArguments, msg
        data = Munch(
            hard_inodes=Null if inode_quota == NONE_CONSTANT else inode_quota,
            soft_inodes=Null if soft_inode_quota == NONE_CONSTANT else soft_inode_quota,
            hard_capacity=Null if capacity_quota == NONE_CONSTANT else capacity_quota,
            soft_capacity=Null if soft_capacity_quota == NONE_CONSTANT else soft_capacity_quota,
        )
        self._update(_api, wanted_treeq, **data)
        return self._result_message(wanted_treeq, "modified in filesystem '{}'".format(fs.name))

    @feature_command(
        feature='treeq',
        status=True,
        description="Rename filesystem TreeQ",
        types=(FilesystemOrSnapshot, TreeQName, TreeQName),
    )
    @roles.INFINIDAT.ADMIN
    def rename(self, _api, fs, treeq, new_name):
        self.ENTITY_TYPE.URI = self.ENTITY_TYPE.URI_TEMPLATE.format(fs.id)
        wanted_treeq = self.ENTITY_TYPE.get_by_name(_api, treeq)
        self._update(_api, wanted_treeq, name=new_name)
        return "TreeQ '{}' name updated to '{}' ".format(treeq, new_name)

    @feature_command(
        feature='treeq',
        status=True,
        description="Delete filesystem TreeQ",
        types=(FilesystemOrSnapshot, TreeQName),
    )
    @roles.INFINIDAT.ADMIN
    def delete(self, _api, fs, treeq):
        self.ENTITY_TYPE.URI = self.ENTITY_TYPE.URI_TEMPLATE.format(fs.id)
        wanted_treeq = self.ENTITY_TYPE.get_by_name(_api, treeq)
        return self._delete(_api, wanted_treeq)

@commandgroup(prefix="nlm", parent=FilesystemCommands)
class FilesystemNlmLockCommands(EntityCommands):
    ENTITY_TYPE = NlmLock

    NLM_LOCK_QUERY_FIELDS          = strlist("client state granted_at lock_type filesystem file_path")
    NLM_LOCK_DETAILED_QUERY_FIELDS = strlist("client state granted_at lock_type filesystem file_path offset length owner lock_id")
    NLM_LOCK_QUERY_SORTABLE        = strlist("client state granted_at")
    NLM_LOCK_QUERY_SUBSTITUTES = dict(
        lock_type="type",
        )
    
    @command(
        name="query",
        description="List NLM locks",
        types=(FilesystemOrSnapshot, FilePath, IPAddress, NlmLockState, NlmLockId),
        options=QueryOptions(fields=NLM_LOCK_QUERY_FIELDS, sortable=NLM_LOCK_QUERY_SORTABLE, substitute=NLM_LOCK_QUERY_SUBSTITUTES, detailed=True),
        )
    @render_collection(
        entity=NlmLock,
        fields=NLM_LOCK_QUERY_FIELDS,
        detailed_fields=NLM_LOCK_DETAILED_QUERY_FIELDS,
        detailed_substitute=NLM_LOCK_QUERY_SUBSTITUTES,
        detailed_individually=True,
        substitute=NLM_LOCK_QUERY_SUBSTITUTES,
        alignment=dict(offset="right", length="right"),
        )
    def query(self, _api, fs=None, file_path=None, client=None, state=None, lock_id=None):
        fs_id = fs.id if fs else None
        if lock_id and (fs or file_path or client or state):
            raise BadArguments, "whenever the Lock ID is specified there is no need to specify other attributes"
        if file_path and not fs:
            raise BadArguments, "using the file_path option requires specifying the filesystem name as well"

        is_count = (True if runtime.options.get("format") == "count" else False)
        if is_count and not yesno(
                "Counting locks requires enumeration of all current locks, which can take a while. Are you sure you want to continue?",
                always_prompt=True):
            raise CommandAborted
        client = client.ip if client else None
        return self._query(_api, filesystem_id=fs_id, file_path=file_path, client=client, state=state, lock_id=lock_id, _hook=partial(self._query_hook, _api, fs=fs))

    def _query_hook(self, _api, data, fs):
        fss = {f.id: f for f in FilesystemOrSnapshot.get_all(_api)} if not fs else {fs.id: fs}
        for nl in data:
            lfs = fss[nl['filesystem_id']] if nl['filesystem_id'] else None
            nl['filesystem'] = lfs.name if lfs else None
            nl.file_path = nl.file_path if nl.file_path else nl.file_path_status
        return data
    
    @command(
        name="break",
        description="Break NLM locks",
        types=(FilesystemOrSnapshot, FilePath, IPAddress, NlmLockId),
        )
    @roles.INFINIDAT.ADMIN
    def break_locks(self, _api, fs=None, file_path=None, client=None, lock_id=None):
        fs_id = fs.id if fs else None
        if lock_id and (fs or file_path or client):
            raise BadArguments, "whenever the Lock ID is specified there is no need to specify other attributes"
        if file_path and not fs:
            raise BadArguments, "using the file_path option requires specifying the filesystem name as well"
        client = client.ip if client else None
        res = self._post(_api, None, "break", data=dict(
            filesystem_id=fs_id,
            file_path=file_path,
            client=client,
            lock_id=lock_id,
            ))
        return '{} lock(s) broken'.format(res.locks_broken)
    
    @command(
        description="Remove locks that are no longer relevant, e.g. because the underlying filesystem was deleted",
        )
    @roles.INFINIDAT.ADMIN
    def remove_orphan(self, _api):
        res = self._post(_api, None, "remove_orphan")
        return '{} lock(s) removed'.format(res.locks_removed)
