from munch import Munch

from infinishell.infishell import commandgroup, command, render
from infinishell.infishell.utils import strlist

from ..system import system
from .types import Node, NodeDrive, NodeType
from .options import QueryOptions
from .utils import strip_newline_chars
from . import EntityCommands, roles

QUERY_FIELDS = strlist("name state core_state core_role mgmt_state mgmt_role bbu_protection")
INFINIDAT_QUERY_FIELDS = QUERY_FIELDS + strlist("vendor model")

DRIVE_QUERY_FIELDS = strlist("name type state state_description")
INFINIDAT_DRIVE_QUERY_FIELDS = strlist("name type state vendor model serial_number firmware state_description")

POWER_QUERY_FIELDS = strlist("name bbu_protection psu1 psu2")
EXTENDED_POWER_QUERY_FIELDS = POWER_QUERY_FIELDS + strlist("psu1_a psu2_a power_w")

POWER_QUERY_SUBSTITUTES = dict(
    power_w="POWER(W)",
    psu1_a="PSU1(A)",
    psu2_a="PSU2(A)",
)


@commandgroup(prefix="node")
class NodeCommands(EntityCommands):
    ENTITY_TYPE = Node

    @command(
        description="List system nodes",
        options=QueryOptions(fields=QUERY_FIELDS, infinidat_fields=INFINIDAT_QUERY_FIELDS),
        )
    @render(
        fields=QUERY_FIELDS,
        infinidat=dict(
            fields=INFINIDAT_QUERY_FIELDS,
            ),
        technician=dict(
            fields=INFINIDAT_QUERY_FIELDS,
            ),
        )
    def query(self, _api):
        return self._query(_api, _hook=self._query_hook)

    def _query_hook(self, nodes):
        for node in nodes:
            services = Munch((s.name, s) for s in node.services)
            yield Munch(name=node.name,
                        state=node.state,
                        vendor=node.vendor,
                        model=node.model,
                        core_state=services.core.state,
                        core_role=services.core.role,
                        mgmt_state=services.mgmt.state,
                        mgmt_role=services.mgmt.role,
                        bbu_protection=node.node_bbu_protection,
                        )

    @command(
        description="Initiate node activation process",
        types=(Node,),
        )
    @roles.INFINIDAT.TECHNICIAN
    def activate(self, _api, node):
        if node.state == "ACTIVE":
            return "%s is already ACTIVE" % node.name
        _api.post("components/nodes/{}/phase_in".format(node.id))
        return "Activating {}".format(node.name)

    @command(
        description="Initiate node deactivation process",
        types=(Node,),
        )
    @roles.INFINIDAT.TECHNICIAN
    def deactivate(self, _api, node):
        _api.post("components/nodes/{}/phase_out".format(node.id))
        return "Deactivating {}".format(node.name)


@commandgroup(prefix="drive", parent=NodeCommands)
class NodeDriveCommands(EntityCommands):
    ENTITY_TYPE = NodeDrive

    @command(
        description="List node drives",
        types=(NodeDrive, Node, NodeType),
        hidden_substitue=dict(name="drive"),
        options=QueryOptions(fields=DRIVE_QUERY_FIELDS, infinidat_fields=INFINIDAT_DRIVE_QUERY_FIELDS),
        )
    @render(
        fields=DRIVE_QUERY_FIELDS,
        infinidat=dict(
            fields=INFINIDAT_DRIVE_QUERY_FIELDS
            ),
        )
    def query(self, _api, drive=None, node=None, type=None):
        return self._query(_api, drive, node_index=node, type=type, _hook=self._node_drive_query_hook)

    def _node_drive_query_hook(self, drives):
        for drive in drives:
            if drive.get("state_description"):
                drive.state_description = strip_newline_chars(drive.state_description)
            if drive.state == "UNCONFIGURED" and system.user.roles.is_customer():
                continue
            yield drive



@commandgroup(prefix="health", parent=NodeCommands)
class NodeHealthCommands(object):
    pass


@commandgroup(prefix="power", parent=NodeHealthCommands)
class NodePowerCommands(EntityCommands):
    ENTITY_TYPE = Node

    @command(
        description="List system nodes",
        options=QueryOptions(fields=POWER_QUERY_FIELDS, substitute=POWER_QUERY_SUBSTITUTES),
    )
    @render(
        fields=POWER_QUERY_FIELDS,
        substitute=POWER_QUERY_SUBSTITUTES,
        infinidat=dict(
            fields=EXTENDED_POWER_QUERY_FIELDS,
        ),
        technician=dict(
            fields=EXTENDED_POWER_QUERY_FIELDS,
        ),
    )
    def query(self, _api):
        return self._query(_api, _hook=self._query_hook)

    def _query_hook(self, nodes):
        for node in nodes:
            services = Munch((s.name, s) for s in node.services)
            power_supplies = node.power_supplies
            yield Munch(name=node.name,
                        bbu_protection=node.node_bbu_protection,
                        power_w=node.power,
                        psu1=power_supplies[0].state if len(power_supplies) > 0 else "N/A",
                        psu2=power_supplies[1].state if len(power_supplies) > 1 else "N/A",
                        psu1_a=power_supplies[0].current if len(power_supplies) > 0 else "N/A",
                        psu2_a=power_supplies[1].current if len(power_supplies) > 1 else "N/A",
                        )
