import datetime
import uuid
from collections import OrderedDict
from functools import partial
from itertools import izip

from infinishell.infishell import commandgroup, command, render, runtime, get_active_shell
from infinishell.infishell.exceptions import BadArguments, CommandAborted, CommandFailed
from infinishell.infishell.types import YesNo, TimeDelta, Str, Hostname
from infinishell.infishell.utils import strlist
from munch import Munch

from . import (
    render_collection,
    render_multi_collection,
    roles,
    EntityCommands,
    LazyIdNameMapping,
    LazyEntityMapping,
    ENTITY,
    set_rendered_unit,
)
from .options import QueryOptions
from .types import (Dataset, MasterFilesystem, MasterVolume,
                    NonReplicatedDataset,
                    NonReplicatedCg,
                    DatasetSnapshot,
                    RemoteDatasetSnapshot,
                    ReplicatedEntity,
                    ReplicatedOrMobilityEntity,
                    NonReplicatedEntity,
                    ReplicaSyncState)
from .types import (
    Name,
    TimeDeltaOrNone,
    ReplicaType,
    ReplicaTypeExcludeActiveActive,
    CgOrDataset,
    CgOrVolume,
    Replica,
    ReplicaRole,
    ReplicaDatasetPair,
    ReplicationLinkBySystemName,
    NetworkSpace,
    ReplicationNetworkSpace,
    Pool,
    ConsistencyGroup,
    ReplicaState,
    ReplicaCapacityUnit,
    HostnameOrNone,
    NONE_CONSTANT,
    ActiveActivePreferred
)
from .utils import is_detailed_query, increased_api_timeout
from ..exceptions import NoSuchEntity, ManagementException, ConnectionFailed
from ..features import feature_command
from ..remote import get_remote_system
from ..renderer import InfiniboxRenderer
from ..system import system
from ...utils import (
    get_username_password,
    humanly_join,
    yesno,
    yesno_ensure_interactive,
)


def generate_entity_name():
    return str(uuid.uuid1())

class RemoteEntity(Str):
    def get_description(self):
        return "Name of a remote entity (auto-completion unavailable)"

class LazyIdNameMappingNoneDetached(LazyIdNameMapping):
    def _get_default(self, id):
        if id is None:
            return 'DETACHED'
        return LazyIdNameMapping._get_default(self, id)

class ReplicationCommandsMixin(object):
    def _should_retry_forcefully(self, exc):
        return exc.is_remote or exc.code in (
            "REMOTE_API_AUTHENTICATION_REQUIRED",
            "REPLICA_DELETION_ERROR",
            "REPLICA_DELETION_NOT_ALLOWED_ON_TARGET",
            "LINK_IO_ERROR",
            "REPLICA_REMOVE_MEMBER_NOT_ALLOWED_ON_TARGET",
            )

class ReplicaCommandsBase(EntityCommands, ReplicationCommandsMixin):
    ENTITY_TYPE = Replica

    def _result_message(self, entity, msg, subject=None, source=None):
        is_cg = entity.entity_type == "CONSISTENCY_GROUP"
        entity_type = ConsistencyGroup if is_cg else MasterVolume if entity.entity_type == "VOLUME" else MasterFilesystem
        if not source:
            if is_cg:
                source = ConsistencyGroup.get_by_id(system.api, entity.local_cg_id)
            elif entity.get("entity_pairs"):
                pair = entity.entity_pairs[0]
                source = Munch(id=pair.local_entity_id, name=pair.local_entity_name)
            else:
                source = entity_type.get_by_id(system.api, entity.local_entity_id)
        return "Replica for {} {}".format(self._result_subject_and_entity(source, entity_type), msg)

    def _get_replica(self, _api, source, link=None):
        is_cg = "members_count" in source
        filter = _api.create_filter().eq("local_cg_id" if is_cg else "local_entity_id", source.id)
        if link:
            filter = filter.eq("link_id", link.id)
        replicas = Replica.get_all(_api, filter=filter)
        if len(replicas) != 1:
            template = "{} is not replicated{}" if not replicas else "{} has more than one replica{}"
            raise NoSuchEntity(template.format(
                self._result_subject_and_entity(source, ConsistencyGroup if is_cg else MasterVolume if source.dataset_type == "VOLUME" else MasterFilesystem, capitalize=True),
                "" if not link else ' on "{}"'.format(link.remote_system_name)
                ))
        return replicas[0]

    def _create_replica_info(self, remote_api, link, source, interval, rpo, replication_type="ASYNC", active_active_preferred=None, remote_pool=None, new_target_name=None):
        info = Munch(
            link_id=link.id,
            replication_type=replication_type,
            entity_pairs=[],
            )
        if replication_type == "ASYNC":
            info.update(rpo_type="TIME")
        if rpo:
            info.rpo_value = rpo*1000 if rpo != TimeDeltaOrNone.CONSTANTS.NONE else 0
        if interval:
            info.sync_interval = interval*1000 if interval != TimeDeltaOrNone.CONSTANTS.NONE else 0
        if remote_pool:
            info.remote_pool_id = Pool.get_by_name(remote_api, remote_pool).id
        if active_active_preferred:
            info.is_preferred = True if active_active_preferred == "local" else False
        if source._is_dataset:
            info.update(
                entity_type="VOLUME" if source.dataset_type == "VOLUME" else "FILESYSTEM",
                )
        else:
            info.update(
                entity_type="CONSISTENCY_GROUP",
                local_cg_id=source.id,
                remote_cg_name=new_target_name,
                )
        return info

    def _old_create_replica_info(self, remote_api, link, source, interval, rpo, remote_pool=None, new_target_name=None):
        info = Munch(
            link_id=link.id,
            sync_interval=interval*1000 if interval != TimeDeltaOrNone.CONSTANTS.NONE else 0,
            rpo_value=rpo*1000 if rpo != TimeDeltaOrNone.CONSTANTS.NONE else 0,
            rpo_type="TIME",
            entity_pairs=[],
            )
        if remote_pool:
            info.remote_pool_id = Pool.get_by_name(remote_api, remote_pool).id
        if source._is_dataset:
            info.update(
                entity_type="VOLUME" if source.dataset_type == "VOLUME" else "FILESYSTEM",
                )
        else:
            info.update(
                entity_type="CONSISTENCY_GROUP",
                local_cg_id=source.id,
                remote_cg_name=new_target_name,
                )
        return info

    def _get_source(self, sources, dataset_type):
        if sources == [] or sources is None:
            return None
        if len(sources) > 1 and dataset_type is None:
            raise BadArguments("There is more than one entity with the name '{}', please set dataset_type".format(sources[0].name))
        if len(sources) > 1 and dataset_type:
            if dataset_type == "CG":
                results = [s for s in sources if s._is_dataset == False]    
            else:
                results = [s for s in sources if s.get("dataset_type", "").lower() == dataset_type.lower()]
            if not results:
                raise BadArguments("There is no {} with name {}".format(dataset_type.lower(), sources[0].name))
            return results and results[0]
        elif len(sources) == 1 and dataset_type:
            if dataset_type == "CG":
                if sources[0]._is_dataset == True:
                    raise BadArguments("Chosen source '{}' is of type {} but dataset_type is set to {}".format(sources[0].name, sources[0].dataset_type, dataset_type))
            else:   
                if sources[0].get("dataset_type","").lower() != dataset_type.lower():
                    raise BadArguments("Chosen source '{}' is of type {} but dataset_type is set to {}".format(sources[0].name, sources[0].get('dataset_type',"CG"), dataset_type))
        return sources[0]


@commandgroup(prefix="replica")
class ReplicaCommands(ReplicaCommandsBase):
    MOBILITY_QUERY_FIELDS = strlist("""
        _local_entity_name
        _entity_display_type
        role
        _remote_system_name
        _remote_entity_name
        sync_state
        throughput
        """)
    MOBILITY_DETAILED_QUERY_FIELDS = strlist("""
        _local_entity_name
        _entity_display_type
        local_pool_name
        role
        _remote_system_name
        _remote_entity_name
        remote_pool_name
        sync_state
        state_reason
        latency
        throughput
        created_at
        updated_at
        pairs
        """)
    ACTIVE_ACTIVE_QUERY_FIELDS = strlist("""
        _local_entity_name
        _entity_display_type
        _remote_system_name
        _remote_entity_name
        is_preferred
        sync_state
        latency
        """)
    ACTIVE_ACTIVE_DETAILED_QUERY_FIELDS = strlist("""
        _local_entity_name
        _entity_display_type
        local_pool_name
        _remote_system_name
        _remote_entity_name
        remote_pool_name
        is_preferred
        sync_state
        latency
        throughput
        created_at
        updated_at
        reserved_for_infinisafe
        pairs
        """)
    SYNC_QUERY_FIELDS = strlist("""
        _local_entity_name
        _entity_display_type
        role
        _remote_system_name
        _remote_entity_name
        state
        sync_state
        latency
        """)
    SYNC_DETAILED_QUERY_FIELDS = strlist("""
        _local_entity_name
        _entity_display_type
        local_pool_name
        role
        _remote_system_name
        _remote_entity_name
        remote_pool_name
        state
        state_reason
        state_description
        sync_state
        latency
        throughput
        rpo_state
        rpo_value
        created_at
        updated_at
        reserved_for_infinisafe
        pairs
        """)
    QUERY_FIELDS = strlist("""
        _local_entity_name
        _entity_display_type
        role
        _remote_system_name
        _remote_entity_name
        state
        job_state
        progress
        rpo_state
        restore_point
        """)
    DETAILED_QUERY_FIELDS = strlist("""
        _local_entity_name
        _entity_display_type
        local_pool_name
        role
        _remote_system_name
        _remote_entity_name
        remote_pool_name
        state
        state_reason
        state_description
        job_state
        progress
        throughput
        rpo_state
        rpo_value
        sync_interval
        created_at
        updated_at
        last_synchronized
        restore_point
        staging_area_allocated_size
        reserved_for_infinisafe
        pairs
        """)
    
    # note that the following COLUMNS property is created to combine existing defined lists,
    # while avoiding duplicates when using --columns=*
    # chose not to use a 'set' so that the existing order is, more or less, preserved
    COLUMNS = list(DETAILED_QUERY_FIELDS)
    for f in (SYNC_DETAILED_QUERY_FIELDS + ACTIVE_ACTIVE_DETAILED_QUERY_FIELDS):
        if f not in COLUMNS:
            COLUMNS.append(f)
    COLUMNS = COLUMNS - strlist('pairs')
    
    QUERY_SUBSTITUTES = dict(
        _remote_system_name="remote_system",
        _entity_display_type="type",
        _local_entity_name="local_dataset",
        _remote_entity_name="remote_dataset",
        staging_area_allocated_size="staging_area_size",
        jos_state="sync_job",
        reserved_for_infinisafe="infinisync_protected",
        sync_interval='interval',
        is_preferred='preferred',
        )
    SORTABLE_FIELDS = strlist("""
        role
        _local_entity_name
        _entity_display_type
        state
        restore_point
        reserved_for_infinisafe
        is_preferred
        sync_state
        """)
    
    def _validate_replication_type_compatibility(self, replication_type, disallowed_replication_types, action, proposition):
        if not (action or replication_type or disallowed_replication_types):
            return
        else:
            proposition = " " + proposition if str(proposition).__len__() > 0 else proposition 
            if replication_type in disallowed_replication_types:
                raise BadArguments, "Cannot {}{} a replica of type {}".format(action, proposition, replication_type)            

    def _validate_non_mobility_replica(self, replica, source):
        if replica.get('mobility_source') != None:
            raise CommandAborted("Dataset {} is part of a mobility replica that cannot be deleted".format(source.get("name", None)))
    
    
    @feature_command(
        feature='sync_replication',
        status=True,
        description="List existing replicas",
        types=(ReplicaType, ReplicaCapacityUnit, ReplicationLinkBySystemName, ReplicatedOrMobilityEntity.List, ReplicaRole, ReplicaState, YesNo, CgOrDataset, ReplicaSyncState),
        options=QueryOptions(fields=COLUMNS, substitute=QUERY_SUBSTITUTES, sortable=SORTABLE_FIELDS, detailed=True),
        substitute=dict(
            link="system",
            datasets="local_dataset",
            ),
        )
    @render_multi_collection(
        entity=Replica,
        fields=QUERY_FIELDS,
        substitute=QUERY_SUBSTITUTES,
        detailed_individually=True,
        detailed_fields=DETAILED_QUERY_FIELDS,
        value_properties=dict(
            skip_value_renderer=['latency'],
            ),
        )
    def query(self, _api, replication_type=None, unit=None, link=None,
              datasets=None, role=None, state=None, infinisync_protected=None, type=None, sync_state=None):
        set_rendered_unit(unit)
        datasets = [i for d in datasets for i in d] if datasets else None
        link_mapping = LazyEntityMapping(ReplicationLinkBySystemName, _api, fields=("remote_system_name", "link_state"))

        if (type == "CG"):
            type = "CONSISTENCY_GROUP"

        sync_data = None
        if replication_type in ["SYNC", None]:
            sync_data = Munch(
                data=self._query(_api,
                    None,
                    link_id=link,
                    role=role,
                    replication_type="SYNC",
                    state=state,
                    reserved_for_infinisafe=infinisync_protected,
                    entity_type=type,
                    sync_state=sync_state,
                    _hook=partial(self._query_hook, _api=_api, lazy_link_by_id=link_mapping, datasets=datasets),
                    _sort_on_client_side=True,
                    ),
                title="SYNC REPLICAS",
                fields=self.SYNC_QUERY_FIELDS, 
                detailed_fields=self.SYNC_DETAILED_QUERY_FIELDS,
            )
        async_data = None
        if replication_type in ["ASYNC", None]:
            async_data = Munch(
                data=self._query(_api,
                    None,
                    link_id=link,
                    role=role,
                    replication_type="ASYNC",
                    state=state,
                    reserved_for_infinisafe=infinisync_protected,
                    entity_type=type,
                    sync_state=sync_state,
                    _hook=partial(self._query_hook, _api=_api, lazy_link_by_id=link_mapping, datasets=datasets),
                    _sort_on_client_side=True,
                    ),
                title="ASYNC REPLICAS",
                fields=self.QUERY_FIELDS, 
                detailed_fields=self.DETAILED_QUERY_FIELDS,
            )
        active_active_data = None
        if replication_type in ["ACTIVE_ACTIVE", None]:
            mobility_filter = _api.create_filter().is_null("mobility_source")
            active_active_data = Munch(
                data=self._query(_api,
                    None,
                    link_id=link,
                    role=role,
                    replication_type="ACTIVE_ACTIVE",
                    state=state,
                    reserved_for_infinisafe=infinisync_protected,
                    entity_type=type,
                    sync_state=sync_state,
                    _hook=partial(self._query_hook, _api=_api, lazy_link_by_id=link_mapping, datasets=datasets),
                    _filter=mobility_filter,
                    _sort_on_client_side=True,
                    ),
                title="ACTIVE-ACTIVE REPLICAS",
                fields=self.ACTIVE_ACTIVE_QUERY_FIELDS,
                detailed_fields=self.ACTIVE_ACTIVE_DETAILED_QUERY_FIELDS,
            )
        mobility_data = None
        if replication_type in [None]:
            mobility_source = True if role == ReplicaRole.SOURCE else False if role == ReplicaRole.TARGET else None
            if role:
                mobility_filter = _api.create_filter().eq("mobility_source", mobility_source)
            else:
                mobility_filter = _api.create_filter().is_not_null("mobility_source")
            mobility_data = Munch(
                data=self._query(_api,
                                 None,
                                 link_id=link,
                                 replication_type="ACTIVE_ACTIVE",
                                 state=state,
                                 reserved_for_infinisafe=infinisync_protected,
                                 entity_type=type,
                                 sync_state=sync_state,
                                 _hook=partial(self._query_hook, _api=_api, lazy_link_by_id=link_mapping,
                                               datasets=datasets),
                                 _filter=mobility_filter,
                                 _sort_on_client_side=True,
                                 ),
                title="MOBILITY REPLICAS",
                fields=self.MOBILITY_QUERY_FIELDS,
                detailed_fields=self.MOBILITY_DETAILED_QUERY_FIELDS,
            )
        control = {
            "allow_json": True,
            "allow_csv": True,
            "additional_fields": strlist("replication_type") if replication_type == None else strlist(""),
            "fields": OrderedDict(((i,i) for i in (self.MOBILITY_QUERY_FIELDS if mobility_data else strlist("")) + (self.ACTIVE_ACTIVE_QUERY_FIELDS if active_active_data else strlist("")) + (self.SYNC_QUERY_FIELDS if sync_data else strlist("")) + (self.QUERY_FIELDS if async_data else strlist("")))).keys(),
            "detailed_fields": OrderedDict(((i,i) for i in (self.MOBILITY_DETAILED_QUERY_FIELDS if mobility_data else strlist("")) + (self.ACTIVE_ACTIVE_DETAILED_QUERY_FIELDS if active_active_data else strlist("")) + (self.SYNC_DETAILED_QUERY_FIELDS if sync_data else strlist("")) + (self.DETAILED_QUERY_FIELDS if async_data else strlist("")))).keys(),
        }
        data =[x for x in [active_active_data, sync_data, async_data, mobility_data] if (x and x.get('data', False))]
        return {"control":control, "data":data}
        

    def _query_hook(self, replicas, _api=None, lazy_link_by_id=None, datasets=None):
        final_replicas = []
        dataset_ids = set((dataset.id for dataset in datasets)) if datasets else None
        for r in replicas:
            is_cg_replica = r.entity_type == "CONSISTENCY_GROUP"
            is_mobility = r.mobility_source != None
            if datasets:
                if is_cg_replica and r.local_cg_id not in dataset_ids \
                or not is_cg_replica and not any(pair.local_entity_id in dataset_ids for pair in r.entity_pairs):
                    continue
            r = Munch((k, v) for k, v in r.iteritems() if k[0] != "_")
            if lazy_link_by_id:
                link = lazy_link_by_id[r.link_id]
                r._remote_system_name = link.remote_system_name
                if link.link_state == "DOWN":
                    r.state_reason = "LINK_DOWN"
            r.rpo_state = r.rpo_state if r.rpo_state else "N/A"
            r._entity_display_type = "CG" if is_cg_replica else r.entity_type
            r._local_entity_name, r._remote_entity_name = [], []
            if not is_mobility:
                r.is_preferred = None if r.is_preferred is None else "LOCAL" if r.is_preferred else "REMOTE"
            r.pairs = ()
            if is_cg_replica:
                r._local_entity_name = r.get("local_cg_name") or "<Local CG: {}>".format(r.local_cg_id)
                r._remote_entity_name = r.get("remote_cg_name") or "<Remote CG: {}>".format(r.remote_cg_id)
                if r.entity_pairs:
                    if r.replication_type == "ACTIVE_ACTIVE":
                        r.pairs = (['Local Dataset','Remote Dataset', 'Initializing'],)
                    else:
                        r.pairs = (['Local Dataset','Remote Dataset', 'Initializing', 'Restore Point'],)
                else:
                    r.pairs = ()
                for pair in r.entity_pairs or ():
                    if r.replication_type == "ACTIVE_ACTIVE":
                        r.pairs = r.pairs + ([pair.local_entity_name, pair.remote_entity_name, "yes" if pair.is_initial else "no"],)
                    else:
                        timestamp = long(pair.restore_point) if pair.restore_point else '-'
                        restore_point =timestamp if timestamp == '-' else datetime.datetime.fromtimestamp(timestamp/1000.0).strftime("%Y-%m-%d %H:%M:%S")
                        r.pairs = r.pairs + ([pair.local_entity_name, pair.remote_entity_name, "yes" if pair.is_initial else "no", restore_point],)
            else:
                r._local_entity_name = r.entity_pairs[0].local_entity_name if r.entity_pairs else "N/A"
                r._remote_entity_name = r.entity_pairs[0].remote_entity_name if r.entity_pairs else "N/A"

            if is_mobility:
                r.replication_type = "MOBILITY"
                r.role = ReplicaRole.SOURCE if r.mobility_source == True else ReplicaRole.TARGET if r.mobility_source == False else None
                if r.role == "TARGET":
                    r.update(
                        state="TARGET",
                        throughput="N/A",
                    )
                else:
                    r.progress = "{}%".format(r['progress']) if r['progress'] is not None else None
            else:
                if r.role == "TARGET":
                    r.update(
                        state="TARGET",
                        progress="N/A",
                        throughput="N/A",
                        )
                else:
                    r.progress = "{}%".format(r['progress']) if r['progress'] is not None else None
            final_replicas.append(r)
        return final_replicas


    @feature_command(
        feature='sync_replication',
        status=False,
        name="query",
        description="List existing replicas",
        types=(ReplicationLinkBySystemName, ReplicatedEntity.List, ReplicaRole),
        options=QueryOptions(fields=DETAILED_QUERY_FIELDS, substitute=QUERY_SUBSTITUTES, sortable=SORTABLE_FIELDS, detailed=True),
        substitute=dict(
            link="system",
            datasets="local_dataset",
            ),
        )
    @render_collection(
        entity=Replica,
        fields=QUERY_FIELDS,
        substitute=QUERY_SUBSTITUTES,
        detailed_individually=True,
        detailed_fields=DETAILED_QUERY_FIELDS,
        )
    def old_query(self, _api, link=None, datasets=None, role=None):
        link_mapping = LazyEntityMapping(ReplicationLinkBySystemName, _api, fields=("remote_system_name", "link_state"))

        # Workaround for staging area allocated size not being shown when querying multiple replicas
        prefetched_replica = None
        if datasets and len(datasets) == 1:
            prefetched_replica_filter = self._create_query_filter(_api, {
                "local_entity_id" if datasets[0]._is_dataset else "local_cg_id" : datasets[0].id,
                "link_id" : link,
                "role" : role
                })
            prefetched_replica = Replica.get_all(_api, filter=prefetched_replica_filter)
            if not prefetched_replica:
                return []
            if len(prefetched_replica) > 1:
                prefetched_replica = None
            else:
                prefetched_replica = [Replica.get_by_id(_api, prefetched_replica[0].id)]
        # End workaround

        return self._query(_api,
            prefetched_replica,
            link_id=link,
            role=role,
            _hook=partial(self._query_hook, _api=_api, lazy_link_by_id=link_mapping, datasets=datasets),
            _sort_on_client_side=True,
            # _default_sort=_api.create_sort().asc("role").asc("_local_entity_name"),
            )


    @feature_command(
        feature='sync_replication',
        status=True,
        description="Create a new replica along with corresponding remote datasets",
        types=(ReplicationLinkBySystemName, NonReplicatedEntity, RemoteEntity, ReplicaType, ActiveActivePreferred, TimeDeltaOrNone, TimeDeltaOrNone, CgOrDataset, Name, Name),
        substitute=dict(
            link="system",
            ),
        )
    @roles.ADMIN.POOLADMIN
    def create(self, _api, link, source, remote_pool, replication_type, active_active_preferred=None, interval=None, rpo=None, dataset_type=None, new_target_name=None, new_datasets_suffix=None):
        source = self._get_source(source, dataset_type)
        if (replication_type == "SYNC" or replication_type == "ASYNC") and (active_active_preferred is not None):
            raise BadArguments("active_active_preferred parameter is applicable only for ACTIVE_ACTIVE replication")
        if (replication_type == "SYNC" or replication_type == "ACTIVE_ACTIVE") and (source.dataset_type == "FILESYSTEM" if source._is_dataset else False):
            raise BadArguments("SYNC and ACTIVE_ACTIVE replication types are not available for Filesystems")
        if (replication_type == "SYNC" or replication_type == "ACTIVE_ACTIVE") and (rpo is not None or interval is not None):
            raise BadArguments("RPO and Interval parameter are applicable only for async replication")
        if replication_type == "ASYNC" and (rpo is None or interval is None):
            raise BadArguments("RPO and Interval parameter are mandatory for async replication")     
        if source._is_dataset and new_datasets_suffix:
            raise BadArguments("new_datasets_suffix is only applicable when source is a consistency group")
        with get_remote_system(_api, link).api as remote_api:
            replica_info = self._create_replica_info(
                remote_api=remote_api,
                link=link,
                source=source,
                interval=interval,
                rpo=rpo,
                replication_type=replication_type,
                active_active_preferred=active_active_preferred,
                remote_pool=remote_pool,
                new_target_name=new_target_name
                )
            if source._is_dataset:
                replica_info.entity_pairs.append(Munch(
                    local_entity_id=source.id,
                    local_base_action="NO_BASE_DATA",
                    remote_base_action="CREATE",
                    remote_entity_name=new_target_name,
                    ))
            else:
                replica_info.entity_pairs.extend(Munch(
                    local_entity_id=member.id,
                    local_base_action="NO_BASE_DATA",
                    remote_base_action="CREATE",
                    remote_entity_name="{}{}".format(member.name, new_datasets_suffix or ""),
                    ) for member in Dataset.iter_all(_api, filter=_api.create_filter(cg_id=source.id)))
            return self._create(_api, **replica_info)

    @feature_command(
        feature='sync_replication',
        status=True,
        name="create",
        description="Create a new replica along with corresponding remote datasets",
        types=(ReplicationLinkBySystemName, NonReplicatedEntity, RemoteEntity, ReplicaType, ActiveActivePreferred, TimeDeltaOrNone, TimeDeltaOrNone, CgOrDataset, Name, Name, YesNo),
        substitute=dict(
            link="system",
            mobility="mobility_replica",
            ),
        )
    @roles.INFINIDAT
    def infinidat_create(self, _api, link, source, remote_pool, replication_type, active_active_preferred=None, interval=None, rpo=None, dataset_type=None, new_target_name=None, new_datasets_suffix=None, mobility=None):
        source = self._get_source(source, dataset_type)
        if (replication_type == "SYNC" or replication_type == "ASYNC") and (active_active_preferred is not None):
            raise BadArguments("active_active_preferred parameter is applicable only for ACTIVE_ACTIVE replication")
        if mobility and replication_type != "ACTIVE_ACTIVE":
            raise BadArguments("The mobility parameter is applicable only for active-active replication")
        if mobility and (source.dataset_type != "VOLUME" if source._is_dataset else False):
            raise BadArguments("The mobility parameter is applicable only for volume replication")
        if (replication_type == "SYNC" or replication_type == "ACTIVE_ACTIVE") and (source.dataset_type == "FILESYSTEM" if source._is_dataset else False):
            raise BadArguments("SYNC and ACTIVE_ACTIVE replication types are not available for Filesystems")
        if (replication_type == "SYNC" or replication_type == "ACTIVE_ACTIVE") and (rpo is not None or interval is not None):
            raise BadArguments("RPO and Interval parameter are applicable only for async replication")
        if replication_type == "ASYNC" and (rpo is None or interval is None):
            raise BadArguments("RPO and Interval parameter are mandatory for async replication")     
        if source._is_dataset and new_datasets_suffix:
            raise BadArguments("new_datasets_suffix is only applicable when source is a consistency group")
        with get_remote_system(_api, link).api as remote_api:
            replica_info = self._create_replica_info(
                remote_api=remote_api,
                link=link,
                source=source,
                interval=interval,
                rpo=rpo,
                replication_type=replication_type,
                active_active_preferred=active_active_preferred,
                remote_pool=remote_pool,
                new_target_name=new_target_name
                )
            if mobility:
                replica_info.mobility_source=mobility
            if source._is_dataset:
                replica_info.entity_pairs.append(Munch(
                    local_entity_id=source.id,
                    local_base_action="NO_BASE_DATA",
                    remote_base_action="CREATE",
                    remote_entity_name=new_target_name,
                    ))
            else:
                replica_info.entity_pairs.extend(Munch(
                    local_entity_id=member.id,
                    local_base_action="NO_BASE_DATA",
                    remote_base_action="CREATE",
                    remote_entity_name="{}{}".format(member.name, new_datasets_suffix or ""),
                    ) for member in Dataset.iter_all(_api, filter=_api.create_filter(cg_id=source.id)))
            return self._create(_api, **replica_info)


    @feature_command(
        feature='sync_replication',
        status=False,
        name="create",
        description="Create a new replica along with corresponding remote datasets",
        types=(ReplicationLinkBySystemName, NonReplicatedEntity, TimeDeltaOrNone, TimeDeltaOrNone, RemoteEntity, Name, Name),
        substitute=dict(
            link="system",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def old_create(self, _api, link, source, interval, rpo, remote_pool, new_target_name=None, new_datasets_suffix=None):
        if source._is_dataset and new_datasets_suffix:
            raise BadArguments("new_datasets_suffix is only applicable when source is a consistency group")
        with get_remote_system(_api, link).api as remote_api:
            replica_info = self._old_create_replica_info(
                remote_api=remote_api,
                link=link,
                source=source,
                interval=interval,
                rpo=rpo,
                remote_pool=remote_pool,
                new_target_name=new_target_name,
                )
            if source._is_dataset:
                replica_info.entity_pairs.append(Munch(
                    local_entity_id=source.id,
                    local_base_action="NO_BASE_DATA",
                    remote_base_action="CREATE",
                    remote_entity_name=new_target_name,
                    ))
            else:
                replica_info.entity_pairs.extend(Munch(
                    local_entity_id=member.id,
                    local_base_action="NO_BASE_DATA",
                    remote_base_action="CREATE",
                    remote_entity_name="{}{}".format(member.name, new_datasets_suffix or ""),
                    ) for member in Dataset.iter_all(_api, filter=_api.create_filter(cg_id=source.id)))
            return self._create(_api, **replica_info)


    @command(
        description="Delete replica",
        types=(ReplicatedEntity, CgOrDataset, YesNo, YesNo, YesNo),
        substitute=dict(
            source="local_dataset",
            ),
        )
    @roles.ADMIN.POOLADMIN
    def delete(self, _api, source, dataset_type=None, retain_staging_area=False, force=False, force_keep_serial=None):
        source = self._get_source(source, dataset_type)
        replica = self._get_replica(_api, source)
        if replica.get('mobility_source') != None:
            raise CommandAborted, "Dataset {} is part of a mobility replica that cannot be deleted".format(source.name)
        replication_type=replica.get('replication_type', None)
        if source._is_dataset and replica.entity_type == "CONSISTENCY_GROUP":
            raise CommandAborted, "Dataset '{}' is part of a replicated CG and cannot be deleted".format(source.name)
        if force_keep_serial != None and replication_type != "ACTIVE_ACTIVE":
            raise BadArguments, "Cannot set force keep serial for {} replication".format(replication_type)
        if force_keep_serial != None and not force:
            raise BadArguments, "Cannot set force keep serial without setting force to 'yes'"
        if retain_staging_area and replication_type == "SYNC":
            raise BadArguments, "Cannot set retain staging area for {} replication".format(replication_type)
        params = dict(retain_staging_area=retain_staging_area,)
        if force:
            if replication_type == "ACTIVE_ACTIVE":
                params.update(force_on_local="true",)
                if force_keep_serial != None:
                    params.update(keep_serial_on_local=force_keep_serial,)
            else:
                params.update(
                    force_if_no_remote_credentials="true",
                    force_if_remote_error="true",
                    force_on_target="true",
                    )
        send_request = partial(self._del, _api, replica, params=params)
        result = None
        try:
            with get_remote_system(_api, replica).api:
                result = send_request()
        except ManagementException as e:
            if self._should_retry_forcefully(e):
                if not force:
                    raise CommandAborted(e.formatted_message + "\nTo perform the operation forcefully please execute the command again with force=yes")
            else:
                raise

        message = ""
        if replication_type == "ACTIVE_ACTIVE":
            on_local_message = " on the local system only" if force == True else ""
            serial_message = ", and the volume serial was kept locally. Make sure that the remote volume serial was changed to avoid having two volumes with the same serial" if force_keep_serial == True else ""
            message = on_local_message + serial_message
        if message != "":
            return self._result_message(replica, "deleted, " + message, source=source)
        else:
            staging_message = "and its staging area " + ("retained" if retain_staging_area else "discarded")
            if retain_staging_area and result and result.entity_pairs:
                non_reclaimed_names = sorted(pair.local_entity_name
                                            for pair in result.entity_pairs
                                            if not pair.get("_local_reclaimed_snapshot_id", False))
                if non_reclaimed_names:
                    staging_message = "but there was nothing in the staging area to retain" if len(result.entity_pairs) == len(non_reclaimed_names) \
                                    else "and its staging area was retained, except for the following datasets: " + humanly_join(non_reclaimed_names)
            return self._result_message(replica, "deleted, " + staging_message, source=source)
    
    
    @command(
        name="delete",
        description="Delete replica",
        types=(ReplicatedOrMobilityEntity, CgOrDataset, YesNo, YesNo, YesNo),
        substitute=dict(
            source="local_dataset",
            ),
        )
    @roles.INFINIDAT
    def infinidat_delete(self, _api, source, dataset_type=None, retain_staging_area=False, force=False, force_keep_serial=None):
        source = self._get_source(source, dataset_type)
        replica = self._get_replica(_api, source)
        replication_type=replica.get('replication_type',None)
        if source._is_dataset and replica.entity_type == "CONSISTENCY_GROUP":
            raise CommandAborted, "Dataset '{}' is part of a replicated CG and cannot be deleted".format(source.name)
        if force_keep_serial != None and replication_type != "ACTIVE_ACTIVE":
            raise BadArguments, "Cannot set force keep serial for {} replication".format(replication_type)
        if force_keep_serial != None and not force:
            raise BadArguments, "Cannot set force keep serial without setting force to 'yes'"
        params = dict(retain_staging_area=retain_staging_area,)
        if force:
            if replication_type == "ACTIVE_ACTIVE":
                params.update(force_on_local="true",)
                if force_keep_serial != None:
                    params.update(keep_serial_on_local=force_keep_serial,)
            else:
                params.update(
                    force_if_no_remote_credentials="true",
                    force_if_remote_error="true",
                    force_on_target="true",
                    )
        send_request = partial(self._del, _api, replica, params=params)
        result = None
        try:
            with get_remote_system(_api, replica).api:
                result = send_request()
        except ManagementException as e:
            if self._should_retry_forcefully(e):
                if not force:
                    raise CommandAborted(e.formatted_message + "\nTo perform the operation forcefully please execute the command again with force=yes")
            else:
                raise

        message = ""
        if replication_type == "ACTIVE_ACTIVE":
            on_local_message = " on the local system only" if force == True else ""
            serial_message = ", and the volume serial was kept locally. Make sure that the remote volume serial was changed to avoid having two volumes with the same serial" if force_keep_serial == True else ""
            message = on_local_message + serial_message
        if message != "":
            return self._result_message(replica, "deleted, " + message, source=source)
        else:
            staging_message = "and its staging area " + ("retained" if retain_staging_area else "discarded")
            if retain_staging_area and result and result.entity_pairs:
                non_reclaimed_names = sorted(pair.local_entity_name
                                            for pair in result.entity_pairs
                                            if not pair.get("_local_reclaimed_snapshot_id", False))
                if non_reclaimed_names:
                    staging_message = "but there was nothing in the staging area to retain" if len(result.entity_pairs) == len(non_reclaimed_names) \
                                    else "and its staging area was retained, except for the following datasets: " + humanly_join(non_reclaimed_names)
            return self._result_message(replica, "deleted, " + staging_message, source=source)
    
    @feature_command(
        feature='sync_replication',
        status=True,
        description="Switch the roles of the replica",
        types=(ReplicatedEntity, CgOrDataset),
        substitute=dict(source="local_dataset",),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def switch_role(self, _api, source, dataset_type=None):
        source = self._get_source(source, dataset_type)
        replica = self._get_replica(_api, source)
        self._validate_non_mobility_replica(replica, source)
        self._validate_replication_type_compatibility(replica.get('replication_type', None), ["ACTIVE_ACTIVE"], "switch role", "of")
        with get_remote_system(_api, replica).api:
            replica = self._post(_api, replica, "switch_role")
        return self._result_message(replica, "role switched to {}".format(replica.role))

    @command(
        description="Set replica sync interval",
        types=(ReplicatedEntity, TimeDelta, CgOrDataset),
        substitute=dict(
            source="local_dataset",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def set_interval(self, _api, source, interval, dataset_type=None):
        source = self._get_source(source, dataset_type)
        replica = self._get_replica(_api, source)
        self._validate_non_mobility_replica(replica, source)
        self._validate_replication_type_compatibility(replica.get('replication_type', None), ["SYNC", "ACTIVE_ACTIVE"], "set interval", "for")
        with get_remote_system(_api, replica).api:
            self._update(_api, replica, sync_interval=interval*1000)
        return self._result_message(replica, "set to synchronize every {} seconds".format(interval), source=source)

    @command(
        description="Set replica recovery point objective (RPO)",
        types=(ReplicatedEntity, TimeDeltaOrNone, CgOrDataset),
        substitute=dict(
            source="local_dataset",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def set_rpo(self, _api, source, rpo, dataset_type=None):
        source = self._get_source(source, dataset_type)
        replica = self._get_replica(_api, source)
        self._validate_non_mobility_replica(replica, source)
        self._validate_replication_type_compatibility(replica.get('replication_type', None), ["SYNC", "ACTIVE_ACTIVE"], "set rpo", "for")
        if replica.rpo_value is 0 and rpo == TimeDeltaOrNone.CONSTANTS.NONE:
            return self._result_message(replica, "recovery point objective already disabled", source=source)
        if rpo != TimeDeltaOrNone.CONSTANTS.NONE and replica.rpo_value == rpo*1000:
            return self._result_message(replica, "recovery point objective already set to {} seconds".format(rpo), source=source)
        with get_remote_system(_api, replica).api:
            self._update(_api, replica,
                rpo_value=rpo*1000 if rpo != TimeDeltaOrNone.CONSTANTS.NONE else 0,
                )
        return self._result_message(replica, "recovery point objective {}".format("set to {} seconds".format(rpo) if rpo != TimeDeltaOrNone.CONSTANTS.NONE else "disabled"), source=source)

    @feature_command(
        feature='sync_replication',
        status=True,
        description="Change replica type",
        types=(ReplicatedEntity, ReplicaTypeExcludeActiveActive, CgOrVolume, TimeDeltaOrNone, TimeDeltaOrNone),
        substitute=dict(
            source="local_dataset",
            replica_type="new_replica_type",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def change_type(self, _api, source, replica_type, dataset_type=None, rpo=None, interval=None):
        source = self._get_source(source, dataset_type)
        replica = self._get_replica(_api, source)
        self._validate_non_mobility_replica(replica, source)
        self._validate_replication_type_compatibility(replica.get('replication_type', None), ["ACTIVE_ACTIVE"], "change type", "of")
        if replica_type == "ASYNC" and (rpo is None or interval is None):
            raise BadArguments("When changing replica type to ASYNC, interval and rpo must be set")
        if replica_type == "SYNC" and (rpo is not None or interval is not None):
            raise BadArguments("When changing replica type to SYNC, interval and rpo must not be set")
        data = dict(
            rpo_type="TIME",
            rpo_value=rpo * 1000 if (rpo and rpo != TimeDeltaOrNone.CONSTANTS.NONE) else 0,
            sync_interval=interval * 1000 if (interval and interval != TimeDeltaOrNone.CONSTANTS.NONE) else 0,
        )
        with get_remote_system(_api, replica).api:
            if replica_type == "SYNC":
                self._post(_api, replica, "change_type_to_sync")
            else:
                self._post(_api, replica, "change_type_to_async", data=data)
        return self._result_message(replica, "type changed to {}".format(replica_type))

    @command(
        description="Suspend replica",
        types=(ReplicatedEntity, CgOrDataset),
        substitute=dict(
            source="local_dataset",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def suspend(self, _api, source, dataset_type=None):
        source = self._get_source(source, dataset_type)
        replica = self._get_replica(_api, source)
        self._validate_non_mobility_replica(replica, source)
        self._validate_replication_type_compatibility(replica.get('replication_type', None), ["ACTIVE_ACTIVE"], "suspend", "")
        if replica.state == "SUSPENDED":
            raise BadArguments, "Replica is already suspended"
        with get_remote_system(_api, replica).api:
            self._post(_api, replica, "suspend")
        return self._result_message(replica, "suspended", source=source)

    @command(
        description="Start a sync job of a replica immediately",
        types=(ReplicatedEntity, CgOrDataset),
        substitute=dict(
            source="local_dataset",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def sync_now(self, _api, source, dataset_type=None):
        source = self._get_source(source, dataset_type)
        replica = self._get_replica(_api, source)
        self._validate_replication_type_compatibility(replica.get('replication_type', None), ["SYNC", "ACTIVE_ACTIVE"], "sync", "")
        with get_remote_system(_api, replica).api:
            self._post(_api, replica, "sync")
        return self._result_message(replica, "sync started", source=source)

    @command(
        description="Resume replica",
        types=(ReplicatedEntity, CgOrDataset, YesNo),
        substitute=dict(
            source="local_dataset",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def resume(self, _api, source=None, dataset_type=None, resume_all=None):
        if dataset_type and not source:
            raise BadArguments, "dataset_type cannot be used without source"
        source = self._get_source(source, dataset_type)
        if source and resume_all:
            raise BadArguments, "when local_dataset is used, resume_all cannot be set to yes"
        if source is None and resume_all is None:
            raise BadArguments, "Please use either local_dataset or resume_all parameters"
        if source:
            replica = self._get_replica(_api, source)
            self._validate_non_mobility_replica(replica, source)
            self._validate_replication_type_compatibility(replica.get('replication_type', None), ["ACTIVE_ACTIVE"], "resume", "")
            if replica.state not in ["SUSPENDED", "AUTO_SUSPENDED", None]:
                raise BadArguments, "Replica with state '{}' cannot be resumed".format(replica.state)
            with get_remote_system(_api, replica).api:
                with increased_api_timeout():
                    self._post(_api, replica, "resume")
            return self._result_message(replica, "resumed", source=source)
        if resume_all:
            active_shell = get_active_shell()
            links = dict()
            for replica in Replica.iter_all(_api):
                if  replica.get('mobility_source') != None:
                    continue 
                if replica.get('replication_type', None) == "ACTIVE_ACTIVE":
                    continue
                if replica.role == "TARGET":
                    continue
                if replica.state not in ["SUSPENDED", "AUTO_SUSPENDED"]:
                    continue
                links.setdefault(replica.link_id,[]).append(replica)
            for link in links.iteritems():
                with get_remote_system(_api, link[0]).api as remote_api:
                    try:
                        remote_api.get("system/version")
                    except:
                        continue
            for _, items in links.iteritems():
                for replica in items:
                    entity = ("%s '%s'" % (replica.entity_type.lower(), replica.entity_pairs[0].local_entity.name)
                    if replica.entity_type != "CONSISTENCY_GROUP"
                    else "consistency group '%s'" % replica.local_cg_name)
                    with get_remote_system(_api, replica).api:
                        try:
                            print '--------------------\n %s' % entity
                            self._post(_api, replica, "resume")
                            active_shell.output("--- Replica resumed")
                        except Exception as e:
                            active_shell.error("--- Replica for %s: %s" % (entity, e))
        if resume_all is False:
            raise BadArguments, 'local_dataset parameter must be used'

    @command(
        description="Change a replica's role one-sidedly",
        types=(ReplicatedEntity, CgOrDataset),
        substitute=dict(
            source="local_dataset",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def change_role(self, _api, source, dataset_type=None):
        source = self._get_source(source, dataset_type)
        replica = self._get_replica(_api, source)
        self._validate_non_mobility_replica(replica, source)
        self._validate_replication_type_compatibility(replica.get('replication_type', None), ["ACTIVE_ACTIVE"], "change role", "of")
        with get_remote_system(_api, replica).api:
            replica = self._post(_api, replica, "change_role")
        return self._result_message(replica, "role changed to {}".format(replica.role), source=source)

@commandgroup(parent=ReplicaCommands, prefix="advanced")
class ReplicaAdvancedCommands(ReplicaCommandsBase):

    @feature_command(
        feature='sync_replication',
        status=True,
        description="Create a new replica to target dataset that already exists on remote system",
        types=(ReplicationLinkBySystemName, NonReplicatedDataset, RemoteEntity, ReplicaType, ActiveActivePreferred, TimeDeltaOrNone, TimeDeltaOrNone, CgOrDataset, DatasetSnapshot, RemoteEntity),
        substitute=dict(
            link="system",
            ),
        )
    @roles.ADMIN.POOLADMIN
    def create(self, _api, link, source, target, replication_type, active_active_preferred=None, interval=None, rpo=None, dataset_type=None, source_snapshot=None, target_snapshot=None):
        source = self._get_source(source, dataset_type)
        if (replication_type == "SYNC" or replication_type == "ASYNC") and (active_active_preferred is not None):
            raise BadArguments("active_active_preferred parameter is applicable only for ACTIVE_ACTIVE replication")
        if source.dataset_type == "FILESYSTEM" and (not target or not source_snapshot or not target_snapshot):
            raise BadArguments("Creating Replica from filesystem requires target, source_snapshot and target_snapshot parameters")
        if source_snapshot:
            if not target_snapshot and not system.user.roles.is_infinidat():
                raise BadArguments("Please specify source_snapshot and target_snapshot, or neither")
        with get_remote_system(_api, link).api as remote_api:
            replica_info = self._create_replica_info(
                remote_api=remote_api,
                link=link,
                source=source,
                interval=interval,
                rpo=rpo,
                replication_type=replication_type,
                active_active_preferred=active_active_preferred
                )
            pair = Munch(
                local_entity_id=source.id,
                remote_entity_id=Dataset.get_by_name(remote_api, target).id,
                local_base_action="NO_BASE_DATA",
                remote_base_action="NO_BASE_DATA",
                )
            if source_snapshot and target_snapshot:
                snapClass =  DatasetSnapshot if int(link.remote_version[0]) > 2 else RemoteDatasetSnapshot
                pair.update(
                    local_base_action="BASE",
                    remote_base_action="BASE",
                    local_base_entity_id=source_snapshot.id,
                    remote_base_entity_id=snapClass.get_by_name(remote_api, target_snapshot).id,
                    )
            replica_info.entity_pairs.append(pair)
            success_message = 'Replica from dataset "{}" to dataset "{}" created'.format(source.name, target)
            return self._create(_api, returns=success_message, **replica_info)

    @feature_command(
        feature='sync_replication',
        status=True,
        name="create",
        description="Create a new replica to target dataset that already exists on remote system",
        types=(ReplicationLinkBySystemName, NonReplicatedDataset, RemoteEntity, ReplicaType, ActiveActivePreferred, TimeDeltaOrNone, TimeDeltaOrNone, CgOrDataset, DatasetSnapshot, RemoteEntity, YesNo),
        substitute=dict(
            link="system",
            mobility="mobility_replica",
            ),
        )
    @roles.INFINIDAT
    def infinidat_create(self, _api, link, source, target, replication_type, active_active_preferred=None, interval=None, rpo=None, dataset_type=None, source_snapshot=None, target_snapshot=None, mobility=None):
        source = self._get_source(source, dataset_type)
        if (replication_type == "SYNC" or replication_type == "ASYNC") and (active_active_preferred is not None):
            raise BadArguments("active_active_preferred parameter is applicable only for ACTIVE_ACTIVE replication")
        if mobility and replication_type != "ACTIVE_ACTIVE":
            raise BadArguments("The mobility parameter is applicable only for active-active replication")
        if mobility and (source.dataset_type != "VOLUME" if source._is_dataset else False):
            raise BadArguments("The mobility parameter is applicable only for volume replication")
        if source.dataset_type == "FILESYSTEM" and (not target or not source_snapshot or not target_snapshot):
            raise BadArguments("Creating Replica from filesystem requires target, source_snapshot and target_snapshot parameters")
        if source_snapshot:
            if not target_snapshot and not system.user.roles.is_infinidat():
                raise BadArguments("Please specify source_snapshot and target_snapshot, or neither")
        with get_remote_system(_api, link).api as remote_api:
            replica_info = self._create_replica_info(
                remote_api=remote_api,
                link=link,
                source=source,
                interval=interval,
                rpo=rpo,
                replication_type=replication_type,
                active_active_preferred=active_active_preferred
                )
            if mobility:
                replica_info.mobility_source=mobility
            pair = Munch(
                local_entity_id=source.id,
                remote_entity_id=Dataset.get_by_name(remote_api, target).id,
                local_base_action="NO_BASE_DATA",
                remote_base_action="NO_BASE_DATA",
                )
            if source_snapshot and target_snapshot:
                snapClass =  DatasetSnapshot if int(link.remote_version[0]) > 2 else RemoteDatasetSnapshot
                pair.update(
                    local_base_action="BASE",
                    remote_base_action="BASE",
                    local_base_entity_id=source_snapshot.id,
                    remote_base_entity_id=snapClass.get_by_name(remote_api, target_snapshot).id,
                    )
            replica_info.entity_pairs.append(pair)
            success_message = 'Replica from dataset "{}" to dataset "{}" created'.format(source.name, target)
            return self._create(_api, returns=success_message, **replica_info)

    @feature_command(
        feature='sync_replication',
        status=True,
        description="Create a new replica to target consistency group that already exists on remote system",
        types=(ReplicationLinkBySystemName, NonReplicatedCg, RemoteEntity, ReplicaTypeExcludeActiveActive, TimeDelta, TimeDelta, ReplicaDatasetPair.List),
        substitute=dict(
            link="system",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def create_cg(self, _api, link, source, target, replication_type, interval=None, rpo=None, cg_pairs=None):
        if (replication_type == "SYNC") and (rpo is not None or interval is not None):
            raise BadArguments("RPO and Interval parameter are applicable only for async replication")
        with get_remote_system(_api, link).api as remote_api:
            source = source[0]
            replica_info = self._create_replica_info(
                remote_api=remote_api,
                link=link,
                source=source,
                interval=interval,
                rpo=rpo,
                replication_type=replication_type,
                )
            remote_cg = ConsistencyGroup.get_by_name(remote_api, target)
            replica_info.update(
                remote_cg_id=remote_cg.id,
                )
            local_members = Dataset.get_all(_api, filter=_api.create_filter(cg_id=source.id))
            remote_members = Dataset.get_all(remote_api, filter=_api.create_filter(cg_id=remote_cg.id))
            if len(local_members) != len(remote_members):
                raise CommandFailed('Invalid consistency groups: Local group "{}" has {} members, but remote group "{}" has {} members'
                                    .format(source.name, len(local_members), remote_cg.name, len(remote_members)))

            def add_pair(local_member, remote_member):
                replica_info.entity_pairs.append(Munch(
                    local_entity_id=local_member.id,
                    local_base_action="NO_BASE_DATA",
                    remote_base_action="NO_BASE_DATA",
                    remote_entity_id=remote_member.id,
                    ))

            def process_member(all_members, member_name, processed_set, locality, cg):
                if member_name in processed_set:
                    raise BadArguments('{} dataset "{}" was specified more than once in cg_pairs'.format(locality, member_name))
                member = all_members.pop(member_name, None)
                if not member:
                    raise NoSuchEntity('No such dataset: {} consistency group "{}" has no such member: "{}"'
                                       .format(locality, cg.name, member_name))
                processed_set.add(member_name)
                return member

            if cg_pairs:  # Manual mapping
                processed_local_members, processed_remote_members = set(), set()
                local_members_by_name = {dataset.name : dataset for dataset in local_members}
                remote_members_by_name = {dataset.name : dataset for dataset in remote_members}
                for local_member_name, remote_member_name in cg_pairs:
                    add_pair(
                        process_member(local_members_by_name, local_member_name, processed_local_members, "Local", source),
                        process_member(remote_members_by_name, remote_member_name, processed_remote_members, "Remote", target),
                        )
            else:  # Automatic mapping
                sort_members = partial(sorted, key=lambda m: (m.size, m.name))
                for local_member, remote_member in izip(sort_members(local_members), sort_members(remote_members)):
                    if local_member.size != remote_member.size:
                        raise CommandFailed("Invalid consistency groups: Sizes of local datasets do not match sizes of remote datasets")
                    add_pair(local_member, remote_member)
                if replica_info.entity_pairs and not self._get_user_confirmation_for_create_cg(replica_info.entity_pairs, local_members, remote_members):
                    raise CommandAborted("If you wish to specify the configuration manually, use cg_pairs")

            replica = self._create(_api, returns=ENTITY, **replica_info)
            return self._result_message(replica, "created" if cg_pairs else "created using automatic configuration", source=source)




    @feature_command(
        feature='sync_replication',
        status=False,
        name="create",
        description="Create a new replica to target dataset that already exists on remote system",
        types=(ReplicationLinkBySystemName, NonReplicatedEntity, TimeDeltaOrNone, TimeDeltaOrNone, RemoteEntity, DatasetSnapshot, RemoteEntity),
        substitute=dict(
            link="system",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def old_create(self, _api, link, source, interval, rpo, target, source_snapshot=None, target_snapshot=None):
        if source.dataset_type == "FILESYSTEM" and (source_snapshot or target_snapshot):
            raise BadArguments("Cannot create replica of filesystem from existing snapshot")
        if source_snapshot:
            if not target_snapshot and not system.user.roles.is_infinidat():
                raise BadArguments("Please specify source_snapshot and target_snapshot, or neither")
        with get_remote_system(_api, link).api as remote_api:
            replica_info = self._old_create_replica_info(
                remote_api=remote_api,
                link=link,
                source=source,
                interval=interval,
                rpo=rpo,
                )
            pair = Munch(
                local_entity_id=source.id,
                remote_entity_id=Dataset.get_by_name(remote_api, target).id,
                local_base_action="NO_BASE_DATA",
                remote_base_action="NO_BASE_DATA",
                )
            if source_snapshot and target_snapshot:
                pair.update(
                    local_base_action="BASE",
                    remote_base_action="BASE",
                    local_base_entity_id=source_snapshot.id,
                    remote_base_entity_id=DatasetSnapshot.get_by_name(remote_api, target_snapshot).id,
                    )
            replica_info.entity_pairs.append(pair)
            success_message = 'Replica from dataset "{}" to dataset "{}" created'.format(source.name, target)
            return self._create(_api, returns=success_message, **replica_info)

    @feature_command(
        feature='sync_replication',
        status=False,
        name="create_cg",
        description="Create a new replica to target consistency group that already exists on remote system",
        types=(ReplicationLinkBySystemName, NonReplicatedCg, TimeDeltaOrNone, TimeDeltaOrNone, RemoteEntity, ReplicaDatasetPair.List),
        substitute=dict(
            link="system",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def old_create_cg(self, _api, link, source, interval, rpo, target, cg_pairs=None):
        source = source[0]
        with get_remote_system(_api, link).api as remote_api:
            replica_info = self._old_create_replica_info(
                remote_api=remote_api,
                link=link,
                source=source,
                interval=interval,
                rpo=rpo,
                )
            remote_cg = ConsistencyGroup.get_by_name(remote_api, target)
            replica_info.update(
                remote_cg_id=remote_cg.id,
                )
            local_members = Dataset.get_all(_api, filter=_api.create_filter(cg_id=source.id))
            remote_members = Dataset.get_all(remote_api, filter=_api.create_filter(cg_id=remote_cg.id))
            if len(local_members) != len(remote_members):
                raise CommandFailed('Invalid consistency groups: Local group "{}" has {} members, but remote group "{}" has {} members'
                                    .format(source.name, len(local_members), remote_cg.name, len(remote_members)))

            def add_pair(local_member, remote_member):
                replica_info.entity_pairs.append(Munch(
                    local_entity_id=local_member.id,
                    local_base_action="NO_BASE_DATA",
                    remote_base_action="NO_BASE_DATA",
                    remote_entity_id=remote_member.id,
                    ))

            def process_member(all_members, member_name, processed_set, locality, cg):
                if member_name in processed_set:
                    raise BadArguments('{} dataset "{}" was specified more than once in cg_pairs'.format(locality, member_name))
                member = all_members.pop(member_name, None)
                if not member:
                    raise NoSuchEntity('No such dataset: {} consistency group "{}" has no such member: "{}"'
                                       .format(locality, cg.name, member_name))
                processed_set.add(member_name)
                return member

            if cg_pairs:  # Manual mapping
                processed_local_members, processed_remote_members = set(), set()
                local_members_by_name = {dataset.name : dataset for dataset in local_members}
                remote_members_by_name = {dataset.name : dataset for dataset in remote_members}
                for local_member_name, remote_member_name in cg_pairs:
                    add_pair(
                        process_member(local_members_by_name, local_member_name, processed_local_members, "Local", source),
                        process_member(remote_members_by_name, remote_member_name, processed_remote_members, "Remote", target),
                        )
            else:  # Automatic mapping
                sort_members = partial(sorted, key=lambda m: (m.size, m.name))
                for local_member, remote_member in izip(sort_members(local_members), sort_members(remote_members)):
                    if local_member.size != remote_member.size:
                        raise CommandFailed("Invalid consistency groups: Sizes of local datasets do not match sizes of remote datasets")
                    add_pair(local_member, remote_member)
                if replica_info.entity_pairs and not self._get_user_confirmation_for_create_cg(replica_info.entity_pairs, local_members, remote_members):
                    raise CommandAborted("If you wish to specify the configuration manually, use cg_pairs")

            replica = self._create(_api, returns=ENTITY, **replica_info)
            return self._result_message(replica, "created" if cg_pairs else "created using automatic configuration", source=source)

    def _get_user_confirmation_for_create_cg(self, entity_pairs, local_members, remote_members):
        if yesno_ensure_interactive():
            # User should not be prompted
            return True
        local_members_by_id = {dataset.id : dataset for dataset in local_members}
        remote_members_by_id = {dataset.id : dataset for dataset in remote_members}
        InfiniboxRenderer().render(
            (Munch(
                local_dataset=local_members_by_id[p.local_entity_id].name,
                local_size=local_members_by_id[p.local_entity_id].size,
                remote_dataset=remote_members_by_id[p.remote_entity_id].name,
                remote_size=remote_members_by_id[p.remote_entity_id].size,
                ) for p in entity_pairs),
            stream=get_active_shell().output_stream,
            fields=strlist("local_dataset local_size remote_dataset remote_size"),
            )
        return yesno("Please examine the table above. Is this the configuration you intended for?")

@commandgroup(prefix="link")
class ReplicationLinkCommands(EntityCommands, ReplicationCommandsMixin):
    ENTITY_TYPE = ReplicationLinkBySystemName
    QUERY_FIELDS = strlist("remote_system_name link_state last_connection_timestamp local_replication_network_space_id remote_host witness_address resiliency_mode link_mode")
    SORT_FIELDS = strlist("remote_system_name local_replication_network_space_id remote_host witness_address link_mode")
    DETAILED_QUERY_FIELDS = QUERY_FIELDS + strlist(" _remote_network_space remote_system_serial_number remote_version link_replication_type local_witness_state remote_witness_state")
    INFINIDAT_DETAILED_QUERY_FIELDS = DETAILED_QUERY_FIELDS + strlist("name")
    QUERY_SUBSTITUTES = dict(
        remote_system_name="remote_system",
        last_connection_timestamp="last_connection",
        remote_host="remote_system_address",
        remote_version="remote_system_version",
        remote_system_serial_number="remote_system_serial",
        local_replication_network_space_id="local_network_space",
        _remote_network_space="remote_network_space",
        name="link_name",
        link_replication_type='supported_replica_types',
        )

    @command(
        description="List existing replication links",
        types=(ReplicationLinkBySystemName.List, NetworkSpace, Hostname),
        options=QueryOptions(fields=DETAILED_QUERY_FIELDS, infinidat_fields=INFINIDAT_DETAILED_QUERY_FIELDS, sortable=SORT_FIELDS, substitute=QUERY_SUBSTITUTES, detailed=True),
        substitute=dict(
            links="remote_system",
            ),
        )
    @render(
        fields=QUERY_FIELDS,
        substitute=QUERY_SUBSTITUTES,
        detailed_individually=True,
        detailed_fields=DETAILED_QUERY_FIELDS,
        alignment=dict(
            local_replication_network_space_id="left",
            _remote_network_space="left",
            ),
        infinidat=dict(
            detailed_individually=True,
            detailed_fields=INFINIDAT_DETAILED_QUERY_FIELDS,
            ),
        )
    def query(self, _api, links=None, net_space=None, witness_address=None):
        netspace_mapping = {net_space.id : net_space.name} if net_space \
                           else LazyIdNameMappingNoneDetached(NetworkSpace, _api)
        return self._query(_api, links,
            local_replication_network_space_id=net_space,
            witness_address=witness_address,
            _hook=partial(self._query_hook, api=_api, lazy_network_spaces_by_id=netspace_mapping),
            )

    def _query_hook(self, links, api=None, lazy_network_spaces_by_id=None):
        should_get_remote_netspace = api and (is_detailed_query() or {"_remote_network_space", "*"}.intersection(runtime.options.get("columns") or ()))
        if not lazy_network_spaces_by_id and not should_get_remote_netspace:
            return links
        for link in links:
            if lazy_network_spaces_by_id:
                link.local_replication_network_space_id = lazy_network_spaces_by_id[link.local_replication_network_space_id]
            if should_get_remote_netspace:
                link._remote_network_space = None
                with get_remote_system(api, link).api as remote_api:
                    try:
                        remote_netspace = remote_api.get(NetworkSpace.get_uri(link.remote_replication_network_space_id), handle_remote_auth_error=False, timeout=5)
                        link._remote_network_space = remote_netspace.name
                    except ManagementException as e:
                        reason = "authentication required" if e.code == api.MGMT_ERROR_CODES.REMOTE_AUTH else e.code
                        link._remote_network_space = "N/A ({})".format(reason) if reason else "N/A"
                    except ConnectionFailed as e:
                        link._remote_network_space = "N/A (timeout)"
        return links

    @command(
        description="Create a new replication link",
        types=(Hostname, ReplicationNetworkSpace, Hostname),
        substitute=dict(
            address="remote_address",
            ),
        )
    @roles.INFINIDAT.ADMIN
    def create(self, _api, address, network_space, witness_address=None):
        name = generate_entity_name()
        with _api.remote_auth_context(address) as ctx:
            link = self._create_link(_api, name, address, network_space, witness_address)
            if ctx.username and ctx.password:
                get_remote_system(_api, link).api.set_basic_auth(ctx.username, ctx.password)
        return self._result_message(link, "created")

    def _create_link(self, _api, name, address, network_space, witness_address):
        return self._create(_api,
            name=name,
            remote_host=address,
            local_replication_network_space_id=network_space.id,
            witness_address = witness_address,
            returns=ENTITY,
            )

    @command(
        description="Delete replication link",
        types=(ReplicationLinkBySystemName, YesNo),
        substitute=dict(
            link="remote_system",
            ),
        )
    @roles.INFINIDAT.ADMIN
    def delete(self, _api, link, force=False):
        send_request = partial(self._del, _api, link)
        try:
            with get_remote_system(_api, link).api:
                send_request()
        except ManagementException as e:
            if not self._should_retry_forcefully(e):
                raise
            elif not force:
                get_active_shell().warning("{}: {}".format(e.code, e.message))
                if not yesno("Would you like to perform the operation forcefully?", always_prompt=True):
                    raise CommandAborted
            send_request(approved=True, params=dict(
                force_if_no_remote_credentials="true",
                force_if_remote_error="true",
                ))
        return self._result_message(link, "deleted")

    @command(
        description="Authenticate on remote system",
        types=(ReplicationLinkBySystemName, Str, Str),
        substitute=dict(
            link="remote_system",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def authenticate(self, _api, link, username=None, password=None):
        if password and not username:
            raise BadArguments("When specifying password, username must be specified")
        if not password:
            username, password = get_username_password(
                username=username,
                prompt="Username for {}".format(link.remote_system_name),
                )
        remote_system = get_remote_system(_api, link)
        remote_system.login(username, password)
        return 'Authenticated on {} as user "{}" with role {}'.format(
            link.remote_system_name,
            remote_system.user.username,
            ", ".join(remote_system.user.roles.roles),
            )

    def _remote_version_greater_then(self, link, version):
        return int(link.remote_version[0]) >= int(version) if \
            link.remote_version else True

    @command(description='Attach network space to link',
             types=(ReplicationLinkBySystemName, ReplicationNetworkSpace),
             substitute={'link': 'remote_system'})
    @roles.ADMIN
    def attach(self, _api, link, network_space):
        if not self._remote_version_greater_then(link, "3"):
            raise BadArguments, "Attach is allowed only when the software version of both systems is 3.0 or higher"
        with increased_api_timeout():
            self._put(_api, link, data={
                'local_replication_network_space_id': network_space.id,
            })
        print('Link that uses network space "%s" is attached to the remote system %s' %
                (network_space.name, self._result_entity_name(link)))
        return self.refresh(_api=_api, link=link)

    @command(description='Attach network space to link',
             types=(ReplicationLinkBySystemName, ReplicationNetworkSpace, YesNo),
             substitute={'link': 'remote_system'},
             name='attach')
    @roles.INFINIDAT
    def infinidat_attach(self, _api, link, network_space, with_refresh=True):
        if not self._remote_version_greater_then(link, "3"):
            raise BadArguments, "Attach is allowed only when the software version of both systems is 3.0 or higher"
        with increased_api_timeout():
            self._put(_api, link, data={
                'local_replication_network_space_id': network_space.id,
            })
        print('Link that uses network space "%s" is attached to the remote system %s' %
              (network_space.name, self._result_entity_name(link)))
        if with_refresh:
            return self.refresh(_api=_api, link=link)

    @command(description='Detach network space from link',
             types=(ReplicationLinkBySystemName,),
             substitute={'link': 'remote_system'})
    @roles.INFINIDAT.ADMIN
    def detach(self, _api, link):
        if not self._remote_version_greater_then(link, "3"):
            raise BadArguments, "Detach is allowed only when the software version of both systems is 3.0 or higher"
        with increased_api_timeout():
            self._put(_api, link, data={
                'local_replication_network_space_id': None,
            })
        print 'Link is detached from the remote system %s' % self._result_entity_name(link)

    @command(description='Refresh link',
             types=(ReplicationLinkBySystemName,),
             substitute={'link': 'remote_system'})
    @roles.INFINIDAT.ADMIN
    def refresh(self, _api, link):
        # The operation requires authentication on the remote system.
        send_request = partial(self._post, _api, link, 'refresh', data={})
        with get_remote_system(_api, link).api as remote_api:
            with increased_api_timeout():
                send_request()
        return self._result_message(link, 'refreshed')

    @command(description='Change remote address',
             types=(ReplicationLinkBySystemName, Hostname),
             substitute={'link': 'remote_system', 'address': 'remote_address'})
    @roles.INFINIDAT.ADMIN
    def change_remote_address(self, _api, link, address):
        self._put(_api, link, data={'remote_host': address})
        print('Link to {} remote address changed'.format(link.remote_system_name))
        return self.refresh(_api=_api, link=link)
    
    @command(description='Update the witness address for a link',
             types=(ReplicationLinkBySystemName, HostnameOrNone),
             substitute={'link': 'remote_system'})
    @roles.INFINIDAT.ADMIN
    def change_witness_address(self, _api, link, witness_address):
        witness_address=None if witness_address == NONE_CONSTANT else witness_address
        send_request = partial(self._post, _api, link, 'set_witness_address', data={'witness_address': witness_address})
        with get_remote_system(_api, link).api as remote_api:
            with increased_api_timeout():
                send_request()
        return 'Witness address updated on Link {}'.format(link.remote_system_name)
