from infinishell.infishell.commands import command, commandgroup
from infinishell.infishell.exceptions import CommandAborted
from infinishell.infishell.types import Int, YesNo
from infinishell.infishell.utils import strlist
from ..exceptions import NoSuchEntity
from . import EntityCommands, render_collection, roles
from .options import QueryOptions
from .types import (
    Name,
    Tenant,
    NFSGroupPolicy,
    NFSUnmappedUsersPolicy,
    UnixUser,
    )

TENANT_QUERY_FIELDS = strlist("name")
TENANT_DETAILED_QUERY_FIELDS = TENANT_QUERY_FIELDS + strlist("short_tenant_key visible_to_sysadmin nfs_group_policy nfs_allow_unmapped_users anonymous_uid anonymous_gid")
TENANT_SORTABLE_FIELDS = TENANT_QUERY_FIELDS
TENANT_SUBSTITUTES = dict(
    short_tenant_key="key",
    nfs_group_policy="group",
    nfs_allow_unmapped_users="allow_unmapped_users",
    anonymous_uid="anonuid",
    anonymous_gid="anongid",
    )

UNIX_USER_QUERY_FIELDS = strlist("uid primary_group_id supplementary_groups")
UNIX_USER_SORTABLE_FIELDS = strlist("uid primary_group_id")
UNIX_USER_SUBSTITUTES = dict(
    primary_group_id="primary_gid",
    supplementary_groups="supplementary_gids",
    )

@commandgroup(prefix="tenant")
class TenantCommands(EntityCommands):
    ENTITY_TYPE = Tenant
    
    @command(
        description="Create a new tenant",
        types=(Name, YesNo,),
        )
    @roles.INFINIDAT
    def create(self, _api, name, visible_to_sysadmin=None):
        return self._create(_api, name=name, visible_to_sysadmin=visible_to_sysadmin)
    
    @command(
        description="Delete a tenant",
        types=(Tenant,),
        hidden_substitue=dict(name="tenant"),
        )
    @roles.INFINIDAT
    def delete(self, _api, tenant):
        return self._delete(_api, tenant)
    
    @command(
        description="Set NFS policies for NFS exports in the tenant",
        types=(Tenant, NFSGroupPolicy, NFSUnmappedUsersPolicy, Int, Int,),
        hidden_substitue=dict(name="tenant"),
        )
    @roles.INFINIDAT
    def set_nfs_policies(self, _api, tenant, group=None, allow_unmapped_users=None, anonuid=None, anongid=None):
        if not (group or allow_unmapped_users or anonuid or anongid):
            raise CommandAborted("no NFS policy parameter provided")
        self._update(_api, tenant, nfs_group_policy=group, nfs_allow_unmapped_users=allow_unmapped_users, anonymous_uid=anonuid, anonymous_gid=anongid)
        return "NFS policies set for tenant '{}'".format(tenant.name)
    
    @command(
        description="Rename a tenant",
        types=(Tenant, Name,),
        hidden_substitue=dict(name="tenant"),
        )
    @roles.INFINIDAT
    def rename(self, _api, tenant, new_name):
        return self._rename(_api, tenant, new_name)
    
    @command(
        description="List existing tenants",
        types=(Tenant.List,),
        hidden_substitue=dict(name="tenant"),
        options=QueryOptions(fields=TENANT_QUERY_FIELDS, sortable=TENANT_SORTABLE_FIELDS, substitute=TENANT_SUBSTITUTES, detailed=True),
    )
    @render_collection(
        entity=Tenant,
        fields=TENANT_QUERY_FIELDS,
        detailed_fields=TENANT_DETAILED_QUERY_FIELDS,
        detailed_individually=True,
        substitute=TENANT_SUBSTITUTES,
    )
    @roles.INFINIDAT
    def query(self, _api, tenant=None):
        return self._query(_api, tenant)


@commandgroup(parent=TenantCommands, prefix="unix_user")
class UnixUserCommands(EntityCommands):
    ENTITY_TYPE = UnixUser
     
    @command(
        description="Create a new unix user",
        types=(Tenant, Int, Int, Int.List,),
        )
    @roles.INFINIDAT
    def create(self, _api, tenant, uid, primary_gid, supplementary_gids=None):
        self.ENTITY_TYPE.URI = self.ENTITY_TYPE.URI_TEMPLATE.format(tenant.id)
        self._create(_api, uid=uid, primary_group_id=primary_gid, supplementary_groups=supplementary_gids)
        return "Unix user '{}' created for tenant '{}'".format(uid, tenant.name)
     
    @command(
        description="Delete a unix user",
        types=(Tenant, Int),
        hidden_substitue=dict(name="unix user"),
        )
    @roles.INFINIDAT
    def delete(self, _api, tenant, uid):
        self.ENTITY_TYPE.URI = self.ENTITY_TYPE.URI_TEMPLATE.format(tenant.id)
        try:
            unix_user = UnixUser.get_by_uid(_api, uid)
        except NoSuchEntity as e:
            raise NoSuchEntity(u'Command rejected: No such {}: "{}"'.format(self.ENTITY_TYPE.get_entity_name(), uid))
        self._delete(_api, unix_user)
        return "Unix user '{}' deleted from tenant '{}'".format(uid, tenant.name)
    
    @command(
        description="Update a unix user",
        types=(Tenant, Int, Int, Int.List,),
        hidden_substitue=dict(name="unix user"),
        )
    @roles.INFINIDAT
    def modify(self, _api, tenant, uid, primary_gid=None, supplementary_gids=None):
        if not (primary_gid or supplementary_gids) or (supplementary_gids is not None and len(supplementary_gids) < 1):
            raise CommandAborted("no unix user parameters provided")
        
        self.ENTITY_TYPE.URI = self.ENTITY_TYPE.URI_TEMPLATE.format(tenant.id)
        try:
            unix_user = UnixUser.get_by_uid(_api, uid)
        except NoSuchEntity as e:
            raise NoSuchEntity(u'Command rejected: No such {}: "{}"'.format(self.ENTITY_TYPE.get_entity_name(), uid))
        return self._update(_api, unix_user, primary_group_id=primary_gid, supplementary_groups=supplementary_gids)
        
    @command(
        description="List existing unix users",
        types=(Tenant, Int, Int,),
        hidden_substitue=dict(name="unix user"),
        options=QueryOptions(fields=UNIX_USER_QUERY_FIELDS, sortable=UNIX_USER_SORTABLE_FIELDS, substitute=UNIX_USER_SUBSTITUTES, detailed=False),
    )
    @render_collection(
        entity=UnixUser,
        fields=UNIX_USER_QUERY_FIELDS,
        detailed_fields=UNIX_USER_QUERY_FIELDS,
        substitute=UNIX_USER_SUBSTITUTES,
    )
    @roles.INFINIDAT
    def query(self, _api, tenant, uid=None, primary_gid=None):
        self.ENTITY_TYPE.URI = self.ENTITY_TYPE.URI_TEMPLATE.format(tenant.id)
        return self._query(_api, uid=uid, primary_group_id=primary_gid)
