from munch import Munch

from infinishell.infishell import availability, command, config
from .runtime_config import RuntimeConfig
from infinishell.infishell.utils import strlist

from ..utils import Version
from .system import system, Roles
from infinishell.infishell.exceptions import CommandFailed


runtimeConfig = RuntimeConfig.get_instance()

class Features(object):

    def _route_commands(self, feature, system):
        try:
            return runtimeConfig['mgmt_config']['mgmt'].get(
                'mgmt.network_space.route_operations_enabled', feature.get('enabled', False))
        except:
            return feature.get('enabled', False)

    def _openldap_commands(self, feature, system):
        try:
            return runtimeConfig['mgmt_config']['mgmt'].get(
                'mgmt.ldap.openldap_enabled', feature.get('enabled', False))
        except:
            return feature.get('enabled', False)

    def _openvms_commands(self, feature, system):
        try:
            return runtimeConfig['mgmt_config']['mgmt'].get(
                'mgmt.openvms.enabled', feature.get('enabled', False))
        except:
            return feature.get('enabled', False)

    def _qos_fs_commands(self, feature, system):
        try:
            return runtimeConfig['mgmt_config']['mgmt'].get(
                'mgmt.qos.filesystem_support_enabled', feature.get('enabled', False))
        except:
            return feature.get('enabled', False)

    def _smb_commands(self, feature, system):
        try:
            plugins = system.api.get('plugins')
            smb = next(x for x in plugins if x.type == 'SMB')
            return smb is not None
        except:
            return feature.get('enabled', False)

    def _fips_commands(self, feature, system):
        try:
            security_mode = system.api.get('config/security/system_security_mode', None)
            return security_mode != "STANDARD"
        except:
            return feature.get('enabled', False)

    #############################
    # enabled = whether the feature is enabled by default
    # persistent = whether feature was already tested with test_func/received
    # from server
    # test_func = test_func to see if feature should be enabled
    #############################
    _features = Munch(
        sync_replication=dict(enabled=False, persistant=False, test_func=None),
        qos=dict(enabled=True, persistant=False, test_func=None),
        iscsi=dict(enabled=True, persistant=False, test_func=None),
        treeq=dict(enabled=True, persistant=False, test_func=None),
    )

    _mgmt_features = Munch(
        qos_fs=dict(enabled=False, persistant=False, test_func=_qos_fs_commands),
        route=dict(enabled=True, persistant=False, test_func=_route_commands),
        openldap=dict(enabled=True, persistant=False, test_func=_openldap_commands),
        openvms=dict(enabled=False, persistant=False, test_func=_openvms_commands),
        smb=dict(enabled=False, persistant=False, test_func=_smb_commands),
        fips=dict(enabled=False, persistant=False, test_func=_fips_commands),
    )

    @classmethod
    def set_mgmt_features(self, system):
        for name, feature in self._mgmt_features.iteritems():
            self.set_feature(name, feature, system)

    @classmethod
    def set_feature(self, feature, info, system):
        self._features.setdefault(feature, dict())
        self._features[feature].update(info)
        test_func = self._features[feature].get('test_func', None)
        persistant = self._features[feature].get('persistant', False)
        if test_func and not persistant:
            self._features[feature] = info
            self._features[feature]['enabled'] = test_func(self, feature=info, system=system)
        elif not persistant:
            self._features[feature] = info
        self._features[feature]['persistant'] = True

    @classmethod
    def get_feature(self, feature_name):
        feature = self._features.get(feature_name, dict(enabled=False, persistant=False, test_func=None))
        assert feature , "Feature {} does not exist in _features list, but is used in a command".format(feature_name)
        test_func = feature.get('test_func', None)
        persistant = feature.get('persistant', False)
        if test_func and not persistant:
            feature['enabled'] = test_func(self, feature=feature, system=system)
            feature['persistant'] = True
        return feature


def feature_command(**kwargs):
    def decorator(_func):
        feature = kwargs.pop('feature')
        status = kwargs.pop('status')
        return availability(command(_func, **kwargs), predicate=lambda: feature_exists(feature) == status)
    if kwargs.get('feature') is None or not isinstance(kwargs.get('status'), (bool)):
        raise CommandFailed("feature_command decorator requires feature and status parameters")
    return decorator

def feature_exists(feature=None):
    return Features.get_feature(feature).get('enabled', False)
