from __future__ import print_function

# Import patches first
from . import patches  # noqa

# Import everything else
import sys
import traceback
import colorama
import pydoc
from requests import get
from functools import partial
from contextlib import contextmanager
import json

from infinishell.infishell import (
    Commands,
    Parser, ShellParser,
    config,
    )
from infinishell.infishell.exceptions import FilesystemException
from infinishell.infishell.formatters import format_exception
from infinishell.infishell.utils.streams import PaginationAborted

from .. import debug
from ..main import BASE_URI  # Imported for testing  # noqa
from ..main import (
    REST_BASE_URI,
    INTERNAL_BASE_URI,
    output,
    error,
    output_stderr,
    get_api_base_url,
    get_history_filename,
    new_client_id,
    get_login_password,
    )
from ..constants import PRODUCT_NAME
from ..environment import variables as env
from ..plugins import plugins
from ..utils import post_mortem
from ..shell import InfiniboxShell, InfiniboxExecutor
from ..exceptions import InvalidOptions

from . import raw_mode, __commands_manifest__
from .api import InfiniboxAPI
from .args_processor import InfiniboxArgsProcessor
from .formatters import InfinidatExceptionFormatter
# Due to cyclic dependencies, types has to be imported before renderers
from .commands import types as ___  # noqa
from .commands.options import DebugOption, AutoApprove
from .renderer import InfiniboxRenderer, toggle_csv_mode, toggle_json_mode, toggle_count_mode
from .system.system import System, system as system_context
from .system.user import Roles
from .runtime_config import RuntimeConfig

from .exceptions import (
    ConnectionFailed,
    ServiceTemporarilyUnavailable,
    InternalServerError,
    ManagementException,
    UnexpectedServerResponse,
    APIFailure,
    )

runtimeConfig = RuntimeConfig.get_instance()

# Import command modules so that they register in the manifest
from .manifest import manifest

COMMANDS = [cls() for cls in manifest.iter_manifest(__commands_manifest__)]
assert COMMANDS, "No commands registered in manifest"

GLOBAL_OPTIONS = (
    DebugOption(),
    AutoApprove(),
    )

LOGOUT_TIMEOUT_SECONDS = 5

def get_default_headers():
    headers = dict()
    for header, var in (
            ("X-Forwarded-For", "X_FORWARDED_FOR"),
            ):
        if env.is_set(var):
            headers[header] = env.get(var)
    return headers

def _prompt_escape(colors):
    return "\x01" + colors + "\x02"

_prompt_template = "{user}@{system}"
_colored_prompt_template = "{user_color}{{user}}{at_color}@{system_color}{{system}}{end_color}".format(
    user_color=_prompt_escape(colorama.Fore.BLUE + colorama.Style.BRIGHT),
    at_color=_prompt_escape(colorama.Fore.RESET + colorama.Style.BRIGHT),
    system_color=_prompt_escape(colorama.Fore.GREEN + colorama.Style.NORMAL),
    end_color=_prompt_escape(colorama.Style.RESET_ALL),
    )

def prompt_hook(shell, address, system):
    template = _colored_prompt_template if shell.has_colors() else _prompt_template
    return template.format(user=system.user.username, system=address)

_prompt_sep = "> "
_colored_prompt_sep = "{sep_color}>{end_color} ".format(
    sep_color=_prompt_escape(colorama.Fore.RESET + colorama.Style.BRIGHT),
    end_color=_prompt_escape(colorama.Style.RESET_ALL),
    )

def prompt_sep_hook(shell):
    return _colored_prompt_sep if shell.has_colors() else _prompt_sep

def run(options):
    return Main(options).run()

class Main(object):

    def __init__(self, options):
        self.options = options
        self.is_interactive = self.options.script is None
        self.out_stream = sys.stdout
        self.system = None
        self.shell = None

    def run(self):
        try:
            try:
                self._before_init()
                return self._init_and_run()
            except:
                # In debug mode, stack trace should be printed for all exceptions
                if env.is_flagged("DEBUG"):
                    traceback.print_exc()
                raise
        except (ConnectionFailed, ManagementException, UnexpectedServerResponse) as e:
            error("Cannot connect to {} system{}: {}", PRODUCT_NAME, " {}".format(self.system._system_name) if self.system._system_name else "",  str(e))
            return 2
        except ServiceTemporarilyUnavailable:
            error("{} temporarily unavailable", PRODUCT_NAME)
            return 2
        except InternalServerError:
            error("An unknown error occurred on the server side: {}", format_exception())
            return 2
        except APIFailure as e:
            msg = "Failed to communicate with server due to product or version mismatch"
            if env.is_flagged("DEBUG") or debug.is_on():
                error(msg + ": {}", format_exception())
            else:
                error(msg)
            return 2
        except FilesystemException as e:
            error(str(e))
            return 1
        except (InvalidOptions, KeyboardInterrupt):
            raise
        except:
            # Save exc_info now, in case other exceptions arise before post mortem
            exc_info = sys.exc_info()
            # For unknown exceptions, stack trace should be printed
            if not env.is_flagged("DEBUG"):
                if debug.is_on():
                    traceback.print_exc()
                error("An unexpected error has occurred - please contact customer support")
            # In debug mode, perform post mortem
            else:
                post_mortem(exc_info)
            return 1

    def _configure_shell_env(self, system):
        # Update the runtimeConfig for later usage
        runtimeConfig.set_system(system)

        # Update the features object to determine what commands will be
        # available to the user
        try:
            response = system.api.get("_features")
        except:
            output_stderr("Experimental features are not available.\n")
            return
        features = {feature['name']:feature for feature in response}
        config['feature_list'] = features
        try:
            from infinishell.infinibox.features import Features
            for feature, params in features.iteritems():
                Features.set_feature(feature, params, system=system)
            Features.set_mgmt_features(system=system)
        except ImportError as e:
            pass
        return

    def _before_init(self):
        if self.options.raw:
            raw_mode.toggle()
        plugins.apply_config()

    def _init_and_run(self):
        self.system = System(self._create_api())
        with system_context.__as__(self.system):
            self._login()
            with self._logout_context():
                self._configure_shell_env(self.system)
                self._create_shell()
                return self._run_shell()

    def _create_api(self):
        base_api = InfiniboxAPI(
                get_api_base_url(self.options.address, self.options.port),
                new_client_id(),
                headers=get_default_headers(),
                encoding=self.options.encoding,
                )
        return base_api.get_sub_resource(REST_BASE_URI, internal=base_api.get_sub_resource(INTERNAL_BASE_URI))

    def _login(self):
        if self.is_interactive:
            output("Connecting to {}...", self.options.address)
        # The no-auth mode comes to work around having to authenticate when
        # using the shell from within the system's OS shell (SSH)
        if env.is_flagged("NO_AUTH"):
            self.system._set_user("INTERNAL", [Roles.INFINIDAT])
        elif env.is_set("EMULATE_LOGIN"):
            login_info = env.get("EMULATE_LOGIN").split(",", 2)
            if len(login_info) == 3:
                [username, password, roles] = login_info
            elif len(login_info) == 1:
                username, password, [roles] = config.username, get_login_password(), login_info
            else:
                raise InvalidOptions('Environment variable {}_EMULATE_LOGIN format is invalid: "{}".\n'
                                     'Expected format (1): "username,password,role[,role,...]"\n'
                                     '                (2): "role"'
                                     .format(env.get_prefix(), login_info))
            self.system._set_user(username, roles.split(","))
            self.system.api.set_basic_auth(username, password)

        elif env.is_set('AUTH_JSESSIONID'):
            try:
                auth_settings = json.loads(env.get('AUTH_JSESSIONID'))
            except ValueError:
                # error('Malformed JSESSIONID cookie: it has to be valid JSON')
                # Continue with the default login sequence
                self.system.login(config.username, get_login_password())
            else:
                self.system._set_user(auth_settings['username'], auth_settings['roles'])
                self.system.api._set_auth_cookie_domain(auth_settings['hostname'])
                self.system.api.set_auth_token(auth_settings['cookie'])

        else:
            self.system.login(config.username, get_login_password())

    @contextmanager
    def _logout_context(self):
        try:
            yield
        finally:
            try:
                if self.is_interactive:
                    output("Logging out...")
                self.system.logout(timeout=LOGOUT_TIMEOUT_SECONDS)
            except Exception:
                if env.is_flagged("DEBUG"):
                    # Don't display this error to the end user:
                    # it's confusing and they can't fix it anyway.
                    error("Failed to log out")
                    traceback.print_exc()

    def _create_commands(self):
        return Commands(
                *(list(COMMANDS) + list(plugins.create_commands())),
                global_options=GLOBAL_OPTIONS,
                args_processor=InfiniboxArgsProcessor(),
                default_renderer=InfiniboxRenderer(),
                api=self.system.api
            )

    def _get_parser_args(self):
        return dict(
            commands=self._create_commands(),
            encoding=self.options.encoding,
            )

    def _get_common_shell_args(self):
        _color_str = config.color.lower()
        _color = True if _color_str == "always" else False if _color_str == "never" else None
        return dict(
            stream=self.out_stream,
            paging=config.paging,
            exc_formatter=InfinidatExceptionFormatter(),
            color=_color,
            errors_color=_color,
            )

    def _create_shell(self):
        if self.is_interactive:
            self.shell = InfiniboxShell(
                parser=ShellParser(**self._get_parser_args()),
                history_filename=get_history_filename(self.options.address, self.system.user.username),
                encoding=self.options.encoding,
                prompt=partial(prompt_hook, address=env.get('PROMPT_ADDRESS', self.options.address), system=self.system),
                prompt_sep=prompt_sep_hook,
                welcome_msg=self.options.welcome_msg,
                exit_msg=self.options.exit_msg,
                error_stream=self.out_stream,  # In interactive mode stderr is not used
                **self._get_common_shell_args()
                )
        else:
            self.shell = InfiniboxExecutor(
                parser=Parser(**self._get_parser_args()),
                keep_going=config.keep_going,
                quiet=config.quiet,
                **self._get_common_shell_args()
                )
        self._prepare_shell()

    def _prepare_shell(self):
        if config.log_file:
            self.shell.start_logging(config.log_file)
        if config.csv_output:
            toggle_csv_mode(self.shell)
        if config.json_output:
            toggle_json_mode(self.shell)
        if config.count_output:
            toggle_count_mode(self.shell)

    def _show_eula(self):
        if self.system.user.has_role("technician","infinidat"):
            return
        config = runtimeConfig.mgmt_config.mgmt
        eula_grace_period = config.get('mgmt.eula_grace_period', None)
        eula_snooze_period = config.get('mgmt.eula_snooze_time', None)
        if not eula_grace_period or not eula_snooze_period:
            return

        def min_to_dhm(m):
            h, m = divmod(m, 60)
            d, h = divmod(h, 24)
            return "{}{}{}".format("{} days".format(d) if d else "", "{}{} hours".format(' ' if d else '', h) if h else "", "{}{} minutes".format(' ' if h else '', m) if m else "")
        
        eula_grace_period = min_to_dhm(eula_grace_period)
        eula_snooze_period = min_to_dhm(eula_snooze_period)

        eula_stats = self.system.api.get("system/eula")
        if eula_stats.status.upper() == "ACCEPTED":
            return True
        if eula_stats.status.upper() == "WAIVED":
            return True
        if eula_stats.status.upper() == "SNOOZED":
            return True
        try:
            eula_text = get("http://{}/pub/eula_infinibox_{}.txt".format(self.system.name(), eula_stats.eula_version))
        except:
            return True
        if eula_text.status_code != 200:
            return True
        eula_text.encoding = "utf-8"
        eula_text = eula_text.text
        
        def _print_eula(eula_text):
            try:
                output(eula_text, paginate=True)
            except PaginationAborted as e:
                return 

        def _get_answer(answers):
            answer = False
            while not answer:
                answer = raw_input("Please select an option: ")
                answer = answer if answer in answers else False
            print("\033[2J\033[1;1f")
            return answer

        def _accept_eula():
            _print_eula(eula_text)
            while True:
                output("""
Do you Accept the EULA?
[1] Accept
[2] Cancel
                """, colors=[colorama.Style.BRIGHT], paginate=True)
                inner_answer = _get_answer(['1','2'])
                if inner_answer == '1':
                    self.system.api.post("system/eula/accepted")
                    output("EULA Accepted.\n")
                    return True
                if inner_answer == '2':
                    return False
            return exit_main_loop

        if self.system.user.has_role("admin","pool_admin"):

            if eula_stats.status.upper() == "SNOOZABLE":
                exit_main_loop = False
                while not exit_main_loop:
                    output("""
EULA (End User License Agreement) was not accepted yet.
EULA needs to be Accepted within {}, after which GUI and CLI will be blocked.
 
You may:
[1] View and Accept the EULA
[2] Snooze EULA reminders for {}
[3] View EULA

Note: To print the EULA, use InfiniBox GUI.
                    """.format(eula_grace_period, eula_snooze_period), colors=[colorama.Style.BRIGHT], paginate=True)
                    answer = _get_answer(['1','2','3'])
                    if answer == '1':
                        exit_main_loop = _accept_eula()
                    if answer == '2':
                        output("""
EULA needs to be Accepted within {}, after which GUI and CLI will be blocked.
EULA will be snoozed for {}.

You may:
[1] Snooze
[2] Back
                            """.format(eula_grace_period, eula_snooze_period), colors=[colorama.Style.BRIGHT], paginate=True)
                        answer = _get_answer(['1','2'])
                        if answer == '1':
                            exit_main_loop = True
                            self.system.api.post("system/eula/snoozed")
                            output("EULA Snoozed.\n")
                    if answer == '3':
                        _print_eula(eula_text)

            if eula_stats.status.upper() == 'EXPIRED':
                while True:
                    output("""
The access to this system was blocked as the End User Licence Agreement was not signed.
 
You may:
[1] View and Accept the EULA
[2] View EULA
[3] Exit InfiniShell

Note: To print the EULA, use InfiniBox GUI.
                    """, colors=[colorama.Style.BRIGHT], paginate=True)
                    answer = _get_answer(['1','2','3'])
                    if answer == '1':
                        if _accept_eula():
                            break
                    if answer == '2':
                        _print_eula(eula_text)
                    if answer == '3':
                        return False

        ### READ_ONLY user ###
        else:
            if eula_stats.status.upper() == "SNOOZABLE":
                output("""
The EULA was not Accepted yet. Please use a non Read Only user to login and accept the EULA.
                """, colors=[colorama.Style.BRIGHT], paginate=True)
                return True
            if eula_stats.status.upper() == 'EXPIRED':
                output("""
The access to this system was blocked as the End User License Agreement was not signed. 
Please use a non Read Only user to login and accept EULA.

Press any key to quit InfiniShell.""", colors=[colorama.Style.BRIGHT], paginate=True)
                raw_input("")
                return False
        return True

        


    def _run_shell(self):
        result_code = 0
        plugins.before_any_session(self.shell)
        if self.is_interactive:
            plugins.before_interactive_session(self.shell)
            self.shell.start(_hook=self._show_eula)
            plugins.after_interactive_session(self.shell)
        else:
            plugins.before_script_session(self.shell)
            success = self.shell.execute(self.options.script)
            result_code = 0 if success else 2
            plugins.after_script_session(self.shell, success)
        plugins.after_any_session(self.shell)
        return result_code
