from __future__ import absolute_import

from itertools import chain, izip
import csv
from numbers import Number
from collections import OrderedDict

from infinishell.infishell import runtime
from infinishell.infishell.shell import get_active_shell

from .renderer import RendererBase
from .values import ValueRenderer
from ..commands.utils import is_detailed_query

class CSVRenderer(RendererBase):
    DEFAULT_SEPARATOR = ","
    def __init__(self, *args, **kwargs):
        super(CSVRenderer, self).__init__(*args, **kwargs)
        self._value_renderer = ValueRenderer(empty_value="")
    def _write_row(self, writer, row):
        self._ensure_stream_open()
        writer.writerow(row)
    def _render_impl(self, obj):
        if obj is None:
            return
        if isinstance(obj, (basestring, Number)):
            self._writeln(str(obj))
            return
        elif self._properties.get("individually", False) and isinstance(obj, dict) and not obj.get('control', False):
            obj = [obj]
        if isinstance(obj, dict) and obj.get('control'):
            chosen_columns = runtime.options.get("columns", ()) or []
            if chosen_columns:
                if '*' in chosen_columns:
                    chosen_columns = self._properties['fields']
                else:
                    chosen_columns = list(OrderedDict.fromkeys(chosen_columns))
            else:
                chosen_columns = []
            if is_detailed_query():
                self._properties['fields'] = (chosen_columns or obj.get('control').get('detailed_fields',  self._properties['fields'])) + obj['control']['additional_fields']
            else:
                self._properties['fields'] = (chosen_columns or obj.get('control').get('fields',  self._properties['fields'])) + obj['control']['additional_fields']
            for table in obj.get('data',[]):
                self._render_collection(self._get_objects(obj))
        else:
            self._render_collection(obj)
        
    def _get_objects(self, objects):
        results = []
        for table in objects.get('data',[]):
            results.append(table.data)
        return chain.from_iterable(results)

    def _render_collection(self, objects):
        it = iter(objects)
        try:
            first_object = next(it)
        except StopIteration:
            return
        field_names, rendered_field_names, _ = self._get_ordered_and_rendered_fields(first_object)
        all_objects = chain([first_object], it)
        writer = csv.DictWriter(self._stream, field_names, lineterminator="\n")
        # Header
        self._write_row(writer, dict(izip(field_names, rendered_field_names)))
        # Rows
        for obj in all_objects:
            self._write_row(writer, { field : self._render_value(obj, field).encode('utf-8', 'ignore') for field in field_names })
    def _render_value(self, obj, field):
        value = obj.get(field, self.ERROR_MISSING_FIELD)
        if value is not self.ERROR_MISSING_FIELD:
            value = self._value_renderer.render(field, value, **self._properties.get("value_properties", {}))
        return value

def toggle_csv_mode(shell=None):
    shell = shell or get_active_shell()
    if not is_csv_mode(shell):
        set_csv_mode(True, shell)
        return True
    set_csv_mode(False, shell)
    return False

def is_csv_mode(shell=None):
    return isinstance((shell or get_active_shell()).get_renderer_override(), CSVRenderer)

def set_csv_mode(enabled, shell=None):
    shell = shell or get_active_shell()
    if enabled:
        shell.set_renderer_override(CSVRenderer())
        shell.disable_stream_encoding()
    else:
        shell.set_renderer_override(None)
        shell.enable_stream_encoding()
