import re
from munch import Munch
from functools import partial

from infinishell.infishell import commandgroup, render, command
from infinishell.infishell.types import (
    Str,
    YesNo,
    Int,
    IPAddress,
    IPAddressCIDR,
    Port,
    )
from infinishell.infishell.utils import strlist
from infinishell.infishell.exceptions import BadValue, BadArguments
from ..types import (
    NONE_CONSTANT,
    NetworkInterfaceName,
    Name,
    EthPort,
    UsableEthPort,
    NetworkInterface,
    PhysicalNetworkInterface,
    NetworkInterfaceType,
    NetworkSpace,
    NetworkSpaceService,
    NetworkRateLimitOrNone,
    Node,
    IPAddressOrNone,
    MTU,
    ISCSIDefaultSecurityMethod,
    ISCSIISNSServer,
    NetworkRateLimit,
    TcpPort,
    )
from ..options import QueryOptions, Grep, InvertedGrep
from ..utils import get_patch_panel_ports
from ...exceptions import NotSupported, NoSuchEntity
from .. import roles, EntityCommands, LazyIdNameMapping, LazyEntityMapping, Null
from . import ConfigCommandsBase

from ...features import feature_command


def get_interfaces(api, param_type, interfaces, node=None, flatten=True):
    if not isinstance(interfaces[0], list):
        interfaces = [interfaces]
    if node:
        for i in interfaces:
            for j in i:
                if j.node_id == node.id:
                    if flatten:
                        return [j]
                    else:
                        return [[j]]
        else:
            raise NoSuchEntity(u'No such {}: "{}" on node-{}'.format(param_type.get_entity_name(), interfaces[0][0].name, node.id))
    if flatten:
        return [i for inter in interfaces for i in inter]
    return interfaces
    


@commandgroup(parent=ConfigCommandsBase, prefix="ethernet")
class EthernetConfigCommands(object):
    pass

@commandgroup(parent=EthernetConfigCommands, prefix="port")
class EthernetPortCommands(EntityCommands):
    ENTITY_TYPE = EthPort
    QUERY_FIELDS =           strlist("name node_index role rendered_state link_state connection_speed max_speed patch_panel_label")
    DETAILED_QUERY_FIELDS =  QUERY_FIELDS + strlist("hw_addr _name")
    INFINIDAT_QUERY_FIELDS = QUERY_FIELDS + strlist("firmware hw_addr")
    INFINIDAT_COLUMNS_FIELDS = INFINIDAT_QUERY_FIELDS + strlist("_name vendor model")
    QUERY_SUBSTITUTES = dict(
        node_index="node",
        _name="device_name",
        rendered_state="state",
        link_state="link",
        connection_speed="current_speed",
        patch_panel_label="patch_panel",
        )

    @command(
        description="List system ethernet ports",
        types=(EthPort, Node),
        hidden_substitue=dict(name="port"),
        options=QueryOptions(fields=DETAILED_QUERY_FIELDS, infinidat_fields=INFINIDAT_COLUMNS_FIELDS, substitute=QUERY_SUBSTITUTES, detailed=True),
        )
    @render(
        fields=QUERY_FIELDS,
        detailed_fields=DETAILED_QUERY_FIELDS,
        substitute=QUERY_SUBSTITUTES,
        infinidat=dict(
            fields=INFINIDAT_QUERY_FIELDS,
            detailed_fields=INFINIDAT_COLUMNS_FIELDS,
            ),
        technician=dict(
            fields=INFINIDAT_QUERY_FIELDS,
            detailed_fields=INFINIDAT_COLUMNS_FIELDS,
            ),
        )
    def query(self, _api, port=None, node=None):
        patch_panel_ports = get_patch_panel_ports("ETH")
        return self._query(_api, port, node_index=node, _hook=partial(self._query_hook, patch_panel_ports=patch_panel_ports))

    def _query_hook(self, ports, patch_panel_ports=None):
        for port in ports:
            port.rendered_state = port.state if not port.state_description \
                                  else "{} ({})".format(port.state, port.state_description)
            port.patch_panel_label = None
            pp_port = None if not patch_panel_ports else patch_panel_ports.get((port.node_index, port.port_number))
            if pp_port:
                port.patch_panel_label = (pp_port.label or pp_port.port_num)
        return ports

    # @command(
    #     description="Enable ethernet port",
    #     types=(EthPort.List,),
    #     substitute=dict(
    #         ports="name",
    #         ),
    #     )
    # def enable(self, _api, ports):
    #     aggregator = _api.create_request_aggregator(
    #         total_success_msg=self._result_message(ports[0], "enabled") if len(ports) == 1 else
    #                           ("Ports enabled: " + ", ".join(p.name for p in ports))
    #         )
    #     with aggregator:
    #         for port in ports:
    #             aggregator.next_task(port.name, "Enabled")
    #             _api.post(EthPort.get_component_uri(port, "enable"))
    #     return aggregator.run()

    # @command(
    #     description="Disable ethernet port",
    #     types=(EthPort.List,),
    #     substitute=dict(
    #         ports="name",
    #         ),
    #     )
    # def disable(self, _api, ports):
    #     aggregator = _api.create_request_aggregator(
    #         total_success_msg=self._result_message(ports[0], "disabled") if len(ports) == 1 else
    #                           ("Ports disabled: " + ", ".join(p.name for p in ports))
    #         )
    #     with aggregator:
    #         for port in ports:
    #             aggregator.next_task(port.name, "Disabled")
    #             _api.post(EthPort.get_component_uri(port, "disable"))
    #     return aggregator.run()

@commandgroup(parent=EthernetConfigCommands, prefix="interface")
class NetworkInterfaceCommands(EntityCommands):
    ENTITY_TYPE = NetworkInterface
    QUERY_FIELDS = strlist("name type node_id port_names physical_interface network_spaces state")
    SORTABLE_FIELDS = strlist("name node_id")
    QUERY_SUBSTITUTES = dict(
        node_id="node",
        port_names="ports",
        )

    @command(
        description="List existing network interfaces",
        types=(NetworkInterface.List, Node),
        hidden_substitue=dict(name="interface"),
        options=QueryOptions(fields=QUERY_FIELDS, sortable=SORTABLE_FIELDS, substitute=QUERY_SUBSTITUTES),
        )
    @render(
        fields=QUERY_FIELDS,
        substitute=QUERY_SUBSTITUTES,
        alignment=dict(
            network_spaces="left",
            ),
        )
    def query(self, _api, interface=None, node=None):
        hook = partial(self._query_hook,
            eth_names_to_port_names={(p.node_index, p._name) : p.name for p in EthPort.iter_all(_api)},
            ns_ids_to_names=LazyIdNameMapping(NetworkSpace, _api),
            interface_ids_to_names=LazyIdNameMapping(NetworkInterface, _api),
            )
        return self._query(_api, interface, node_id=node, _hook=hook)

    def _query_hook(self, interfaces, eth_names_to_port_names=None, ns_ids_to_names=None, interface_ids_to_names=None):
        for interface in interfaces:
            if interface.type == "PORT_GROUP":
                interface.port_names = [eth_names_to_port_names.get((interface.node_id, p.name), "<N/A>") for p in interface.ports]
                interface.physical_interface = "N/A"
            elif interface.type == 'VLAN':
                interface.port_names = "N/A"
                interface.physical_interface = [interface_ids_to_names[interface.underlying_interface_id]]
            elif interface.type == 'PORT':
                interface.port_names = "N/A"
                interface.physical_interface = "N/A"
            else:
                raise NotSupported, interface.type
            interface.network_spaces = [ns_ids_to_names[ns_id] for ns_id in interface.network_space_ids]
        return interfaces

    @command(
        description="Create network interface",
        types=(UsableEthPort.List, NetworkInterfaceName, NetworkInterfaceType, Node),
        )
    @roles.INFINIDAT.ADMIN.TECHNICIAN
    def create(self, _api, ports, name, type=None, node=None):
        port_objects = [dict(name=port[0]._name) for port in ports]
        if node is None:
            aggregator = _api.create_request_aggregator()
            with aggregator:
                for node in Node.iter_all(_api):
                    aggregator.next_task("Interface {}".format(name), "Interface {} created on node-{}".format(name, node.id))
                    self._create(_api,
                        name=name,
                        type=type or 'PORT_GROUP',
                        node_id=node.id,
                        ports=port_objects,
                        )
            return aggregator.run()
        else:
            self._create(_api,
                name=name,
                type=type or 'PORT_GROUP',
                node_id=node.id,
                ports=port_objects,
                )
            return "Interface {} created on node-{}".format(name, node.id)

    @feature_command(
        feature="vlans",
        status=True,
        description="Create network interface of type VLAN",
        types=(PhysicalNetworkInterface, Int, Node),
        )
    @roles.INFINIDAT.ADMIN.TECHNICIAN
    def create_vlan(self, _api, physical_interface, vlan_id, node=None):
        interfaces = get_interfaces(_api, PhysicalNetworkInterface, physical_interface, node=node)
        aggregator = _api.create_request_aggregator()
        with aggregator:
            for interface in interfaces:
                name = '{}.{}'.format(interface.name, vlan_id)
                aggregator.next_task("VLAN {}".format(name), "VLAN {} created on interface {} (node-{})".format(name, interface.name, interface.node_id))
                params = dict(
                    name=name,
                    type="VLAN",
                    underlying_interface_id=interface.id,
                    vlan=vlan_id,
                    )
                self._create(_api,
                    **params
                    )
        return aggregator.run()

    @command(
        description="Create a single-port network interface (for special purposes only)",
        types=(UsableEthPort, Node,),
        )
    @roles.INFINIDAT
    def create_single_port(self, _api, port, node):
        return self._create(_api,
            name=port[0]._name,
            type="PORT",
            node_id=node.id,
            )

    @command(
        description="Delete network interface",
        types=(NetworkInterface, Node),
        hidden_substitue=dict(name="interface"),
        )
    @roles.INFINIDAT.ADMIN.TECHNICIAN
    def delete(self, _api, interface, node=None):
        interfaces = get_interfaces(_api, NetworkInterface, interface, node=node)
        aggregator = _api.create_request_aggregator()
        with aggregator:
            for i in interfaces:
                aggregator.next_task("Interface {}".format(i.name), "Interface {} deleted on node-{}".format(i.name, i.node_id))
                self._delete(_api, i)
        return aggregator.run()

    @command(
        description="Rename network interface",
        types=(PhysicalNetworkInterface, NetworkInterfaceName, Node),
        hidden_substitue=dict(name="interface"),
        )
    @roles.INFINIDAT.ADMIN.TECHNICIAN
    def rename(self, _api, interface, new_name, node=None):
        interfaces = get_interfaces(_api, PhysicalNetworkInterface, interface, node=node)
        name = interfaces[0].name if interfaces else ""
        aggregator = _api.create_request_aggregator(
            total_success_msg="Interface {}: Interface name changed to '{}'".format(name, new_name),
            )
        with aggregator:
            for interface in interfaces:
                aggregator.next_task("Interface {}".format(interface.name), "Interface {} on node-{} renamed to {}".format(interface.name, interface.node_id, new_name))
                self._rename(_api, interface, new_name)
        return aggregator.run()


    @command(
        description="Enable network interface",
        types=(PhysicalNetworkInterface, Node),
        hidden_substitue=dict(name="interface"),
        )
    @roles.INFINIDAT.ADMIN.TECHNICIAN
    def enable(self, _api, interface, node=None):
        interfaces = get_interfaces(_api, PhysicalNetworkInterface, interface, node=node)
        aggregator = _api.create_request_aggregator()
        with aggregator:
            for interface in interfaces:
                aggregator.next_task("Interface {}".format(interface.name), "Interface {} enabled on node-{}".format(interface.name, interface.node_id))
                self._post(_api, interface, "enable")
        return aggregator.run()

    @command(
        description="Disable network interface",
        types=(NetworkInterface, Node),
        hidden_substitue=dict(name="interface"),
        )
    @roles.INFINIDAT.ADMIN.TECHNICIAN
    def disable(self, _api, interface, node=None):
        interfaces = get_interfaces(_api, NetworkInterface, interface, node=node)
        aggregator = _api.create_request_aggregator()
        with aggregator:
            for interface in interfaces:
                aggregator.next_task("Interface {}".format(interface.name), "Interface {} disabled on node-{}".format(interface.name, interface.node_id))
                self._post(_api, interface, "disable")
        return aggregator.run()

    @command(
        description="Add ethernet port to network interface",
        types=(PhysicalNetworkInterface, UsableEthPort.List, Node),
        hidden_substitue=dict(name="interface"),
        substitute=dict(
            ports="port",
            ),
        )
    @roles.INFINIDAT.ADMIN.TECHNICIAN
    def add_port(self, _api, interface, ports, node=None):
        interfaces = get_interfaces(_api, PhysicalNetworkInterface, interface, node=node)
        aggregator = _api.create_request_aggregator(
            total_success_msg="Port{} added to {}: {}".format("" if len(ports) == 1 else "s", self._result_subject_and_entity(interfaces[0]),
                                                              ", ".join(p[0].name for p in ports)),
            )
        with aggregator:
            for interface in interfaces:
                for port in ports:
                    aggregator.next_task("Port {}".format(port[0].name), "Added on node-{}".format(interface.node_id))
                    self._post(_api, interface, "ports", data=dict(name=port[0]._name))
        return aggregator.run()

    @command(
        description="Remove ethernet port from network interface",
        types=(PhysicalNetworkInterface, UsableEthPort.List, Node),
        hidden_substitue=dict(name="interface"),
        substitute=dict(
            ports="port",
            ),
        )
    @roles.INFINIDAT.ADMIN.TECHNICIAN
    def remove_port(self, _api, interface, ports, node=None):
        interfaces = get_interfaces(_api, PhysicalNetworkInterface, interface, node=node)
        aggregator = _api.create_request_aggregator(
            total_success_msg="Port{} removed from {}: {}".format("" if len(ports) == 1 else "s", self._result_subject_and_entity(interfaces[0]),
                                                                  ", ".join(p[0].name for p in ports)),
            )
        with aggregator:
            for interface in interfaces:
                for port in ports:
                    aggregator.next_task("Port {}".format(port[0].name), "Removed on node-{}".format(interface.node_id))
                    self._del(_api, interface, "ports", port[0]._name)
        return aggregator.run()

@commandgroup(parent=ConfigCommandsBase, prefix="net_space")
class NetworkSpaceCommands(EntityCommands):
    ENTITY_TYPE = NetworkSpace
    QUERY_FIELDS = strlist("name service network mtu rate_limit gateway ips interface_names automatic_ip_fallback")
    SORTABLE_FIELDS = strlist("name")
    QUERY_SUBSTITUTES = dict(
        services="service",
        automatic_ip_fallback="auto_ip_fallback",
        interface_names="interfaces",
        )
    MGMT_SERVICE_TO_USER_SERVICE = {v : k for k, v in NetworkSpaceService()._get_mapping().iteritems()}

    @command(
        description="List existing network spaces",
        types=(NetworkSpace, NetworkSpaceService),
        hidden_substitue=dict(name="net_space"),
        options=QueryOptions(fields=QUERY_FIELDS, sortable=SORTABLE_FIELDS, substitute=QUERY_SUBSTITUTES, detailed=True),
        )
    @render(
        fields=QUERY_FIELDS,
        detailed_fields=QUERY_FIELDS,
        detailed_individually=True,
        substitute=QUERY_SUBSTITUTES,
        )
    def query(self, _api, net_space=None, service=None):
        lazy_interface_name_by_id = LazyIdNameMapping(NetworkInterface, _api)
        return self._query(_api, net_space, _hook=partial(self._query_hook, lazy_interface_name_by_id=lazy_interface_name_by_id, service=service))

    def _query_hook(self, spaces, lazy_interface_name_by_id, service=None):
        res_spaces = []
        for net_space in spaces:
            if service and net_space.service != service:
                continue
            res_spaces.append(net_space)
            net_space.ips = len(net_space.ips)
            net_space.network, net_space.gateway = "{}/{}".format(net_space.network_config.network, net_space.network_config.netmask), net_space.network_config.default_gateway
            net_space.interface_names = list(set([lazy_interface_name_by_id[id] for id in net_space.interfaces]))
            net_space.service = self.MGMT_SERVICE_TO_USER_SERVICE.get(net_space.service, net_space.service)
        return res_spaces

    @command(
        description="""Sets a bandwidth limitation on a Network Space per node. This limit is configured
        on the Interface used by the Network space to allow traffic shaping and avoid bandwidth starvation""",
        types=(NetworkSpace, NetworkRateLimitOrNone),
        hidden_substitue=dict(name="net_space"),
        )
    @roles.INFINIDAT.ADMIN.TECHNICIAN.POOLADMIN
    def set_rate_limit(self, _api, net_space, limit_per_node):
        if limit_per_node is NONE_CONSTANT:
            limit_per_node = Null
        self._update(_api, net_space, rate_limit=limit_per_node)
        return self._result_message(net_space, "bandwidth limitation (per node): {}".format("Limited to {}Mbps".format(limit_per_node) if limit_per_node is not Null else "Limit removed"))


    @command(
        description="Create network space",
        types=(Name, NetworkSpaceService, NetworkInterface.List, IPAddressCIDR, IPAddress, MTU, NetworkRateLimit),
        )
    @roles.INFINIDAT.ADMIN
    def create(self, _api, name, service, interface, network, default_gateway=None, mtu=None, rate_limit=None):
        interfaces = get_interfaces(_api, NetworkInterface, interface, node=None)
        params = dict(
            name=name,
            service=service,
            rate_limit=rate_limit,
            network_config=dict(
                netmask=int(network.cidr),  # infi.shell bug, cidr is a string - fixed in v1.12.6
                network=network.ip,
                default_gateway=default_gateway.ip if default_gateway else None,
                ),
            interfaces=[i.id for i in interfaces],
            mtu=mtu,
        )
        return self._create(_api, **params)

    @command(
        description="Rename network space",
        types=(NetworkSpace, Name),
        hidden_substitue=dict(name="net_space"),
        )
    @roles.INFINIDAT.ADMIN
    def rename(self, _api, net_space, new_name):
        return self._rename(_api, net_space, new_name)

    @command(
        description="Set default gateway",
        types=(NetworkSpace, IPAddressOrNone),
        hidden_substitue=dict(name="net_space"),
        )
    @roles.INFINIDAT.ADMIN
    def set_default_gateway(self, _api, net_space, default_gateway):
        return self._update(_api, net_space,
            network_config=dict(
                netmask=net_space.network_config.netmask,
                network=net_space.network_config.network,
                default_gateway=None if default_gateway is NONE_CONSTANT else default_gateway.ip,
                ),
            returns=self._result_message(net_space, 'default gateway unset' if default_gateway is NONE_CONSTANT
                                             else 'default gateway set to "{}"'.format(default_gateway.ip)),
            )

    @command(
        description="Set network space port groups",
        types=(NetworkSpace, NetworkInterface.List),
        hidden_substitue=dict(name="net_space"),
        )
    @roles.INFINIDAT.ADMIN
    def set_interfaces(self, _api, net_space, interfaces):
        interfaces = get_interfaces(_api, NetworkInterface, interfaces, node=None, flatten=True)
        return self._update(_api, net_space,
            interfaces=[i.id for i in interfaces],
            returns=self._result_message(net_space, "interfaces set to " +
                ", ".join('"{}"'.format(i.name) for i in interfaces))
            )

    @command(
        description="Set network space network configuration",
        types=(NetworkSpace, IPAddressCIDR),
        hidden_substitue=dict(name="net_space"),
        )
    @roles.INFINIDAT.ADMIN
    def set_network(self, _api, net_space, network):
        return self._update(_api, net_space,
            network_config=dict(
                netmask=int(network.cidr),  # infi.shell bug, cidr is a string - fixed in v1.12.6
                network=network.ip,
                default_gateway=net_space.network_config.default_gateway,
                ),
            returns=self._result_message(net_space, 'network set to "{}"'.format(network.ip_and_cidr)),
            )

    @command(
        description="Set network space MTU",
        types=(NetworkSpace, MTU),
        hidden_substitue=dict(name="net_space"),
        )
    @roles.INFINIDAT.ADMIN
    def set_mtu(self, _api, net_space, mtu):
        return self._update(_api, net_space,
            mtu=mtu,
            returns=self._result_message(net_space, 'MTU set to {}'.format(mtu)),
            )


    # @feature_command(
    #     feature="iscsi",
    #     status=True,
    #     description="Enables a Network Space, activates its service and IP addresses.",
    #     types=(NetworkSpace,),
    #     hidden_substitue=dict(name="net_space"),
    #     )
    # @roles.INFINIDAT.ADMIN
    # def enable(self, _api, net_space):
    #     if net_space.enabled == True:
    #         return self._result_message(net_space, "already enabled")
    #     self._update(_api, net_space, enabled=True)
    #     return self._result_message(net_space, "enabled")

    # @feature_command(
    #     feature="iscsi",
    #     status=True,
    #     description="Disabled a Network Space, deactivates its service and IP addresses.",
    #     types=(NetworkSpace,),
    #     hidden_substitue=dict(name="net_space"),
    #     )
    # @roles.INFINIDAT.ADMIN
    # def disable(self, _api, net_space):
    #     if net_space.enabled == False:
    #         return self._result_message(net_space, "already disabled")
    #     self._update(_api, net_space, enabled=False)
    #     return self._result_message(net_space, "disabled")



    @command(
        description="Delete network space",
        types=(NetworkSpace,),
        hidden_substitue=dict(name="net_space"),
        )
    @roles.INFINIDAT.ADMIN
    def delete(self, _api, net_space):
        return self._delete(_api, net_space)

@commandgroup(parent=NetworkSpaceCommands, prefix="params")
class NetworkSpaceParamsCommands(EntityCommands):
    ENTITY_TYPE = NetworkSpace
    _QUERY_LSTRIP_PATTERN = re.compile(r"^iscsi_")
    QUERY_STRIPPED_FIELDS = strlist("""
        name
        iscsi_iqn
        iscsi_tcp_port
        iscsi_isns_servers
        """)
    QUERY_STRIPPED_SUBSTITUTES = dict(
        iscsi_iqn="iqn",
        iscsi_tcp_port="tcp_port",
        iscsi_isns_servers="isns_servers",
        )

    @feature_command(
        feature="iscsi",
        status=True,
        description="List network space service parameters",
        types=(NetworkSpace, NetworkSpaceService),
        substitute=dict(
            ns="net_space",
            ),
        options=QueryOptions(fields=QUERY_STRIPPED_FIELDS, substitute=QUERY_STRIPPED_SUBSTITUTES, detailed=True),
        )
    @render(
        fields=QUERY_STRIPPED_FIELDS,
        detailed_individually=True,
        substitute=QUERY_STRIPPED_SUBSTITUTES,
        )
    def query(self, _api, ns=None, service=None):
        if not ns and not service:
            raise BadArguments("Please specify either net_space or service")
        return self._query(_api, ns, service=service, _hook=self._query_hook)

    def _query_hook(self, netspaces):
        for ns in netspaces:
            if ns.service == "ISCSI_SERVICE":
                ns.properties.name = ns.name
                yield ns.properties

    @feature_command(
        feature="iscsi",
        status=True,
        description="Set network space iSCSI parameters",
        types=(NetworkSpace, TcpPort, ISCSIISNSServer.List),
        substitute=dict(
            ns="net_space",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def set_iscsi(self, _api, ns, tcp_port=None, isns_servers=None):
        if all(arg is None for arg in (tcp_port, isns_servers)):
            raise BadArguments("At least one more argument must be specified")
        if isns_servers:
            # If only one value was specified in the list, and that value is NONE, then empty the list. See CONSTANTS of ISCSIISNSServer.
            if len(isns_servers) == 1 and isns_servers[0] is None:
                isns_servers = ()
            else:
                isns_servers = [hostname for hostname in isns_servers if hostname is not None]
        props = {}
        for attr, value in (
            ("iscsi_tcp_port", tcp_port),
            ("iscsi_isns_servers", isns_servers),
            ):
            if value is not None:
                props[attr] = value
        return self._update(_api, ns, properties=props)

@commandgroup(parent=NetworkSpaceCommands, prefix="ip")
class NetworkSpaceIpCommands(EntityCommands):
    ENTITY_TYPE = NetworkSpace

    QUERY_FIELDS = strlist("network_space ip_address enabled node network_interface reserved")
    QUERY_SUBSTITUTES = dict(
        reserved="replication_management_ip",
        )

    @command(
        description="Query network space IPs",
        types=(NetworkSpace,),
        substitute=dict(
            ns="net_space",
            ),
        options=(Grep(), InvertedGrep()),
        )
    @render(
        fields=QUERY_FIELDS,
        substitute=QUERY_SUBSTITUTES,
        alignment=dict(
            network_space="left",  # Because "space" is usually used for capacities
            ),
        )
    def query(self, _api, ns=None):
        lazy_interface_by_id = LazyEntityMapping(NetworkInterface, _api, fields=strlist("name node_id"))
        return self._query(_api, ns, _hook=partial(self._query_hook, lazy_interface_by_id=lazy_interface_by_id))

    def _query_hook(self, spaces, lazy_interface_by_id=None):
        for ns in spaces:
            for ip in ns.ips:
                interface = None if not ip.interface_id else lazy_interface_by_id[ip.interface_id]
                yield Munch(
                    network_space=ns.name,
                    ip_address=ip.ip_address,
                    enabled=ip.enabled,
                    reserved=ip.reserved,
                    node=None if not interface else interface.node_id,
                    network_interface=None if not interface else interface.name,
                    )

    def _aggregated_result_message(self, action, ns, ips):
        if len(ips) == 1:
            return 'IP address {} {} in {} {}'.format(ips[0].ip, action, self._result_subject(), self._result_entity_name(ns))
        return "IP addresses {} in {} {}: {}".format(action, self._result_subject(), self._result_entity_name(ns), ", ".join(ip.ip for ip in ips))

    @command(
        description="Create IP address in network space",
        types=(NetworkSpace, IPAddress.List),
        substitute=dict(
            ns="net_space",
            ips="ip_address",
            )
        )
    @roles.INFINIDAT.ADMIN
    def create(self, _api, ns, ips):
        aggregator = _api.create_request_aggregator(
            total_success_msg=self._aggregated_result_message("created", ns, ips)
            )
        with aggregator:
            for ip in ips:
                aggregator.next_task(ip.ip, "Created")
                self._post(_api, ns, "ips", data=ip.ip)
        return aggregator.run()

    def _ensure_ip_exists(self, ns, ip):
        if ip.ip not in (ns_ip.ip_address for ns_ip in ns.ips):
            raise BadValue("{} {} has no such IP address: {}".format(self._result_subject(), self._result_entity_name(ns), ip.ip))

    @command(
        description="Delete IP address from network space",
        types=(NetworkSpace, IPAddress.List),
        substitute=dict(
            ns="net_space",
            ips="ip_address",
            )
        )
    @roles.INFINIDAT.ADMIN
    def delete(self, _api, ns, ips):
        for ip in ips:
            self._ensure_ip_exists(ns, ip)
        aggregator = _api.create_request_aggregator(
            total_success_msg=self._aggregated_result_message("deleted", ns, ips)
            )
        with aggregator:
            for ip in ips:
                aggregator.next_task(ip.ip, "Deleted")
                self._del(_api, ns, "ips", ip.ip)
        return aggregator.run()

    @command(
        description="Enable network space IP address",
        types=(NetworkSpace, IPAddress.List),
        substitute=dict(
            ns="net_space",
            ips="ip_address",
            )
        )
    @roles.INFINIDAT.ADMIN
    def enable(self, _api, ns, ips):
        for ip in ips:
            self._ensure_ip_exists(ns, ip)
        aggregator = _api.create_request_aggregator(
            total_success_msg=self._aggregated_result_message("enabled", ns, ips)
            )
        with aggregator:
            for ip in ips:
                aggregator.next_task(ip.ip, "Enabled")
                self._post(_api, ns, "ips", ip.ip, "enable")
        return aggregator.run()

    @command(
        description="Disable network space IP address",
        types=(NetworkSpace, IPAddress.List),
        substitute=dict(
            ns="net_space",
            ips="ip_address",
            )
        )
    @roles.INFINIDAT.ADMIN
    def disable(self, _api, ns, ips):
        for ip in ips:
            self._ensure_ip_exists(ns, ip)
        aggregator = _api.create_request_aggregator(
            total_success_msg=self._aggregated_result_message("disabled", ns, ips)
            )
        with aggregator:
            for ip in ips:
                aggregator.next_task(ip.ip, "Disabled")
                self._post(_api, ns, "ips", ip.ip, "disable")
        return aggregator.run()
