from munch import Munch

from infinishell.infishell import render
from infinishell.infishell.commands import commandgroup, command
from infinishell.infishell.types import Str
from infinishell.infishell.utils import strlist
from infinishell.infishell.exceptions import BadArguments

from .options import QueryOptions
from .types import ObjType, Name, Metadata
from . import EntityCommands, roles, render_collection, set_renderer_query_fields

@commandgroup(prefix="metadata")
class MetadataCommands(EntityCommands):
    ENTITY_TYPE = object()  # Dummy

    def _get_object(self, _api, objtype, objname):
        if objtype.get_name() == 'SYSTEM':
            return Munch(
                id='system',
                )
        return objtype.get_by_name(_api, objname)

    @command(
        description="Set metadata on object",
        types=(ObjType, Str, Str, Name),
        substitute=dict(objname="object"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def set(self, _api, objtype, key, value, objname=None):
        if objtype.get_name() == 'SYSTEM' and objname:
            raise BadArguments, "Cannot use object parameter with objtype=system"
        if objtype.get_name() != 'SYSTEM' and not objname:
            raise BadArguments, "object parameter is mandatory with objtype={}".format(objtype.get_name().lower())
        obj = self._get_object(_api, objtype, objname)
        data = {key : value}
        _api.put("metadata/{}".format(obj.id), data)
        if objtype.get_name() != 'SYSTEM':
            return self._result_message(obj, 'metadata key "{}" set to "{}"'.format(key, value), subject=objtype)
        else:
            return "System metadata key '{}' set to '{}'".format(key, value)

    @command(
        description="Get metadata from object",
        types=(ObjType, Name, Str),
        substitute=dict(objname="object"),
        options=QueryOptions(fields=strlist("key value"), sortable=strlist("key"), detailed=True),
        )
    @render_collection(
        entity=Metadata,
        fields=strlist("key value"),
        detailed_fields=strlist("key value"),
        detailed_individually=True,
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def get(self, _api, objtype, objname=None, key=None):
        if objtype.get_name() == 'SYSTEM' and objname:
            raise BadArguments, "Cannot use object parameter with objtype=system"
        if objtype.get_name() != 'SYSTEM' and not objname:
            raise BadArguments, "object parameter is mandatory with objtype={}".format(objtype.get_name().lower())
        set_renderer_query_fields()
        obj = self._get_object(_api, objtype, objname)
        if key is None:
            return _api.paged_query("metadata/{}".format(obj.id),
                params=_api.create_sort().asc("key").to_params(),
                hook=self._query_hook,
                )
        return self._query_hook([_api.get(u"metadata/{}/{}".format(obj.id, key))])

    def _query_hook(self, metadata_objects):
        for m in metadata_objects:
            yield Munch(key=m.key, value=m.value)

    @command(
        description="Remove a single metadata key from object",
        types=(ObjType, Name, Str),
        substitute=dict(objname="object"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def remove(self, _api, objtype, key, objname=None):
        if objtype.get_name() == 'SYSTEM' and objname:
            raise BadArguments, "Cannot use object parameter with objtype=system"
        if objtype.get_name() != 'SYSTEM' and not objname:
            raise BadArguments, "object parameter is mandatory with objtype={}".format(objtype.get_name().lower())
        obj = self._get_object(_api, objtype, objname)
        _api.delete(u"metadata/{}/{}".format(obj.id, key))
        if objtype.get_name() != 'SYSTEM':
            return self._result_message(obj, 'metadata key "{}" removed'.format(key), subject=objtype)
        else:
            return 'system metadata key "{}" removed'.format(key)

    @command(
        description="Clear all metadata keys from object",
        types=(ObjType, Name),
        substitute=dict(objname="object"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def clear(self, _api, objtype, objname=None):
        if objtype.get_name() == 'SYSTEM' and objname:
            raise BadArguments, "Cannot use object parameter with objtype=system"
        if objtype.get_name() != 'SYSTEM' and not objname:
            raise BadArguments, "object parameter is mandatory with objtype={}".format(objtype.get_name().lower())
        obj = self._get_object(_api, objtype, objname)
        _api.delete("metadata/{}".format(obj.id))
        if objtype.get_name() != 'SYSTEM':
            return self._result_message(obj, "metadata cleared", subject=objtype)
        else:
            return 'System metadata cleared'
