from munch import Munch

from infinishell.infishell.commands import commandgroup, command, deprecated
from infinishell.infishell.exceptions import BadValue
from infinishell.infishell.types import YesNo
from infinishell.infishell.utils import strlist

from ..exceptions import ManagementException
from ..renderer import render_capacity
from ..system import system
from ..features import feature_command

from .hierarchical_base import HierarchicalEntityCommandsBase
from .types import (
    Pool,
    Filesystem,
    FilesystemSnapshot,
    FilesystemSnapshotWP,
    FilesystemSnapshotNotWP,
    FilesystemOrSnapshot,
    FilesystemOrSnapshotWP,
    FilesystemOrSnapshotNotWP,
    MasterFilesystem,
    Name,
    Size,
    SizeDelta,
    CapacityUnit,
    FSSeries,
    )
from .options import QueryOptions
from .utils import get_target_pool, format_savings
from . import render_collection, roles, set_rendered_unit, set_renderer_query_fields


FS_QUERY_FIELDS                     = strlist("name                        thin size used allocated tree_allocated pool write_protected mapped compression_enabled data_reduction")
DETAILED_FS_QUERY_FIELDS            = strlist("name nfs_filesystem_id      thin size used allocated tree_allocated pool write_protected mapped compression_enabled data_reduction ssd_enabled root_mode created_at updated_at")

SNAP_QUERY_FIELDS                   = strlist("name                        thin size used allocated                pool write_protected mapped                                                          created_at")
DETAILED_SNAP_QUERY_FIELDS          = strlist("name nfs_filesystem_id      thin size used allocated                pool write_protected mapped                                    ssd_enabled root_mode created_at updated_at")

CLONE_QUERY_FIELDS            = strlist("name             thin size used allocated tree_allocated pool write_protected mapped                              created_at")
DETAILED_CLONE_QUERY_FIELDS   = DETAILED_FS_QUERY_FIELDS

FS_TREE_FIELDS                      = strlist("name                   type      size used allocated                pool write_protected                                                                 created_at")
DETAILED_FS_TREE_FIELDS             = strlist("name nfs_filesystem_id type thin size used allocated tree_allocated pool write_protected mapped compression_enabled data_reduction ssd_enabled root_mode created_at updated_at")

SORTABLE_FIELDS = strlist("name serial size write_protected mapped ssd_enabled created_at updated_at compression_enabled")

QUERY_SUBSTITUTES = dict(
    write_protected="wp",
    ssd_enabled="ssd_cache",
    root_mode="default_filesystem_permissions",
    nfs_filesystem_id="nfs_fs_id",
    mapped="exported",
    compression_enabled="compressed",
    tree_allocated="Snapshots",
    )

DETAILED_QUERY_SUBSTITUTES = dict(
    write_protected="write_protected",
    )

class FilesystemCommandsMixin(object):
    FS_TYPE_NAMES = Munch(
        MASTER="filesystem",
        SNAPSHOT="snapshot",
        )

    def _get_fs_type(self, fs):
        if fs.type == "MASTER":
            return "FILESYSTEM"
        return fs.type

    def _get_fs_type_name(self, volume):
        return self.FS_TYPE_NAMES.get(volume.type) or "filesystem"

class FilesystemCommandsBase(FilesystemCommandsMixin, HierarchicalEntityCommandsBase):
    HIERARCHICAL_TYPE = Filesystem

    def _query_hook(self, filesystems, _api, pool_names_by_id=None):
        for fs in filesystems:
            if pool_names_by_id:
                fs.pool = pool_names_by_id[fs.pool_id]
            fs.type = self._get_fs_type(fs)
            fs.thin = fs.provtype == self.THIN
            if 'capacity_savings' in fs:
                fs.data_reduction = format_savings(fs)
        return filesystems

def create_common_filesystem_commands_mixin(FilesystemType, fs_arg):

    class CommonFilesystemCommandsMixin(object):

        @command(
            description="Rename {}".format(FilesystemType.DESCRIPTION_SUBJECT),
            types=(FilesystemType, Name),
            hidden_substitue=dict(name="fs"),
            substitute=dict() if FilesystemType == MasterFilesystem else dict(fs="snap"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def rename(self, _api, fs, new_name):
            return self._rename(_api, fs, new_name)

        @command(
            description="Delete {}".format(FilesystemType.DESCRIPTION_SUBJECT),
            types=(FilesystemType.List,),
            hidden_substitue=dict(name="fs"),
            substitute=dict() if FilesystemType == MasterFilesystem else dict(fs="snap"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def delete(self, _api, fs):
            return self._delete(_api, fs)

        @command(
            description="Enable or disable SSD cache",
            types=(FilesystemType, YesNo, YesNo),
            hidden_substitue=dict(name="fs"),
            substitute=dict() if FilesystemType == MasterFilesystem else dict(fs="snap"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def cache(self, _api, fs, ssd_cache, recursive=False):
            return self._update_recursive(_api,
                entity=fs,
                recurse=recursive,
                success_msg="{} cached on SSD".format("now" if ssd_cache else "no longer"),
                success_msg_with_is=True,
                noop_msg="already{} cached on SSD".format("" if ssd_cache else " not"),
                noop_msg_with_is=True,
                props=dict(
                    ssd_enabled=ssd_cache,
                    ),
                predicate=lambda dataset: dataset.ssd_enabled != ssd_cache,
                )

    return CommonFilesystemCommandsMixin

def create_writable_filesystem_commands_mixin(FilesystemType, fs_arg):

    class WritableFilesystemCommandsMixin(object):

        @command(
            description="Resize {}".format(FilesystemType.DESCRIPTION_SUBJECT),
            types=(FilesystemType, SizeDelta),
            hidden_substitue=dict(name="fs"),
            substitute=dict() if FilesystemType == MasterFilesystem else dict(fs="snap"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def resize(self, _api, fs, size):
            size = size.compute_size(fs.size)
            if fs.size == size:
                return self._result_message(fs, "size is already {}".format(render_capacity(fs.size)))
            self._update(_api, fs, size=size)
            return self._result_message(fs, "resized")

        @command(
            description="Restore data state of {} from previous snapshot".format(FilesystemType.DESCRIPTION_SUBJECT),
            types=(FilesystemType, FilesystemSnapshot),
            hidden_substitue=dict(target="fs"),
            substitute=dict() if FilesystemType == MasterFilesystem else dict(fs="snap"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def restore(self, _api, fs, source):
            if fs.id == source.id:
                raise BadValue('"{}" was specified as both arguments'.format(fs.name))
            # _api.post(FilesystemType.get_uri(fs.id, "restore"), source.id)
            self._post(_api, fs, "restore", data=dict(
                source_id=source.id
                ))
            return self._result_message(fs, 'restored from {} "{}"'.format(self._get_fs_type_name(source), source.name))

        @command(
            description="Disable write access on {}".format(FilesystemType.DESCRIPTION_SUBJECT),
            types=(FilesystemType, YesNo),
            hidden_substitue=dict(name="fs"),
            substitute=dict() if FilesystemType == MasterFilesystem else dict(fs="snap"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def write_protect(self, _api, fs, recursive=False):
            return self._update_recursive(_api,
                entity=fs,
                recurse=recursive,
                success_msg="now write-protected",
                success_msg_with_is=True,
                noop_msg="already write-protected",
                noop_msg_with_is=True,
                props=dict(
                    write_protected=True,
                    ),
                predicate=lambda dataset: dataset.type != "SNAP" and not dataset.write_protected,
                )

        @command(
            description="Enable write access on {}".format(FilesystemType.DESCRIPTION_SUBJECT),
            types=(FilesystemType, YesNo),
            hidden_substitue=dict(name="fs"),
            substitute=dict() if FilesystemType == MasterFilesystem else dict(fs="snap"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def write_enable(self, _api, fs, recursive=False):
            return self._update_recursive(_api,
                entity=fs,
                recurse=recursive,
                success_msg="now write-enabled",
                success_msg_with_is=True,
                noop_msg="already write-enabled",
                noop_msg_with_is=True,
                props=dict(
                    write_protected=False,
                    ),
                predicate=lambda dataset: dataset.type != "SNAP" and dataset.write_protected,
                )

    return WritableFilesystemCommandsMixin

def create_child_filesystem_commands_mixin(FilesystemType, ParentType, fs_arg, parent_arg, create_parent_arg):

    class ChildFilesystemCommandsMixin(object):

        @command(
            description="Create a new {} from {}".format(FilesystemType.DESCRIPTION_SUBJECT, ParentType.DESCRIPTION_SUBJECT),
            types=(ParentType, Name, YesNo),
            hidden_substitue=dict(source="fs"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def create(self, _api, fs, name, ssd_cache=None):
            parent = self._create_and_return_entity(_api, parent_id=fs.id, name=name, ssd_enabled=ssd_cache)
            return self._result_message(parent, "created", subject=FilesystemType.get_entity_name())

        @command(
            description="Display hierarchy of snapshots",
            types=(FilesystemType, CapacityUnit),
            hidden_substitue=dict(name="fs"),
            substitute=dict() if FilesystemType == MasterFilesystem else dict(fs="snap"),
            options=QueryOptions(fields=DETAILED_FS_TREE_FIELDS, sortable=FS_TREE_FIELDS, substitute=DETAILED_QUERY_SUBSTITUTES),
            )
        @render_collection(
            entity=Filesystem,
            fields=FS_TREE_FIELDS,
            substitute=QUERY_SUBSTITUTES,
        )
        def tree(self, _api, fs, unit=None):
            set_rendered_unit(unit)
            set_renderer_query_fields()
            return self._get_tree(_api, fs)

    return ChildFilesystemCommandsMixin

@commandgroup(prefix="fs")
class FilesystemCommands(FilesystemCommandsBase,
                         create_common_filesystem_commands_mixin(MasterFilesystem, "fs"),
                         create_writable_filesystem_commands_mixin(MasterFilesystem, "fs")):
    ENTITY_TYPE = MasterFilesystem
    @command(
        description="List existing filesystems",
        types=(MasterFilesystem.List, Pool, CapacityUnit, YesNo, YesNo, YesNo, YesNo),
        hidden_substitue=dict(name="fs"),
        options=QueryOptions(fields=DETAILED_FS_QUERY_FIELDS, sortable=SORTABLE_FIELDS, substitute=QUERY_SUBSTITUTES, detailed=True),
        )
    @render_collection(
        entity=Filesystem,
        fields=FS_QUERY_FIELDS,
        detailed_fields=DETAILED_FS_QUERY_FIELDS,
        detailed_substitute=DETAILED_QUERY_SUBSTITUTES,
        detailed_individually=True,
        substitute=QUERY_SUBSTITUTES,
        )
    def query(self, _api, fs=None, pool=None, unit=None, has_exports=None, write_protected=None, thin=None, compression_enabled=None):
        set_rendered_unit(unit)
        provtype = None
        if thin is not None:
            provtype = self.THIN if thin else self.THICK
        return self._common_query(_api, fs, pool_id=pool, mapped=has_exports, write_protected=write_protected, provtype=provtype, compression_enabled=compression_enabled)

    def _query_hook(self, entities, _api, pool_names_by_id=None):
        for e in entities:
            if system.user.has_role('infinidat','technician'):
                if e.get('compression_enabled', False) and e.get('compression_suppressed', False):
                    e.compression_enabled = "yes (suppressed)"
        return super(FilesystemCommands, self)._query_hook(entities, _api, pool_names_by_id)

    @feature_command(
        feature="compression",
        status=True,
        description="Create a new filesystem",
        types=(Name.List, Size, Pool, YesNo, YesNo, FSSeries, YesNo),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def create(self, _api, name, size, pool=None, thin=False, ssd_cache=None, series=None, compression=None):
        pool = get_target_pool(_api, pool)
        if series:
            # In case of multiple filesystems, since this is not an atomic
            # operation, ensure there's enough free space for all requested
            # filesystems.
            total_size = size * series.total
            if total_size > pool.free_virtual_space:
                raise ManagementException(409, Munch(code="POOL_VIRTUAL_CAPACITY", message="Pool virtual capacity reached."))
            if not thin and total_size > pool.free_physical_space:
                raise ManagementException(409, Munch(code="POOL_PHYSICAL_CAPACITY", message="Pool physical capacity reached."))
        return self._create_single_or_multiple_or_series(_api, name, series,
            size=size,
            pool_id=pool.id,
            provtype=self.THIN if thin else self.THICK,
            write_protected=False,
            ssd_enabled=ssd_cache,
            compression_enabled=compression
            )








    @feature_command(
        feature="compression",
        status=False,
        name='create',
        description="Create a new filesystem",
        types=(Name.List, Size, Pool, YesNo, YesNo, FSSeries),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def old_create(self, _api, name, size, pool=None, thin=False, ssd_cache=None, series=None):
        pool = get_target_pool(_api, pool)
        if series:
            # In case of multiple filesystems, since this is not an atomic
            # operation, ensure there's enough free space for all requested
            # filesystems.
            total_size = size * series.total
            if total_size > pool.free_virtual_space:
                raise ManagementException(409, Munch(code="POOL_VIRTUAL_CAPACITY", message="Pool virtual capacity reached."))
            if total_size > pool.free_physical_space:
                raise ManagementException(409, Munch(code="POOL_PHYSICAL_CAPACITY", message="Pool physical capacity reached."))
        return self._create_single_or_multiple_or_series(_api, name, series,
            size=size,
            pool_id=pool.id,
            provtype=self.THIN if thin else self.THICK,
            write_protected=False,
            ssd_enabled=ssd_cache,
            )








    @command(
        description="Move filesystem to a different pool",
        types=(MasterFilesystem, Pool, YesNo),
        hidden_substitue=dict(name="fs"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def move(self, _api, fs, pool, move_capacity=False):
        if fs.pool_id == pool.id:
            return self._result_message(fs, 'already in {} "{}"'.format(Pool.get_entity_name(), pool.name))
        _api.post(self.ENTITY_TYPE.get_uri(fs.id, "move"), dict(
            pool_id=pool.id,
            with_capacity=move_capacity,
            ))
        return self._result_message(fs, 'moved to pool "{}" {} its capacity'
                                        .format(pool.name, "along with" if move_capacity else "without"))

    @command(
        description="Set filesystem provisioning type",
        types=(MasterFilesystem, YesNo),
        hidden_substitue=dict(name="fs"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def set_thin(self, _api, fs, thin):
        prov_name = "thin" if thin else "thick"
        # if fs.provtype.upper() == self._get_provtype(thin):
        #     raise CommandAborted(self._result_message(fs, "already set to {} provisioning".format(prov_name)))
        self._update(_api, fs, provtype=self._get_provtype(thin))
        return self._result_message(fs, "set to {} provisioning".format(prov_name))

    @feature_command(
        feature="compression",
        status=True,
        description="Set filesystem compression",
        types=(MasterFilesystem, YesNo),
        hidden_substitue=dict(name="fs"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN.TECHNICIAN
    def set_compression(self, _api, fs, compression):
        if fs.compression_enabled  == compression:
            return self._result_message(fs, "compression is already {}".format("enabled" if compression else "disabled"))
        self._update(_api, fs, compression_enabled=compression)
        return 'Compression was {} for filesystem "{}"'.format("enabled" if compression else "disabled", fs.name)

    @command(
        description="Estimate reclaimed capacity (deletion simulation)",
        types=(Filesystem,),
        hidden_substitue=dict(name="fs"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def reclaimable(self, _api, fs):
        data = dict(entities=[fs.id])
        result = _api.post(Filesystem.get_uri("delete_simulation"), data)
        return 'Estimated reclaimable space for filesystem "{}" and its tree descendants: {}'.format(fs.name, render_capacity(result.space_reclaimable))

    @command(
        description="Display hierarchy of filesystems and snapshots",
        types=(MasterFilesystem, Pool, CapacityUnit),
        hidden_substitue=dict(name="fs"),
        options=QueryOptions(fields=DETAILED_FS_TREE_FIELDS, sortable=FS_TREE_FIELDS, substitute=DETAILED_QUERY_SUBSTITUTES),
        )
    @render_collection(
        entity=Filesystem,
        fields=FS_TREE_FIELDS,
        substitute=QUERY_SUBSTITUTES,
        )
    def tree(self, _api, fs=None, pool=None, unit=None):
        set_rendered_unit(unit)
        set_renderer_query_fields()
        return self._get_tree(_api, fs, pool)

@commandgroup(prefix="fs.snap")
class FilesystemSnapshotCommands(FilesystemCommandsBase,
                                 create_common_filesystem_commands_mixin(
                                     FilesystemType=FilesystemSnapshot,
                                     fs_arg="snap",
                                     ),
                                 create_writable_filesystem_commands_mixin(
                                     FilesystemType=FilesystemSnapshot,
                                     fs_arg="snap",
                                     ),
                                 create_child_filesystem_commands_mixin(
                                     FilesystemType=FilesystemSnapshot,
                                     ParentType=FilesystemOrSnapshot,
                                     fs_arg="snap",
                                     parent_arg="fs",
                                     create_parent_arg="fs",
                                     )):
    ENTITY_TYPE = FilesystemSnapshot

    @command(
        description="List existing snapshots",
        types=(FilesystemSnapshot.List, FilesystemOrSnapshot, Pool, CapacityUnit, YesNo, YesNo),
        hidden_substitue=dict(name="snap"),
        options=QueryOptions(fields=DETAILED_SNAP_QUERY_FIELDS, sortable=SORTABLE_FIELDS, substitute=QUERY_SUBSTITUTES, detailed=True),
        )
    @render_collection(
        entity=Filesystem,
        fields=SNAP_QUERY_FIELDS,
        detailed_fields=DETAILED_SNAP_QUERY_FIELDS,
        detailed_substitute=DETAILED_QUERY_SUBSTITUTES,
        detailed_individually=True,
        substitute=QUERY_SUBSTITUTES,
        )
    def query(self, _api, snap=None, source=None, pool=None, unit=None, has_exports=None, write_protected=None):
        set_rendered_unit(unit)
        return self._common_query(_api, snap, parent_id=source, pool_id=pool, type=FilesystemSnapshot.ENTITY_MGMT_SUBTYPE, mapped=has_exports, write_protected=write_protected)

    @command(description='Refresh a snapshot',
             types=(FilesystemSnapshot,),
             hidden_substitue=dict(name="snap"),
             )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def refresh(self, _api, snap):
        _api.post(self.ENTITY_TYPE.get_uri(snap.id, 'refresh'),
                  {'source_id': snap.parent_id})
        return self._result_message(snap, 'refreshed')


@commandgroup(prefix="clone", parent=FilesystemCommands, hidden=True)
class FilesystemCloneCommands(object):


    @command(
        description="List existing clones",
        types=(FilesystemOrSnapshot.List, FilesystemOrSnapshot, Pool, CapacityUnit, YesNo),
        hidden_substitue=dict(name="clone"),
        options=QueryOptions(fields=DETAILED_CLONE_QUERY_FIELDS, sortable=SORTABLE_FIELDS, substitute=QUERY_SUBSTITUTES, detailed=True),
        )
    @render_collection(
        entity=Filesystem,
        fields=CLONE_QUERY_FIELDS,
        detailed_fields=DETAILED_CLONE_QUERY_FIELDS,
        detailed_substitute=DETAILED_QUERY_SUBSTITUTES,
        detailed_individually=True,
        substitute=QUERY_SUBSTITUTES,
        )
    @deprecated
    def query(self, _api, clone=None, source=None, pool=None, unit=None, has_exports=None):
        data = FilesystemSnapshotCommands().query(_api, snap=clone, source=source, pool=pool, unit=unit, has_exports=has_exports, write_protected=False)
        return data        

    @command(
        description="Resize clone",
        types=(FilesystemOrSnapshot, SizeDelta),
        hidden_substitue=dict(name="clone"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    @deprecated
    def resize(self, _api, clone, size):
        data = FilesystemSnapshotCommands().resize(_api, fs=clone, size=size)
        return data

    @command(
        description="Restore data state of clone from previous snapshot",
        types=(FilesystemOrSnapshot, FilesystemSnapshotWP),
        hidden_substitue=dict(name="clone"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    @deprecated
    def restore(self, _api, clone, source):
        data = FilesystemSnapshotCommands().restore(_api, fs=clone, source=source)
        return data
 
    @command(
        description="Disable write access on clone",
        types=(FilesystemOrSnapshot, YesNo),
        hidden_substitue=dict(name="clone"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    @deprecated
    def write_protect(self, _api, clone, recursive=False):
        data = FilesystemSnapshotCommands().write_protect(_api, fs=clone, recursive=recursive)
        return data
 
    @command(
        description="Enable write access on clone",
        types=(FilesystemOrSnapshot, YesNo),
        substitute=dict(fs="name"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    @deprecated
    def write_enable(self, _api, fs, recursive=False):
        data = FilesystemSnapshotCommands().write_enable(_api, fs=fs, recursive=recursive)
        return data

    @command(
        description="Create a new clone from snapshot",
        types=(FilesystemOrSnapshot, Name, YesNo),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    @deprecated
    def create(self, _api, snap, name, ssd_cache=None):
        data = FilesystemSnapshotCommands().create(_api, fs=snap, name=name, ssd_cache=ssd_cache)
        clone = FilesystemOrSnapshot.get_by_name(_api, name)
        FilesystemSnapshotCommands().write_enable( _api, fs=clone, recursive=False)
        return data

    @command(
        description="Display hierarchy of snapshots and clones",
        types=(FilesystemOrSnapshot, CapacityUnit),
        hidden_substitue=dict(name="clone"),
        options=QueryOptions(fields=DETAILED_FS_TREE_FIELDS, sortable=FS_TREE_FIELDS, substitute=DETAILED_QUERY_SUBSTITUTES),
        )
    @render_collection(
        entity=Filesystem,
        fields=FS_TREE_FIELDS,
        substitute=QUERY_SUBSTITUTES,
        )
    @deprecated
    def tree(self, _api, clone, unit=None):
        data = FilesystemSnapshotCommands().tree(_api, fs=clone, unit=unit)
        return data
    
    @command(
        description="Rename clone",
        types=(FilesystemOrSnapshot, Name),
        hidden_substitue=dict(name="clone"),
        substitute=dict(),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    @deprecated
    def rename(self, _api, clone, new_name):
        data = FilesystemSnapshotCommands().rename(_api, fs=clone, new_name=new_name)
        return data

    @command(
        description="Delete clone",
        types=(FilesystemOrSnapshot.List,),
        hidden_substitue=dict(name="clone"),
        substitute=dict(),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    @deprecated
    def delete(self, _api, clone):
        data = FilesystemSnapshotCommands().delete(_api, fs=clone)
        return data

    @command(
        description="Enable or disable SSD cache",
        types=(FilesystemOrSnapshot, YesNo, YesNo),
        hidden_substitue=dict(name="clone"),
        substitute=dict(),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    @deprecated
    def cache(self, _api, clone, ssd_cache, recursive=False):
        data = FilesystemSnapshotCommands().cache(_api, fs=clone, ssd_cache=ssd_cache, recursive=recursive)
        return data

