from munch import Munch
from itertools import chain
from functools import partial
from functools32 import lru_cache

import time
import os
import ipaddress
import re
from collections import defaultdict

from ..renderer import CollectionRenderer
from infinishell.infishell import commandgroup, command, get_active_shell, render
from infinishell.infishell.utils import strlist
from infinishell.infishell.exceptions import BadArguments
from infinishell.infishell.types import (
    Str,
    Int,
    IPAddress,
)

from ...utils import Version
from ..exceptions import InvalidSystemState, ConnectionFailed, ManagementException
from .types import (
    TimeDelta,
    MetricsNasSortByType,
    MetricsResults,
    MetricsCount,
    MetricsSanSortByType,
    Pool,
    FilesystemOrSnapshotNotWP,
    DatasetOrSnapshot,
    User,
    MetricsInterval,
    VolumeOrSnapshotNotWP,
    Host,
    UsedFCLoggedInInitiator,
    UsediSCSILoggedInInitiator,
    FCHardPort,
    FCSoftTarget,
    ReplicationLinkBySystemName,
    Initiator,
    )
from ..system import system
from .options import QueryOptions
from .mapping import MappingCommandsBase
from . import EntityCommands, render_collection, set_renderer_query_fields, roles, LazyIdNameMapping
from .utils import getch_pause



class MetricsException(ManagementException):
    def __init__(self):
        error = Munch(
            code="METRICS_NOT_COLLECTED",
            message="The requested metrics could not be collected"
            )
        super(MetricsException, self).__init__(http_code="409",error_obj=error)


ALIASES = dict(
    system="system",
    read="READ",
    write="WRITE",
    xcopy="XCOPY",
    write_same="WRITE_SAME",
    )
ALIASES['meta_data'] = "METADATA"


class CounterBuilder(object):

    active_filters = dict()
    active_collectors = dict()
    active_collectors_data = dict()
    collected_fields = []

    def __init__(self, *args, **kwargs):
        if not kwargs.get('api', None):
            raise Exception, "api must be provided"
        self.api = kwargs['api']

    def _set_filter(self, filter_type):
        try:
            if self.protocol_field is None:
                response = self.api.post("metrics/filters", dict(protocol=self.protocol))
            else:
                params = dict()
                params[self.protocol_field] = self.protocol
                response = self.api.post("metrics/filters", params)
        except ManagementException as e:
            # overide mgmt error since mgmt doesnt give any usefull information
            if e.code == "INVALID_FILTER_HANDLE_ID":
                raise MetricsException()
            raise e
        self.active_filters.update({filter_type:response.id})

    def _add_filters_to_filter(self, filter_id):
        for field, value in self.filter_fields.iteritems():
            if value:
                values = dict()
                values[field] = value
                response = self.api.put("metrics/filters/%s" % filter_id, values)
                if response.id != filter_id:
                    raise Exception, "could not build filter with %s" % values
        return True

    def _set_collector(self, filter_type, filter_id, filters=None):
        collector_dict = dict(
            filter_id=filter_id,
            filters=filters,
            type=self.collector_type,
            collected_fields=self.collected_fields,
            sorting_field=self.sorting_field,
            grouping_field=self.grouping_field,
            max_results=self.max_results,
            )
        collector_dict = {k: v for k, v in collector_dict.items() if v is not None}
        try:
            response = self.api.post("metrics/collectors",collector_dict)
        except ManagementException as e:
            # overide mgmt error since mgmt doesnt give any usefull information
            if e.code == "INVALID_COLLECTOR_HANDLE_ID":
                raise MetricsException()
            raise e
        self.active_collectors.update({filter_type:response})

    def _get_data_from_collectors(self):
        self.active_collectors_data = dict()
        ids = "in:({})".format(",".join([str(v.id) for _, v in self.active_collectors.iteritems()]))
        response = self.api.get('metrics/collectors/data', params=dict(collector_id=ids))
        for collector in response.collectors:
            self.active_collectors_data[collector.id] = collector

    def get_message(self):
        return self.message if self.message else None

    # Overide this method
    def build(self):
        pass

    # Overide this method
    def get_data(self):
        pass
    
    def destroy_collectors(self):
        for ctype, collector in self.active_collectors.items():
            try:
                self.api.delete("metrics/collectors/{}".format(collector.id))
            except:
                continue



class TopCounterBuilder(CounterBuilder):

    collector_type = "TOP"

    def __init__(self, *args, **kwargs):
        super(TopCounterBuilder, self).__init__(*args, **kwargs)
        self.grouping_field = kwargs['grouping_field']
        self.message = kwargs['message']

    def get_filters_params(self, *args, **kwargs):
        self.protocol = kwargs['protocol']
        self.protocol_field = kwargs.get('protocol_field' , None)
        self.filter_fields = kwargs.get('filter_fields' , None)
        self.filters = kwargs.get('filters', None)

    def get_collectors_params(self, *args, **kwargs):
        self.collected_fields = kwargs.get('collected_fields', self.collected_fields)
        self.convert_data_function = kwargs.get('convert_data_function', None)
        self.sorting_field = kwargs['sorting_field']
        self.max_results = kwargs['max_results']

    def add_filters(self):
        self.active_filters = dict()
        self._set_filter("top")

    def add_filters_to_filters(self):
        if self.filter_fields is None:
            return
        for filter_type, filter_id in self.active_filters.iteritems():
            self._add_filters_to_filter(filter_id)

    def _add_collectors(self):
        self.active_collectors = dict()
        for filter_type, filter_id in self.active_filters.iteritems():
            self._set_collector(filter_type, filter_id)

    def build(self):
        self.add_filters()
        self.add_filters_to_filters()
        self._add_collectors()

    def get_data(self):
        self._get_data_from_collectors()
        self._format_data()
        return [v for _, values in self.active_collectors_data.iteritems() for v in values]

    def _format_data(self):
        result_data = {}
        data = self.active_collectors_data[self.active_collectors['top'].id]
        fields = data.fields
        if data.data == [] or data.data[0] == []:
            result_data = [Munch(zip(fields, ["-"] * len(fields)))]
        else:
            result_data = [Munch(zip(fields, i)) for i in data.data[0] if len(i)]
        result = self.convert_data_function(data=result_data) if self.convert_data_function else result_data
        self.active_collectors_data[data.id] = result


class ShowCounterBuilder(CounterBuilder):

    collector_type = "COUNTER"
    filters = dict()
    params_substitute = dict()


    def __init__(self, *args, **kwargs):
        super(ShowCounterBuilder, self).__init__(*args, **kwargs)
        self.grouping_field = kwargs['grouping_field']
        self.message = kwargs['message']
        self.filters = kwargs.get('filters', self.filters)

    def get_filters_params(self, *args, **kwargs):
        self.protocol = kwargs['protocol']
        self.protocol_field = kwargs.get('protocol_field' , None)
        self.filters_fields = kwargs.get('filters_fields', None)
        self.sorting_field = None
        self.grouping_field = None

    def get_collectors_params(self, *args, **kwargs):
        self.collected_fields = kwargs.get('collected_fields', self.collected_fields)
        self.grouping_field = kwargs.get('grouping_field', self.grouping_field)
        self.convert_data_function = kwargs.get('convert_data_function', None)
        self.max_results = None

    def _add_collectors(self):
        self.active_collectors = dict()
        if not self.filters:
            fields = dict()
            fields[self.protocol_field or 'protocol'] = self.protocol
            self._set_collector(filter_type="system", filter_id=None, filters=fields)
        filters = [[key, subval] for key, value in self.filters.iteritems() for subval in value]
        for f in filters:
            fields = dict()
            fields[self.protocol_field or 'protocol'] = self.protocol
            fields[f[0]] = f[1]
            if self.filters_fields:
                for key, val in self.filters_fields.iteritems():
                    if val is not None:
                        fields[key] = val
            self._set_collector(filter_type=f[1], filter_id=None, filters=fields)

    def build(self):
        self._add_collectors()

    def get_data(self):
        self._get_data_from_collectors()
        self._format_data()
        return [v for k, v in self.active_collectors_data.iteritems()]

    def _format_data(self):
        data = self.active_collectors_data
        for collector_type, collector_params in self.active_collectors.iteritems():
            data[collector_params.id]["metrics_collector_type"] = self.params_substitute.get(collector_type, ALIASES[collector_type])
        for _, collector in data.iteritems():
            if collector.data == []:
                for field in collector.fields:
                    collector[field] = 0
            else:
                for field, value in zip(collector.fields, collector.data[0]):
                    collector[field] = value
        self.active_collectors_data = data


class TopNasCounterBuilder(TopCounterBuilder):

    collected_fields = ["throughput", "ops", "latency"]


class ShowNasCounterBuilder(ShowCounterBuilder):

    collected_fields = ["throughput", "ops", "latency"]
    filters = dict(
        operation_type=['read', 'write'],
        operation_category=['meta_data'],
        )


class TopSanCounterBuilder(TopCounterBuilder):

    collected_fields = ['throughput', 'ops', 'external_latency_wout_err', 'errors']


class ShowSanCounterBuilder(ShowCounterBuilder):

    collected_fields = ['throughput', 'ops', 'external_latency_wout_err', 'errors']
    filters = dict(
        operation_type=['read', 'write', 'xcopy', 'write_same'],
        )

    def __init__(self, *args, **kwargs):
        super(ShowSanCounterBuilder, self).__init__(*args, **kwargs)
        self.collected_fields = ['throughput', 'ops', 'external_latency_wout_err', 'internal_latency_wout_err', 'errors']
        # set more collected fields for internal usage
        if system.user.has_role('infinidat','technician'):
            self.collected_fields = self.collected_fields + \
            ["internal_latency", "external_latency"]


class TopAsyncCounterBuilder(TopCounterBuilder):

    collected_fields = ['throughput']


class ShowAsyncCounterBuilder(ShowCounterBuilder):

    collected_fields = ['throughput']
    filters = dict(
        operation_type=['read', 'write'],
        )
    params_substitute = dict(
        read='sent',
        write='received'
        )





params = {
        'api' : None,
        'element_name' : None,
        'initiators': None
        }

@lru_cache(maxsize=1000)
def get_name(element_id):
    global params
    return params['element_name'].get_by_id(params['api'], element_id)

def add_element_names_to_data(_api, element_name, element_type, data, filter=None):
    global params
    params['api'] = _api
    params['element_name'] = element_name
    results = []
    for i in data:
        if i[element_type] == '-':
            i["metrics_{}_name".format(element_name.__name__).lower()] = i[element_type]
            results.append(i)
            continue
        try:
            element = get_name(int(i[element_type]))
        except ManagementException:
            i["metrics_{}_name".format(element_name.__name__).lower()] = "<ID_NOT_FOUND={}>".format(i[element_type])
        else:
            if filter:
                if element.get(filter['attribute'], None) != filter['value']:
                    continue
            i["metrics_{}_name".format(element_name.__name__).lower()] = element.name
        results.append(i)
    return results

def parse_file_path(_api, element_name, element_type, data):
    global params
    params['api'] = _api
    params['element_name'] = element_name
    for i in data:
        if i[element_type] in ['-',{}]:
            i["metrics_{}_name".format(element_name.__name__).lower()] = i[element_type]
            continue
        filesystem = get_name(int(i[element_type]['fs_entity_id']))
        path = i[element_type]['path']
        file_path = "{}{}".format(filesystem.name, path)
        i["file_path"] = file_path
    return data    

def initiator_decimal_to_address(data):
    for i in data:
        if i['initiator_WWPN'] == '-':
            continue
        decimal_address = i['initiator_WWPN']
        address = '%x' % int(decimal_address)
        address = hex_to_colon_seperated_address(address)
        i['initiator_WWPN'] = address
    return data

def get_target_node_port(target):
    node = target >> 8
    port = target & 0xff
    return node , port

def target_decimal_to_address(api, data):
    for i in data:
        # decimal_target = i['target_port']
        if i['target_WWPN'] == '-':
            continue
        # node, port = get_target_node_port(int(decimal_target))
        # target_address = FCHardPort._get_entity_by_ids(api, node, port)
        # i['target_port'] = target_address.wwpn
        i['target_WWPN'] = hex_to_colon_seperated_address('%x' % int(i['target_WWPN']))
    return data

def initiator_port_key_to_address(_api, type, data):
    global params
    params['api'] = _api
    params['initiators'] =params['initiators'] or {initiator.port_key : initiator for initiator in Initiator.iter_all(_api)}
    port_keys = []
    for i in data:
        if i['initiator'] != '-':
            port_keys.append(i.initiator)
    initiators = dict()
    for key in port_keys:
        initiators[key] = params['initiators'][long(key)]
    for i in data:
        if i['initiator'] == '-':
            i["metrics_initiator_name"] = i['initiator']
            continue
        element = initiators[str(i['initiator'])]
        i["metrics_initiator_name"] = element.address if type == 'iscsi' else hex_to_colon_seperated_address(element.address)
    return data

def hex_to_decimal(hex):
    line = re.sub(':', '', hex)
    return int(line, 16)

def hex_to_colon_seperated_address(element):
    return ':'.join(a+b for a,b in zip(element[::2], element[1::2]))

def decimal_to_hex(element_name, element_type, data):
    for i in data:
        if i[element_type] == '-':
            i["metrics_{}_name".format(element_name).lower()] = i[element_type]
            continue
        element = '%x' % int(i[element_type])
        element = hex_to_colon_seperated_address(element)
        i[element_name.lower()] = element
    return data

#convert int to ip form 1.1.1.1
def int_ip_to_long(element_type, data):
    for i in data:
        if i[element_type] == '-':
            continue
        ip = ipaddress.IPv4Address(int(i[element_type])).compressed
        ip = ip.split('.')
        ip.reverse()
        ip = '.'.join(ip)
        i[element_type] = ip
    return data

#reverse the int sent to mgmt since the FM stores it in a reversed order :/
def ip_to_reverse_int(ip):
    ip = ip.split('.')
    ip.reverse()
    o = map(int, ip)
    res = (16777216 * o[0]) + (65536 * o[1]) + (256 * o[2]) + o[3]
    return res


def run_loop(builder, renderer_kwargs, interval=5, error_interval=10, count=None):
    output_stream = get_active_shell().output_stream
    render = partial(
        CollectionRenderer().render,
        stream=output_stream,
        header=True,
        **renderer_kwargs
        )

    context = Munch(
        print_data=[],
        render_flag=False,
        time_passed=0,
        retry=True,
        )

    def loop(last_iteration=False, waiting=False):
        try:
            print("\033[2J\033[1;1f")
            message = builder.get_message()
            if message:
                print("System {}\n".format(system.name()))
                print(message + "\n\n")
            if waiting:
                print("Collecting data...")
                time.sleep(2)
                return
            else:
                context.print_data = builder.get_data()
                # print("********************************")
                # print(context.print_data)
                # print("********************************\n\n")
                render(context.print_data, header=True)
            if last_iteration:
                return
            if getattr(getch_pause, 'prefer_Q', False):
                print('\n\nPress Q to return to the shell')
            getch_pause(interval)
            context.time_passed = 0
        except (ConnectionFailed, ManagementException) as e:
            # overide mgmt error since mgmt doesnt know if this collector ever existed
            try:
                if e.code in ("INVALID_COLLECTOR_HANDLE_ID", "INVALID_FILTER_HANDLE_ID"):
                    if context.retry:
                        builder.build()
                        context.retry = False
                    else:
                        raise MetricsException()
            except Exception: 
                pass
            getch_pause(error_interval)
            context.time_passed += error_interval
            if context.time_passed > 500:
                raise e

    try:
        #First loop enterance, only sleep
        loop(waiting=True)

        if count:
            for i in xrange(count):
                loop(last_iteration=i == count - 1)
            builder.destroy_collectors()
            return

        while True:
            loop()
        builder.destroy_collectors()
    except KeyboardInterrupt:
        builder.destroy_collectors()


SYSTEM_QUERY_FIELDS = strlist("metrics_collector_type ops throughput latency")
SYSTEM_QUERY_SUBSTITUTES = dict(
    metrics_collector_type="---",
    ops="operations",
    )
SYSTEM_SHOW_RENDERER_KWARGS = dict(
        fields=SYSTEM_QUERY_FIELDS,
        substitute=SYSTEM_QUERY_SUBSTITUTES,
    )

POOL_QUERY_FIELDS = strlist("metrics_pool_name ops throughput latency")
POOL_QUERY_SUBSTITUTES = dict(
    metrics_pool_name="pool_name",
    ops="operations",
    )
POOL_TOP_RENDERER_KWARGS = dict(
        fields=POOL_QUERY_FIELDS,
        substitute=POOL_QUERY_SUBSTITUTES,
    )

FS_QUERY_FIELDS = strlist("metrics_filesystemorsnapshotnotwp_name ops throughput latency")
FS_QUERY_SUBSTITUTES = dict(
    metrics_filesystemorsnapshotnotwp_name="fs_name",
    ops="operations",
    )
FS_TOP_RENDERER_KWARGS = dict(
        fields=FS_QUERY_FIELDS,
        substitute=FS_QUERY_SUBSTITUTES,
    )

UID_QUERY_FIELDS = strlist("metrics_user_name user_id ops throughput latency")
UID_QUERY_SUBSTITUTES = dict(
    metrics_user_name="user_name",
    ops="operations",
    )
UID_TOP_RENDERER_KWARGS = dict(
        fields=UID_QUERY_FIELDS,
        substitute=UID_QUERY_SUBSTITUTES,
    )

GID_QUERY_FIELDS = strlist("metrics_group_name group_id ops throughput latency")
GID_QUERY_SUBSTITUTES = dict(
    metrics_group_name="group_name",
    ops="operations",
    )
GID_TOP_RENDERER_KWARGS = dict(
        fields=GID_QUERY_FIELDS,
        substitute=GID_QUERY_SUBSTITUTES,
    )

SOURCE_IP_QUERY_FIELDS = strlist("metrics_source_ip_name source_ip ops throughput latency")
SOURCE_IP_QUERY_SUBSTITUTES = dict(
    metrics_source_ip_name="source_ip",
    ops="operations",
    )
SOURCE_IP_TOP_RENDERER_KWARGS = dict(
        fields=SOURCE_IP_QUERY_FIELDS,
        substitute=SOURCE_IP_QUERY_SUBSTITUTES,
    )

DESTINATION_IP_QUERY_FIELDS = strlist("metrics_destination_ip_name destination_ip ops throughput latency")
DESTINATION_IP_QUERY_SUBSTITUTES = dict(
    metrics_destination_ip_name="destination_ip",
    ops="operations",
    )
DESTINATION_IP_TOP_RENDERER_KWARGS = dict(
        fields=DESTINATION_IP_QUERY_FIELDS,
        substitute=DESTINATION_IP_QUERY_SUBSTITUTES,
    )

NAS_FILE_PATH_QUERY_FIELDS = strlist("file_path ops throughput latency")
NAS_FILE_PATH_QUERY_SUBSTITUTES = dict(
    ops="operations",
    )
NAS_FILE_PATH_TOP_RENDERER_KWARGS = dict(
        fields=NAS_FILE_PATH_QUERY_FIELDS,
        substitute=NAS_FILE_PATH_QUERY_SUBSTITUTES,
    )


def generate_title_message(pool=None, fs=None, host=None, vol=None):
    pre_message = []
    if pool:
        pre_message.append("pool '%s'" % pool.name)
    if fs:
        pre_message.append("fileSystem '%s'" % fs.name)
    if host:
        pre_message.append("host '%s'" % host.name)
    if vol:
        pre_message.append("volume '%s'" % vol.name)
    
    message = ', '.join(pre_message)
    return "in {}".format(message if message != "" else "system")


@commandgroup(prefix="metrics")
class MetricsCommands(object):
    pass

@commandgroup(parent=MetricsCommands, prefix="nas")
class NasCommands(object):
    pass


@commandgroup(parent=NasCommands, prefix="system")
class NasSystemCommands(object):

    @command(
        description="Show metrics for system",
        types=(MetricsInterval, MetricsCount),
        )
    @render(
        fields=SYSTEM_QUERY_FIELDS,
        substitute=SYSTEM_QUERY_SUBSTITUTES,
        )
    def show(self, _api, interval=1, count=5):

        message = "System '%s' NAS performance" % system.name()
        builder = ShowNasCounterBuilder(api=_api, grouping_field=None, message=message)
        builder.get_filters_params(protocol="NFS")
        builder.get_collectors_params()
        builder.build()
        run_loop(builder=builder, renderer_kwargs=SYSTEM_SHOW_RENDERER_KWARGS, interval=interval, error_interval=10, count=count)


@commandgroup(parent=NasCommands, prefix="pool")
class NasPoolCommands(object):

    ELEMENT = Pool
    ELEMENT_ID = "pool_id"

    @command(
        description="Show top metrics for pools in system",
        types=(MetricsNasSortByType, MetricsResults),
        )
    def top(self, _api, sort_by=MetricsNasSortByType.TYPES['operations'], results=10):

        message = "Top pools NAS performance in system By NAS %s" % (sort_by.upper())
        builder = TopNasCounterBuilder(api=_api, grouping_field=self.ELEMENT_ID, message=message)
        builder.get_filters_params(protocol="NFS", filter_fields=None)
        builder.get_collectors_params(sorting_field=sort_by, max_results=results,
                                      convert_data_function=partial(add_element_names_to_data, _api=_api, element_name=self.ELEMENT, element_type=self.ELEMENT_ID))
        builder.build()
        run_loop(builder=builder, renderer_kwargs=POOL_TOP_RENDERER_KWARGS, interval=1, error_interval=10)

    @command(
        description="Show metrics for specific pool in system",
        types=(Pool, MetricsInterval, MetricsCount),
        )
    def show(self, _api, pool, interval=1, count=5):

        message = "Pool '%s' NAS performance" % pool.name
        builder = ShowNasCounterBuilder(api=_api, element_type=self.ELEMENT, grouping_field=self.ELEMENT_ID, message=message)
        filters_fields = dict(pool_id=pool.id)
        builder.get_filters_params(protocol="NFS", filters_fields=filters_fields)
        builder.get_collectors_params()
        builder.build()
        run_loop(builder=builder, renderer_kwargs=SYSTEM_SHOW_RENDERER_KWARGS, interval=interval, error_interval=10, count=count)


@commandgroup(parent=NasCommands, prefix="fs")
class NasFsCommands(object):

    ELEMENT = FilesystemOrSnapshotNotWP
    ELEMENT_ID = "fs_entity_id"

    @command(
        description="Show top metrics for filesystems in system",
        types=(Pool, MetricsNasSortByType, MetricsResults),
        )
    def top(self, _api,pool=None, sort_by=MetricsNasSortByType.TYPES['operations'], results=10):

        pre_message = generate_title_message(pool=pool)
        message = "Top filesystems %s by NAS %s" % (pre_message, sort_by.upper())
        builder = TopNasCounterBuilder(api=_api, grouping_field=self.ELEMENT_ID, message=message)
        filter_fields = dict(pool_id=pool.id if pool else None)
        builder.get_filters_params(protocol="NFS", filter_fields=filter_fields)
        builder.get_collectors_params(sorting_field=sort_by, max_results=results,
                                      convert_data_function=partial(add_element_names_to_data, _api=_api, element_name=self.ELEMENT, element_type=self.ELEMENT_ID))
        builder.build()
        run_loop(builder=builder, renderer_kwargs=FS_TOP_RENDERER_KWARGS, interval=1, error_interval=10)

    @command(
        description="Show metrics for specific filesystem in system",
        types=(FilesystemOrSnapshotNotWP, MetricsInterval, MetricsCount),
        )
    def show(self, _api, fs, interval=1, count=5):

        message = "Filesystem '%s' NAS performance" % fs.name
        builder = ShowNasCounterBuilder(api=_api, element_type=self.ELEMENT, grouping_field=self.ELEMENT_ID, message=message)
        filters_fields = dict(fs_entity_id=fs.id)
        builder.get_filters_params(protocol="NFS", filters_fields=filters_fields)
        builder.get_collectors_params()
        builder.build()
        run_loop(builder=builder, renderer_kwargs=SYSTEM_SHOW_RENDERER_KWARGS, interval=interval, error_interval=10, count=count)


@commandgroup(parent=NasCommands, prefix="uid")
class NasUidCommands(object):

    ELEMENT = ''
    ELEMENT_ID = "user_id"

    @command(
        description="Show top metrics for uid in system",
        types=(Pool, FilesystemOrSnapshotNotWP, MetricsNasSortByType, MetricsResults),
        )
    def top(self, _api, pool=None, fs=None, sort_by=MetricsNasSortByType.TYPES['operations'], results=10):

        pre_message = generate_title_message(pool=pool , fs=fs)
        message = "Top UIDs %s by NAS %s" % (pre_message, sort_by.upper())
        builder = TopNasCounterBuilder(api=_api, grouping_field=self.ELEMENT_ID, message=message)
        filter_fields = dict(
            pool_id=pool.id if pool else None,
            fs_entity_id=fs.id if fs else None,
            )
        builder.get_filters_params(protocol="NFS", filter_fields=filter_fields)
        builder.get_collectors_params(sorting_field=sort_by, max_results=results, convert_data_function=None)
        builder.build()
        run_loop(builder=builder, renderer_kwargs=UID_TOP_RENDERER_KWARGS, interval=1, error_interval=10)

    @command(
        description="Show metrics for specific uid in system",
        types=(Int, FilesystemOrSnapshotNotWP, MetricsInterval, MetricsCount),
        )
    def show(self, _api, uid, fs=None, interval=1, count=5):

        pre_message = generate_title_message(fs=fs)
        message = "UID '%s' NAS performance %s" % (uid, pre_message)
        builder = ShowNasCounterBuilder(api=_api, element_type=self.ELEMENT, grouping_field=self.ELEMENT_ID, message=message)
        filters_fields = dict(
            user_id=uid,
            fs_entity_id=fs.id if fs else None,
            )
        builder.get_filters_params(protocol="NFS", filters_fields=filters_fields)
        builder.get_collectors_params()
        builder.build()
        run_loop(builder=builder, renderer_kwargs=SYSTEM_SHOW_RENDERER_KWARGS, interval=interval, error_interval=10, count=count)


@commandgroup(parent=NasCommands, prefix="gid")
class NasGidCommands(object):

    ELEMENT = ''
    ELEMENT_ID = "group_id"

    @command(
        description="Show top metrics for gid in system",
        types=(Pool, FilesystemOrSnapshotNotWP, MetricsNasSortByType, MetricsResults),
        )
    def top(self, _api, pool=None, fs=None, sort_by=MetricsNasSortByType.TYPES['operations'], results=10):

        pre_message = generate_title_message(pool=pool , fs=fs)
        message = "Top GIDs %s by NAS %s" % (pre_message, sort_by.upper())
        builder = TopNasCounterBuilder(api=_api, grouping_field=self.ELEMENT_ID, message=message)
        filter_fields = dict(
            pool_id=pool.id if pool else None,
            fs_entity_id=fs.id if fs else None,
            )
        builder.get_filters_params(protocol="NFS", filter_fields=filter_fields)
        builder.get_collectors_params(sorting_field=sort_by, max_results=results, convert_data_function=None)
        builder.build()
        run_loop(builder=builder, renderer_kwargs=GID_TOP_RENDERER_KWARGS, interval=1, error_interval=10)

    @command(
        description="Show metrics for specific gid in system",
        types=(Int, FilesystemOrSnapshotNotWP, MetricsInterval, MetricsCount),
        )
    def show(self, _api, gid, fs=None, interval=1, count=5):

        pre_message = generate_title_message(fs=fs)
        message = "GID '%s' NAS performance %s" % (gid, pre_message)
        builder = ShowNasCounterBuilder(api=_api, element_type=self.ELEMENT, grouping_field=self.ELEMENT_ID, message=message)
        filters_fields = dict(
            group_id=gid,
            fs_entity_id=fs.id if fs else None,
            )
        builder.get_filters_params(protocol="NFS", filters_fields=filters_fields)
        builder.get_collectors_params()
        builder.build()
        run_loop(builder=builder, renderer_kwargs=SYSTEM_SHOW_RENDERER_KWARGS, interval=interval, error_interval=10, count=count)


@commandgroup(parent=NasCommands, prefix="source_ip")
class NasSourceIPCommands(object):

    ELEMENT = ''
    ELEMENT_ID = "source_ip"

    @command(
        description="Show top metrics for source ips in system",
        types=(Pool, FilesystemOrSnapshotNotWP, MetricsNasSortByType, MetricsResults),
        )
    def top(self, _api, pool=None, fs=None, sort_by=MetricsNasSortByType.TYPES['operations'], results=10):

        pre_message = generate_title_message(pool=pool , fs=fs)
        message = "Top NAS source IPs %s by NAS %s" % (pre_message, sort_by.upper())
        builder = TopNasCounterBuilder(api=_api, grouping_field=self.ELEMENT_ID, message=message)
        filter_fields = dict(
            pool_id=pool.id if pool else None,
            fs_entity_id=fs.id if fs else None,
            )
        builder.get_filters_params(protocol="NFS", filter_fields=filter_fields)
        builder.get_collectors_params(sorting_field=sort_by, max_results=results,
                                      convert_data_function=partial(int_ip_to_long, element_type=self.ELEMENT_ID))
        builder.build()
        run_loop(builder=builder, renderer_kwargs=SOURCE_IP_TOP_RENDERER_KWARGS, interval=1, error_interval=10)


    @command(
        description="Show metrics for specific source ip in system",
        types=(IPAddress, Pool, FilesystemOrSnapshotNotWP, MetricsInterval, MetricsCount),
        )
    def show(self, _api, source_ip, pool=None, fs=None, interval=1, count=5):

        pre_message = generate_title_message(pool=pool , fs=fs)
        message = "Source IP '%s' NAS performance %s" % (source_ip.ip, pre_message)
        builder = ShowNasCounterBuilder(api=_api, element_type=self.ELEMENT, grouping_field=self.ELEMENT_ID, message=message)
        filters_fields = dict(
            source_ip=ip_to_reverse_int(source_ip.ip),
            pool_id=pool.id if pool else None,
            fs_entity_id=fs.id if fs else None,
            )
        builder.get_filters_params(protocol="NFS", filters_fields=filters_fields)
        builder.get_collectors_params()
        builder.build()
        run_loop(builder=builder, renderer_kwargs=SYSTEM_SHOW_RENDERER_KWARGS, interval=interval, error_interval=10, count=count)


@commandgroup(parent=NasCommands, prefix="destination_ip")
class NasDestinationIPCommands(object):

    ELEMENT = ''
    ELEMENT_ID = "destination_ip"

    @command(
        description="Show top metrics for destination ips in system",
        types=(Pool, FilesystemOrSnapshotNotWP, MetricsNasSortByType, MetricsResults),
        )
    def top(self, _api, pool=None, fs=None, sort_by=MetricsNasSortByType.TYPES['operations'], results=10):

        pre_message = generate_title_message(pool=pool , fs=fs)
        message = "Top NAS destination IPs %s by NAS %s" % (pre_message, sort_by.upper())
        builder = TopNasCounterBuilder(api=_api, grouping_field=self.ELEMENT_ID, message=message)
        filter_fields = dict(
            pool_id=pool.id if pool else None,
            fs_entity_id=fs.id if fs else None,
            )
        builder.get_filters_params(protocol="NFS", filter_fields=filter_fields)
        builder.get_collectors_params(sorting_field=sort_by, max_results=results,
                                      convert_data_function=partial(int_ip_to_long, element_type=self.ELEMENT_ID))
        builder.build()
        run_loop(builder=builder, renderer_kwargs=DESTINATION_IP_TOP_RENDERER_KWARGS, interval=1, error_interval=10)

    @command(
        description="Show metrics for specific destination ip in system",
        types=(IPAddress, Pool, FilesystemOrSnapshotNotWP, MetricsInterval, MetricsCount),
        )
    def show(self, _api, destination_ip, pool=None, fs=None, interval=1, count=5):

        pre_message = generate_title_message(pool=pool , fs=fs)
        message = "Destination IP '%s' NAS performance %s" % (destination_ip.ip, pre_message)
        builder = ShowNasCounterBuilder(api=_api, element_type=self.ELEMENT, grouping_field=self.ELEMENT_ID, message=message)
        filters_fields = dict(
            destination_ip=ip_to_reverse_int(destination_ip.ip),
            pool_id=pool.id if pool else None,
            fs_entity_id=fs.id if fs else None,
            )
        builder.get_filters_params(protocol="NFS", filters_fields=filters_fields)
        builder.get_collectors_params()
        builder.build()
        run_loop(builder=builder, renderer_kwargs=SYSTEM_SHOW_RENDERER_KWARGS, interval=interval, error_interval=10, count=count)


@commandgroup(parent=NasCommands, prefix="file")
class NasFileommands(object):

    ELEMENT = FilesystemOrSnapshotNotWP
    ELEMENT_ID = "file_path"

    @command(
        description="Show top metrics for file paths in system",
        types=(Pool, FilesystemOrSnapshotNotWP, MetricsNasSortByType, MetricsResults),
        )
    def top(self, _api, pool=None, fs=None, sort_by=MetricsNasSortByType.TYPES['operations'], results=10):

        pre_message = generate_title_message(pool=pool , fs=fs)
        message = "Top NAS file paths %s by NAS %s" % (pre_message, sort_by.upper())
        builder = TopNasCounterBuilder(api=_api, grouping_field=self.ELEMENT_ID, message=message)
        filter_fields = dict(
            pool_id=pool.id if pool else None,
            fs_entity_id=fs.id if fs else None,
            )
        builder.get_filters_params(protocol="NFS", filter_fields=filter_fields)
        builder.get_collectors_params(sorting_field=sort_by, max_results=results,
                                      convert_data_function=partial(parse_file_path, _api=_api, element_name=self.ELEMENT, element_type=self.ELEMENT_ID))
        builder.build()
        run_loop(builder=builder, renderer_kwargs=NAS_FILE_PATH_TOP_RENDERER_KWARGS, interval=1, error_interval=10)

    





































































SAN_SYSTEM_QUERY_FIELDS = strlist("metrics_collector_type ops throughput errors external_latency_wout_err internal_latency_wout_err")
SAN_SYSTEM_QUERY_SUBSTITUTES = dict(
    metrics_collector_type="---",
    ops="operations",
    external_latency_wout_err="external_latency",
    internal_latency_wout_err="internal_latency",
    )
SAN_SYSTEM_SHOW_RENDERER_KWARGS = dict(
        fields=SAN_SYSTEM_QUERY_FIELDS,
        substitute=SAN_SYSTEM_QUERY_SUBSTITUTES,
    )

INFINIDAT_SAN_SYSTEM_QUERY_FIELDS = strlist("metrics_collector_type ops throughput errors external_latency external_latency_wout_err internal_latency internal_latency_wout_err")
INFINIDAT_SAN_SYSTEM_QUERY_SUBSTITUTES = dict(
    metrics_collector_type="---",
    ops="operations",
    external_latency="external latency",
    external_latency_wout_err="external latency w/o err",
    internal_latency_wout_err="internal latency w/o err",
    )
INFINIDAT_SAN_SYSTEM_SHOW_RENDERER_KWARGS = dict(
        fields=INFINIDAT_SAN_SYSTEM_QUERY_FIELDS,
        substitute=INFINIDAT_SAN_SYSTEM_QUERY_SUBSTITUTES,
    )

SAN_POOL_QUERY_FIELDS = strlist("metrics_pool_name ops throughput external_latency_wout_err internal_latency_wout_err errors")
SAN_POOL_QUERY_SUBSTITUTES = dict(
    metrics_pool_name="pool_name",
    ops="operations",
    external_latency_wout_err="external_latency",
    internal_latency_wout_err="internal_latency",
    )
SAN_POOL_TOP_RENDERER_KWARGS = dict(
        fields=SAN_POOL_QUERY_FIELDS,
        substitute=SAN_POOL_QUERY_SUBSTITUTES,
    )

SAN_VOL_QUERY_FIELDS = strlist("metrics_volumeorsnapshotnotwp_name ops throughput external_latency_wout_err internal_latency_wout_err errors")
SAN_VOL_QUERY_SUBSTITUTES = dict(
    metrics_volumeorsnapshotnotwp_name="vol_name",
    ops="operations",
    external_latency_wout_err="external_latency",
    internal_latency_wout_err="internal_latency",
    )
SAN_VOL_TOP_RENDERER_KWARGS = dict(
        fields=SAN_VOL_QUERY_FIELDS,
        substitute=SAN_VOL_QUERY_SUBSTITUTES,
    )

SAN_FS_QUERY_FIELDS = strlist("metrics_filesystemorsnapshotnotwp_name ops throughput external_latency_wout_err internal_latency_wout_err  errors")
SAN_FS_QUERY_SUBSTITUTES = dict(
    metrics_filesystemorsnapshotnotwp_name="fs_name",
    ops="operations",
    external_latency_wout_err="external_latency",
    internal_latency_wout_err="internal_latency",
    )
SAN_FS_TOP_RENDERER_KWARGS = dict(
        fields=SAN_FS_QUERY_FIELDS,
        substitute=SAN_FS_QUERY_SUBSTITUTES,
    )

SAN_HOST_QUERY_FIELDS = strlist("metrics_host_name ops throughput external_latency_wout_err internal_latency_wout_err errors")
SAN_HOST_QUERY_SUBSTITUTES = dict(
    metrics_host_name="host_name",
    ops="operations",
    external_latency_wout_err="external_latency",
    internal_latency_wout_err="internal_latency",
    )
SAN_HOST_TOP_RENDERER_KWARGS = dict(
        fields=SAN_HOST_QUERY_FIELDS,
        substitute=SAN_HOST_QUERY_SUBSTITUTES,
    )

SAN_INITIATOR_QUERY_FIELDS = strlist("initiator_IQN initiator_WWPN ops throughput external_latency_wout_err internal_latency_wout_err errors")
SAN_INITIATOR_QUERY_SUBSTITUTES = dict(
    initiator_IQN="iqn",
    initiator_WWPN="wwpn",
    ops="operations",
    external_latency_wout_err="external_latency",
    internal_latency_wout_err="internal_latency",
    )
SAN_INITIATOR_TOP_RENDERER_KWARGS = dict(
        fields=SAN_INITIATOR_QUERY_FIELDS,
        substitute=SAN_INITIATOR_QUERY_SUBSTITUTES,
    )


ISCSI_TARGET_QUERY_FIELDS = strlist("metrics_target_port_name source_ip ops throughput latency")
ISCSI_TARGET_QUERY_SUBSTITUTES = dict(
    metrics_target_port_name="source_ip",
    ops="operations",
    )
ISCSI_TARGET_TOP_RENDERER_KWARGS = dict(
        fields=ISCSI_TARGET_QUERY_FIELDS,
        substitute=ISCSI_TARGET_QUERY_SUBSTITUTES,
    )

FC_TARGET_QUERY_FIELDS = strlist("target_WWPN ops throughput external_latency_wout_err internal_latency_wout_err errors")
FC_TARGET_QUERY_SUBSTITUTES = dict(
    metrics_target_WWPN_name="target",
    target_WWPN='target',
    ops="operations",
    external_latency_wout_err="external_latency",
    internal_latency_wout_err="internal_latency",
    )
FC_TARGET_TOP_RENDERER_KWARGS = dict(
        fields=FC_TARGET_QUERY_FIELDS,
        substitute=FC_TARGET_QUERY_SUBSTITUTES,
    )

ISCSI_DESTINATION_IP_QUERY_FIELDS = strlist(" destination_ip ops throughput external_latency_wout_err internal_latency_wout_err errors")
ISCSI_DESTINATION_IP_QUERY_SUBSTITUTES = dict(
    metrics_destination_ip_name="target_ip",
    destination_ip="target_ip",
    ops="operations",
    external_latency_wout_err="external_latency",
    internal_latency_wout_err="internal_latency",
    )
ISCSI_DESTINATION_IP_TOP_RENDERER_KWARGS = dict(
        fields=ISCSI_DESTINATION_IP_QUERY_FIELDS,
        substitute=ISCSI_DESTINATION_IP_QUERY_SUBSTITUTES,
    )

ISCSI_SOURCE_IP_QUERY_FIELDS = strlist(" source_ip ops throughput external_latency_wout_err internal_latency_wout_err errors")
ISCSI_SOURCE_IP_QUERY_SUBSTITUTES = dict(
    metrics_source_ip_name="source_ip",
    ops="operations",
    external_latency_wout_err="external_latency",
    internal_latency_wout_err="internal_latency",
    )
ISCSI_SOURCE_IP_TOP_RENDERER_KWARGS = dict(
        fields=ISCSI_SOURCE_IP_QUERY_FIELDS,
        substitute=ISCSI_SOURCE_IP_QUERY_SUBSTITUTES,
    )

SAN_SOFT_TARGET_QUERY_FIELDS = strlist(" target ops throughput external_latency_wout_err internal_latency_wout_err errors")
SAN_SOFT_TARGET_QUERY_SUBSTITUTES = dict(
    metrics_target_name="soft_target_port",
    ops="operations",
    external_latency_wout_err="external_latency",
    internal_latency_wout_err="internal_latency",
    target = "soft_target",
    )
SAN_SOFT_TARGET_TOP_RENDERER_KWARGS = dict(
        fields=SAN_SOFT_TARGET_QUERY_FIELDS,
        substitute=SAN_SOFT_TARGET_QUERY_SUBSTITUTES,
    )


SORT_OPTIONS = {v : k for k,v in MetricsSanSortByType.TYPES.items()}


@commandgroup(parent=MetricsCommands, prefix="san")
class SanCommands(object):
    pass


@commandgroup(parent=SanCommands, prefix="system")
class SanSystemCommands(object):

    @command(
        description="Show metrics for system",
        types=(MetricsInterval, MetricsCount),
        )
    def show(self, _api, interval=1, count=5):
        message = "System %s SAN performance" % system.name()
        builder = ShowSanCounterBuilder(api=_api, grouping_field=None, message=message)
        builder.get_filters_params(protocol="SAN", protocol_field='protocol_type')
        builder.get_collectors_params()
        builder.build()
        renderer_kwargs = INFINIDAT_SAN_SYSTEM_SHOW_RENDERER_KWARGS if system.user.has_role('infinidat','technician') else SAN_SYSTEM_SHOW_RENDERER_KWARGS
        run_loop(builder=builder, renderer_kwargs=renderer_kwargs, interval=interval, error_interval=10, count=count)

@commandgroup(parent=SanCommands, prefix="pool")
class SanPoolCommands(object):

    ELEMENT = Pool
    ELEMENT_ID = "pool_id"

    @command(
        description="Show top metrics for pools in system",
        types=(MetricsSanSortByType, MetricsResults),
        )
    def top(self, _api, sort_by=MetricsSanSortByType.TYPES['operations'], results=10):

        message = "Top pools SAN performance in system By SAN %s" % (SORT_OPTIONS[sort_by].upper())
        builder = TopSanCounterBuilder(api=_api, grouping_field=self.ELEMENT_ID, message=message)
        builder.get_filters_params(protocol="SAN", filter_fields=None, protocol_field='protocol_type')
        builder.get_collectors_params(sorting_field=sort_by, max_results=results,
                                      convert_data_function=partial(add_element_names_to_data, _api=_api, element_name=self.ELEMENT, element_type=self.ELEMENT_ID))
        builder.build()
        run_loop(builder=builder, renderer_kwargs=SAN_POOL_TOP_RENDERER_KWARGS, interval=1, error_interval=10)


    @command(
        description="Show metrics for specific pool in system",
        types=(Pool, MetricsInterval, MetricsCount),
        )
    def show(self, _api, pool, interval=1, count=5):
        message = "Pool '%s' SAN performance" % pool.name
        builder = ShowSanCounterBuilder(api=_api, element_type=self.ELEMENT, grouping_field=self.ELEMENT_ID, message=message)
        filters_fields = dict(pool_id=pool.id)
        builder.get_filters_params(protocol="SAN", filters_fields=filters_fields, protocol_field='protocol_type')
        builder.get_collectors_params()
        builder.build()
        renderer_kwargs = INFINIDAT_SAN_SYSTEM_SHOW_RENDERER_KWARGS if system.user.has_role('infinidat','technician') else SAN_SYSTEM_SHOW_RENDERER_KWARGS
        run_loop(builder=builder, renderer_kwargs=renderer_kwargs, interval=interval, error_interval=10, count=count)


@commandgroup(parent=SanCommands, prefix="vol")
class SanVolCommands(object):

    ELEMENT = VolumeOrSnapshotNotWP
    ELEMENT_ID = "vol_id"

    @command(
        description="Show top metrics for volumes in system",
        types=(Pool, MetricsSanSortByType, MetricsResults),
        )
    def top(self, _api,pool=None, sort_by=MetricsSanSortByType.TYPES['operations'], results=10):

        pre_message = generate_title_message(pool=pool)
        message = "Top volumes %s by SAN %s" % (pre_message, SORT_OPTIONS[sort_by].upper())
        builder = TopSanCounterBuilder(api=_api, grouping_field=self.ELEMENT_ID, message=message)
        filter_fields = dict(pool_id=pool.id if pool else None)
        builder.get_filters_params(protocol="SAN", filter_fields=filter_fields, protocol_field='protocol_type')
        builder.get_collectors_params(sorting_field=sort_by, max_results=results,
                                      convert_data_function=partial(add_element_names_to_data, _api=_api, element_name=self.ELEMENT, element_type=self.ELEMENT_ID))
        builder.build()
        run_loop(builder=builder, renderer_kwargs=SAN_VOL_TOP_RENDERER_KWARGS, interval=1, error_interval=10)

    @command(
        description="Show metrics for specific volume in system",
        types=(VolumeOrSnapshotNotWP, MetricsInterval, MetricsCount),
        )
    def show(self, _api, vol, interval=1, count=5):
        message = "Volume '%s' SAN Performance" % vol.name
        builder = ShowSanCounterBuilder(api=_api, element_type=self.ELEMENT, grouping_field=self.ELEMENT_ID, message=message)
        filters_fields = dict(vol_id=vol.id)
        builder.get_filters_params(protocol="SAN", filters_fields=filters_fields, protocol_field='protocol_type')
        builder.get_collectors_params()
        builder.build()
        renderer_kwargs = INFINIDAT_SAN_SYSTEM_SHOW_RENDERER_KWARGS if system.user.has_role('infinidat','technician') else SAN_SYSTEM_SHOW_RENDERER_KWARGS
        run_loop(builder=builder, renderer_kwargs=renderer_kwargs, interval=interval, error_interval=10, count=count)


@commandgroup(parent=SanCommands, prefix="host")
class SanHostCommands(object):

    ELEMENT = Host
    ELEMENT_ID = "host_id"

    @command(
        description="Show top metrics for hosts in system",
        types=(Pool, MetricsSanSortByType, MetricsResults),
        )
    def top(self, _api,pool=None, sort_by=MetricsSanSortByType.TYPES['operations'], results=10):

        pre_message = generate_title_message(pool=pool)
        message = "Top hosts %s by SAN %s" % (pre_message, SORT_OPTIONS[sort_by].upper())
        builder = TopSanCounterBuilder(api=_api, grouping_field=self.ELEMENT_ID, message=message)
        filter_fields = dict(pool_id=pool.id if pool else None)
        builder.get_filters_params(protocol="SAN", filter_fields=filter_fields, protocol_field='protocol_type')
        builder.get_collectors_params(sorting_field=sort_by, max_results=results,
                                      convert_data_function=partial(add_element_names_to_data, _api=_api, element_name=self.ELEMENT, element_type=self.ELEMENT_ID))
        builder.build()
        run_loop(builder=builder, renderer_kwargs=SAN_HOST_TOP_RENDERER_KWARGS, interval=1, error_interval=10)

    @command(
        description="Show metrics for specific host in system",
        types=(Host, MetricsInterval, MetricsCount),
        )
    def show(self, _api, host, interval=1, count=5):
        message = "Host '%s' SAN Performance" % host.name
        builder = ShowSanCounterBuilder(api=_api, element_type=self.ELEMENT, grouping_field=self.ELEMENT_ID, message=message)
        filters_fields = dict(host_id=host.id)
        builder.get_filters_params(protocol="SAN", filters_fields=filters_fields, protocol_field='protocol_type')
        builder.get_collectors_params()
        builder.build()
        renderer_kwargs = INFINIDAT_SAN_SYSTEM_SHOW_RENDERER_KWARGS if system.user.has_role('infinidat','technician') else SAN_SYSTEM_SHOW_RENDERER_KWARGS
        run_loop(builder=builder, renderer_kwargs=renderer_kwargs, interval=interval, error_interval=10, count=count)


# @commandgroup(parent=SanCommands, prefix="soft_target")
# class SanInitiatorCommands(object):

#     ELEMENT = ''
#     ELEMENT_ID = "target"

#     @command(
#         description="Show top metrics for soft_targets in system",
#         types=(Pool, VolumeOrSnapshotNotWP, MetricsSanSortByType, MetricsResults),
#         )
#     def top(self, _api, pool=None, vol=None, sort_by=MetricsSanSortByType.TYPES['operations'], results=10):


#         pre_message = generate_title_message(pool=pool , vol=vol)
#         message = "Top targets %s by %s" % (pre_message, SORT_OPTIONS[sort_by].upper())
#         builder = TopSanCounterBuilder(api=_api, grouping_field=self.ELEMENT_ID, message=message)
#         filter_fields = dict(
#             pool_id=pool.id if pool else None,
#             vol_id=vol.id if vol else None,
#             )
#         builder.get_filters_params(protocol="SAN", filter_fields=filter_fields, protocol_field='protocol_type')
#         builder.get_collectors_params(sorting_field=sort_by, max_results=results,
#                                       convert_data_function=partial(decimal_to_hex, element_name=self.ELEMENT_ID, element_type=self.ELEMENT_ID))
#         builder.build()
#
#         run_loop(builder=builder, renderer_kwargs=SAN_SOFT_TARGET_TOP_RENDERER_KWARGS, interval=1, error_interval=10)

#     @command(
#         description="Show metrics for specific soft_target in system",
#         types=(FCSoftTarget, MetricsInterval, MetricsCount),
#         )
#     def show(self, _api, soft_port, interval=1, count=5):

#         target = soft_port
#         target.address = hex_to_decimal(target.address.replace(":",""))
#         message = "Initiator %s SAN Performance" % (soft_target.address)
#         builder = ShowSanCounterBuilder(api=_api, element_type=self.ELEMENT, grouping_field=self.ELEMENT_ID, message=message)
#         filters_fields = dict(
#             target=target.address,
#             )
#         builder.get_filters_params(protocol="SAN", filters_fields=filters_fields, protocol_field='protocol_type')
#         builder.get_collectors_params()
#         builder.build()
#         renderer_kwargs = INFINIDAT_SAN_SYSTEM_SHOW_RENDERER_KWARGS if system.user.has_role('infinidat','technician') else SAN_SYSTEM_SHOW_RENDERER_KWARGS
#         run_loop(builder=builder, renderer_kwargs=renderer_kwargs, interval=interval, error_interval=10, count=count)


@commandgroup(parent=SanCommands, prefix="iscsi")
class SaniSCSICommands(object):

    ELEMENT = ''
    ELEMENT_ID = ""

    @command(
        description="Show metrics for iSCSI in system",
        types=(Pool, VolumeOrSnapshotNotWP, MetricsInterval, MetricsCount),
        )
    def show(self, _api, pool=None, vol=None, interval=1, count=5):
        pre_message = generate_title_message(pool=pool , vol=vol)
        message = "iSCSI SAN Performance %s" % (pre_message)
        builder = ShowSanCounterBuilder(api=_api, element_type=self.ELEMENT, grouping_field=self.ELEMENT_ID, message=message)
        filters_fields = dict(
            pool_id=pool.id if pool else None,
            vol_id=vol.id if vol else None,
            )
        builder.get_filters_params(protocol="iSCSI", filters_fields=filters_fields)
        builder.get_collectors_params()
        builder.build()
        renderer_kwargs = INFINIDAT_SAN_SYSTEM_SHOW_RENDERER_KWARGS if system.user.has_role('infinidat','technician') else SAN_SYSTEM_SHOW_RENDERER_KWARGS
        run_loop(builder=builder, renderer_kwargs=renderer_kwargs, interval=interval, error_interval=10, count=count)


@commandgroup(parent=SanCommands, prefix="iscsi_initiator")
class SanInitiatorCommands(object):

    ELEMENT = ''
    ELEMENT_ID = "initiator_IQN"

    @command(
        description="Show top metrics for iSCSI initiators in system",
        types=(Pool, Host, MetricsSanSortByType, MetricsResults),
        )
    def top(self, _api, pool=None, host=None, sort_by=MetricsSanSortByType.TYPES['operations'], results=10):

        pre_message = generate_title_message(pool=pool , host=host)
        message = "Top iSCSI initiators %s by SAN %s" % (pre_message, SORT_OPTIONS[sort_by].upper())
        builder = TopSanCounterBuilder(api=_api, grouping_field=self.ELEMENT_ID, message=message)
        filter_fields = dict(
            pool_id=pool.id if pool else None,
            host_id=host.id if host else None,
            )
        builder.get_filters_params(protocol="iSCSI", filter_fields=filter_fields)
        builder.get_collectors_params(sorting_field=sort_by, max_results=results)
        builder.build()
        run_loop(builder=builder, renderer_kwargs=SAN_INITIATOR_TOP_RENDERER_KWARGS, interval=1, error_interval=10)

    @command(
        description="Show metrics for specific iSCSI initiator in system",
        types=(UsediSCSILoggedInInitiator, MetricsInterval, MetricsCount),
        )
    def show(self, _api, initiator, interval=1, count=5):
        message = "iSCSI initiator '%s' SAN Performance" % (initiator.address)
        builder = ShowSanCounterBuilder(api=_api, element_type=self.ELEMENT, grouping_field=self.ELEMENT_ID, message=message)
        filters_fields = dict(
            initiator_IQN=initiator.address,
            )
        builder.get_filters_params(protocol="iSCSI", filters_fields=filters_fields)
        builder.get_collectors_params()
        builder.build()
        renderer_kwargs = INFINIDAT_SAN_SYSTEM_SHOW_RENDERER_KWARGS if system.user.has_role('infinidat','technician') else SAN_SYSTEM_SHOW_RENDERER_KWARGS
        run_loop(builder=builder, renderer_kwargs=renderer_kwargs, interval=interval, error_interval=10, count=count)


@commandgroup(parent=SanCommands, prefix="iscsi_target_ip")
class SanInitiatorCommands(object):

    ELEMENT = ''
    ELEMENT_ID = "destination_ip"

    @command(
        description="Show top metrics for iscsi target IPs in system",
        types=(Pool, VolumeOrSnapshotNotWP, Host, MetricsSanSortByType, MetricsResults),
        )
    def top(self, _api, pool=None, vol=None, host=None, sort_by=MetricsSanSortByType.TYPES['operations'], results=10):


        pre_message = generate_title_message(pool=pool , vol=vol, host=host)
        message = "Top iSCSI target IPs %s by SAN %s" % (pre_message, SORT_OPTIONS[sort_by].upper())
        builder = TopSanCounterBuilder(api=_api, grouping_field=self.ELEMENT_ID, message=message)
        filter_fields = dict(
            pool_id=pool.id if pool else None,
            vol_id=vol.id if vol else None,
            host_id=host.id if host else None,
            )
        builder.get_filters_params(protocol="iSCSI", filter_fields=filter_fields)
        builder.get_collectors_params(sorting_field=sort_by, max_results=results,
                                    convert_data_function=partial(int_ip_to_long, element_type=self.ELEMENT_ID))
        builder.build()
        run_loop(builder=builder, renderer_kwargs=ISCSI_DESTINATION_IP_TOP_RENDERER_KWARGS, interval=1, error_interval=10)

    @command(
        description="Show metrics for specific iscsi target IP in system",
        types=(IPAddress, MetricsInterval, MetricsCount),
        )
    def show(self, _api, ip_address, interval=1, count=5):
        message = "iSCSI target IP '%s' SAN Performance" % (ip_address.ip)
        builder = ShowSanCounterBuilder(api=_api, element_type=self.ELEMENT, grouping_field=self.ELEMENT_ID, message=message)
        filters_fields = dict(
            destination_ip=ip_to_reverse_int(ip_address.ip),
            )
        builder.get_filters_params(protocol="iSCSI", filters_fields=filters_fields)
        builder.get_collectors_params()
        builder.build()
        renderer_kwargs = INFINIDAT_SAN_SYSTEM_SHOW_RENDERER_KWARGS if system.user.has_role('infinidat','technician') else SAN_SYSTEM_SHOW_RENDERER_KWARGS
        run_loop(builder=builder, renderer_kwargs=renderer_kwargs, interval=interval, error_interval=10, count=count)

@commandgroup(parent=SanCommands, prefix="iscsi_initiator_ip")
class SanInitiatorCommands(object):

    ELEMENT = ''
    ELEMENT_ID = "source_ip"

    @command(
        description="Show top metrics for iscsi initiator IPs in system",
        types=(Pool, VolumeOrSnapshotNotWP, Host, MetricsSanSortByType, MetricsResults),
        )
    def top(self, _api, pool=None, vol=None, host=None, sort_by=MetricsSanSortByType.TYPES['operations'], results=10):


        pre_message = generate_title_message(pool=pool , vol=vol, host=host)
        message = "Top iSCSI initiator IPs %s by SAN %s" % (pre_message, SORT_OPTIONS[sort_by].upper())
        builder = TopSanCounterBuilder(api=_api, grouping_field=self.ELEMENT_ID, message=message)
        filter_fields = dict(
            pool_id=pool.id if pool else None,
            vol_id=vol.id if vol else None,
            host_id=host.id if host else None,
            )
        builder.get_filters_params(protocol="iSCSI", filter_fields=filter_fields)
        builder.get_collectors_params(sorting_field=sort_by, max_results=results,
                                    convert_data_function=partial(int_ip_to_long, element_type=self.ELEMENT_ID))
        builder.build()
        run_loop(builder=builder, renderer_kwargs=ISCSI_SOURCE_IP_TOP_RENDERER_KWARGS, interval=1, error_interval=10)

    @command(
        description="Show metrics for specific iscsi initiator IP in system",
        types=(IPAddress, MetricsInterval, MetricsCount),
        )
    def show(self, _api, ip_address, interval=1, count=5):
        message = "iSCSI initiator IP '%s' SAN Performance" % (ip_address.ip)
        builder = ShowSanCounterBuilder(api=_api, element_type=self.ELEMENT, grouping_field=self.ELEMENT_ID, message=message)
        filters_fields = dict(
            source_ip=ip_to_reverse_int(ip_address.ip),
            )
        builder.get_filters_params(protocol="iSCSI", filters_fields=filters_fields)
        builder.get_collectors_params()
        builder.build()
        renderer_kwargs = INFINIDAT_SAN_SYSTEM_SHOW_RENDERER_KWARGS if system.user.has_role('infinidat','technician') else SAN_SYSTEM_SHOW_RENDERER_KWARGS
        run_loop(builder=builder, renderer_kwargs=renderer_kwargs, interval=interval, error_interval=10, count=count)

@commandgroup(parent=SanCommands, prefix="fc")
class SaniSCSICommands(object):

    ELEMENT = ''
    ELEMENT_ID = ""

    @command(
        description="Show metrics for FC in system",
        types=(Pool, VolumeOrSnapshotNotWP, MetricsInterval, MetricsCount),
        )
    def show(self, _api, pool=None, vol=None, interval=1, count=5):
        pre_message = generate_title_message(pool=pool , vol=vol)
        message = "FC SAN Performance %s" % (pre_message)
        builder = ShowSanCounterBuilder(api=_api, element_type=self.ELEMENT, grouping_field=self.ELEMENT_ID, message=message)
        filters_fields = dict(
            pool_id=pool.id if pool else None,
            vol_id=vol.id if vol else None,
            )
        builder.get_filters_params(protocol="FC", filters_fields=filters_fields)
        builder.get_collectors_params()
        builder.build()
        renderer_kwargs = INFINIDAT_SAN_SYSTEM_SHOW_RENDERER_KWARGS if system.user.has_role('infinidat','technician') else SAN_SYSTEM_SHOW_RENDERER_KWARGS
        run_loop(builder=builder, renderer_kwargs=renderer_kwargs, interval=interval, error_interval=10, count=count)


@commandgroup(parent=SanCommands, prefix="fc_initiator")
class SanInitiatorCommands(object):

    ELEMENT = ''
    ELEMENT_ID = "initiator_WWPN"

    @command(
        description="Show top metrics for FC initiators in system",
        types=(Pool, Host, MetricsSanSortByType, MetricsResults),
        )
    def top(self, _api, pool=None, host=None, sort_by=MetricsSanSortByType.TYPES['operations'], results=10):

        pre_message = generate_title_message(pool=pool , host=host)
        message = "Top FC initiators %s by SAN %s" % (pre_message, SORT_OPTIONS[sort_by].upper())
        builder = TopSanCounterBuilder(api=_api, grouping_field=self.ELEMENT_ID, message=message)
        filter_fields = dict(
            pool_id=pool.id if pool else None,
            host_id=host.id if host else None,
            )
        builder.get_filters_params(protocol="FC", filter_fields=filter_fields)
        builder.get_collectors_params(sorting_field=sort_by, max_results=results, 
                                      convert_data_function=initiator_decimal_to_address)
        builder.build()
        run_loop(builder=builder, renderer_kwargs=SAN_INITIATOR_TOP_RENDERER_KWARGS, interval=1, error_interval=10)

    @command(
        description="Show metrics for specific FC initiator in system",
        types=(UsedFCLoggedInInitiator, MetricsInterval, MetricsCount),
        )
    def show(self, _api, initiator, interval=1, count=5):
        message = "FC initiator '%s' SAN Performance" % (initiator.address)
        builder = ShowSanCounterBuilder(api=_api, element_type=self.ELEMENT, grouping_field=self.ELEMENT_ID, message=message)
        filters_fields = dict(
            initiator_WWPN=str(initiator.port_key),
            )
        builder.get_filters_params(protocol="FC", filters_fields=filters_fields)
        builder.get_collectors_params()
        builder.build()
        renderer_kwargs = INFINIDAT_SAN_SYSTEM_SHOW_RENDERER_KWARGS if system.user.has_role('infinidat','technician') else SAN_SYSTEM_SHOW_RENDERER_KWARGS
        run_loop(builder=builder, renderer_kwargs=renderer_kwargs, interval=interval, error_interval=10, count=count)


@commandgroup(parent=SanCommands, prefix="fc_target")
class SanInitiatorCommands(object):

    ELEMENT = ''
    ELEMENT_ID = "target_WWPN"

    @command(
        description="Show top metrics for fc targets in system",
        types=(Pool, VolumeOrSnapshotNotWP, Host, MetricsSanSortByType, MetricsResults),
        )
    def top(self, _api, pool=None, vol=None, host=None, sort_by=MetricsSanSortByType.TYPES['operations'], results=10):


        pre_message = generate_title_message(pool=pool , vol=vol, host=host)
        message = "Top FC targets %s by SAN %s" % (pre_message, SORT_OPTIONS[sort_by].upper())
        builder = TopSanCounterBuilder(api=_api, grouping_field=self.ELEMENT_ID, message=message)
        filter_fields = dict(
            pool_id=pool.id if pool else None,
            vol_id=vol.id if vol else None,
            host_id=host.id if host else None,
            )
        builder.get_filters_params(protocol="FC", filter_fields=filter_fields)
        builder.get_collectors_params(sorting_field=sort_by, max_results=results,
                                    convert_data_function=partial(target_decimal_to_address, _api))
        builder.build()
        run_loop(builder=builder, renderer_kwargs=FC_TARGET_TOP_RENDERER_KWARGS, interval=1, error_interval=10)

    @command(
        description="Show metrics for specific port in system",
        types=(FCHardPort, MetricsInterval, MetricsCount),
        )
    def show(self, _api, port, interval=1, count=5):
        target = port
        target.wwpn = hex_to_decimal(target.wwpn.replace(":",""))
        message = "Port '%s' SAN Performance" % (hex_to_colon_seperated_address('%x' % int(port.wwpn)))
        builder = ShowSanCounterBuilder(api=_api, element_type=self.ELEMENT, grouping_field=self.ELEMENT_ID, message=message)
        filters_fields = dict(
            target_WWPN=str(target.wwpn),
            )
        builder.get_filters_params(protocol="FC", filters_fields=filters_fields)
        builder.get_collectors_params()
        builder.build()
        renderer_kwargs = INFINIDAT_SAN_SYSTEM_SHOW_RENDERER_KWARGS if system.user.has_role('infinidat','technician') else SAN_SYSTEM_SHOW_RENDERER_KWARGS
        run_loop(builder=builder, renderer_kwargs=renderer_kwargs, interval=interval, error_interval=10, count=count)




















ASYNC_VOL_QUERY_FIELDS = strlist("metrics_datasetorsnapshot_name ops throughput external_latency_wout_err errors")
ASYNC_VOL_QUERY_SUBSTITUTES = dict(
    metrics_datasetorsnapshot_name="vol_name",
    ops="operations",
    external_latency_wout_err="latency",
    )
ASYNC_VOL_TOP_RENDERER_KWARGS = dict(
        fields=ASYNC_VOL_QUERY_FIELDS,
        substitute=ASYNC_VOL_QUERY_SUBSTITUTES,
    )

ASYNC_FS_QUERY_FIELDS = strlist("metrics_datasetorsnapshot_name ops throughput external_latency_wout_err errors")
ASYNC_FS_QUERY_SUBSTITUTES = dict(
    metrics_datasetorsnapshot_name="fs_name",
    ops="operations",
    external_latency_wout_err="latency",
    )
ASYNC_FS_TOP_RENDERER_KWARGS = dict(
        fields=ASYNC_FS_QUERY_FIELDS,
        substitute=ASYNC_FS_QUERY_SUBSTITUTES,
    )


@commandgroup(parent=MetricsCommands, prefix="async")
class AsyncCommands(object):
    pass


@commandgroup(parent=AsyncCommands, prefix="system")
class AsyncSystemCommands(object):


    @command(
        description="Show metrics for system",
        types=(MetricsInterval, MetricsCount),
        )
    @render(
        fields=SYSTEM_QUERY_FIELDS,
        substitute=SYSTEM_QUERY_SUBSTITUTES,
        )
    def show(self, _api, interval=1, count=5):

        message = "System '%s' ASYNC performance" % system.name()
        builder = ShowAsyncCounterBuilder(api=_api, grouping_field=None, message=message)
        builder.get_filters_params(protocol="RMR", protocol_field='protocol_type')
        builder.get_collectors_params()
        builder.build()
        run_loop(builder=builder, renderer_kwargs=SYSTEM_SHOW_RENDERER_KWARGS, interval=interval, error_interval=10, count=count)


@commandgroup(parent=AsyncCommands, prefix="vol")
class AsyncVolCommands(object):

    ELEMENT = VolumeOrSnapshotNotWP
    ELEMENT_ID = "dataset_id"

    @command(
        description="Show top ASYNC metrics for replicating volumes in system",
        types=(MetricsResults,),
        )
    def top(self, _api, results=10):

        sort_by = MetricsSanSortByType.TYPES['throughput']
        pre_message = generate_title_message()
        message = "Top replicating volumes %s by ASYNC %s" % (pre_message, sort_by)
        builder = TopAsyncCounterBuilder(api=_api, grouping_field=self.ELEMENT_ID, message=message)
        builder.get_filters_params(protocol="RMR", protocol_field='protocol_type')
        result_filter = {'attribute':'dataset_type', 'value':'VOLUME'}
        builder.get_collectors_params(sorting_field=sort_by, max_results=results,
                                      convert_data_function=partial(add_element_names_to_data, _api=_api, element_name=DatasetOrSnapshot, element_type=self.ELEMENT_ID, filter=result_filter))
        builder.build()
        run_loop(builder=builder, renderer_kwargs=ASYNC_VOL_TOP_RENDERER_KWARGS, interval=1, error_interval=10)

    @command(
        description="Show metrics for specific replicating volume in system",
        types=(VolumeOrSnapshotNotWP, MetricsInterval, MetricsCount),
        )
    def show(self, _api, vol, interval=1, count=5):

        message = "Volume '%s' ASYNC Performance" % vol.name
        builder = ShowAsyncCounterBuilder(api=_api, element_type=self.ELEMENT, grouping_field=self.ELEMENT_ID, message=message)
        filters_fields = dict(dataset_id=vol.id)
        builder.get_filters_params(protocol="RMR", filters_fields=filters_fields, protocol_field='protocol_type')
        builder.get_collectors_params()
        builder.build()
        run_loop(builder=builder, renderer_kwargs=SAN_SYSTEM_SHOW_RENDERER_KWARGS, interval=interval, error_interval=10, count=count)


@commandgroup(parent=AsyncCommands, prefix="fs")
class AsyncFsCommands(object):

    ELEMENT = FilesystemOrSnapshotNotWP
    ELEMENT_ID = "dataset_id"

    @command(
        description="Show top ASYNC metrics for replicating filesystems in system",
        types=(MetricsResults,),
        )
    def top(self, _api, results=10):

        sort_by = MetricsSanSortByType.TYPES['throughput']
        pre_message = generate_title_message()
        message = "Top replicating filesystems %s by ASYNC %s" % (pre_message, sort_by)
        builder = TopAsyncCounterBuilder(api=_api, grouping_field=self.ELEMENT_ID, message=message)
        builder.get_filters_params(protocol="RMR", protocol_field='protocol_type')
        result_filter = {'attribute':'dataset_type', 'value':'FILESYSTEM'}
        builder.get_collectors_params(sorting_field=sort_by, max_results=results,
                                      convert_data_function=partial(add_element_names_to_data, _api=_api, element_name=DatasetOrSnapshot, element_type=self.ELEMENT_ID, filter=result_filter))
        builder.build()
        run_loop(builder=builder, renderer_kwargs=ASYNC_FS_TOP_RENDERER_KWARGS, interval=1, error_interval=10)

    @command(
        description="Show metrics for specific replicating filesystem in system",
        types=(FilesystemOrSnapshotNotWP, MetricsInterval, MetricsCount),
        )
    def show(self, _api, fs, interval=1, count=5):

        message = "Filesystem '%s' ASYNC Performance" % fs.name
        builder = ShowAsyncCounterBuilder(api=_api, element_type=self.ELEMENT, grouping_field=self.ELEMENT_ID, message=message)
        filters_fields = dict(dataset_id=fs.id)
        builder.get_filters_params(protocol="RMR", filters_fields=filters_fields, protocol_field='protocol_type')
        builder.get_collectors_params()
        builder.build()
        run_loop(builder=builder, renderer_kwargs=SAN_SYSTEM_SHOW_RENDERER_KWARGS, interval=interval, error_interval=10, count=count)
