from munch import Munch
from functools import partial
from functools32 import lru_cache
import capacity

from infinishell.infishell import commandgroup, command
from infinishell.infishell.utils import strlist
from infinishell.infishell.exceptions import BadArguments, BadValue
from infinishell.infishell.types import Str, YesNo, Int, Positive

from .types import QOS, QOSPolicy, DatasetType, Name, QOSAssociatedEntity, Pool, DatasetOrSnapshot, Float, MaxMiBpsOrNone, MaxMiBps,QosPolicyType, QosOpsOrNone, QosOps, Burst, BurstFactor, NONE_CONSTANT, VolumeOrSnapshot, FilesystemOrSnapshot, DatasetQosOps, PoolQosOps, DatasetMaxMiBps, PoolMaxMiBps
from .options import QueryOptions
from ..features import feature_command
from ..system import system
from . import render_collection, roles, Null, EntityCommands, LazyIdNameMapping
from .config import ConfigCommandsBase


POLICY_QUERY_FIELDS = strlist("name type max_ops max_bps burst_factor entities")
POLICY_QUERY_SUBSTITUTES = dict(
    max_bps="max_throughput",
    )
POLICY_SORTABLE_FIELDS = strlist("name type")

ENTITY_QUERY_FIELDS = strlist("entity_name type  policy entities")
ENTITY_QUERY_SUBSTITUTES = dict(
    entity_name='name',
    entities="policed_entities",
    )

ENTITY_SORTABLE_FIELDS = strlist("entity_name entity_type policy")


@commandgroup(parent=ConfigCommandsBase, prefix="qos_policy")
class PolicyCommands(EntityCommands, ConfigCommandsBase):
    ENTITY_TYPE = QOSPolicy

    @feature_command(
        feature='qos',
        status=True,
        description="List existing policies",
        types=(QOSPolicy, QosPolicyType),
        options=QueryOptions(fields=POLICY_QUERY_FIELDS, sortable=POLICY_SORTABLE_FIELDS, substitute=POLICY_QUERY_SUBSTITUTES, detailed=True),
        hidden_substitue=dict(name="policy"),
        substitute=dict(),
        )
    @render_collection(
        entity=QOSPolicy,
        fields=POLICY_QUERY_FIELDS,
        detailed_fields=POLICY_QUERY_FIELDS,
        detailed_substitute=POLICY_QUERY_SUBSTITUTES,
        detailed_individually=True,
        substitute=POLICY_QUERY_SUBSTITUTES,
        value_properties=dict(
            capacity_unit=(capacity.MiB,'MiB'),
            ),
        alignment=dict(
            burst_factor="right",
            ),
        )
    def query(self, _api, policy=None, type=None):
        return self._query(_api, policy, type=type, _hook=self._query_hook)

    def _query_hook(self, policies):
        for p in policies:
            p.burst = "x{} , {} seconds".format(p.burst_factor, p.burst_duration_seconds) if p.burst_enabled else "disabled"
            p.max_ops = p.max_ops or "UNLIMITED"
            p.max_bps = p.max_bps or "UNLIMITED"
            p.burst_factor = p.burst_factor
            # p.entities = system.api.paged_query("qos/policies/{}/assigned_entities".format(p.id, page_size=1)).total_objects
        return policies

    def _verify_ops_and_throughput(self, qos_type, ops, throughput):
        if not(ops or throughput) or not qos_type:
            return
        max_ops , min_ops, max_throughput, min_throughput = 0, 0, 0, 0
        if qos_type in ("VOLUME", "FILESYSTEM"):
            min_ops = DatasetQosOps.MIN or DatasetQosOps._get_min()
            max_ops = DatasetQosOps.MAX or DatasetQosOps._get_max()
            min_throughput = DatasetMaxMiBps.MIN or DatasetMaxMiBps._get_min()
            max_throughput = DatasetMaxMiBps.MAX or DatasetMaxMiBps._get_max()
        else:
            min_ops = PoolQosOps.MIN or PoolQosOps._get_min()
            max_ops = PoolQosOps.MAX or PoolQosOps._get_max()
            min_throughput = PoolMaxMiBps.MIN or PoolMaxMiBps._get_min()
            max_throughput = PoolMaxMiBps.MAX or PoolMaxMiBps._get_max()
        if ops and ops != NONE_CONSTANT:
            if not (ops <= max_ops and ops >= min_ops):
                raise BadArguments, "max_ops value ({} ops) is out of range.(Min: {} ops - Max: {} ops)".format(ops, min_ops, max_ops)
        if throughput and throughput != NONE_CONSTANT:
            if not (throughput <= max_throughput and throughput >= min_throughput ):
                raise BadArguments, "max_throughput value ({0:MiB} MiB) is out of range.(Min: {1:MiB} MiB - Max: {2:MiB} MiB)".format(throughput * capacity.byte, min_throughput * capacity.byte, max_throughput * capacity.byte)
        return



    @feature_command(
        feature='qos',
        status=True,
        description="Create a new QoS policy",
        types=(Name, QosPolicyType, QosOpsOrNone, MaxMiBpsOrNone, Burst, BurstFactor),
        )
    @roles.INFINIDAT.ADMIN
    def create(self, _api, name, type, max_ops=None, max_throughput=None, burst=None, burst_factor=None):
        if not max_ops and not max_throughput:
            raise BadArguments, "max_ops or max_throughput parameter must be used"
        if not burst and (burst_factor):
            raise BadArguments, "burst_factor can be used only with burst parameter enabled"
        if burst and not (burst_factor):
            raise BadArguments, "burst_factor must be used when burst parameter enabled"
        self._verify_ops_and_throughput(type, max_ops, max_throughput)
        data = dict(
            type=type,
            max_ops=None if max_ops == NONE_CONSTANT else max_ops,
            max_bps=None if max_throughput == NONE_CONSTANT else max_throughput,
            burst_enabled=burst,
            burst_factor=burst_factor,
            burst_duration_seconds=60 if burst else None,
            )
        return self._create(_api, name=name, **data)

    @feature_command(
        feature='qos',
        status=True,
        description="Delete a policy",
        types=(QOSPolicy,),
        )
    @roles.INFINIDAT.ADMIN
    def delete(self, _api, policy):
        return self._delete(_api, policy)

    @feature_command(
        feature='qos',
        status=True,
        description="Modify a policy",
        types=(QOSPolicy, Name, QosOpsOrNone, MaxMiBpsOrNone, Burst, BurstFactor),
        )
    @roles.INFINIDAT.ADMIN
    def modify(self, _api, policy, new_name=None, max_ops=None, max_throughput=None, burst=None, burst_factor=None):
        if (not burst and not policy.burst_enabled) and (burst_factor):
            raise BadArguments, "burst_factor can be used only when burst parameter enabled"
        if burst and not (burst_factor):
            raise BadArguments, "burst_factor must be used when burst parameter enabled"
        self._verify_ops_and_throughput(policy.type, max_ops, max_throughput)
        data = dict(
            name=new_name,
            max_ops=Null if max_ops == NONE_CONSTANT else max_ops,
            max_bps=Null if max_throughput == NONE_CONSTANT else max_throughput,
            burst_enabled=burst,
            burst_factor=burst_factor if burst_factor else Null if burst is False else None,
            burst_duration_seconds=60 if burst or burst_factor else Null if burst is False else None,
            )
        self._update(_api, policy, _multi_field_put=True, **data)
        return "QoS policy '{}' was updated".format(new_name if new_name else policy.name)


    @lru_cache(maxsize=1000)
    def get_policy(self, qos_policy_id):
        return QOSPolicy.get_by_id(system.api, qos_policy_id)

    @feature_command(
        feature='qos',
        status=True,
        description="List assosiated entities",
        types=(QOSPolicy, QosPolicyType),
        options=QueryOptions(fields=ENTITY_QUERY_FIELDS, substitute=ENTITY_QUERY_SUBSTITUTES, detailed=True),
        substitute=dict(),
        )
    @render_collection(
        entity=QOSPolicy,
        fields=ENTITY_QUERY_FIELDS,
        detailed_fields=ENTITY_QUERY_FIELDS,
        detailed_substitute=ENTITY_QUERY_SUBSTITUTES,
        detailed_individually=True,
        substitute=ENTITY_QUERY_SUBSTITUTES,
        )
    def assignment_query(self, _api, policy=None, type=None):
        if policy and type:
            raise BadArguments, "policy and type are mutually exclusive"
        if policy:
            return _api.paged_query("qos/policies/{}/assigned_entities".format(policy.id), hook=partial(self._entity_query_hook, api=_api, type=type, policy=policy))
        entity_type = type if type not in ['POOL_VOLUME', 'POOL_FILESYSTEM'] else 'POOL'
        return self._query(_api, None, _entity_type=QOSAssociatedEntity, entity_type=entity_type,
                           _hook=partial(self._entity_query_hook, api=_api, type=type, policy=policy))

    def _entity_query_hook(self, entities, api=None, type=None, policy=None):
        results = []
        self.get_policy.cache_clear()
        for e in entities:
            e_policy = policy if policy else self.get_policy(e.qos_policy_id)
            e.policy = e_policy.name 
            if type and e_policy.type != type:
                continue
            e.type = type if type else e_policy.type
            e.entities = None
            if e.type.lower() in ['pool_volume', 'pool_filesystem']:
                pool = Pool.get_by_id(api, e.entity_id)
                if e.type.lower() == 'pool_volume':
                    res = pool.volumes_count + pool.snapshots_count
                else:
                    res = pool.filesystems_count + pool.filesystem_snapshots_count
                e.entities = res
            results.append(e)
        return results

    @feature_command(
        feature='qos_fs',
        status=True,
        description="Assign entity to policy",
        types=(QOSPolicy, VolumeOrSnapshot.List, FilesystemOrSnapshot.List, Pool.List),
        )
    @roles.INFINIDAT.ADMIN
    def assign(self, _api, policy, vol=None, fs=None, pool=None):
        if not vol and not fs and not pool:
            raise BadArguments("vol or fs or pool parameter must be used")
        if sum(x is not None  for x in [vol, fs, pool]) > 1:
            raise BadArguments("Only one parameter from vol, fs, pool is allowed")
        entities = vol or fs or pool
        with _api.aggregation() as aggregator:
            for e in entities:
                aggregator.next_task(e.name, "Assigned")
                data = dict(entity_id=e.id)
                result = _api.post("{}/{}/assign".format(QOSPolicy.get_uri(), policy.id), data)
        entity_name = entities[0]['dataset_type'].lower() if entities[0].get('dataset_type', None) else "pool"
        if len(entities) == 1:
            return "{} '{}' was assigned to policy '{}'".format(entity_name.upper(), entities[0].name, policy.name)
        else:
            return "{} {}s were assigned to policy '{}'".format(len(entities), entity_name, policy.name)

    @feature_command(
        feature='qos_fs',
        status=False,
        name="assign",
        description="Assign entity to policy",
        types=(QOSPolicy, VolumeOrSnapshot.List, Pool.List),
    )
    @roles.INFINIDAT.ADMIN
    def old_assign(self, _api, policy, vol=None, pool=None):
        if not vol and not pool:
            raise BadArguments("vol or pool parameter must be used")
        if sum(x is not None  for x in [vol, pool]) > 1:
            raise BadArguments("Only one parameter from vol, pool is allowed")
        entities = vol or pool
        with _api.aggregation() as aggregator:
            for e in entities:
                aggregator.next_task(e.name, "Assigned")
                data = dict(entity_id=e.id)
                result = _api.post("{}/{}/assign".format(QOSPolicy.get_uri(), policy.id), data)
        entity_name = entities[0]['dataset_type'].lower() if entities[0].get('dataset_type', None) else "pool"
        if len(entities) == 1:
            return "{} '{}' was assigned to policy '{}'".format(entity_name.upper(), entities[0].name, policy.name)
        else:
            return "{} {}s were assigned to policy '{}'".format(len(entities), entity_name, policy.name)


    @feature_command(
        feature='qos_fs',
        status=True,
        description="Remove entity from policy",
        types=(VolumeOrSnapshot.List, FilesystemOrSnapshot.List, Pool.List, QOSPolicy),
        )
    @roles.INFINIDAT.ADMIN
    def unassign(self, _api, vol=None, fs=None, pool=None, policy=None):
        if not vol and not fs and not pool:
            raise BadArguments("vol or fs or pool parameter must be used")
        if sum(x is not None for x in [vol, fs, pool]) > 1 and policy:
            raise BadArguments("Only one parameter from vol, fs, pool is allowed when policy selected")
        aggregator = _api.create_request_aggregator()
        with aggregator:
            if vol or fs:
                entities = (vol or []) + (fs or [])
                for e in entities:
                    aggregator.next_task(e.name, "Unassigned")
                    data = dict(entity_id=e.id)
                    policy_id =  policy.id if policy else e.qos_policy_id
                    if policy_id:
                        result = _api.post("{}/{}/unassign".format(QOSPolicy.get_uri(),policy_id), data)
            if pool:
                entities = pool
                for e in entities:
                    aggregator.next_task(e.name, "Unassigned")
                    data = dict(entity_id=e.id)
                    if policy:
                        result = _api.post("{}/{}/unassign".format(QOSPolicy.get_uri(), policy.id), data)
                    else:
                        for qos_policy in e.qos_policies:                   
                            result = _api.post("{}/{}/unassign".format(QOSPolicy.get_uri(), qos_policy.id), data)
        if aggregator.is_empty():
            return  "Entities are not assigned to QoS policy"
        return aggregator.run()

    @feature_command(
        feature='qos_fs',
        status=False,
        name="unassign",
        description="Remove entity from policy",
        types=(VolumeOrSnapshot.List, Pool.List, QOSPolicy),
    )
    @roles.INFINIDAT.ADMIN
    def old_unassign(self, _api, vol=None, pool=None, policy=None):
        if not vol and not pool:
            raise BadArguments("vol or pool parameter must be used")
        if sum(x is not None for x in [vol, pool]) > 1 and policy:
            raise BadArguments("Only one parameter from vol, pool is allowed when policy selected")
        aggregator = _api.create_request_aggregator()
        with aggregator:
            if vol:
                entities = (vol or [])
                for e in entities:
                    aggregator.next_task(e.name, "Unassigned")
                    data = dict(entity_id=e.id)
                    policy_id =  policy.id if policy else e.qos_policy_id
                    if policy_id:
                        result = _api.post("{}/{}/unassign".format(QOSPolicy.get_uri(),policy_id), data)
            if pool:
                entities = pool
                for e in entities:
                    aggregator.next_task(e.name, "Unassigned")
                    data = dict(entity_id=e.id)
                    if policy:
                        result = _api.post("{}/{}/unassign".format(QOSPolicy.get_uri(), policy.id), data)
                    else:
                        for qos_policy in e.qos_policies:
                            result = _api.post("{}/{}/unassign".format(QOSPolicy.get_uri(), qos_policy.id), data)
            if aggregator.is_empty():
                return  "Entities are not assigned to QoS policy"
        return aggregator.run()