import textwrap
import time
import os
import json
import operator
from collections import OrderedDict
from functools import partial
from munch import Munch
from numbers import Number

from infinishell.infishell import (
    commandgroup,
    command,
    render,
    config,
    get_active_shell,
    runtime,
    CommandAborted,
    BadValue,
    )

from infinishell.infishell.types import Str, Int, YesNo, Positive
from infinishell.infishell.utils.python import getattr_deep
from infinishell.infishell.utils import strlist
from infinishell.infishell.renderer import set_renderer_properties, get_default_renderer_properties
from infinishell.infishell.commands import hidden

from ...utils import SafeNestedDict, yesno
from ..exceptions import APIFailure, InvalidSystemState
from ..system import system
from ..renderer.values import StorageCapacityRenderer
from .utils import input_password_check_credentials
from .types import SystemService, Node, Pool, ExperimentalFeature, MetricsInterval
from .utils import increased_api_timeout, format_savings
from .options import QueryOptions, flag, Grep, InvertedGrep

from .metrics import ShowSanCounterBuilder, ShowNasCounterBuilder, run_loop

from . import roles, EntityCommands, render_collection
from ..features import feature_exists


class FieldGetter(object):
    def __init__(self, obj, default='N/A'):
        self.obj = obj
        self.default = default

    def __getitem__(self, item):
        return getattr_deep(self.obj, item, self.default)


SYSTEM_QUERY_FIELDS = strlist("type ops throughput latency")
SYSTEM_QUERY_SUBSTITUTES = dict(
    ops="operations",
    )
SYSTEM_RENDERER_KWARGS = dict(
        fields=SYSTEM_QUERY_FIELDS,
        substitute=SYSTEM_QUERY_SUBSTITUTES,
    )

class SystemMetrics(object):

    def __init__(self, *args, **kwargs):
        self.message = kwargs.get('message', "")
        self.api = kwargs['api']
        self.nasBuilder = ShowNasCounterBuilder(api=self.api, element_type=None, grouping_field=None, message="", filters=dict())
        self.sanBuilder = ShowSanCounterBuilder(api=self.api, element_type=None, grouping_field=None, message="", filters=dict())

    def build(self):
        self.nasBuilder.get_filters_params(protocol="NFS", filters_fields=None)
        self.nasBuilder.get_collectors_params()
        self.nasBuilder.build()
        self.sanBuilder.get_filters_params(protocol="SAN", filters_fields=None, protocol_field='protocol_type')
        self.sanBuilder.get_collectors_params()
        self.sanBuilder.build()

    def get_message(self):
        return self.message

    def get_data(self):
        nas_data = self.nasBuilder.get_data()
        nas_data[0].type = "NAS"
        san_data = self.sanBuilder.get_data()
        san_data[0].type = "SAN"
        san_data[0].latency = san_data[0].external_latency_wout_err
        result = [nas_data[0], san_data[0]]
        return result
    
    def destroy_collectors(self):
        self.nasBuilder.destroy_collectors()
        self.sanBuilder.destroy_collectors()


@commandgroup(prefix="system")
class SystemCommands(object):

    @command
    @render(
        disable_renderer_override=True,
        )
    @roles.INFINIDAT
    def stats_query(self, _api):
        return json.dumps(_api.get("system/stats"), indent=4)

    INFO_SUBSTITUTES = dict(
        rebuild_2_inprogress="rebuild_2_in_progress",
        rebuild_1_inprogress="rebuild_1_in_progress",
        )

    @command(
        description="Display general system information",
        options=(Grep(), InvertedGrep(),),
        )
    @render(
        individually=True,
        substitute=INFO_SUBSTITUTES,
        )
    def info(self, _api):
        default = "N/A"
        _system = SafeNestedDict(_api.get("system"), default=default)
        _config = SafeNestedDict(_api.get("config"), default=default)
        _mgmt = _config['mgmt']
        _core = _config['core']
        prop = partial(_system.get_nested, replace_null=True)
        false_prop = partial(_system.get_nested)
        int_prop = partial(prop, default=0, replace_null=True)
        pool_stats = self._get_pool_stats(_api, default=default)
        arithmetic = partial(self._safe_arithmetic, default=default)
        is_privileged = system.user.roles.is_infinidat() or system.user.roles.is_technician()
        charges = self._compute_bbu(_api)
        protections = self._compute_node_bbu_protection(_api)
        has_compression = feature_exists('compression')
        state = "{}{}".format(prop("operational_state.state")," (Read Only)" if false_prop("operational_state.read_only_system") else "")
        result = OrderedDict()
        for is_shown, field, value in (
            (True,              "name", prop("name")),
            (True,              "serial", prop("serial_number")),
            (True,              "product_id", prop("product_id")),
            (True,              "model_name", _mgmt["system.model_short_name"]),
            (True,              "version", prop("version")),
            (is_privileged,     "revision", prop("release.system.revision") if "release" in _system else prop("revision")),
            (is_privileged,     "deployment_id", prop("deployment_id")),
            (is_privileged,     "installed_on", prop("install_timestamp")),
            (is_privileged,     "updated_on", prop("upgrade_timestamp")),
            (True,              "uptime", prop("uptime")),
            (True,              "operational_state", state),
            (is_privileged,     "inactive_nodes", prop("health_state.inactive_nodes")),
            (True,              "disk_encryption", prop("health_state.encryption_enabled")),
            (True,              "wwnn", prop("wwnn")),

            (True,              "ssd_cache_drives", prop("health_state.active_cache_ssd_devices")),

            (True,              "total_physical_capacity",      prop("capacity.total_physical_capacity")),
            (True,              "free_physical_capacity",       prop("capacity.free_physical_space")),
            (True,              "pools_physical_capacity",      prop("capacity.total_allocated_physical_space")),
            (True,              "pools_allocated_physical_capacity",  pool_stats.allocated_physical_capacity),

            (True,              "total_virtual_capacity",       prop("capacity.total_virtual_capacity")),
            (True,              "free_virtual_capacity",        prop("capacity.free_virtual_space")),
            (True,              "pools_virtual_capacity",       arithmetic(
                                                                    prop("capacity.total_virtual_capacity"),
                                                                    operator.sub,
                                                                    prop("capacity.free_virtual_space"))),
            (True,              "pools_allocated_virtual_capacity",   pool_stats.allocated_virtual_capacity),
            (has_compression,   "compression_savings",                pool_stats.capacity_savings),

            (True,              "pools",        prop('entity_counts.pools')),
            (True,              "volumes",      "{} ({} snapshots)".format(prop('entity_counts.volumes'), prop('entity_counts.volume_snapshots'))),
            (True,              "filesystems",  "{} ({} snapshots)".format(prop('entity_counts.filesystems'), prop('entity_counts.filesystem_snapshots'))),
            (True,              "consistency_groups",       "{} ({} snap-groups)".format(prop('entity_counts.consistency_groups'), prop('entity_counts.snapshot_groups'))),
            (True,              "replicas",     self._get_total_objects(_api, "replicas")),
            (True,              "exports",      self._get_total_objects(_api, "exports")),
            (True,              "mappings",        prop('entity_counts.mapped_volumes')),
            (True,              "hosts",        prop('entity_counts.hosts')),

            (True,     "bbu_charge_percent", charges),
            (True,     "node_bbu_protection", protections),
            (is_privileged,     "enclosure_failure_safe_distribution", prop("health_state.enclosure_failure_safe_distribution")),

            (is_privileged,     "total_spare_drives",           arithmetic(
                                                                    prop('capacity.total_spare_partitions'),
                                                                    operator.div,
                                                                    _core['infinispares.partitions_per_drive'])),
            (is_privileged,     "available_spare_drives",       arithmetic(
                                                                    arithmetic(
                                                                        prop('capacity.total_spare_partitions'),
                                                                        operator.sub,
                                                                        prop('capacity.used_spare_partitions')),
                                                                    operator.div,
                                                                    _core['infinispares.partitions_per_drive'])),
            (is_privileged,     "infinispare_drives_in_use",    arithmetic(
                                                                    prop('capacity.used_dynamic_spare_partitions'),
                                                                    operator.div,
                                                                    _core['infinispares.partitions_per_drive'])),
            (is_privileged,     "infinispare_capacity_in_use",  prop('capacity.used_dynamic_spare_bytes')),

            (not is_privileged, "inactive_drives",      sum(int_prop("health_state.{}_drives".format(state))
                                                            for state in ("failed", "missing", "phasing_out", "ready", "testing", "unknown",))),
            (is_privileged,     "active_drives",        prop("health_state.active_drives")),
            (is_privileged,     "failed_drives",        prop("health_state.failed_drives")),
            (is_privileged,     "missing_drives",       prop("health_state.missing_drives")),
            (is_privileged,     "deactivating_drives",   prop("health_state.phasing_out_drives")),
            (is_privileged,     "ready_drives",         prop("health_state.ready_drives")),
            (is_privileged,     "testing_drives",       prop("health_state.testing_drives")),
            (is_privileged,     "unknown_drives",       prop("health_state.unknown_drives")),
            ):
            if is_shown:
                result[field] = value

        result.update(self._get_rebuild_data(prop("health_state", {})))
        return result

    def _safe_arithmetic(self, left, operator, right, default=None):
        if not (isinstance(left, Number) and isinstance(right, Number)):
            return default
        return operator(left, right)

    def _compute_bbu(self, api):
        ups = api.get("components/ups")
        charges = []
        for u in ups:
            charges.append("BBU {}: {}".format(u.id,"{}%".format(u.battery_charge) if u.battery_charge else "N/A" ))
        return ", ".join(charges)

    def _compute_node_bbu_protection(self, api):
        nodes = api.get("components/nodes")
        result = []
        for n in nodes:
            result.append("Node {}: {}".format(n.id,"{}".format(n.node_bbu_protection) if n.node_bbu_protection else "N/A" ))
        return ", ".join(result)

    def _get_rebuild_data(self, health_state):
        result = OrderedDict()
        field = "rebuild_{}_inprogress"
        group_field = "raid_groups_pending_rebuild_{}"
        for i in (1, 2):
            rebuild_field = field.format(i)
            display_field = "rebuild_{}_in_progress".format(i)
            rebuild = health_state.get(rebuild_field, None)
            groups = health_state.get(group_field.format(i), False)
            if rebuild is not None:
                result[display_field] = rebuild and "yes ({} RAID groups pending)".format(groups)
        return result

    def _get_pool_stats(self, _api, default=None):
        try:
            query = Pool.paged_query(_api, page_size=config.api_max_page_size)
        except APIFailure:
            return Munch(
                count=default,
                allocated_physical_capacity=default,
                allocated_virtual_capacity=default,
                capacity_savings=default,
                )
        physical, virtual, savings = 0, 0, 0
        for pool in query:
            physical += pool.allocated_physical_space
            virtual += pool.virtual_capacity - pool.free_virtual_space
            savings += getattr(pool, 'capacity_savings', 0)
        return Munch(
            count=query.total_objects,
            allocated_physical_capacity=physical,
            allocated_virtual_capacity=virtual,
            capacity_savings=format_savings(Munch(allocated=physical,
                                                  tree_allocated=0,
                                                  capacity_savings=savings)),
            )

    def _get_total_objects(self, _api, collection):
        try:
            return _api.get("{}/?page_size=1".format(collection), unwrap_result=False).metadata.number_of_objects
        except APIFailure:
            return "N/A"

    BOOT_PROGRESS_API_TIMEOUT = 5
    BOOT_PROGRESS_REQUEST_INTERVAL = 5
    BOOT_ACTIVATION_TIMEOUT = 120

    @command(
        description="Display system boot progress",
        )
    def boot_progress(self, _api):
        shell = get_active_shell()
        get_info = partial(_api.get, "system/operational_state", timeout=self.BOOT_PROGRESS_API_TIMEOUT)
        erase_line = lambda _line: shell.output("\r" + " "*len(line) + "\r", nl=False, flush=True) if shell.is_interactive_terminal() else shell.output("")
        # Wait for system to respond first
        line = "Waiting for system to respond..."
        shell.output(line, nl=False, flush=True)
        info = None
        while True:
            try:
                info = get_info()
            except APIFailure:
                time.sleep(5)
                continue
            break
        erase_line(line)
        # Loop until boot is complete
        while info.state == "STARTING":
            if info.init_state:
                line = "({boot.phase_number}/{boot.total_phases}) {boot.phase}: {boot.description} [{boot.progress_percent}%]".format(boot=info.init_state)
            else:
                line = "State N/A"
            shell.output(line, nl=False, flush=True)
            time.sleep(self.BOOT_PROGRESS_REQUEST_INTERVAL)
            info = get_info()
            erase_line(line)
        if info.state == "STANDBY":
            timeout = time.time() + self.BOOT_ACTIVATION_TIMEOUT
            line = "Waiting for system to complete activation..."
            shell.output(line, nl=False, flush=True)
            while info.state == "STANDBY":
                if time.time() > timeout:
                    erase_line(line)
                    shell.output("Timed out waiting for system to complete activation.")
                    break
                time.sleep(self.BOOT_PROGRESS_REQUEST_INTERVAL)
                info = get_info()
            erase_line(line)
        # Show final system state
        shell.output("System state: {} ({})".format(info.state, info.description), nl=True)

    @command(
        description="Create a system log bookmark for later review by Infinidat customer support",
        )
    @roles.INFINIDAT
    def log_mark(self, _api):
        _api.internal.post("logs/mark", {})
        return "Logs marked"

    @command(
        description="Save latest system log records for later review by Infinidat customer support",
        )
    @roles.INFINIDAT
    def log_freeze(self, _api):
        _api.internal.post("logs/freeze", dict(cause=None, duration=None))
        return "Logs frozen"

    def _pre_shutdown(self):
        if not (config.auto_confirm_prompts and system.user.roles.is_infinidat()):
            prompt_parts = (
                textwrap.fill("Warning: You are about to shutdown the system. This will stop all I/O services to hosted volumes."),
                textwrap.fill("To avoid potential loss of data, please make sure that no application is accessing the system's volumes. "),
                "Are you sure you wish to continue?",
                )
            if not yesno("\n\n".join(prompt_parts), always_prompt=True):
                raise CommandAborted("Shutdown aborted by user")
            if not system.user.username == "INTERNAL":
                if not input_password_check_credentials("Please enter your password: "):
                    raise CommandAborted("Wrong password, shutdown aborted")

    def _shutdown(self, _api, poweroff):
        self._pre_shutdown()
        operation = "shutdown_and_poweroff" if poweroff else "shutdown"
        _api.post("system/operational_state/" + operation)
        msg = "Shutting down system"
        if not poweroff:
            msg += ", keeping power on"
        return msg

    @command(
        name="shutdown",
        description="Initiate system shutdown process",
        types=(YesNo,),
        )
    @roles.INFINIDAT
    def infinidat__shutdown(self, _api, keep_power_on=False):
        return self._shutdown(_api, not keep_power_on)

    @command(
        description="Initiate system shutdown process",
        )
    @roles.ADMIN.TECHNICIAN
    def shutdown(self, _api):
        return self._shutdown(_api, True)

    @command(
        description="Initiate emergency system shutdown process",
        )
    @roles.INFINIDAT
    def emergency_shutdown(self, _api):
        self._pre_shutdown()
        _api.post("system/operational_state/emergency_shutdown")
        return "SYSTEM GOING INTO EMERGENCY SHUTDOWN"

    @command
    @roles.INFINIDAT
    def activate(self, _api):
        if self._get_system_state(_api) != "STANDBY":
            raise InvalidSystemState("System can be activated only while in STANDBY state")
        _api.post("system/operational_state/activate")
        return "System undergoing activation"

    def _get_system_state(self, _api):
        system = _api.get("system")
        return system.operational_state.state

    @command(
        types=(Str,),
        description="Upload an upgrade package",
        )
    @roles.INFINIDAT.ADMIN
    def upgrade(self, _api, filename):
        if not os.path.isfile(filename):
            raise BadValue("File {} was not found".format(filename))
        if os.path.getsize(filename) > 50 * 1024 * 1024:
            raise BadValue("Allowed package size is at most 50 MiB")
        with increased_api_timeout():
            _api.upload("system/upgrade", (filename,))
        return "Package uploaded successfully"

    @command(
        options = (flag("total", "Show only the total billable capacity"),),
        description = "Query Capacity On Demand (COD) consumption in the system.",
        )
    @render(
        fields=strlist("pool consumed_cod")
        )
    @roles.INFINIDAT.ADMIN
    @hidden
    def cod_query(self, _api):
        total = 0
        results = []
        for pool in Pool.iter_all(_api, page_size=config.api_max_page_size):
            used_virtual_capacity = pool.virtual_capacity - pool.free_virtual_space
            capacity = max(pool.allocated_physical_space, used_virtual_capacity)
            total += capacity
            results.append(Munch(
                pool=pool.name,
                consumed_cod=capacity,
                ))
        results.append(Munch(
            pool="______",
            consumed_cod="______",
            ))
        results.append(Munch(
            pool="TOTAL",
            consumed_cod=total,
            ))
        if runtime.options.get("total"):
            return StorageCapacityRenderer().render(total, "")
        return results


    @command(
        types=(MetricsInterval, Positive),
        description="",
        )
    def show_metrics(self, _api, interval=1, count=5):
        message = "System metrics stats"
        builder = SystemMetrics(api=_api, message=message)
        builder.build()
        run_loop(builder=builder, renderer_kwargs=SYSTEM_RENDERER_KWARGS, interval=interval, error_interval=10, count=count)



@commandgroup(prefix="service", parent=SystemCommands)
class SystemServiceCommands(EntityCommands):
    ENTITY_TYPE = SystemService
    DETAILED_QUERY_FIELDS = strlist("""
        name
        cluster_state
        """)
    QUERY_SUBSTITUTES = dict(
        name="service",
        cluster_state="state",
        )

    @command(
        name="query",
        types=(SystemService.List,),
        description="report state on services",
        hidden_substitue=dict(name="services"),
        options=QueryOptions(fields=DETAILED_QUERY_FIELDS, substitute=QUERY_SUBSTITUTES),
        )
    @render(
        fields=strlist("name cluster_state"),
        substitute=dict(
            name="service",
            cluster_state="state",
            ),
        )
    @roles.INFINIDAT.ADMIN.TECHNICIAN
    def query(self, _api, services=None):
        return self._query(_api, services, _hook=self._query_hook if system.user.roles.is_infinidat() or system.user.has_role("TECHNICIAN") else None)

    def _query_hook(self, services):
        fields = list(get_default_renderer_properties()['fields'])
        fields_set = runtime.options.get("columns", None) is not None
        results = []
        for service in services:
            res = Munch(
                name=service.name,
                cluster_state=service.cluster_state,
                )
            results.append(res)
            states = Munch(("node_{}".format(s.node_id), s) for s in service.node_states)
            for node in sorted(states.keys()):
                res[node] = states[node].role if states[node].state == "ACTIVE" else states[node].state
                if not fields_set:
                    fields.append(node)
            if not fields_set:
                set_renderer_properties(fields=fields)
                fields_set = True
        return results

    @command(
        types=(SystemService, Node),
        description="activate service",
        hidden_substitue=dict(name="service"),
        )
    @roles.INFINIDAT.TECHNICIAN
    def activate(self, _api, service, node=None):
        if service.name == "management":
            raise BadValue("Management service activation and deactivation are not supported")
        data = dict(node_id=node.id if node else None)
        self._post(_api, service, "start", data=data)
        return self._result_message(service, "commencing activation{}".format("" if not node else " on {}".format(node.name)))

    @command(
        types=(SystemService, Node),
        description="deactivate service",
        hidden_substitue=dict(name="service"),
        )
    @roles.INFINIDAT.TECHNICIAN
    def deactivate(self, _api, service, node=None):
        if service.name == "management":
            raise BadValue("Management service activation and deactivation are not supported")
        if service.name in ['core', 'clm'] and node is None:
            raise CommandAborted("When deactivating clm or core services, a node must be specified. For system wide deactivation use system.shutdown")
        data = dict(node_id=node.id if node else None)
        self._post(_api, service, "stop", data=data)
        return self._result_message(service, "commencing deactivation{}".format("" if not node else " on {}".format(node.name)))


@commandgroup(prefix="experimental", parent=SystemCommands)
class ExperimentalCommands(EntityCommands):

    ENTITY_TYPE = ExperimentalFeature

    @command
    @render(
        fields=strlist("name version enabled"),
        )
    @roles.INFINIDAT
    def query(self, _api):
        return self._query(
            _api,
            _hook=self._experimental_query_hook,
            )

    def _experimental_query_hook(self, features):
        for f in features:
            if f.experimental:
                yield f

    @command(
        types=(ExperimentalFeature,),
        description="Activate feature",
        )
    @roles.INFINIDAT
    def enable(self, _api, feature):
        if feature.enabled == True:
            return "Feature already enabled"
        self._put(_api, feature, data=dict(enabled=True))
        return self._result_message(feature, "enabled\nPlease restart the shell to enable all relevant commands.")