from munch import Munch
from functools import partial
from collections import defaultdict
from datetime import datetime

from infinishell.infishell import config, runtime
from infinishell.infishell.utils import NOTHING
from infinishell.infishell.exceptions import BadArguments

from ..system import system
from ..exceptions import ManagementException

from .types import Pool
from . import EntityCommands, LazyIdNameMapping, Null


class TreeNode(object):
    def __init__(self, entity=None, children=None):
        self.entity = entity
        self.children = [] if children is None else children
        self.parent = None
    def add_child(self, treenode):
        self.children.append(treenode)
        treenode.parent = self
    def __getitem__(self, key):
        return self.entity.get(key, None)

class HierarchicalEntityCommandsBase(EntityCommands):
    HIERARCHICAL_TYPE = None  # Subclasses must override this

    THIN = "THIN"
    THICK = "THICK"

    def _get_default_provtype(self):
        defaultProvType = self.THICK
        try:
            dft_prov_md = system.api.get('metadata/system/ui-dataset-default-provisioning')
            if dft_prov_md:
                if dft_prov_md.value and (self.THICK == dft_prov_md.value.upper() or self.THIN == dft_prov_md.value.upper()):
                    return dft_prov_md.value.upper()
        except ManagementException as e:
            return defaultProvType
        return defaultProvType

    def _get_provtype(self, is_thin):
        return self.THIN if is_thin else self.THICK

    def _create_lock_filter(self, _api, lock=None, from_time=None, to_time=None):
        filter =  _api.create_filter()
        #If no filter was requested
        if not lock and not from_time and not to_time:
            return filter
        #Calculate utc_timestamp for current time
        epoch_time = datetime.utcfromtimestamp(0)
        utc_time = datetime.utcnow()
        utc_delta = utc_time - epoch_time
        utc_timestamp = int(utc_delta.days * 86400 + utc_delta.seconds + utc_delta.microseconds / 1e6) * 1000

        if lock == "LOCKED":
            from_time = utc_timestamp if from_time is None else (utc_timestamp if utc_timestamp > from_time else from_time)
        elif lock == "EXPIRED":
            to_time = utc_timestamp if to_time is None else (utc_timestamp if utc_timestamp < to_time else to_time)
        elif lock == "UNLOCKED":
            return filter.eq('lock_expires_at', 'null')

        return self._filter_range(filter, "lock_expires_at", from_time, to_time + 999 if to_time else to_time)

    def _filter_range(self, filter, field, from_value, to_value):
        if from_value is not None:
            if to_value is not None:
                return filter.between(field, from_value, to_value)
            return filter.ge(field, from_value)
        if to_value is not None:
            return filter.le(field, to_value)
        return filter

    def _common_query(self, _api, entities, _filter=None, **params):
        hook = params.pop('_hook', None)
        return self._query(_api, entities, _hook=hook if hook else self._get_actual_query_hook(_api), _filter=_filter, **params)

    def _get_actual_query_hook(self, _api):
        return partial(self._query_hook, _api=_api)

    def _query_hook(self, entities, _api):
        return entities

    def _get_tree(self, _api, entity=None, pool=None):

        sort_fields = runtime.options.get("sort", None)
        sort = self._create_sort(_api, sort_fields) if sort_fields else None
        if entity and pool:
            raise BadArguments("Either {}, pool, or neither may be specified".format(self.ENTITY_TYPE.get_entity_name()))
        if entity is None:
            roots = self._get_all_trees(_api, pool, sort)
        else:
            roots = self._get_single_tree(_api, entity, sort)
        return self._flatten_tree_for_rendering(roots, sort=sort)

    def _get_all_trees(self, _api, pool=None, sort=None):
        treenodes = defaultdict(TreeNode)
        roots = []
        for v in self.HIERARCHICAL_TYPE.paged_query(_api,
            hook=self._get_actual_query_hook(_api),
            filter=None if not pool else _api.create_filter(pool_id=pool.id),
            page_size=config.api_max_page_size,
            sort=sort,
            ):
            treenode = treenodes[v.id]
            treenode.entity = v
            if not v.parent_id:
                roots.append(treenode)
            else:
                treenodes[v.parent_id].add_child(treenode)
        return roots

    def _get_single_tree(self, _api, entity, sort=None):
        # Apply query hook on single entity
        [entity] = self._get_actual_query_hook(_api)([entity])
        # Compute tree
        root = TreeNode(entity)
        self._fill_single_tree(_api, root, sort)
        return [root]

    def _fill_single_tree(self, _api, parentnode, sort=None):
        for entity in self.HIERARCHICAL_TYPE.get_all(_api, hook=self._get_actual_query_hook(_api), sort=sort, filter=_api.create_filter(parent_id=parentnode.entity.id)):
            childnode = TreeNode(entity)
            parentnode.add_child(childnode)
            self._fill_single_tree(_api, childnode)

    def _flatten_tree_for_rendering(self, treenodes, depth=0, sort=None, result=None):
        if result is None:
            result = []

        for treenode in treenodes:
            treenode.entity.name = ('|   ' * depth) + treenode.entity.name
            result.append(treenode.entity)
            self._flatten_tree_for_rendering(treenode.children, depth=depth+1, result=result)
        return result

    def _walk_tree(self, _api, entity):
        for treenode in self._walk_treenodes(self._get_single_tree(_api, entity)):
            yield treenode.entity

    def _walk_treenodes(self, nodes):
        for node in nodes:
            yield node
            for child in self._walk_treenodes(node.children):
                yield child

    def _walk_tree_or_single_entity(self, _api, entity, recurse):
        if recurse:
            return self._walk_tree(_api, entity)
        return (entity,)

    def _update_recursive(self, _api, entity, recurse, success_msg, props, predicate=None, success_msg_with_is=False, noop_msg=None, noop_msg_with_is=False):
        success_msg = self._result_message(entity, success_msg, with_tree=recurse, with_is=success_msg_with_is)
        if noop_msg:
            noop_msg = self._result_message(entity, noop_msg, with_tree=recurse, with_is=noop_msg_with_is)
        if not recurse:
            if predicate and not predicate(entity):
                return noop_msg or success_msg
            return self._update(_api, entity, returns=success_msg, **props)
        aggregator = _api.create_request_aggregator(total_success_msg=success_msg)
        single_success_msg = success_msg[0].upper() + success_msg[1:]
        with aggregator:
            for e in self._walk_tree(_api, entity):
                if predicate and not predicate(e):
                    continue
                aggregator.next_task(e.name, single_success_msg)
                self._update(_api, e, returns=NOTHING, **props)
        if aggregator.is_empty():
            return noop_msg or success_msg
        return aggregator.run()

    def _result_message(self, entity, msg, subject=None, with_tree=False, with_is=False):
        return '{} {}{}{} {}'.format(
            self._result_subject(subject, capitalize=True),
            self._result_entity_name(entity),
            "" if not with_tree else " and all of its descendants",
            "" if not with_is else " are" if with_tree else " is",
            msg,
            )
