import re
import string
import random
from munch import Munch
from itertools import chain
from functools import partial

from infinishell.infishell import commandgroup, command, render, get_active_shell
from infinishell.infishell.utils import strlist
from infinishell.infishell.exceptions import BadArguments
from infinishell.infishell.types import Str

from ...utils import Version, get_confirmed_password

from .types import (
    LoggedInInitiator,
    WWN,
    Name,
    NetworkSpace,
    Host,
    Cluster,
    SCSIAddress,
    DefinedSCSIAddress,
    SCSIPortType,
    ScsiInitiator,
    Volume,
    MasterVolume,
    VolumeSnapshot,
    Lun,
    Initiator,
    AvailableLoggedInInitiator,
    HostMappingType,
    ISCSISecurityMethod,
    HostType,
    WWN_PATTERN,
    IQN_PATTERN,
    )
from .options import QueryOptions
from .system import system
from .mapping import MappingCommandsBase
from . import EntityCommands, render_collection, set_renderer_query_fields, roles, LazyIdNameMapping, Null

from ..features import feature_command, feature_exists


class HostCommandsBase(EntityCommands, MappingCommandsBase):
    def _map(self, _api, mappable, lun, *volume_collections):
        all_volumes = list(chain(*((c or ()) for c in volume_collections)))
        if not len(all_volumes):
            raise BadArguments("At least one volume or snapshot must be specified")
        if lun is not None and len(all_volumes) != 1:
            raise BadArguments("Specifying LUN in conjunction with more than one volume or snapshot is not allowed")
        return super(HostCommandsBase, self)._map(_api, self.ENTITY_TYPE, mappable, all_volumes, lun)

    def _unmap(self, _api, mappable, lun, *volume_collections):
        all_volumes = list(chain(*((c or ()) for c in volume_collections)))
        if lun is None:
            if not len(all_volumes):
                raise BadArguments("At least one volume or snapshot must be specified")
        else:
            if len(all_volumes):
                raise BadArguments("Specifying LUN in conjunction with volumes or snapshots is not allowed")
            return self._unmap_lun(_api, mappable, lun)
        return super(HostCommandsBase, self)._unmap(_api, self.ENTITY_TYPE, mappable, all_volumes)

    def _map_query(self, _api, lun_container, volume=None, map_type=None, host_type=None):
        set_renderer_query_fields()
        name = lun_container.name if lun_container else None
        lun_containers = self._query(_api, name=name, host_type=host_type)
        if volume:
            return [Munch({self.ENTITY_TYPE.get_entity_name() : lc.name,
                           "lun" : mapping.lun,
                           "volume" : volume.name,
                           "target_type" : "CLUSTER" if mapping.clustered else "HOST",
                           "serial" : volume.serial,
                           'udid' : volume.udid,
                           })
                    for lc in lun_containers
                    for mapping in lc.luns
                    if ((mapping.volume_id == volume.id) and ((mapping.clustered == map_type) or (map_type is None)))]
        else:
            results = []
            for lc in lun_containers:
                for mapping in lc.luns: 
                    if (mapping.clustered == map_type) or (map_type is None):
                        vol = Volume.get_by_id(_api, mapping.volume_id)
                        item = Munch({self.ENTITY_TYPE.get_entity_name() : lc.name,
                                "lun" : mapping.lun,
                                "volume" : vol.name,
                                "target_type" : "CLUSTER" if mapping.clustered else "HOST",
                                "serial" : vol.serial,
                                'udid' : vol.udid,
                                })
                        if feature_exists('openvms'):
                            item["host_type"] = '-' if (lc.get('host_type',None) == "DEFAULT") else lc.get('host_type')
                            item["cluster_type"] = '-' if (lc.get('host_type',None) == "DEFAULT") else lc.get('host_type')
                        results.append(item)
            return results


    def _generate_chap_secret(self, length=16):
        return "".join(random.SystemRandom().choice(string.ascii_letters + string.digits) for _ in range(length))

    def _prepare_host_iscsi_security_props(self, host, method):
        if host.security_method == method:
            return None, "iSCSI authentication type already set to {}".format(ISCSISecurityMethod.MGMT_METHODS[method])
        props = dict(
            security_method=method,
            )
        result_msg = ["iSCSI authentication type set to {}".format(ISCSISecurityMethod.MGMT_METHODS[method])]
        if method in (ISCSISecurityMethod.METHODS.CHAP, ISCSISecurityMethod.METHODS.MUTUAL_CHAP) \
        and not host.security_chap_has_inbound_secret:
            inbound_secret = self._generate_chap_secret()
            result_msg.append("auto-generated inbound secret: '{}'".format(inbound_secret))
            props.update(
                security_chap_inbound_secret=inbound_secret,
                )
        if method == ISCSISecurityMethod.METHODS.MUTUAL_CHAP and not host.security_chap_has_outbound_secret:
            outbound_secret = self._generate_chap_secret()
            result_msg.append("auto-generated outbound secret: '{}'".format(outbound_secret))
            props.update(
                security_chap_outbound_secret=outbound_secret,
                )
        return props, ", ".join(result_msg)

    def _get_secrets_interactively(self, inbound_user, inbound_secret, outbound_user, outbound_secret):
        if inbound_user:
            inbound_secret = get_confirmed_password(inbound_secret, "inbound secret")
        if outbound_user:
            outbound_secret = get_confirmed_password(outbound_secret, "outbound secret")
        return {'inbound_user':inbound_user, 'inbound_secret':inbound_secret, 'outbound_user':outbound_user, 'outbound_secret':outbound_secret}

    def _set_host_chap_settings(self, _api, host, inbound_user, inbound_secret, outbound_user, outbound_secret, aggregator_task=None):
        if all(arg is None for arg in (inbound_user, inbound_secret, outbound_user, outbound_secret)):
            raise BadArguments("At least one argument must be specified")
        params = dict(
            security_chap_inbound_username=inbound_user,
            security_chap_inbound_secret=inbound_secret,
            security_chap_outbound_username=outbound_user,
            security_chap_outbound_secret=outbound_secret,
            )
        params = dict((k, v) for k, v in params.iteritems() if v)
        self._update(_api, host, _entity_type=Host, returns=None, _multi_field_put=True, **params)
        warned = self._verify_chap_config_valid_security_method(host, (outbound_user, outbound_secret), "MUTUAL_CHAP",
            (ISCSISecurityMethod.METHODS.MUTUAL_CHAP,), aggregator_task)
        if not warned:
            self._verify_chap_config_valid_security_method(host, (inbound_user, inbound_secret), "CHAP",
                (ISCSISecurityMethod.METHODS.CHAP, ISCSISecurityMethod.METHODS.MUTUAL_CHAP), aggregator_task)

    def _verify_chap_config_valid_security_method(self, host, values, config_name, acceptable_security_methods, aggregator_task):
        if any(value is not None for value in values) and host.security_method not in acceptable_security_methods:
            warning_func = partial(self._warn_chap_secrets_invalid_security_method, host, config_name)
            if aggregator_task:
                aggregator_task.add_success_hook(warning_func)
            else:
                warning_func()
            return True
        return False

    def _warn_chap_secrets_invalid_security_method(self, host, config_name):
        get_active_shell().warning(
            "{} {} settings updated, but current authentication type is {}. Use host.set_auth_type to set the authentication type."
            .format(
                self._result_subject_and_entity(host, entity_type=Host, capitalize=True),
                config_name,
                ISCSISecurityMethod.MGMT_METHODS[host.security_method])
            )

QUERY_FIELDS =          strlist("name cluster lun_count fc_ports iscsi_iqns security_method                                                                created_at")
DETAILED_QUERY_FIELDS = strlist("name cluster lun_count allocated_capacity fc_ports iscsi_iqns security_method security_chap_inbound_username security_chap_outbound_username created_at updated_at")
SORTABLE_FIELDS = strlist("name created_at")
QUERY_SUBSTITUTES = dict(
    lun_count="luns",
    security_method="auth_type",
    security_chap_inbound_username="chap_inbound_user",
    security_chap_outbound_username="chap_outbound_user",
    )

HOST_PORT_QUERY_FIELDS = strlist("host type port logged_in")

MAP_QUERY_FIELDS = strlist("host volume lun target_type")
DETAILED_MAP_QUERY_FIELDS = strlist("host volume lun target_type serial")

@commandgroup(prefix="host", provides=(Host,))
class HostCommands(HostCommandsBase):
    ENTITY_TYPE = Host
    version = Version('2.2.0.0')

    @feature_command(
        feature="openvms",
        status=False,
        name='query',
        description="List existing hosts",
        types=(Host.List, Cluster.List, ISCSISecurityMethod.List),
        hidden_substitue=dict(name="host"),
        substitute=dict(
            clusters="cluster",
            security_methods="auth_type",
            ),
        options=QueryOptions(fields=DETAILED_QUERY_FIELDS, sortable=SORTABLE_FIELDS, substitute=QUERY_SUBSTITUTES, detailed=True),
        )
    @render_collection(
        entity=Host,
        fields=QUERY_FIELDS,
        detailed_fields=DETAILED_QUERY_FIELDS,
        detailed_individually=True,
        substitute=QUERY_SUBSTITUTES,
        )
    def old_query(self, _api, host=None, clusters=None, security_methods=None):
        return self._query(_api,
            host,
            host_cluster_id=clusters,
            security_method=security_methods,
            _hook=partial(self._host_query_hook, cluster_names_by_id=LazyIdNameMapping(Cluster, _api, preloaded_entities=clusters)),
            )

    @feature_command(
        feature="openvms",
        status=True,
        description="List existing hosts",
        types=(Host.List, Cluster.List, ISCSISecurityMethod.List, HostType),
        hidden_substitue=dict(name="host"),
        substitute=dict(
            clusters="cluster",
            security_methods="auth_type",
            ),
        options=QueryOptions(fields=DETAILED_QUERY_FIELDS + strlist('host_type'), sortable=SORTABLE_FIELDS, substitute=QUERY_SUBSTITUTES, detailed=True),
        )
    @render_collection(
        entity=Host,
        fields=QUERY_FIELDS,
        detailed_fields=DETAILED_QUERY_FIELDS + strlist('host_type'),
        detailed_individually=True,
        substitute=QUERY_SUBSTITUTES,
        )
    def query(self, _api, host=None, clusters=None, security_methods=None, host_type=None):
        return self._query(_api,
            host,
            host_cluster_id=clusters,
            security_method=security_methods,
            host_type=host_type,
            _hook=partial(self._host_query_hook, cluster_names_by_id=LazyIdNameMapping(Cluster, _api, preloaded_entities=clusters)),
            )

    def _host_query_hook(self, hosts, cluster_names_by_id):
        #Only query for volumes if --detailed was set
        shell = get_active_shell()
        sorted_volumes = dict()
        if (shell.current_command.options.get('detailed', False)):
            filter = system.api.create_filter().eq("mapped", True)
            volumes = Volume.get_all(system.api, filter=filter, pluck=system.api.create_pluck("id", "size"))
            sorted_volumes = {v.id:v.size for v in volumes}
        for host in hosts:
            if feature_exists('openvms'):
                host.host_type = '-' if (host.host_type == "DEFAULT") else host.host_type
            allocated_capacity = None
            if (shell.current_command.options.get('detailed', False)):
                allocated_capacity = self._sum_capacity(host, sorted_volumes)
            fc_port_count, iscsi_port_count = self._count_ports(host)
            host.update(
                cluster=cluster_names_by_id[host.host_cluster_id] if host.host_cluster_id else None,
                lun_count=len(host.luns),
                fc_ports=fc_port_count,
                iscsi_iqns=iscsi_port_count,
                allocated_capacity=allocated_capacity,
                )
            yield host

    def _sum_capacity(self, host, volumes):
        return sum(volumes.get(lun.volume_id) for lun in host.luns)

    def _count_ports(self, host):
        fc, iscsi = 0, 0
        for port in host.ports:
            if port.type.lower() == SCSIPortType.TYPES.FC.lower():
                fc += 1
            elif port.type.lower() == SCSIPortType.TYPES.ISCSI.lower():
                iscsi += 1
        return fc, iscsi

    @feature_command(
        feature='openvms',
        status=False,
        name='create',
        description="Create a new host",
        types=(Name,),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def old_create(self, _api, name):
        return self._create(_api, name=name)

    @feature_command(
        feature='openvms',
        status=True,
        description="Create a new host",
        types=(Name, HostType),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def create(self, _api, name, host_type=None):
        return self._create(_api, name=name, host_type=host_type)

    @feature_command(
        feature="iscsi",
        status=True,
        description="Add port to host",
        types=(Host, ScsiInitiator.List, SCSIPortType),
        hidden_substitue=dict(name="host"),
        substitute=dict(
            ports="port",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def add_port(self, _api, host, ports, type=None):
        with _api.aggregation() as aggregator:
            for port in ports:
                aggregator.next_task(port, "Added")
                self._post(_api, host, "ports", data=dict(
                    address=port,
                    type=type if type is not None else SCSIPortType.TYPES.FC if WWN_PATTERN.match(port) else SCSIPortType.TYPES.ISCSI,
                    ))
        type = type if type is not None else SCSIPortType.TYPES.FC if WWN_PATTERN.match(port) else SCSIPortType.TYPES.ISCSI
        if len(ports) == 1:
            print('{} port "{}" added to host "{}"'.format(type, port, host.name))
        print('{} ports added to host "{}"'.format(len(ports), host.name))
        if type == SCSIPortType.TYPES.ISCSI:
            if not any(NetworkSpace.get_all(_api, filter=_api.create_filter().eq("service","ISCSI_SERVICE"))):
                get_active_shell().warning("iSCSI is not configured - please create a network space to allow this host to connect (use config.net_space.create)")

    INITIATOR_QUERY_FIELDS = strlist("address type")
    @feature_command(
        feature="iscsi",
        status=True,
        description="List logged-in initiators which are not assigned to any host in the system",
        types=(SCSIPortType,),
        options=[QueryOptions(fields=INITIATOR_QUERY_FIELDS)],
        substitute=dict(
            address="WWPN",
            type="protocol",
            ),
        )
    @render(
        fields=INITIATOR_QUERY_FIELDS,
        substitute=dict(
            address="WWPN",
            type="protocol"
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def initiator_query(self, _api, type=None):
        if type:
            return [i for i in Initiator.iter_all(_api, page_size=1000) if i.host_id == 0 and i.type == type]
        else:
            return [i for i in Initiator.iter_all(_api, page_size=1000) if i.host_id == 0]

    @feature_command(
        feature="iscsi",
        status=True,
        description="Remove port from host",
        types=(Host, DefinedSCSIAddress.List, SCSIPortType),
        hidden_substitue=dict(name="host"),
        substitute=dict(
            ports="port",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def remove_port(self, _api, host, ports, type=None):
        with _api.aggregation() as aggregator:
            for port in ports:
                aggregator.next_task(port, "Removed")
                old_type = type
                type = type if type is not None else SCSIPortType.TYPES.FC if WWN_PATTERN.match(port) else SCSIPortType.TYPES.ISCSI
                self._del(_api, host, "ports", type, port)
                type = old_type
        type = type if type is not None else SCSIPortType.TYPES.FC if WWN_PATTERN.match(port) else SCSIPortType.TYPES.ISCSI
        if len(ports) == 1:
            return '{} port "{}" removed from host "{}"'.format(type.upper(), port, host.name)
        return '{} ports removed from host "{}"'.format(len(ports), host.name)

    @command(
        description="Rename host",
        types=(Host, Name),
        hidden_substitue=dict(name="host"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def rename(self, _api, host, new_name):
        return self._rename(_api, host, new_name)

    @command(
        description="Delete host",
        types=(Host,),
        hidden_substitue=dict(name="host"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def delete(self, _api, host):
        return self._delete(_api, host)

    @command(
        description="Map {} to host".format(Volume.DESCRIPTION_SUBJECT),
        types=(Host, MasterVolume.List, VolumeSnapshot.List, Lun),
        hidden_substitue=dict(name="host"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def map(self, _api, host, vol=None, snap=None, lun=None):
        return self._map(_api, host, lun, vol, snap)

    @command(
        description="Unmap {} from host".format(Volume.DESCRIPTION_SUBJECT),
        types=(Host, MasterVolume.List, VolumeSnapshot.List, Lun.List),
        hidden_substitue=dict(name="host"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def unmap(self, _api, host, vol=None, snap=None, lun=None):
        return self._unmap(_api, host, lun, vol, snap)

    @feature_command(
        feature="iscsi",
        status=True,
        description="List ports of existing hosts",
        types=(Host, SCSIAddress),
        hidden_substitue=dict(name="host"),
        options=QueryOptions(fields=HOST_PORT_QUERY_FIELDS, detailed=True),
        )
    @render_collection(
        entity=Host,
        fields=HOST_PORT_QUERY_FIELDS,
        substitute=dict(host="name"),
        detailed_fields=HOST_PORT_QUERY_FIELDS,
        detailed_substitute=dict(host="name"),
        detailed_individually=True,
        alignment=dict(
            port="left",
            ),
        )
    def port_query(self, _api, host=None, port=None):
        WWN_PATTERN = re.compile(r"([0-9a-fA-F]{2}:?){7}[0-9a-fA-F]{2}")
        name = host.name if host else None
        hosts = self._query(_api, name=name)
        logged_in_initiators = {fc.address:fc for fc in Initiator.iter_all(_api, page_size=1000)}
        if port:
            port = port.replace(":", "") if WWN_PATTERN.match(port) else port
            logged_in = "yes" if port in logged_in_initiators else "no"
            return [Munch(host=_host.name,
                          port=hostport.address,
                          type=hostport.type,
                          logged_in=logged_in)
                    for _host in hosts
                    for hostport in _host.ports if port == (hostport.address.replace(":", "") if hostport.type == 'FC' else hostport.address)]
        else:
            return [Munch(host=_host.name,
                          port=hostport.address,
                          type=hostport.type,
                          logged_in="yes" if (hostport.address.replace(":", "") if hostport.type == 'FC' else hostport.address) in logged_in_initiators else "no")
                    for _host in hosts
                    for hostport in _host.ports]

    @feature_command(
        feature='openvms',
        status=False,
        name='map_query',
        description="List existing host mappings",
        types=(Host, HostMappingType),
        hidden_substitue=dict(name="host"),
        options=QueryOptions(fields=DETAILED_MAP_QUERY_FIELDS, detailed=True),
        )
    @render_collection(
        fields=MAP_QUERY_FIELDS,
        detailed_fields=DETAILED_MAP_QUERY_FIELDS,
        detailed_individually=True,
        substitute=dict(
            target_type="mapping_type",
            lun="lun_id",
            ),
        )
    def old_map_query(self, _api, host=None, map_type=None):
        return self._map_query(_api, host, map_type=map_type)

    @feature_command(
        feature='openvms',
        status=True,
        description="List existing host mappings",
        types=(Host, HostMappingType, HostType),
        hidden_substitue=dict(name="host"),
        options=QueryOptions(fields=DETAILED_MAP_QUERY_FIELDS + strlist('host_type udid'), detailed=True),
        )
    @render_collection(
        fields=MAP_QUERY_FIELDS,
        detailed_fields=DETAILED_MAP_QUERY_FIELDS + strlist('host_type udid'),
        detailed_individually=True,
        substitute=dict(
            target_type="mapping_type",
            lun="lun_id",
            ),
        )
    def map_query(self, _api, host=None, map_type=None, host_type=None):
        return self._map_query(_api, host, map_type=map_type, host_type=host_type)


    @feature_command(
        feature="iscsi",
        status=True,
        description="Set iSCSI host authentication method",
        types=(Host, ISCSISecurityMethod),
        hidden_substitue=dict(name="host"),
        substitute=dict(
            method="type",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def set_auth_type(self, _api, host, method):
        props, message = self._prepare_host_iscsi_security_props(host, method)
        message = self._result_message(host, message)
        if not props:
            return message
        return self._update(_api, host, returns=message, **props)

    @feature_command(
        feature="iscsi",
        status=True,
        description="Set iSCSI host CHAP usernames and secrets",
        types=(Host, Str, Str, Str, Str),
        hidden_substitue=dict(name="host"),
        substitute=dict(),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def set_chap(self, _api, host, inbound_user=None, inbound_secret=None, outbound_user=None, outbound_secret=None):
        params = self._get_secrets_interactively(inbound_user, inbound_secret, outbound_user, outbound_secret)
        self._set_host_chap_settings(_api, host, **params)
        return self._result_message(host, "CHAP settings updated")

    @feature_command(
        feature="iscsi",
        status=True,
        description="Clear CHAP usernames and secrets",
        types=(Host,),
        hidden_substitue=dict(name="host"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def clear_chap(self, _api, host):
        self._update(_api, host, _entity_type=Host, _multi_field_put=True, returns=None,
            security_method="NONE",
            security_chap_inbound_username=Null,
            security_chap_inbound_secret=Null,
            security_chap_outbound_username=Null,
            security_chap_outbound_secret=Null,
            )
        return self._result_message(host, "CHAP settings cleared")

