from munch import Munch
import time

from infinishell.infishell import commandgroup, command, render
from infinishell.infishell.utils import strlist
from infinishell.infishell.types import YesNo

from ..system import system
from .types import Host, MetricsInterval, Initiator, MaintenanceModeDuration

from .options import QueryOptions, Grep, InvertedGrep
from .metrics import initiator_decimal_to_address, TopCounterBuilder
from . import EntityCommands, roles



class TopSanSupportAggregatorCounterBuilder(TopCounterBuilder):

    collector_type = "AGGREGATOR"
    collected_fields = ['throughput', 'ops', 'external_latency', 'internal_latency']
    aggregated_results = Munch()

    def _format_data(self):
        aggregated_results = Munch()
        collector_data = self.active_collectors_data[self.active_collectors['top'].id]
        data, fields = collector_data.data, collector_data.fields
        if data == [] or data[0] == []:  
            return
        else:
            results = self.aggregated_results
            intervals = len(data)
            for d in data:
                for row in d:
                    #get results[initiator][target] and create if doesnt exist
                    current_row = results.setdefault(row[0][0], Munch()).setdefault(row[0][1], Munch())
                    for ind,f in enumerate(fields):
                        if f == "FC_I_T":
                            current_row['initiator'] = '%x' % int(row[0][0])
                            current_row['target'] = '%x' % int(row[0][1])
                        else:
                            current_row.setdefault(f,0)
                            current_row[f] += row[ind]
            self.aggregated_results = results

    def _aggregate_results(self):
        results = self.aggregated_results
        final_data = []
        for initiator, targets in results.iteritems():
            row = Munch()
            row['initiator'] = '%x' % int(initiator)
            initiator_info = Initiator.get_by_name(system.api, row['initiator'])
            row['host'] = Host.get_by_id(system.api, initiator_info['host_id']).name
            for i in self.collected_fields:
                row[i] = 0
            nodes = Munch()
            for _, values in targets.iteritems():
                target_node = values['target'][-2]
                node_data = nodes.setdefault(target_node, Munch())
                node_data['target'] = values['target']
                for f in self.collected_fields:
                    node_data[f] = values[f]
            for n in ['1', '2', '3']:
                row["node{}".format(n)] = "ACTIVE" if nodes.get(n, None) else "INACTIVE"
                for f in self.collected_fields:
                    row["node{}_{}".format(n, f)] = nodes[n][f] if nodes.get(n, None) else None
            final_data.append(row)
        return final_data

    def get_data(self):
        self._get_data_from_collectors()
        self._format_data()
        return self._aggregate_results()



@commandgroup(prefix="support")
class SupportCommands(object):
	pass

@commandgroup(parent=SupportCommands, prefix="counters")
class SupportCommands(object):

    ELEMENT = ''
    ELEMENT_ID = "FC_I_T"

    FIELDS = strlist("host initiator node1 node2 node3 node1_ops node2_ops node3_ops node1_throughput node2_throughput node3_throughput node1_internal_latency node2_internal_latency node3_internal_latency node1_external_latency node2_external_latency node3_external_latency")

    @command(
        description="Show top metrics for FC initiators in system",
        types=(Host, MetricsInterval),
        options=(Grep(), InvertedGrep(),),
        )
    @render(
        fields=FIELDS,
        substitute=dict(
            ops="operations",
            node1_throughput="node1_tp",
            node2_throughput="node2_tp",
            node3_throughput="node3_tp",
            node1_internal_latency="node1_in_latency",
            node2_internal_latency="node2_in_latency",
            node3_internal_latency="node3_in_latency",
            node1_external_latency="node1_ex_latency",
            node2_external_latency="node2_ex_latency",
            node3_external_latency="node3_ex_latency",
            ),
        )
    @roles.INFINIDAT
    def show_active_paths(self, _api, host=None, interval=5):

        message = "Support - Top FC initiators and targets"
        builder = TopSanSupportAggregatorCounterBuilder(api=_api, grouping_field=self.ELEMENT_ID, message=message)
        filter_fields = dict(
            # pool_id=pool.id if pool else None,
            host_id=host.id if host else None,
            )
        builder.get_filters_params(protocol="FC", filter_fields=filter_fields)
        builder.get_collectors_params(sorting_field=None, max_results=None, 
                                      convert_data_function=initiator_decimal_to_address)
        builder.build()
        time.sleep(interval)
        return builder.get_data()


    @commandgroup(prefix="power_maintenance_mode", parent=SupportCommands)
    class PowerMaintenanceCommands(object):

        @command(
            description="Enable power maintenance mode or extend its duration",
            types=(MaintenanceModeDuration,),
        )
        @roles.INFINIDAT.TECHNICIAN
        def enable(self, _api, duration):
            _api.put('/system/power_maintenance_mode',data=dict(duration_seconds=duration *60, in_maintenance=True))
            return "Power Maintenance Mode is enabled for {} minutes.".format(duration)


        @command(
            description="Disable the Power Maintenance Mode",
            types=(YesNo,),
        )
        @roles.INFINIDAT.TECHNICIAN
        def disable(self, _api, force=None):
            _api.put('/system/power_maintenance_mode',data=dict(force=force, in_maintenance=False))
            return "Power Maintenance Mode is now disabled."


        @command(
            description="Query power maintenance mode"
        )
        @roles.INFINIDAT.TECHNICIAN
        def query(self, _api):
            res = _api.get('/system/power_maintenance_mode')
            if res.get('in_maintenance', False):
                if res.get('duration_seconds',0) > 0:
                    str1 = "Power Maintenance Mode is enabled for additional {} (hours:minutes:seconds).".format(time.strftime("%H:%M:%S", time.gmtime(res['duration_seconds'])))
                    str2 = "\nYou can extend the time by using support.power_maintenance_mode.enable"
                    str = str1 + str2
                else:
                    str = "Power Maintenance Mode is enabled although the timer elapsed."
                    "\n\nThis means that the Nodes are not properly protected by the BBUs."
                    "\nCheck the systems hardware ASAP."
            else:
                str = "Power Maintenance Mode is disabled."

            return str

