from munch import Munch
from itertools import chain

from infinishell.infishell import commandgroup, command, render
from infinishell.infishell.types import Str, YesNo, Port
from infinishell.infishell.utils import strlist
from infinishell.infishell.exceptions import BadArguments

from ....utils import input_password
from ..types import (
    Name,
    UsersRepository,
    LdapType,
    )
from ..options import QueryOptions
from ..utils import increased_api_timeout
from .. import roles, EntityCommands
from . import ConfigCommandsBase

from ...features import feature_command

@commandgroup(parent=ConfigCommandsBase, prefix="ldap")
class LdapConfigCommands(EntityCommands, ConfigCommandsBase):
    ENTITY_TYPE = UsersRepository
    QUERY_FIELDS = strlist("""
        name repository_type domain_name servers
        """)
    DETAILED_QUERY_FIELDS = strlist("""
        name repository_type domain_name bind_username
        user_class username_attribute users_basedn
        group_class group_name_attribute group_memberof_attribute groups_basedn
        use_ldaps ldap_port servers
        """)
    SORTABLE_FIELDS = strlist("name")
    QUERY_SUBSTITUTES = dict(
        ldap_port="port",
        )
    @command(
        description="List existing Active Directory servers",
        types=(UsersRepository, LdapType),
        hidden_substitue=dict(name="ldap"),
        substitute=dict(repo_type="type",),
        options=QueryOptions(fields=QUERY_FIELDS, infinidat_fields=DETAILED_QUERY_FIELDS, sortable=SORTABLE_FIELDS,  detailed=True),
        )
    @render(
        fields=QUERY_FIELDS,
        detailed_individually=True,
        detailed_fields=DETAILED_QUERY_FIELDS,
        substitute=QUERY_SUBSTITUTES,
        )
    def query(self, _api, ldap=None, repo_type=None):
        return self._query(_api, ldap, repository_type=repo_type, _hook=self._query_hook)

    def _query_hook(cls, results):
        return (cls._object_hook(r) for r in results)

    def _object_hook(cls, ldap):
        if "schema_definition" in ldap:
            ldap.update(ldap.schema_definition)
            del ldap.schema_definition
        ldap.pop("role_mappings", None)
        return ldap

    def _create_ldap_params(_self, username, password,
                                 schema_user_class, port, use_ldaps, schema_username_attribute, schema_user_basedn,
                                 schema_group_class, schema_group_attribute, schema_member_attribute, schema_group_basedn, domain=None,  server=None, **_):
        return Munch(
            domain_name=domain,
            servers=server,
            bind_username=username,
            bind_password=password,
            ldap_port=port,
            use_ldaps=use_ldaps,
            schema_definition=Munch((k, v) for k, v in dict(
                user_class=schema_user_class,
                username_attribute=schema_username_attribute,
                users_basedn=schema_user_basedn,
                group_class=schema_group_class,
                group_name_attribute=schema_group_attribute,
                group_memberof_attribute=schema_member_attribute,
                groups_basedn=schema_group_basedn,
                ).iteritems() if v is not None),
            )

    @feature_command(
        feature="openldap",
        status=False,
        name="define",
        description="Define Active Directory server",
        types=(Name, Str, Str, Str,
               Str, Port, YesNo, Str, Str,
               Str, Str, Str, Str),
        substitute=dict(
            repo_type="type",
            ),
        )
    @roles.INFINIDAT.ADMIN
    def old_define(self, _api, name, username, domain=None, password=None,
                           schema_user_class=None, port=None, use_ldaps=True, schema_username_attribute=None, schema_user_basedn=None,
                           schema_group_class=None, schema_group_attribute=None, schema_member_attribute=None, schema_group_basedn=None):
        if password is None:
            password = input_password('Please enter the password for {} "{}": '.format(self.ENTITY_TYPE.get_entity_name(), name),
                                      interaction_exc_type=BadArguments, interaction_error="Please specify a password")                
        params = Munch(
            name=name,
            repository_type="ActiveDirectory",
            )
        if port is None:
            port = 636 if use_ldaps else 389
        params.update(self._create_ldap_params(**locals()))
        with increased_api_timeout():
            return self._create(_api, **params)

    @feature_command(
        feature="openldap",
        status=True,
        description="Define Active Directory server",
        types=(Name, Str, LdapType, Str, Str, Str.List,
               Str, Port, YesNo, Str, Str,
               Str, Str, Str, Str),
        substitute=dict(
            repo_type="type",
            ),
        )
    @roles.INFINIDAT.ADMIN
    def define(self, _api, name, username, repo_type, domain=None, password=None, server=None,
                           schema_user_class=None, port=None, use_ldaps=True, schema_username_attribute=None, schema_user_basedn=None,
                           schema_group_class=None, schema_group_attribute=None, schema_member_attribute=None, schema_group_basedn=None):
        if password is None:
            password = input_password('Please enter the password for {} "{}": '.format(self.ENTITY_TYPE.get_entity_name(), name),
                                      interaction_exc_type=BadArguments, interaction_error="Please specify a password")
        if repo_type == "ACtiveDirectory" and server:
            raise BadArguments, "server parameter not allowed with repo_type ACtiveDirectory"                    
        if repo_type == "LDAP" and domain:
            raise BadArguments, "domain parameter not allowed with repo_type LDAP"                    
        params = Munch(
            name=name,
            repository_type=repo_type,
            )
        if port is None:
            port = 636 if use_ldaps else 389
        params.update(self._create_ldap_params(**locals()))
        with increased_api_timeout():
            return self._create(_api, **params)

    @command(
        description="Modify Active Directory server",
        types=(UsersRepository, Str, Str, Str.List,
               Str, Port, YesNo, Str, Str,
               Str, Str, Str, Str),
        hidden_substitue=dict(name="ldap"),
        )
    @roles.INFINIDAT.ADMIN
    def modify(self, _api, ldap, username=None, password=None, server=None,
                           schema_user_class=None, port=None, use_ldaps=None, schema_username_attribute=None, schema_user_basedn=None,
                           schema_group_class=None, schema_group_attribute=None, schema_member_attribute=None, schema_group_basedn=None):
        if all(arg is None for arg in (
            username, password, server,
            schema_user_class, port, use_ldaps, schema_username_attribute, schema_user_basedn,
            schema_group_class, schema_group_attribute, schema_member_attribute, schema_group_basedn,
            )):
            raise BadArguments("At least one more argument must be specified")
        if port is None:
            port = 636 if (use_ldaps is True) else 389 if (use_ldaps is False) else None
        if ldap.repository_type == "ActiveDirectory" and server:
            raise BadArguments, "server parameter is not allowed for ActiveDirectory repository"
        if username is not None and password is None:
            password = input_password('Please enter the password for {} "{}": '.format(self.ENTITY_TYPE.get_entity_name(), ldap.name),
                                      interaction_exc_type=BadArguments, interaction_error="Please specify a password")
        params = self._create_ldap_params(**locals())
        # Since schema definition must be specified as a whole, it's values are
        # based on the existing definition, and are updated with the specified
        # definition
        params.schema_definition = Munch(chain(
            ldap.schema_definition.iteritems(),
            params.schema_definition.iteritems(),
            ))
        with increased_api_timeout():
            if use_ldaps is not None and port is not None:
                self._update(_api, ldap, _multi_field_put=True,
                             use_ldaps=use_ldaps, ldap_port=port)
                del params['use_ldaps']
                del params['ldap_port']
            return self._update(_api, ldap, **params)

    @command(
        description="Rename Active Directory server",
        types=(UsersRepository, Name),
        hidden_substitue=dict(name="ldap"),
        )
    @roles.INFINIDAT.ADMIN
    def rename(self, _api, ldap, new_name):
        return self._rename(_api, ldap, new_name)

    @command(
        description="Undefine Active Directory server",
        types=(UsersRepository,),
        hidden_substitue=dict(name="ldap"),
        )
    @roles.INFINIDAT.ADMIN
    def remove(self, _api, ldap):
        return self._delete(_api, ldap)

    @command(
        description="Test defined Active Directory server",
        types=(UsersRepository,),
        hidden_substitue=dict(name="ldap"),
        )
    @roles.INFINIDAT.ADMIN
    def test(self, _api, ldap):
        self._post(_api, ldap, "test")
        return self._result_message(ldap, "tested successfully")

    @command(
        description="Set LDAP & Active Directory resolution order",
        types=(UsersRepository.List,),
        substitute=dict(repos="order"),
        )
    @roles.INFINIDAT.ADMIN
    def set_order(self, _api, repos):
        self._post(_api, None, "set_order", data=[r.id for r in repos])
        return "Order set"

    @command(
        description="LDAP & Active Directory resolution order",
        types=(),
        substitute=dict(search_order="index"),
        )
    @render(
        fields=strlist('''search_order name'''),
        substitute=dict(search_order="index"),
        )
    def order_query(self, _api):
        return self._query(_api, _hook=self._order_query_hook,_default_sort=_api.create_sort().asc("search_order"))

    def _order_query_hook(cls, results):
        def incr_search_order(x):
            x.search_order += 1
            return x
        return map(incr_search_order, results)

    @command(
        description="Flush LDAP authentication cache",
        )
    @roles.INFINIDAT.ADMIN
    def flush_cache(self, _api):
        self._post(_api, None, "reload")
        return "LDAP authentication cache flushed"
