from functools import partial
from collections import defaultdict
from munch import Munch

from infinishell.infishell import runtime

from infinishell.infishell.exceptions import BadArguments, BadValue
from infinishell.infishell.commands import commandgroup, command, deprecated
from infinishell.infishell.types import YesNo
from infinishell.infishell.utils import strlist

from ..exceptions import ManagementException
from ..renderer import render_capacity
from .hierarchical_base import HierarchicalEntityCommandsBase
from .types import (
    QOSPolicy,
    QOSAssociatedEntity,
    Pool,
    Volume,
    VolumeSnapshot,
    VolumeSnapshotNotWP,
    VolumeSnapshotWP,
    VolumeOrSnapshot,
    VolumeOrSnapshotWP,
    VolumeOrSnapshotNotWP,
    MasterVolume,
    ConsistencyGroup,
    SnapGroup,
    Cluster,
    Host,
    Lun,
    Name,
    Size,
    SizeDelta,
    SizePositiveDelta,
    CapacityUnit,
    VolumeSeries,
    ReplicaRoleOrNone,
    Udid,
    QueryUdid,
    UdidUpdate,
    NONE_CONSTANT,
    )
from .options import QueryOptions
from .mapping import MappingCommandsBase
from .utils import get_target_pool, format_savings
from . import LazyIdNameMapping, render_collection, roles, set_rendered_unit, set_renderer_query_fields, Null

from ..system import system
from ..features import feature_command, feature_exists

VOLUME_QUERY_FIELDS           = strlist("name             thin size used allocated tree_allocated pool write_protected compression_enabled data_reduction mapped")
DETAILED_VOLUME_QUERY_FIELDS  = strlist("name      serial thin size used allocated tree_allocated pool cg policy write_protected compression_enabled data_reduction mapped ssd_enabled replication_role created_at updated_at")


SNAP_QUERY_FIELDS             = strlist("name             thin size used allocated                pool write_protected                                    mapped                              created_at")
DETAILED_SNAP_QUERY_FIELDS    = strlist("name      serial thin size used allocated                pool sg write_protected                                    mapped policy ssd_enabled replication_role created_at updated_at parent")

CLONE_QUERY_FIELDS            = strlist("name             thin size used allocated tree_allocated pool write_protected mapped                              created_at")
DETAILED_CLONE_QUERY_FIELDS   = DETAILED_VOLUME_QUERY_FIELDS

VOLUME_TREE_FIELDS            = strlist("name type             size used allocated                pool write_protected                                                                        created_at")
DETAILED_VOLUME_TREE_FIELDS   = strlist("name type serial thin size used allocated tree_allocated pool write_protected compression_enabled data_reduction mapped ssd_enabled replication_role created_at updated_at")

MAP_QUERY_FIELDS = strlist("name target_type target_name luns")
DETAILED_MAP_QUERY_FIELDS = strlist("name target_type target_name luns udid")
MAP_QUERY_FIELDS_SUBSTITUTES = dict(
    luns="lun_id",
    target_name="mapped_to",
    target_type="mapping_type",
    )

SORTABLE_FIELDS = strlist("name serial size write_protected mapped ssd_enabled created_at updated_at compression_enabled")

QUERY_SUBSTITUTES = dict(
    write_protected="wp",
    ssd_enabled="ssd_cache",
    compression_enabled="compressed",
    tree_allocated="Snapshots",
    )

DETAILED_QUERY_SUBSTITUTES = dict(
    write_protected="write_protected",
    )

class VolumeCommandsBase(HierarchicalEntityCommandsBase):
    HIERARCHICAL_TYPE = Volume
    VOLUME_TYPE_NAMES = Munch(
        MASTER="volume",
        SNAPSHOT="snapshot",
        )

    def _query_hook(self, volumes, _api, pool_names_by_id=None):
        has_rmr_fields = None
        cgs = LazyIdNameMapping(ConsistencyGroup, _api)
        snap_groups = LazyIdNameMapping(SnapGroup, _api)
        for v in volumes:
            if pool_names_by_id:
                v.pool = pool_names_by_id[v.pool_id]
            v.cg = cgs.get(v.cg_id) if v.cg_id else None
            if v.cg == None or 'UNKNOWN' in v.cg:
                v.sg = snap_groups.get(v.cg_id) if v.cg_id else None
            v.type = self._get_volume_type(v)
            v.thin = v.provtype == self.THIN
            if has_rmr_fields is None:
                has_rmr_fields = "rmr_source" in v and "rmr_target" in v
            if has_rmr_fields:
                v.replication_role = ReplicaRoleOrNone.SOURCE if v.rmr_source \
                                     else ReplicaRoleOrNone.TARGET if v.rmr_target \
                                     else None
            if 'capacity_savings' in v:
                v.data_reduction = format_savings(v)
        return volumes

    def _get_volume_type(self, volume):
        if volume.type == "MASTER":
            return "VOLUME"
        return volume.type

    def _get_volume_type_name(self, volume):
        return self.VOLUME_TYPE_NAMES.get(volume.type) or "volume"

def create_common_volume_commands_mixin(VolumeType):

    class CommonVolumeCommandsMixin(MappingCommandsBase):

        @command(
            description="Rename {}".format(VolumeType.DESCRIPTION_SUBJECT),
            types=(VolumeType, Name),
            hidden_substitue=dict(name="vol"),
            substitute=dict() if VolumeType == MasterVolume else dict(vol="snap"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def rename(self, _api, vol, new_name):
            return self._rename(_api, vol, new_name)

        @command(
            description="Delete {}".format(VolumeType.DESCRIPTION_SUBJECT),
            types=(VolumeType.List,),
            hidden_substitue=dict(name="vol"),
            substitute=dict() if VolumeType == MasterVolume else dict(vol="snap"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def delete(self, _api, vol):
            return self._delete(_api, vol)

        @command(
            description="Map {} to host or cluster".format(VolumeType.DESCRIPTION_SUBJECT),
            types=(VolumeType.List, Host, Cluster, Lun),
            hidden_substitue=dict(name="vol"),
            substitute=dict() if VolumeType == MasterVolume else dict(vol="snap"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def map(self, _api, vol, host=None, cluster=None, lun=None):
            if host and cluster or not (host or cluster):
                raise BadArguments("Either host or cluster must be specified")
            if lun and len(vol) != 1:
                raise BadArguments("LUN may be specified only when mapping a single " + VolumeType.get_entity_name())
            mappable_type, mappable = self._get_mappable(host, cluster)
            return self._map(_api, mappable_type, mappable, vol, lun)

        @command(
            description="Unmap {} from host or cluster".format(VolumeType.DESCRIPTION_SUBJECT),
            types=(VolumeType.List, Host, Cluster),
            hidden_substitue=dict(name="vol"),
            substitute=dict() if VolumeType == MasterVolume else dict(vol="snap"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def unmap(self, _api, vol, host=None, cluster=None):
            if not host and not cluster:
                raise BadArguments("Either host or cluster must be specified")
            mappable_type, mappable = self._get_mappable(host, cluster)
            return self._unmap(_api, mappable_type, mappable, vol)

        def _get_mappable(self, host, cluster):
            if host:
                return Host, host
            if cluster:
                return Cluster, cluster
            raise AssertionError("Undefined mappable type")


        @command(
            description="List host and cluster mappings by volume",
            types=(VolumeType.List, Pool, Host, Cluster),
            hidden_substitue=dict(name="vol"),
            substitute=dict() if VolumeType == MasterVolume else dict(vol="snap"),
            options=QueryOptions(fields=MAP_QUERY_FIELDS + strlist('udid'), substitute=MAP_QUERY_FIELDS_SUBSTITUTES, detailed=True),
            )
        @render_collection(
            entity=Volume,
            fields=MAP_QUERY_FIELDS,
            substitute=MAP_QUERY_FIELDS_SUBSTITUTES,
            detailed_fields=DETAILED_MAP_QUERY_FIELDS,
            detailed_substitute=MAP_QUERY_FIELDS_SUBSTITUTES,
            detailed_individually=True,
            )
        def map_query(self, _api, vol=None, pool=None, host=None, cluster=None):
            if host and cluster:
                raise BadArguments("Either host, cluster, or neither may be specified")
            show_all_mappables = all(mappable is None for mappable in (host, cluster))
            def get_mappables(mappable_type, specified_mappable):
                return mappable_type.iter_all(_api) if show_all_mappables else (specified_mappable,) if specified_mappable else ()
            mappings = defaultdict(partial(defaultdict, partial(defaultdict, Munch)))
            mapping_infos = defaultdict(list)
            for mappable_type, specified_mappable, lun_predicate in (
                (Host,       host,      lambda lun: not lun.clustered),
                (Cluster,    cluster,   None),
                ):
                for mappable in get_mappables(mappable_type, specified_mappable):
                    for lun in mappable.luns:
                        if not lun_predicate or lun_predicate(lun):
                            mapping_info = mappings[lun.volume_id][mappable_type][mappable.id]
                            if not mapping_info:
                                mapping_info.target_name = mappable.name
                                mapping_info.target_type = mappable_type.get_entity_name().upper()
                                mapping_info.luns = []
                                mapping_infos[lun.volume_id].append(mapping_info)
                            mapping_info.luns.append(lun.lun)
            return self._query(_api, vol, pool_id=pool, _hook=partial(self._map_query_hook, mapping_infos=mapping_infos))

        def _map_query_hook(self, volumes, mapping_infos):
            for v in volumes:
                for info in sorted(mapping_infos.get(v.id, ()), key=lambda i: i.target_name):
                    info.name = v.name
                    info.udid = v.udid
                    info.luns.sort()
                    yield info

        @command(
            description="Enable or disable SSD cache",
            types=(VolumeType, YesNo, YesNo),
            hidden_substitue=dict(name="vol"),
            substitute=dict() if VolumeType == MasterVolume else dict(vol="snap"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def cache(self, _api, vol, ssd_cache, recursive=False):
            return self._update_recursive(_api,
                entity=vol,
                recurse=recursive,
                success_msg="{} cached on SSD".format("now" if ssd_cache else "no longer"),
                success_msg_with_is=True,
                noop_msg="already{} cached on SSD".format("" if ssd_cache else " not"),
                noop_msg_with_is=True,
                props=dict(
                    ssd_enabled=ssd_cache,
                    ),
                predicate=lambda dataset: dataset.ssd_enabled != ssd_cache,
                    )

        @feature_command(
            feature='qos',
            status=True,
            description="Assign volume to policy",
            types=(VolumeType.List, QOSPolicy),
            hidden_substitue=dict(name="vol"),
            substitute=dict() if VolumeType == MasterVolume else dict(vol="snap"),
            )
        @roles.INFINIDAT.ADMIN
        def assign_qos(self, _api, vol, policy):
            with _api.aggregation() as aggregator:
                for v in vol:
                    aggregator.next_task(v.name, "Assigned")
                    data = dict(entity_id=v.id)
                    result = _api.post("{}/{}/assigned_entities".format(QOSPolicy.get_uri(), policy.id), data)
            if len(vol) == 1:
                return "Volume '{}' was assigned to policy '{}'".format(vol[0].name, policy.name)
            else:
                return "{} volumes were assigned to policy '{}'".format(len(vol), policy.name)
        
        @feature_command(
            feature='qos',
            status=True,
            description="Unassign volume from policy",
            types=(VolumeType.List,),
            hidden_substitue=dict(name="vol"),
            substitute=dict() if VolumeType == MasterVolume else dict(vol="snap"),
            )
        @roles.INFINIDAT.ADMIN
        def unassign_qos(self, _api, vol):
            with _api.aggregation() as aggregator:
                for v in vol:
                    if not v.qos_policy_id:
                        raise BadArguments("Volume '{}' does not belong to QoS policy".format(v.name))
                    data = dict(entity_id=v.id)
                    aggregator.next_task(v.name, "Unassigned")
                    result = _api.post("{}/{}/unassign".format(QOSPolicy.get_uri(), v.qos_policy_id), data)
            if len(vol) == 1:
                return "Volume '{}' was unassigned from policy '{}'".format(v.name, v.qos_policy_name)
            else:
                return "{} volumes were unassigned from their policies".format(len(vol))


    return CommonVolumeCommandsMixin

def create_writable_volume_commands_mixin(VolumeType):

    class WritableVolumeCommandsMixin(object):

        @command(
            description="Resize {}".format(VolumeType.DESCRIPTION_SUBJECT),
            types=(VolumeType, SizePositiveDelta),
            hidden_substitue=dict(name="vol"),
            substitute=dict() if VolumeType == MasterVolume else dict(vol="snap"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def resize(self, _api, vol, size):
            if vol.type == "SNAPSHOT" and vol.write_protected:
                raise BadValue("Snapshot {}: Cannot resize a write protected snapshot".format(vol.name))
            size = size.compute_size(vol.size)
            if vol.size == size:
                return self._result_message(vol, "size is already {}".format(render_capacity(vol.size)))
            self._update(_api, vol, size=size)
            return self._result_message(vol, "resized")

        @command(
            description="Restore data state of {} from previous snapshot".format(VolumeType.DESCRIPTION_SUBJECT),
            types=(VolumeType, VolumeSnapshot),
            hidden_substitue=dict(target="vol"),
            substitute=dict() if VolumeType == MasterVolume else dict(vol="snap"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def restore(self, _api, vol, source):
            if vol.id == source.id:
                raise BadValue('"{}" was specified as both arguments'.format(vol.name))
            data = dict(
                source_id=source.id
                )
            self._post(_api, vol, "restore", data=data)
            return self._result_message(vol, 'restored from {} "{}"'.format(self._get_volume_type_name(source), source.name))

        @command(
            description="Disable write access on {}".format(VolumeType.DESCRIPTION_SUBJECT),
            types=(VolumeType, YesNo),
            hidden_substitue=dict(name="vol"),
            substitute=dict() if VolumeType == MasterVolume else dict(vol="snap"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def write_protect(self, _api, vol, recursive=False):
            return self._update_recursive(_api,
                entity=vol,
                recurse=recursive,
                success_msg="now write-protected",
                success_msg_with_is=True,
                noop_msg="already write-protected",
                noop_msg_with_is=True,
                props=dict(
                    write_protected=True,
                    ),
                predicate=lambda dataset: dataset.type != "SNAP" and not dataset.write_protected,
                )

        @feature_command(
            feature='openvms',
            status=True,
            description="Change UDID on a {}".format(VolumeType.DESCRIPTION_SUBJECT),
            types=(VolumeType, UdidUpdate),
            hidden_substitue=dict(name="vol"),
            substitute=dict() if VolumeType == MasterVolume else dict(vol="snap"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def set_udid(self, _api, vol, udid):
            return self._update(_api, vol, udid=Null if udid == NONE_CONSTANT else udid)

        @command(
            description="Enable write access on {}".format(VolumeType.DESCRIPTION_SUBJECT),
            types=(VolumeType, YesNo),
            hidden_substitue=dict(name="vol"),
            substitute=dict() if VolumeType == MasterVolume else dict(vol="snap"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def write_enable(self, _api, vol, recursive=False):
            return self._update_recursive(_api,
                entity=vol,
                recurse=recursive,
                success_msg="now write-enabled",
                success_msg_with_is=True,
                noop_msg="already write-enabled",
                noop_msg_with_is=True,
                props=dict(
                    write_protected=False,
                    ),
                predicate=lambda dataset: dataset.type != "SNAP" and dataset.write_protected,
                )

    return WritableVolumeCommandsMixin

def create_child_volume_commands_mixin(VolumeType, ParentType, parent_arg, create_parent_arg):

    class ChildVolumeCommandsMixin(object):

        @command(
            description="Create a new {} from {}".format(VolumeType.DESCRIPTION_SUBJECT, ParentType.DESCRIPTION_SUBJECT),
            types=(ParentType, Name, YesNo),
            hidden_substitue=dict(source="vol"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def create(self, _api, vol, name, ssd_cache=None):
            parent = self._create_and_return_entity(_api, parent_id=vol.id, name=name, ssd_enabled=ssd_cache)
            return self._result_message(parent, "created", subject=VolumeType.get_entity_name())

        @command(
            description="Display hierarchy of snapshots",
            types=(VolumeType, CapacityUnit),
            hidden_substitue=dict(name="vol"),
            substitute=dict() if VolumeType == MasterVolume else dict(vol="snap"),
            options=QueryOptions(fields=DETAILED_VOLUME_TREE_FIELDS, sortable=VOLUME_TREE_FIELDS, substitute=DETAILED_QUERY_SUBSTITUTES),
            )
        @render_collection(
            entity=Volume,
            fields=VOLUME_TREE_FIELDS,
            substitute=QUERY_SUBSTITUTES,
            )
        def tree(self, _api, vol, unit=None):
            set_rendered_unit(unit)
            set_renderer_query_fields()
            return self._get_tree(_api, vol)

    return ChildVolumeCommandsMixin

@commandgroup(prefix="vol")
class VolumeCommands(VolumeCommandsBase,
                     create_common_volume_commands_mixin(MasterVolume),
                     create_writable_volume_commands_mixin(MasterVolume)):
    ENTITY_TYPE = MasterVolume

    @feature_command(
        feature='openvms',
        status=False,
        name='query',
        description="List existing volumes",
        types=(MasterVolume.List, ConsistencyGroup, Pool, CapacityUnit, YesNo, YesNo, YesNo, YesNo, YesNo),
        hidden_substitue=dict(name="vol"),
        options=QueryOptions(fields=DETAILED_VOLUME_QUERY_FIELDS + strlist("cg_id cg"), sortable=SORTABLE_FIELDS, substitute=QUERY_SUBSTITUTES, detailed=True),
        )
    @render_collection(
        entity=Volume,
        fields=VOLUME_QUERY_FIELDS,
        detailed_fields=DETAILED_VOLUME_QUERY_FIELDS,
        detailed_substitute=DETAILED_QUERY_SUBSTITUTES,
        detailed_individually=True,
        substitute=QUERY_SUBSTITUTES,
        alignment=dict(
            data_reduction="right",
            ),
        )
    def old_query(self, _api, vol=None, cg=None, pool=None, unit=None, mapped=None, write_protected=None, thin=None, compression_enabled=None, ssd_enabled=None):
        set_rendered_unit(unit)
        provtype = None
        if thin is not None:
            provtype = self.THIN if thin else self.THICK
        return self._common_query(_api, vol, cg_id=cg, pool_id=pool, mapped=mapped, write_protected=write_protected, provtype=provtype, compression_enabled=compression_enabled, ssd_enabled=ssd_enabled)

    @feature_command(
        feature='openvms',
        status=True,
        description="List existing volumes",
        types=(MasterVolume.List, ConsistencyGroup, Pool, CapacityUnit, YesNo, YesNo, YesNo, YesNo, YesNo, QueryUdid),
        hidden_substitue=dict(name="vol"),
        options=QueryOptions(fields=DETAILED_VOLUME_QUERY_FIELDS + strlist("cg_id cg udid"), sortable=SORTABLE_FIELDS, substitute=QUERY_SUBSTITUTES, detailed=True),
        )
    @render_collection(
        entity=Volume,
        fields=VOLUME_QUERY_FIELDS,
        detailed_fields=DETAILED_VOLUME_QUERY_FIELDS + strlist('cg udid'),
        detailed_substitute=DETAILED_QUERY_SUBSTITUTES,
        detailed_individually=True,
        substitute=QUERY_SUBSTITUTES,
        alignment=dict(
            data_reduction="right",
            ),
        )
    def query(self, _api, vol=None, cg=None, pool=None, unit=None, mapped=None, write_protected=None, thin=None, compression_enabled=None, ssd_enabled=None, udid=None):
        set_rendered_unit(unit)
        provtype = None
        if thin is not None:
            provtype = self.THIN if thin else self.THICK
        return self._common_query(_api, vol, cg_id=cg, pool_id=pool, mapped=mapped, write_protected=write_protected, provtype=provtype, compression_enabled=compression_enabled, ssd_enabled=ssd_enabled, udid=udid)

    def _query_hook(self, entities, _api, pool_names_by_id=None):
        for e in entities:
            if system.user.has_role('infinidat','technician'):
                if e.get('compression_enabled', False) and e.get('compression_suppressed', False):
                    e.compression_enabled = "yes (suppressed)"
            e.policy = e.qos_policy_name
        return super(VolumeCommands, self)._query_hook(entities, _api, pool_names_by_id)
        
    @feature_command(
        feature='openvms',
        status=False,
        name='create',
        description="Create a new volume",
        types=(Name.List, Size, Pool, YesNo, YesNo, YesNo, VolumeSeries),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def old_create(self, _api, name, size, pool=None, thin=False, ssd_cache=None, compression=None, series=None):
        pool = get_target_pool(_api, pool)
        if series:
            # In case of multiple volumes, since this is not an atomic
            # operation, ensure there's enough free space for all requested
            # volumes.
            total_size = size * series.total
            if total_size > pool.free_virtual_space:
                raise ManagementException(409, Munch(code="POOL_VIRTUAL_CAPACITY", message="Pool virtual capacity reached."))
            if not thin and total_size > pool.free_physical_space:
                raise ManagementException(409, Munch(code="POOL_PHYSICAL_CAPACITY", message="Pool physical capacity reached."))
        return self._create_single_or_multiple_or_series(_api, name, series,
            size=size,
            pool_id=pool.id,
            provtype=self._get_provtype(thin),
            ssd_enabled=ssd_cache,
            compression_enabled=compression,
            )

    @feature_command(
        feature='openvms',
        status=True,
        description="Create a new volume",
        types=(Name.List, Size, Pool, YesNo, YesNo, YesNo, VolumeSeries, Udid),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def create(self, _api, name, size, pool=None, thin=False, ssd_cache=None, compression=None, series=None, udid=None):
        pool = get_target_pool(_api, pool)
        if series:
            # In case of multiple volumes, since this is not an atomic
            # operation, ensure there's enough free space for all requested
            # volumes.
            total_size = size * series.total
            if total_size > pool.free_virtual_space:
                raise ManagementException(409, Munch(code="POOL_VIRTUAL_CAPACITY", message="Pool virtual capacity reached."))
            if not thin and total_size > pool.free_physical_space:
                raise ManagementException(409, Munch(code="POOL_PHYSICAL_CAPACITY", message="Pool physical capacity reached."))
        return self._create_single_or_multiple_or_series(_api, name, series,
            size=size,
            pool_id=pool.id,
            provtype=self._get_provtype(thin),
            ssd_enabled=ssd_cache,
            compression_enabled=compression,
            udid=udid,
            )        

    @command(
        description="Move volume to a different pool",
        types=(MasterVolume, Pool, YesNo),
        hidden_substitue=dict(name="vol"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def move(self, _api, vol, pool, move_capacity=False):
        if vol.pool_id == pool.id:
            return self._result_message(vol, 'already in {} "{}"'.format(Pool.get_entity_name(), pool.name))
        _api.post(self.ENTITY_TYPE.get_uri(vol.id, "move"), dict(
            pool_id=pool.id,
            with_capacity=move_capacity,
            ))
        return self._result_message(vol, 'moved to pool "{}" {} its capacity'
                                            .format(pool.name, "along with" if move_capacity else "without"))

    @command(
        description="Set volume provisioning type",
        types=(MasterVolume, YesNo),
        hidden_substitue=dict(name="vol"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def set_thin(self, _api, vol, thin):
        prov_name = "thin" if thin else "thick"
        if vol.provtype.upper() == self._get_provtype(thin):
            return self._result_message(vol, "already set to {} provisioning".format(prov_name))
        self._update(_api, vol, provtype=self._get_provtype(thin))
        return self._result_message(vol, "set to {} provisioning".format(prov_name))

    @command(
        description="Set volume compression",
        types=(MasterVolume, YesNo),
        hidden_substitue=dict(name="vol"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN.TECHNICIAN
    def set_compression(self, _api, vol, compression):
        if vol.compression_enabled  == compression:
            return self._result_message(vol, "compression is already {}".format("enabled" if compression else "disabled"))
        self._update(_api, vol, compression_enabled=compression)
        return 'Compression was {} for volume "{}"'.format("enabled" if compression else "disabled", vol.name)

    @command(
        description="Estimate reclaimed capacity (deletion simulation)",
        types=(Volume,),
        hidden_substitue=dict(name="vol"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def reclaimable(self, _api, vol):
        data = dict(entities=[vol.id])
        result = _api.post(Volume.get_uri("delete_simulation"), data)
        return 'Estimated reclaimable space for volume "{}" and its tree descendants: {}'.format(vol.name, render_capacity(result.space_reclaimable))

    @command(
        description="Display hierarchy of volumes and snapshots",
        types=(MasterVolume, Pool, CapacityUnit),
        hidden_substitue=dict(name="vol"),
        options=QueryOptions(fields=DETAILED_VOLUME_TREE_FIELDS, sortable=VOLUME_TREE_FIELDS, substitute=DETAILED_QUERY_SUBSTITUTES),
        )
    @render_collection(
        entity=Volume,
        fields=VOLUME_TREE_FIELDS,
        substitute=QUERY_SUBSTITUTES,
        )
    def tree(self, _api, vol=None, pool=None, unit=None):
        set_rendered_unit(unit)
        set_renderer_query_fields()
        return self._get_tree(_api, vol, pool)

@commandgroup(prefix="snap", parent=VolumeCommands)
class VolumeSnapshotCommands(VolumeCommandsBase,
                             create_common_volume_commands_mixin(VolumeSnapshot),
                             create_writable_volume_commands_mixin(VolumeSnapshot),
                             create_child_volume_commands_mixin(
                                 VolumeType=VolumeSnapshot,
                                 ParentType=VolumeOrSnapshot,
                                 parent_arg="vol",
                                 create_parent_arg="vol",
                                 )):
    ENTITY_TYPE = VolumeSnapshot

    @feature_command(
        feature='openvms',
        status=False,
        name='query',
        description="List existing snapshots",
        types=(VolumeSnapshot.List, VolumeOrSnapshot, SnapGroup, Pool, CapacityUnit, YesNo, YesNo),
        hidden_substitue=dict(name="snap"),
        options=QueryOptions(fields=DETAILED_SNAP_QUERY_FIELDS + strlist("master"), sortable=SORTABLE_FIELDS, substitute=QUERY_SUBSTITUTES, detailed=True),
        )
    @render_collection(
        entity=Volume,
        fields=SNAP_QUERY_FIELDS,
        substitute=QUERY_SUBSTITUTES,
        detailed_individually=True,
        detailed_fields=DETAILED_SNAP_QUERY_FIELDS,
        detailed_substitute=DETAILED_QUERY_SUBSTITUTES,
        )
    def old_query(self, _api, snap=None, source=None, sg=None, pool=None, unit=None, mapped=None, write_protected=None):
        set_rendered_unit(unit)
        return self._common_query(_api, snap, parent_id=source, cg_id=sg, pool_id=pool, type=VolumeSnapshot.ENTITY_MGMT_SUBTYPE, mapped=mapped, write_protected=write_protected)

    @feature_command(
        feature='openvms',
        status=True,
        description="List existing snapshots",
        types=(VolumeSnapshot.List, VolumeOrSnapshot, SnapGroup, Pool, CapacityUnit, YesNo, YesNo, QueryUdid),
        hidden_substitue=dict(name="snap"),
        options=QueryOptions(fields=DETAILED_SNAP_QUERY_FIELDS + strlist("master udid"), sortable=SORTABLE_FIELDS + strlist('udid'), substitute=QUERY_SUBSTITUTES, detailed=True),
        )
    @render_collection(
        entity=Volume,
        fields=SNAP_QUERY_FIELDS,
        substitute=QUERY_SUBSTITUTES,
        detailed_individually=True,
        detailed_fields=DETAILED_SNAP_QUERY_FIELDS + strlist('udid'),
        detailed_substitute=DETAILED_QUERY_SUBSTITUTES,
        )
    def query(self, _api, snap=None, source=None, sg=None, pool=None, unit=None, mapped=None, write_protected=None, udid=None):
        set_rendered_unit(unit)
        return self._common_query(_api, snap, parent_id=source, cg_id=sg, pool_id=pool, type=VolumeSnapshot.ENTITY_MGMT_SUBTYPE, mapped=mapped, write_protected=write_protected, udid=udid)

    def _query_hook(self, entities, _api, pool_names_by_id=None):
        for e in entities:
            e.policy = e.qos_policy_name
            if runtime.options.get('detailed', False) or runtime.options.get('columns'):
                e.parent = VolumeOrSnapshot.get_by_id(_api, e.parent_id).name
                if e.family_id:
                    e.master = VolumeOrSnapshot.get_by_id(_api, e.family_id).name
                else:
                    e.master = None
        return super(VolumeSnapshotCommands, self)._query_hook(entities, _api, pool_names_by_id)

    @command(
        description="Create a consistent set of snapshots of multiple volumes",
        types=(VolumeOrSnapshotWP.List, Name, Name),
        )
    @roles.INFINIDAT
    def group_create(self, _api, source, snap_prefix=None, snap_suffix=None):
        if not snap_suffix and not snap_prefix:
            raise BadArguments("Either snap_prefix, snap_suffix, or both must be specified")
        _api.post(MasterVolume.get_uri("group_snapshot"), dict(
            entities=[dict(id=entity.id) for entity in source],
            snap_prefix=snap_prefix,
            snap_suffix=snap_suffix,
            ))
        return "Consistent snapshots created"

    @command(
        description='Refresh a snapshot',
        types=(VolumeSnapshot, YesNo),
        auto_complete_hidden_args=['force'],
        hidden_substitue=dict(name="snap"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def refresh(self, _api, snap, force=None):
        if force:
            parent = Volume.get_by_id(_api, snap.parent_id)
            if not parent.get('rmr_target'):
                force = False

        _api.post(self.ENTITY_TYPE.get_uri(snap.id, 'refresh')+ ("?force_if_replicated_on_target=True" if force else ''),
                  {'source_id': snap.parent_id})
        return self._result_message(snap, 'refreshed')


@commandgroup(prefix="clone", parent=VolumeCommands, hidden=True)
class VolumeCloneCommands(object):


    @command(
        description="List existing clones",
        types=(VolumeOrSnapshot.List, VolumeOrSnapshot, ConsistencyGroup, Pool, CapacityUnit, YesNo),
        hidden_substitue=dict(name="clone"),
        options=QueryOptions(fields=DETAILED_CLONE_QUERY_FIELDS, sortable=SORTABLE_FIELDS, substitute=QUERY_SUBSTITUTES, detailed=True),
        )
    @render_collection(
        entity=Volume,
        fields=CLONE_QUERY_FIELDS,
        detailed_fields=DETAILED_CLONE_QUERY_FIELDS,
        detailed_substitute=DETAILED_QUERY_SUBSTITUTES,
        detailed_individually=True,
        substitute=QUERY_SUBSTITUTES,
        )
    @deprecated
    def query(self, _api, clone=None, snap=None, cg=None, pool=None, unit=None, mapped=None):
        set_rendered_unit(unit)
        data = VolumeSnapshotCommands().query(_api, snap=clone, source=snap, sg=cg, pool=pool, unit=unit, mapped=mapped, write_protected=False)
        return data        

    @command(
        description="Resize clone",
        types=(VolumeOrSnapshot, SizeDelta),
        hidden_substitue=dict(name="clone"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    @deprecated
    def resize(self, _api, clone, size):
        data = VolumeSnapshotCommands().resize(_api, vol=clone, size=size)
        return data

    @command(
        description="Restore data state of clone from previous snapshot",
        types=(VolumeOrSnapshot, VolumeSnapshotWP),
        hidden_substitue=dict(name="clone"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    @deprecated
    def restore(self, _api, clone, source):
        data = VolumeSnapshotCommands().restore(_api, vol=clone, source=source)
        return data
 
    @command(
        description="Disable write access on clone",
        types=(VolumeOrSnapshot, YesNo),
        hidden_substitue=dict(name="clone"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    @deprecated
    def write_protect(self, _api, clone, recursive=False):
        data = VolumeSnapshotCommands().write_protect(_api, vol=clone, recursive=recursive)
        return data
 
    @command(
        description="Enable write access on clone",
        types=(VolumeOrSnapshot, YesNo),
        substitute=dict(volume="name"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    @deprecated
    def write_enable(self, _api, volume, recursive=False):
        data = VolumeSnapshotCommands().write_enable(_api, vol=volume, recursive=recursive)
        return data

    @command(
        description="Create a new clone from snapshot",
        types=(VolumeOrSnapshot, Name, YesNo),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    @deprecated
    def create(self, _api, snap, name, ssd_cache=None):
        data = VolumeSnapshotCommands().create(_api, vol=snap, name=name, ssd_cache=ssd_cache)
        clone = VolumeOrSnapshot.get_by_name(_api, name)
        VolumeSnapshotCommands().write_enable( _api, vol=clone, recursive=False)
        return data

    @command(
        description="Display hierarchy of snapshots and clones",
        types=(VolumeOrSnapshot, CapacityUnit),
        hidden_substitue=dict(name="clone"),
        options=QueryOptions(fields=DETAILED_VOLUME_TREE_FIELDS, sortable=VOLUME_TREE_FIELDS, substitute=DETAILED_QUERY_SUBSTITUTES),
        )
    @render_collection(
        entity=Volume,
        fields=VOLUME_TREE_FIELDS,
        substitute=QUERY_SUBSTITUTES,
        )
    @deprecated
    def tree(self, _api, clone, unit=None):
        data = VolumeSnapshotCommands().tree(_api, vol=clone, unit=unit)
        return data
    
    @command(
        description="Rename clone",
        types=(VolumeOrSnapshot, Name),
        hidden_substitue=dict(name="clone"),
        substitute=dict(),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    @deprecated
    def rename(self, _api, clone, new_name):
        data = VolumeSnapshotCommands().rename(_api, vol=clone, new_name=new_name)
        return data

    @command(
        description="Delete clone",
        types=(VolumeOrSnapshot.List,),
        hidden_substitue=dict(name="clone"),
        substitute=dict(),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    @deprecated
    def delete(self, _api, clone):
        data = VolumeSnapshotCommands().delete(_api, vol=clone)
        return data

    @command(
        description="Map clone to host or cluster",
        types=(VolumeOrSnapshot.List, Host, Cluster, Lun),
        hidden_substitue=dict(name="clone"),
        substitute=dict(),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    @deprecated
    def map(self, _api, clone, host=None, cluster=None, lun=None):
        data = VolumeSnapshotCommands().map(_api, vol=clone, host=host, cluster=cluster, lun=lun)
        return data

    @command(
        description="Unmap clone from host or cluster",
        types=(VolumeOrSnapshot.List, Host, Cluster),
        hidden_substitue=dict(name="clone"),
        substitute=dict(),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    @deprecated
    def unmap(self, _api, clone, host=None, cluster=None):
        data = VolumeSnapshotCommands().unmap(_api, vol=clone, host=host, cluster=cluster)
        return data

    @command(
        description="List host and cluster mappings by volume",
        types=(VolumeOrSnapshot.List, Host, Cluster),
        hidden_substitue=dict(name="clone"),
        substitute=dict(),
        options=QueryOptions(fields=MAP_QUERY_FIELDS),
        )
    @render_collection(
        entity=Volume,
        fields=MAP_QUERY_FIELDS,
        substitute=dict(
            luns="lun_id",
            target_name="mapped_to",
            target_type="mapping_type",
        ),
        )
    @deprecated
    def map_query(self, _api, clone=None, host=None, cluster=None):
        data = VolumeSnapshotCommands().map_query(_api, vol=clone, host=host, cluster=cluster)
        return data

    @command(
        description="Enable or disable SSD cache",
        types=(VolumeOrSnapshot, YesNo, YesNo),
        hidden_substitue=dict(name="clone"),
        substitute=dict(),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    @deprecated
    def cache(self, _api, clone, ssd_cache, recursive=False):
        data = VolumeSnapshotCommands().cache(_api, vol=clone, ssd_cache=ssd_cache, recursive=recursive)
        return data

