from munch import Munch
from functools import partial
import datetime

from infinishell.infishell import commandgroup, command
from infinishell.infishell.utils import strlist

from .types import (
    Replica,
    ReplicaRole,
    ReplicationLinkBySystemName,
    MobilityEntity
    )

from .options import QueryOptions

from . import (
    render_collection,
    EntityCommands,
    LazyEntityMapping,
    )


MOBILITY_QUERY_FIELDS = strlist("""
    _local_entity_name
    _entity_display_type
    role
    _remote_entity_name
    _remote_system_name
    sync_state
    throughput
    """)

MOBILITY_DETAILED_QUERY_FIELDS = strlist("""
    _local_entity_name
    _entity_display_type
    local_pool_name
    role
    _remote_system_name
    _remote_entity_name
    remote_pool_name
    sync_state
    state_reason
    latency
    throughput
    created_at
    updated_at
    pairs
    """)

SORTABLE_FIELDS = strlist("""
    _local_entity_name
    role
    _entity_display_type
    sync_state
    """)

QUERY_SUBSTITUTES = dict(
    _local_entity_name="local_dataset",
    _entity_display_type="type",
    _remote_system_name="remote_system",
    _remote_entity_name="remote_dataset",
    sync_state="local_sync_state",
    )


@commandgroup(prefix="mobility")
class MobilityCommands(EntityCommands):
    ENTITY_TYPE = Replica
    
    @command(
        description="List existing Active-Active mobility replicas",
        types=(MobilityEntity.List, ReplicationLinkBySystemName, ReplicaRole),
        options=QueryOptions(fields=MOBILITY_DETAILED_QUERY_FIELDS, substitute=QUERY_SUBSTITUTES, sortable=SORTABLE_FIELDS, detailed=True),
        substitute=dict(
            datasets="local_dataset",
            link="remote_system",
            ),
        )
    @render_collection(
        entity=Replica,
        fields=MOBILITY_QUERY_FIELDS,
        detailed_fields=MOBILITY_DETAILED_QUERY_FIELDS,
        substitute=QUERY_SUBSTITUTES,
        detailed_substitute=QUERY_SUBSTITUTES,
        detailed_individually=True,
        value_properties=dict(
            skip_value_renderer=['latency'],
            ),
        )
    def query(self, _api, datasets=None, link=None, role=None):
        datasets = [i for d in datasets for i in d] if datasets else None
        link_mapping = LazyEntityMapping(ReplicationLinkBySystemName, _api, fields=("remote_system_name", "link_state"))
        mobility_source = True if role == ReplicaRole.SOURCE else False if role == ReplicaRole.TARGET else None

        # Workaround for staging area allocated size not being shown when querying multiple replicas
        prefetched_replica = None
        if datasets and len(datasets) == 1:
            prefetched_replica_filter = self._create_query_filter(_api, {
                "local_entity_id" if datasets[0]._is_dataset else "local_cg_id" : datasets[0].id,
                "link_id" : link,
                "replication_type": "ACTIVE_ACTIVE",
                })
            if role:
                prefetched_replica_filter.eq("mobility_source", mobility_source)
            else:
                prefetched_replica_filter.is_not_null("mobility_source")
            prefetched_replica = Replica.get_all(_api, filter=prefetched_replica_filter)
            if not prefetched_replica:
                return []
            if len(prefetched_replica) > 1:
                prefetched_replica = None
            else:
                prefetched_replica = [Replica.get_by_id(_api, prefetched_replica[0].id)]
        # End workaround

        if role:
            mobility_filter = _api.create_filter().eq("mobility_source", mobility_source)
        else:
            mobility_filter = _api.create_filter().is_not_null("mobility_source")
        return self._query(_api,
                prefetched_replica,
                link_id=link,
                replication_type="ACTIVE_ACTIVE",
                _hook=partial(self._query_hook, _api=_api, lazy_link_by_id=link_mapping, datasets=datasets),
                _filter=mobility_filter,
                _sort_on_client_side=True,
                )
        

    def _query_hook(self, replicas, _api=None, lazy_link_by_id=None, datasets=None):
        final_replicas = []
        dataset_ids = set((dataset.id for dataset in datasets)) if datasets else None
        for r in replicas:
            is_cg_replica = r.entity_type == "CONSISTENCY_GROUP"
            if datasets:
                if is_cg_replica and r.local_cg_id not in dataset_ids \
                or not is_cg_replica and not any(pair.local_entity_id in dataset_ids for pair in r.entity_pairs):
                    continue
            r = Munch((k, v) for k, v in r.iteritems() if k[0] != "_")
            if lazy_link_by_id:
                link = lazy_link_by_id[r.link_id]
                r._remote_system_name = link.remote_system_name
                if link.link_state == "DOWN":
                    r.state_reason = "LINK_DOWN"
            r.rpo_state = r.rpo_state if r.rpo_state else "N/A"
            r._entity_display_type = "CG" if is_cg_replica else r.entity_type
            r._local_entity_name, r._remote_entity_name = [], []
            r.pairs = ()
            if is_cg_replica:
                r._local_entity_name = r.get("local_cg_name") or "<Local CG: {}>".format(r.local_cg_id)
                r._remote_entity_name = r.get("remote_cg_name") or "<Remote CG: {}>".format(r.remote_cg_id)
                r.pairs = (['Local Dataset','Remote Dataset', 'Initializing', 'Restore Point'],) if r.entity_pairs else ()
                for pair in r.entity_pairs or ():
                    timestamp = long(pair.restore_point) if pair.restore_point else '-'
                    restore_point =timestamp if timestamp == '-' else datetime.datetime.fromtimestamp(timestamp/1000.0).strftime("%Y-%m-%d %H:%M:%S")
                    r.pairs = r.pairs + ([pair.local_entity_name, pair.remote_entity_name, "yes" if pair.is_initial else "no", restore_point],)
            else:
                r._local_entity_name = r.entity_pairs[0].local_entity_name if r.entity_pairs else "N/A"
                r._remote_entity_name = r.entity_pairs[0].remote_entity_name if r.entity_pairs else "N/A"
            r.role = ReplicaRole.SOURCE if r.mobility_source == True else ReplicaRole.TARGET if r.mobility_source == False else None
            if r.role == "TARGET":
                r.update(
                    state="TARGET",
                    throughput="N/A",
                    )
            else:
                r.progress = "{}%".format(r['progress']) if r['progress'] is not None else None
            final_replicas.append(r)
        return final_replicas



