import inspect
import itertools
import sys
import logging
from munch import Munch
from collections import OrderedDict, defaultdict
from functools import wraps
from contextlib import contextmanager
from importlib import import_module

from ..exceptions import NoSuchCommand, BadArguments, BadCommandDefinition
from ..renderer import Renderer
from ..utils import SpellingCorrector, NOTHING, DictProxy, DictEmulator, ContextProxy, NULL_TIMER
from ..parser import COMMAND_PREFIX_SEPARATOR, MAGIC_COMMAND_PREFIX, format_option
from ..rlcontext import get_active_rlcontext, rlcontext_disabled
from ..options import OptionTypeBase, OptionSet
from ..environment import variables as env
from .._compat import reraise, iteritems, nested, string_types, xrange, iterkeys, itervalues, izip, PY2

logger = logging.getLogger("commands")

runtime = ContextProxy("command")
""" Command runtime context. This object holds contextual information about the
currently executed command. """

class Manifest(object):
    DEFAULT_MANIFEST = "default"
    def __init__(self):
        self.__manifests = defaultdict(lambda: defaultdict(list))
    def _register(self, cls, manifest):
        self.__manifests[manifest][cls._command_prefix].append(cls)
    def get(self, prefix, manifest=DEFAULT_MANIFEST):
        """
        Get command-group classes for the specified prefix (category or nested
        category), and manifest name.

        :param prefix: Command group classes of this prefix (category or nested
                       category) will be returned.
        :param manifest: Command group classes of this manifest will be
                         returned. When omitted, the default manifest is used.
        """
        return list(self.__manifests[manifest].get(prefix, ()))
    def iter_manifest(self, manifest=DEFAULT_MANIFEST):
        """
        Iterate all command-group classes of the specified manifest. If a
        manifest was not specified, the default manifest is used.

        :param manifest: Command group classes of this manifest will be
                         returned. When omitted, the default manifest is used.
        """
        return (cls for category, classes in iteritems(self.__manifests[manifest]) for cls in classes)
    def __iter__(self):
        return self.iter_manifest()
    def __repr__(self):
        return "<{}>".format(type(self).__name__)

manifest = Manifest()
""" A manifest of command-group classes. Every class that is decorated with
:py:func:`~commandgroup` is registered here. This class supports iteration,
and offers a ``get(category, manifest=None)`` method, for retrieving a list of
command-group classes for a given category and manifest. This class also offers
an ``iter_manifest(manifest=None)`` method, for iterating all command-group
classes of a manifest. When a manifest is not specified in
:py:func:`~commandgroup`, the default manifest is used. A commands group can be
excluded from all manifests by passing ``manifest=None`` to the decorator.
Default manifest can also be declared in package modules by declaring a
``__commands_manifest__`` attribute, assigned with the desired default name.
``iter(manifest)`` is equivalent to calling ``iter_manifest()`` with no
arguments. """
del Manifest

MANIFEST_NAME_ATTR = "__commands_manifest__"

def _get_object_manifest_name(obj):
    assert obj
    name = getattr(obj, MANIFEST_NAME_ATTR, NOTHING)
    # If:
    #   - There's a manifest attribute
    #   - And it's a string
    #   - Or it's None, which translates to the default manifest name
    if name is not NOTHING and name is None or isinstance(name, string_types):
        # print('Found manifest attribute in {!r}: {}'.format(obj.__name__, name))  # TODO: REMOVE THIS LINE
        return name or manifest.DEFAULT_MANIFEST
    return None

_package_default_manifests = {}
def _get_class_manifest_name(cls):
    assert cls
    module = inspect.getmodule(cls)
    name = _get_object_manifest_name(cls) or _get_object_manifest_name(module) or _package_default_manifests.get(module.__package__)
    if name:
        # print('(1) Manifest for {!r}: {}'.format(cls.__name__, name))  # TODO: REMOVE THIS LINE
        return name
    package_parts = module.__name__.split(".")
    package_modules_without_manifest = []
    try:
        for i in xrange(len(package_parts), 0, -1):
            curr_package_name = ".".join(package_parts[:i])
            # If this module is a package module, we've already checked it
            if curr_package_name == module.__name__:
                continue
            name  = _package_default_manifests.get(curr_package_name)
            if name:
                break
            # print('Importing ' + curr_package_name)  # TODO: REMOVE THIS LINE
            ancestor_package_module = import_module(curr_package_name)
            name = _get_object_manifest_name(ancestor_package_module)
            if name:
                _package_default_manifests[ancestor_package_module.__name__] = name
                break
            package_modules_without_manifest.append(ancestor_package_module)
    except ImportError:
        # import traceback; traceback.print_exc()  # TODO: REMOVE THIS LINE
        pass
    name = name or manifest.DEFAULT_MANIFEST
    for m in package_modules_without_manifest:
        # print('{} -> {}'.format(m.__name__, name))  # TODO: REMOVE THIS LINE
        _package_default_manifests[m.__name__] = name
        # setattr(m, MANIFEST_NAME_ATTR, name)
    # print('(2) Manifest for {!r}: {}'.format(cls.__name__, name))  # TODO: REMOVE THIS LINE
    return name

def commandgroup(cls=None, prefix=None, parent=None, provides=(), hidden=False, manifest=NOTHING):
    """
    Decorator for command-group classes.

    :param prefix: A prefix for this command, resulting the command to belong to
                   a commands category.
    :param parent: Parent commands class. If specified, then the commands of
                   this group will be nested in the parent's category. This
                   parameter may be used in conjunction with ``prefix``.
    :param provides: An iterable of :py:class:`infi.shell.types.ArgType`
                     subclasses, for this class to provide auto-completion for.
                     When specified, the class must implement a
                     ``get_possible_values(self, type_cls, api, starts_with=None)``
                     method.
    :param hidden: Whether or not this group is hidden from the interactive
                   help. This is the command-group equivalent of the
                   :py:meth:`~hidden` decorator.
    :param manifest: The manifest to which this group belongs. When omitted, the
                     group's class is registered to the default manifest. See
                     :py:attr:`~manifest`.
    """
    if parent:
        prefix = "{}{}{}".format(parent._command_prefix, COMMAND_PREFIX_SEPARATOR, prefix) if prefix else parent._command_prefix
    def _decorate_group(_cls):
        for attr_name in dir(_cls):
            attr = getattr(_cls, attr_name)
            if inspect.ismethod(attr):
                assert PY2
                attr = attr.im_func
                assert inspect.isfunction(attr)

            if inspect.isfunction(attr) and not attr_name.startswith('__'):
                attr.is_hidden = hidden
        _cls._command_prefix = prefix
        _cls._parent_group = parent
        _cls._value_provider_types = tuple(provides)
        # Note that None is a valid value for the manifest
        globals()["manifest"]._register(_cls, manifest if manifest is not NOTHING else _get_class_manifest_name(_cls))
        return _cls
    if cls is None:
        return _decorate_group
    elif not inspect.isclass(cls):
        raise TypeError("Expected class, but got : {!r}".format(type(cls)))
    return _decorate_group(cls)

class _Metadata(object):
    def __init__(self, func):
        self.func = func
        self.name = func.__name__
        self.availability_predicates = []
    def __getattr__(self, attr):
        setattr(self, attr, None)
        return getattr(self, attr)
    def __str__(self):
        return "<Metadata {}>".format(", ".join("{}={!r}".format(k, v) for k, v in iteritems(self.__dict__) if k[0] != '_'))
def _get_command_metadata(func):
    if not hasattr(func, "_command_metadata"):
        func._command_metadata = _Metadata(func)
    return func._command_metadata

def metadata(func, **kwargs):
    """
    Decorator for setting command metadata, which is accessible via
    ``Commands.get_spec(name).metadata``.

    :param func: Function object or ``None``, same as any other decorator
    :param kwargs: Attributes and values to save in metadata object
    """
    def decorator(_func):
        metadata = _get_command_metadata(_func)
        for attr, value in iteritems(kwargs):
            setattr(metadata, attr, value)
        return _func
    if func is None:
        return decorator
    elif not inspect.isroutine(func):
        raise TypeError("Expected function or method, but got: {!r}".format(type(func)))
    return decorator(func)
def command(func=None, name=None, description=None, types=(), substitute={}, hidden_substitue={},
            options=(), auto_complete_hidden_args=[]):
    """ Command method decorator.
        :param name: Command name (default: method name)
        :param description: Command description (default: empty)
        :param types: An iterable of :py:class:`~infi.shell.types.ArgType`
                      classes (default: no arguments)
        :param substitute: A dict of argument name substitutions. It maps method
                           argument names to command argument names. This is
                           optional, and when specified, may include only those
                           argument names that need to be changed.
        :param hidden_substitue: A dict of argument name hidden possible substitutes.
                           This is optional, and when specified, may include only those
                           argument names that need to be changed.
        :param options: An iterable of :py:class:`~infi.shell.options.Option`,
                        :py:class:`~infi.shell.options.Flag`, and
                        :py:class:`~infi.shell.options.OptionSet` instances, or
                        a single :py:class:`~infi.shell.options.OptionSet`
                        instance.
        : param auto_complete_hidden_args: A list of arg names that are hidden from auto_complete and help
    """
    assert not types or hasattr(types, "__iter__"), '"types" expected to be iterable'
    assert not substitute or isinstance(substitute, dict), '"substitute" expected to be a dict'
    assert not hidden_substitue or isinstance(hidden_substitue, dict), '"hidden_substitue" expected to be a dict'
    assert not auto_complete_hidden_args or isinstance(auto_complete_hidden_args, list), '"auto_complete_hidden_args" expected to be a list'
    assert isinstance(options, OptionSet) or not options or hasattr(types, "__iter__"), \
        '"options" expected to be an iterable, or an instance of {!r}, but instead got {!r}'.format(OptionSet, options)
    description = (description.strip().rstrip(".") + ".") if description else None
    kwargs = dict(
        description=description,
        arg_types=tuple(types),
        auto_complete_hidden_args=auto_complete_hidden_args,
        arg_substitutions=dict(substitute),
        arg_hidden_substitutions=dict(hidden_substitue),
        options=(options,) if isinstance(options, OptionSet) else tuple(options),
        )
    if name:
        kwargs.update(name=name)
    return metadata(func, **kwargs)
def render(func=None, renderer=None, disable_renderer_override=None, **kwargs):
    """ Command method decorator for rendering options.
        :param renderer: Renderer instance to use, instead of default renderer
        :param disable_renderer_override: Use specified renderer even if a renderer-override is enabled
        :param kwargs: Rendering options
    """
    return metadata(func, renderer=renderer, disable_renderer_override=disable_renderer_override, render_kwargs=kwargs)
def hidden(func=None):
    """ Command method decorator for defining hidden commands. """
    return metadata(func, is_hidden=True)
def deprecated(func=None):
    """ Command method decorator for defining deprecated commands. """
    return metadata(func, is_deprecated=True)
def internal(func=None):
    """ Command method decorator for defining internal commands. """
    return metadata(func, is_internal=True)
def magic(func=None):
    """ Command method decorator for defining magic commands. """
    return metadata(func, is_magic=True, is_internal=True, is_hidden=True)
def invalidates_autocompletion_cache(func):
    """ Command method decorator causing auto-completion cache to be
    invalidated after successful command invocations. """
    @wraps(func)
    def decorator(*args, **kwargs):
        result = func(*args, **kwargs)
        rlcontext = get_active_rlcontext(safe=True)
        if rlcontext is not None:
            rlcontext.clear_completion_cache()
        return result
    return decorator
def availability(func=None, predicate=None):
    """ Command method decorator for defining availability predicate. When the
    specified predicate returns False, the shell acts as if the command does
    not exist. """
    assert predicate, "Predicate must be specified"
    def decorator(_func):
        _get_command_metadata(_func).availability_predicates.append(predicate)
        return _func
    if func is None:
        return decorator
    elif not inspect.isroutine(func):
        raise TypeError("Expected function or method, but got: {!r}".format(type(func)))
    return decorator(func)

class ArgsProcessor(object):
    """ Default argument processor implementation, and base class for custom
    argument processors. """
    def process(self, spec, args=None, options=None, flags=None):
        """
        Process command spec and raw arguments, and return interpreted
        arguments.

        :param spec: Command spec
        :param args: Dict of raw (string) arguments and their values
        :param options: Dict of raw (string) options and their values
        :param flags: Iterable of raw (string) flags
        :returns: An object with the following attributes:
        ============    ========================================================
        Attribute       Value
        ============    ========================================================
        args            A dict of argument names and computed values
        options         A dict of option & flag names and computed values
        ============    ========================================================
        """
        args, options, flags = args or {}, options or {}, flags or ()
        # Substitue hidden argumets for visible arguments
        args = self._sub_hidden_args(spec, args)
        # Verify validity and uniqueness of argument, option and flag names
        self._verify_args(spec, iterkeys(args))
        self._verify_options(spec, iterkeys(options), flags)
        # Process arguments
        actual_args = self._process(args, DictEmulator(lambda arg: spec.argspec[arg].type))
        # Identify each option by its unique name, and process options
        actual_options = {spec.options[name].get_identifying_name() : value for name, value in iteritems(options)}
        # Process options
        actual_options = self._process(actual_options, DictEmulator(lambda option: spec.options[option].get_argument_type()))
        # Flags are options with a boolean value, so add them to the result
        for option_type in spec.options:
            if option_type.is_flag() and any(name in flags for name in option_type.get_option_names()):
                actual_options[option_type.get_identifying_name()] = True
        return Munch(
            args=actual_args,
            options=actual_options,
            )
    def _verify_args(self, spec, arg_names):
        argspec = dict(spec.argspec)
        unexpected = []
        for arg in arg_names:
            popped = argspec.pop(arg, None)
            if popped is None:
                unexpected.append(arg)
        missing = [arg for (arg, info) in iteritems(argspec) if info.is_mandatory]
        if missing or unexpected:
            err = ""
            if missing:
                err = "Missing mandatory argument{}: {}{}"\
                      .format("" if len(missing) == 1 else "s", ", ".join(missing), "; " if unexpected else "")
            if unexpected:
                err = "{}Unexpected argument{}: {}"\
                      .format(err, "" if len(unexpected) == 1 else "s", ", ".join(unexpected))
            raise BadArguments(err)
    def _sub_hidden_args(self, spec, arg_names):
        subs = {v: k for k, v in spec.substitutions.items()}
        args = {}
        for arg, value in iteritems(arg_names):
            if arg in spec.arg_hidden_substitutions:
                if spec.arg_hidden_substitutions[arg] in arg_names:
                    continue
                # check whether there was a substitute dict with the hidden arg's name substitute.
                # i.e 
                # hidden_substitue=dict(name="vol"),
                #        substitute=dict(vol="snap"),
                # when using a name parameter it should eventually be changed to snap instead of vol 
                sub_arg = spec.arg_hidden_substitutions[arg]
                args[subs.get(sub_arg, sub_arg )] = value
            else:
                args[arg] = value
        return args
    def _verify_options(self, spec, option_names, flag_names):
        added = []
        unexpected = []
        for name, specified_as_flag in itertools.chain(
            izip(option_names, itertools.cycle((False,))),
            izip(flag_names, itertools.cycle((True,))),
            ):
            option_type = spec.options.get(name)
            if option_type is None:
                unexpected.append(name)
                continue
            if option_type.is_flag() and not specified_as_flag:
                raise BadArguments("{} is a flag, not an option, and cannot have a value".format(format_option(name)))
            if specified_as_flag and not option_type.is_flag():
                raise BadArguments("{} is an option, not a flag, and must have a value".format(format_option(name)))
            for added_name, added_option_type in added:
                if option_type is added_option_type:
                    raise BadArguments("{} and {} are synonymous".format(format_option(added_name), format_option(name)))
        if unexpected:
            raise BadArguments("Unexpected option{}: {}"
                               .format("" if len(unexpected) == 1 else "s", ", ".join(map(format_option, unexpected))))
    def _process(self, raw_values, types):
        actual_values = {}
        # Separate constant args from the rest, and get their constant values
        constants = {}
        for arg, value in iteritems(raw_values):
            const_value = types[arg].get_constant(value)
            if const_value is not NOTHING:
                constants[arg] = const_value
            else:
                actual_values[arg] = value
        # Verify raw (textual) values, convert to actual values, then verify those
        self._verify_raw_values(types, actual_values)
        actual_values = self._get_actual_args(types, actual_values)
        self._verify_actual_values(types, actual_values)
        # Add constant values
        actual_values.update(constants)
        return actual_values
    def _verify_raw_values(self, types, argvals):
        for arg, value, in iteritems(argvals):
            types[arg].verify_raw(arg, value)
    def _get_actual_args(self, types, argvals):
        return dict((arg, types[arg].get_actual(arg, value))
                    for arg, value in iteritems(argvals))
    def _verify_actual_values(self, types, argvals):
        for arg, value, in iteritems(argvals):
            types[arg].verify(arg, value)

class Commands(object):
    """
    Commands container. An instance of this class is used by the
    :py:class:`~infi.shell.parser.Parser` in order to instantiate new command
    instances.

    :param command_groups: Iterable of instances of classes decorated with
                           :py:meth:`~commandgroup`.
    :param global_options: Iterable of
                           :py:class:`infi.shell.options.OptionTypeBase` and
                           :py:class:`infi.shell.options.OptionSet` instances.
    :param api: API object, to be passed to command methods
    :param args_processor: Defaults to a new instance of
                           :py:class:`ArgsProcessor`.
    :param default_renderer: Defaults to a new instance of
                            :py:class:`infi.shell.renderer.Renderer`.
    """
    def __init__(self, *command_groups, **kwargs):
        self._global_options = tuple(kwargs.pop("global_options", ()))
        assert all(isinstance(o, (OptionTypeBase, OptionSet)) for o in self._global_options)
        for option in self._global_options:
            option._mark_as_global_option()
        self._api = kwargs.pop("api", None) # TODO: Remove API argument - it's not part of the infra
        self._args_processor = kwargs.pop("args_processor", None) or ArgsProcessor()
        default_renderer = kwargs.pop("default_renderer", None) or Renderer()
        if kwargs:
            raise TypeError("Unexpected keyword arguments: {}".format(", ".join(iterkeys(kwargs))))
        self._types = {}
        self._commands = {}
        self.spec_tree = None  # See self._initialize_spec_tree()
        self._initialize_specs(command_groups, default_renderer)
    def _initialize_specs(self, command_groups, default_renderer):
        for command_group in itertools.chain(self._create_default_command_groups(), command_groups):
            # Create spec
            for command_method in [getattr(command_group, name)
                                   for name in dir(command_group) if not name.startswith("_")]:
                if not hasattr(command_method, "_command_metadata"):
                    continue
                spec = CommandSpec(command_method, command_group._command_prefix, self, default_renderer, self._global_options)
                self._commands.setdefault(spec.command_name, []).append(spec)
            # If value provider, register in types
            if hasattr(command_group, "_value_provider_types"):
                for type in (self._get_type(t) for t in command_group._value_provider_types):
                    type._register_value_provider(command_group)
        self._initialize_spec_tree()
    def _create_default_command_groups(self):
        from .help import HelpCommands    # Import explicitly to register in manifest  # noqa
        from .magic import MagicCommands  # Import explicitly to register in manifest  # noqa
        return tuple(group_cls(self) for group_cls in manifest.iter_manifest("__internal__"))
    class TreeNode(object):
        ROOT = '__root__'
        def __init__(self, category=ROOT):
            self.category = category
            self.parent = None
            self.command_prefix = None
            self.category_prefix = None
            self.spec_lists = {}
            self.children = {}
        def add_spec_list(self, specs):
            self.spec_lists[specs[0].command_name] = specs
        def add_child(self, child):
            assert child.category is not self.ROOT
            self.children[child.category] = child
            child.parent = self
            child.command_prefix = child.category if self.category is self.ROOT \
                                   else COMMAND_PREFIX_SEPARATOR.join((self.command_prefix, child.category))
            child.category_prefix = None if self.category is self.ROOT else self.command_prefix
        def get_child(self, category):
            return self.children.get(category)
        def iter_specs(self):
            for command_name, specs in sorted(iteritems(self.spec_lists)):
                for spec in specs:
                    if spec.is_available():
                        yield spec
                        break
        def iter_non_deprecated_specs(self):
            for command_name, specs in sorted(iteritems(self.spec_lists)):
                for spec in specs:
                    if not spec.is_deprecated:
                        yield spec
                        break
        def iter_all_categories(self):
            for category, node in sorted(iteritems(self.children)):
                yield node
        def iter_available_categories(self):
            for node in self.iter_all_categories():
                if node.has_available_nested_specs():
                    yield node
        def has_available_nested_specs(self):
            return self._has_available_nested_specs(self)
        def _has_available_nested_specs(self, node):
            if node.has_available_specs():
                return True
            for child_node in node.iter_all_categories():
                if self._has_available_nested_specs(child_node):
                    return True
            return False
        def has_available_specs(self):
            return any(self.iter_specs())
        def should_be_visible(self):
            return any(self.iter_non_deprecated_specs()) or (True if self.children else False)
        def has_non_deprecated_nested_specs(self):
            return self._has_non_deprecated_nested_specs(self)
        def _has_non_deprecated_nested_specs(self, node):
            if node.has_available_specs():
                return True
            for child_node in node.iter_all_categories():
                if self.has_non_deprecated_nested_specs(child_node):
                    return True
            return False
        def get_category_names(self):
            return sorted(self.children)
        def get_nested_category(self, nested_category):
            node = self
            for category in nested_category.split(COMMAND_PREFIX_SEPARATOR):
                node = node.get_child(category)
                if not node:
                    return None
            return node
        def search(self, text):
            """
            Search for command specs and categories, optionally in a nested
            level, and return matches from the relevant depth only.

            :param text: Text to search for in command and category names
            :returns: A 2-tuple of which first element is a list of matching
                      specs, and the second is a list of matching categories.
            """
            return self._search(text, self, ([], []))
        def _search(self, text, node, result):
            for spec in node.iter_specs():
                if spec.command_name.startswith(text) and spec.is_available():
                    result[0].append(spec)
            for child in node:
                if not child.has_available_nested_specs():
                    continue
                if child.command_prefix.startswith(text):
                    result[1].append(child)
                elif text.startswith(child.command_prefix):
                    node._search(text, child, result)
            return result
        def __getitem__(self, category):
            return self.get_child(category)
        def __iter__(self):
            return self.iter_all_categories()
        def __repr__(self):
            return "<Category {}: commands={}, categories={}>".format(self.category, sorted(self.spec_lists), sorted(self.children))
    def _initialize_spec_tree(self):
        root = self.TreeNode()
        node_by_prefix = {}
        # For each spec
        for command_name, specs in iteritems(self._commands):
            node = root
            command_prefix = specs[0].command_prefix
            # If spec has a prefix, it doesn't belong in the root node
            if command_prefix:
                # Get initialized node by its perfix
                node = node_by_prefix.get(command_prefix, root)
                if node is root:
                    # Node was not initialized, so initialize it
                    for category in command_prefix.split(COMMAND_PREFIX_SEPARATOR):
                        parent, node = node, node.get_child(category)
                        if not node:
                            node = self.TreeNode(category)
                            parent.add_child(node)
                    # Store initialized node
                    node_by_prefix[command_prefix] = node
            node.add_spec_list(specs)
        self.spec_tree = root
    def _get_available_specs(self):
        specs = {}
        for name, specs_list in iteritems(self._commands):
            for spec in specs_list:
                if spec.is_available():
                    specs[name] = spec
                    break
        return specs
    def get_api(self):
        return self._api
    def _get_type(self, cls):
        return self._types.get(cls) or self._types.setdefault(cls, cls())
    def get_types(self):
        return sorted(itervalues(self._types), key=lambda t: t.get_name())
    def get_specs(self):
        return self._get_available_specs().values()
    def get_spec(self, name, safe=False):
        commands = self._get_available_specs()
        spec = commands.get(name)
        if spec is None:
            if safe:
                return None
            if COMMAND_PREFIX_SEPARATOR not in name and self._is_prefix_available(commands, name):
                raise NoSuchCommand("{} is a category, not a command".format(name))
            raise NoSuchCommand(name, suggestions=self._get_spelling_corrector(commands).correct(name))
        return spec
    def search(self, text):
        return self.spec_tree.search(text)
    def _get_spelling_corrector(self, commands):
        return SpellingCorrector(
             [name for name in commands]
             + list(set(spec.command_prefix for spec in itervalues(commands)))
             )
    def _format_argval(self, a, v):
        if " " in v:
            v = u'"{}"'.format(v.replace(u'"', u'\\"'))
        return u"{}={}".format(a, v)
    def create_command(self, name, text=None, args=None, options=None, flags=None):
        args, options, flags = args or {}, options or {}, flags or set()
        if not text:
            text = name
            if args:
                text = u"{} {}".format(text, u" ".join(self._format_argval(a, v) for a, v in iteritems(args)))
            if options:
                text = u"{} {}".format(text, u" ".join(self._format_argval(format_option(o), v) for o, v in iteritems(options)))
            if flags:
                text = u"{} {}".format(text, u" ".join(format_option(flag) for flag in flags))
        spec = self.get_spec(name)
        data = self._args_processor.process(spec, args, options, flags)
        return Command(spec, text, data.args, data.options)
    def help(self, subject=None):
        command_args = {}
        if subject:
            command_args["on"] = subject
        return self.create_command("help", args=command_args)
    def _is_prefix_available(self, commands, prefix):
        return any(spec.command_prefix == prefix for spec in itervalues(commands))

class CommandSpec(object):
    # TODO: There's a kind of duplicity here: command-metadata and command-spec
    # are kinda the same thing, so why separate them? The code would look much
    # better if the method would have the spec as its attribute, and the
    # decorators would each update that, instead of the huge amount of code in
    # the constructor below.
    def __init__(self, command_method, prefix, commands_manager, default_renderer, global_options):
        self.command_method = command_method
        # Argspec validation
        argspec = inspect.getargspec(command_method)
        if not argspec.args or "_api" not in argspec.args:
            raise BadCommandDefinition("Method {!r}: Command methods must define a non-keyword '_api' argument"
                                       .format(command_method))
        metadata = _get_command_metadata(command_method)
        self.metadata = metadata
        # Command name, prefix and description
        self.command_prefix = prefix
        self.command_name = self.command_basename = metadata.name
        if metadata.is_magic:
            if self.command_prefix:
                raise BadCommandDefinition("Command {}: Magic commands must not have a group prefix"
                                           .format(self.command_name))
            self.command_name = self.command_basename = "{}{}".format(MAGIC_COMMAND_PREFIX, self.command_name)
        if self.command_prefix:
            self.command_name = "{}{}{}".format(self.command_prefix, COMMAND_PREFIX_SEPARATOR, self.command_name)
        self.command_description = metadata.description
        # Arguments and types
        num_types, num_args = len(metadata.arg_types), len(argspec.args) - 2
        if num_types != num_args:
            raise BadCommandDefinition("Command {}: Invalid argument definition: {} arguments, {} types"
                                       .format(self.command_name, num_types, num_args))
        # Argument name substitutions
        method_arg_to_command_arg = metadata.arg_substitutions
        self.substitutions = dict((value, arg) for arg, value in iteritems(method_arg_to_command_arg))
        # Argspec compilation
        arg_types = [commands_manager._get_type(type_cls) for type_cls in metadata.arg_types]
        args = [method_arg_to_command_arg.get(arg, arg) for arg in argspec.args[1:] if not arg.startswith("_")]
        optional_index = len(args) - len(argspec.defaults or ())
        self.argspec = OrderedDict((args[i], Munch(name=args[i],
                                                   type=arg_types[i],
                                                   is_mandatory=i<optional_index),
                                                   )
                                   for i in range(len(args)))
        # Options
        try:
            self.options = OptionSpec(itertools.chain(metadata.options, global_options))
        except BadCommandDefinition as e:
            raise BadCommandDefinition("Command {}: {}".format(self.command_name, e))
        # Flags
        self.is_hidden = metadata.is_hidden or getattr(command_method, "is_hidden", False)
        self.is_deprecated = metadata.is_deprecated or getattr(command_method, "is_deprecated", False)
        self.is_internal = metadata.is_internal
        self.is_magic = metadata.is_magic
        self.arg_hidden_substitutions = metadata.arg_hidden_substitutions
        # Rendering
        self.renderer = metadata.renderer or default_renderer
        self.disable_renderer_override = False if metadata.disable_renderer_override is None else metadata.disable_renderer_override
        self.render_kwargs = metadata.render_kwargs or {}
        # Misc
        self.availability_predicates = metadata.availability_predicates
        self._str = None
    def is_available(self):
        if not self.availability_predicates:
            return True
        for predicate in self.availability_predicates:
            if not predicate():
                return False
        return True
    def is_deprecated(self):
        return self.is_deprecated
    def iter_argspec(self):
        return itervalues(self.argspec)
    def __iter__(self):
        return iter(self.iter_argspec())
    def __repr__(self):
        return "<CommandSpec {}>".format(self.command_name)
    def __str__(self):
        if not self._str:
            self._str = "{} {}".format(self.command_name,
                                       " ".join(("{}={}" if info.is_mandatory else "[{}={}]")\
                                                .format(arg, info.type.get_name())
                                                for arg, info in iteritems(self.argspec)))
        return self._str

class OptionSpec(object):
    def __init__(self, options=()):
        # It's important to keep the options in a consistent order, there's
        # code that depends on this (e.g OptionHooksContext). It's preferable
        # to keep the options in the order by which they were defined in the
        # command.
        self._options = []
        self._option_name_to_option = {}
        for option in options:
            self._add_option(option)
    def _add_option(self, option):
        if isinstance(option, OptionSet):
            for nested_option in option.get_options():
                self._add_option(nested_option)
            return
        self._options.append(option)
        for name in option.get_option_names():
            if name in self._option_name_to_option:
                raise BadCommandDefinition("Conflicting options were defined: {} and {}"
                                           .format(self._option_name_to_option[name], option))
            self._option_name_to_option[name] = option
    def __len__(self):
        return len(self._options)
    def __iter__(self):
        return iter(self._options)
    def __contains__(self, option_name):
        return  option_name in self._option_name_to_option
    def __getitem__(self, option_name):
        return self._option_name_to_option[option_name]
    def get(self, option_name):
        return self._option_name_to_option.get(option_name)

class DynamicValue(object):
    """ Wrapper for a value that is computed within the context of a command.
    """
    def __init__(self, value, func):
        """
        :param value: Raw, pre-computed value.
        :param func: ``func(api, arg_name, raw_value)``, returning a
                     2-tuple ``(arg_name, computed_value)``. The function may
                     raise an exception (that should derive from
                     :py:class:`~infi.shell.exceptions.CommandFailed` in most cases).
        """
        self._value = value
        self._func = func
    def get(self, api, arg):
        argval_tuple = self._func(api, arg, self._value)
        assert isinstance(argval_tuple, (list, tuple)) and len(argval_tuple) == 2
        return argval_tuple

class Command(object):
    def __init__(self, spec, text=None, args=None, options=None):
        self.spec = spec
        self.text = text
        self.args = args or {}
        self.options = options or {}
        self._fill_missing_option_values()
        self._actual_args = None
        self._actual_options = None
        self._option_hooks_context = None
        self._is_within_context = False
    def _fill_missing_option_values(self):
        for option_type in self.spec.options:
            name = option_type.get_identifying_name()
            if name not in self.options:
                self.options[name] = option_type.get_default_value()
    @contextmanager
    def context(self, api):
        if self._is_within_context:
            yield
            return
        self._actual_args = self._compute_args(api)
        self._actual_options = self._compute_options(api)
        self._option_hooks_context = OptionHooksContext(self.spec, self._actual_options)
        try:
            with nested(
                runtime.__as__(Munch(
                    spec=self.spec,
                    text=self.text,
                    args=self._actual_args,
                    options=self._actual_options,
                    )),
                self._option_hooks_context,
                ):
                self._is_within_context = True
                yield
        finally:
            self._actual_args = None
            self._actual_options = None
            self._option_hooks_context = None
            self._is_within_context = False
    def execute(self, api, timer=NULL_TIMER):
        with self.context(api):
            with timer:
                if env.is_flagged("COMMAND_DEBUG"):
                    try:
                        import ipdb as pdb
                    except ImportError:
                        import pdb
                    with rlcontext_disabled():
                        result = pdb.runcall(self.spec.command_method, _api=api, **self._actual_args)
                else:
                    result = self.spec.command_method(_api=api, **self._actual_args)
            self._option_hooks_context.set_command_result(result)
        return result
    def _compute_args(self, api):
        args = {}
        for arg, value in iteritems(self.args):
            arg = self.spec.substitutions.get(arg, arg)
            arg, value = self._compute_dynamic_value(api, arg, value)
            args[arg] = value
        return args
    def _compute_options(self, api):
        option_values = {option : self._compute_dynamic_value(api, option, value)[1]
                         for option, value in iteritems(self.options)}
        options = CommandOptions(self.spec, option_values)
        # TODO: ADD GLOBAL OPTIONS
        return options
    def _compute_dynamic_value(self, api, arg, value):
        if isinstance(value, DynamicValue):
            arg, value = value.get(api, arg)
        return arg, value
    def render(self, result, stream=None, renderer=None, color=True, custom_properties=None):
        renderer = self.spec.renderer if self.spec.disable_renderer_override else (renderer or self.spec.renderer)
        properties = DictProxy(self.spec.render_kwargs, custom_properties or {})
        return renderer.render(result, stream=stream, color=color, **properties)
    def __str__(self):
        return self.text
    def __repr__(self):
        return "<Command '{}'>".format(self.spec.command_name)

class CommandOptions(object):
    ''' A container of option names and their corresponding values, intended for
    use during the runtime of a command. The values of all of the command's
    options are available - even for options that were not specified by the
    user, in which case the value is the return value of
    :py:meth:`~infi.shell.options.OptionType.get_default_value`.

    Option values are accessible by key, and as attributes, such that for a
    given option ``--my-option``, the value is accessible via
    ``runtime.options["my-option"]`` and ``runtime.options.my_option`` (see
    :py:attr:`~runtime`).

    This object also supports iteration, which yields all option keys.
    '''
    def __init__(self, spec, option_values):
        self._options = {}
        for identifying_name, value in iteritems(option_values):
            self.__add_option__(identifying_name, value)
    def __add_option__(self, identifying_name, value):
        self._options[identifying_name] = value
        attr = identifying_name.lstrip("-").replace("-", "_")
        setattr(self, attr, value)
    def get(self, option_key, default=None):
        ''' Similar to item-access (e.g ``runtime.options["my-option"]``).
        Returns the option's value, if such option is defined for the current
        command. Otherwise, returns ``default``. '''
        return self._options.get(option_key, default)
    def __iter__(self):
        return iter(self._options)
    def __getitem__(self, option_key):
        try:
            return self._options[option_key]
        except KeyError:
            raise KeyError("No such option key: {}. Available option keys: {}"
                           .format(option_key, ", ".join(sorted(self._options))))
    def __getattr__(self, option_attr):
        if option_attr[0] == "_":
            return object.__getattr__(self, option_attr)
        raise AttributeError("No such option attribute: {}. Available option attributes: {}"
                             .format(option_attr, ", ".join(a for a in dir(self) if a[0] != "_") or "<none>"))

class OptionHooksContext(object):
    def __init__(self, spec, option_values):
        self._spec = spec
        self._option_valuess = option_values
        self._invoked_option_type_value_pairs = []
        self._command_result = NOTHING
        self._after_command_hooks_invoked = False
    def set_command_result(self, result):
        assert result is not NOTHING, "Command result must not be NOTHING"
        assert self._command_result is NOTHING, "Command result already set"
        self._command_result = result
    def __enter__(self):
        self._before_command()
        return self
    def __exit__(self, exc_type, exc_value, traceback):
        exc_info = None
        if exc_type is not None:
            exc_info = (exc_type, exc_value, traceback)
            # Result may have been set if the command itself was successful,
            # in which case the exception occurred by the renderer or the infra.
            if self._command_result is NOTHING:
                self.set_command_result(None)
        self._after_command(exc_info)
    def _before_command(self):
        # For each option type and value
        for option_type in self._spec.options:
            option_value = self._option_valuess[option_type.get_identifying_name()]
            try:
                # Invoke the option's before-command hook
                option_type.before_command(option_value)
            except Exception:
                # In case of an exception,
                hook_exc_info = sys.exc_info()
                logger.error("Option %s before-command hook raised an exception -"
                             " aborting command and running after-command hooks",
                             option_type, exc_info=hook_exc_info)
                self.set_command_result(None)
                try:
                    self._after_command(sys.exc_info())
                except Exception:
                    pass
                reraise(*hook_exc_info)
            self._invoked_option_type_value_pairs.append((option_type, option_value))
    def _after_command(self, command_exc_info):
        if self._after_command_hooks_invoked:
            return
        self._after_command_hooks_invoked = True
        hook_exc_info = None
        # For each option type and value
        for option_type, option_value in reversed(self._invoked_option_type_value_pairs):
            try:
                option_type.after_command(option_value, self._command_result, command_exc_info)
            except Exception:
                hook_exc_info = hook_exc_info or sys.exc_info()
                logger.error("Option %s after-command hook raised an exception",
                             option_type, exc_info=hook_exc_info)
        if hook_exc_info:
            reraise(*hook_exc_info)

@commandgroup(manifest="__internal__")
class InternalCommands(object):
    def __init__(self, manager):
        super(InternalCommands, self).__init__()
        self._manager = manager
    @internal
    @command(description="End current session")
    def exit(self, _api):
        raise EOFError
    @internal
    @command(description="End current session")
    def quit(self, _api):
        raise EOFError
