import itertools
import textwrap
from munch import Munch

from . import commandgroup, command, render, internal
from ..types import Str, TypeChoicesHelpMixin
from ..exceptions import BadValue, BadArguments
from ..renderer import Renderer
from ..constants import MAX_HELP_LINE_LENGTH, COLORS
from ..parser import COMMAND_PREFIX_SEPARATOR, format_option
from .._compat import nested

class HelpSubject(Str):

    def get_description(self):
        return "Name of a command or a category"

    @classmethod
    def get_name(cls):
        return "SUBJECT"

class FlagArgTypeEmulator(object):
    _IS_LIST = False
    def __init__(self, name, description):
        self._name = name
        self._description = description
    def get_name(self):
        return self._name
    def get_description(self):
        return self._description

class HelpRenderer(Renderer):
    MANDATORY = ""
    OPTIONAL = "(optional)"
    LIST_REPEAT_PREFIX = "[,"
    LIST_REPEAT_SUFFIX = ",...]"
    MAX_CHOICES_LENGTH = 40

    def _render(self, result):
        assert isinstance(result, Munch)
        self._writeln()
        if result.is_single_command():
            return self.render_single_command(result.specs[0])
        return self.render_multiple_commands_and_categories(result)

    def render_multiple_commands_and_categories(self, result):
        general_indent_context = []
        if result.get("description"):
            self._writeln(result.description, colors=COLORS.DESCRIPTION)
            self._writeln()
            general_indent_context.append(self._indent())
        with nested(*general_indent_context):
            if result.get("categories"):
                self._writeln("Categories:\n", colors=COLORS.HEADER)
                with self._indent():
                    for category in result.categories:
                        if category.category_prefix:
                            self._write(category.category_prefix + COMMAND_PREFIX_SEPARATOR, colors=COLORS.CATEGORY_PREFIX)
                        self._writeln(category.category, colors=COLORS.CATEGORY_NAME)
                self._writeln()
            if result.get("specs"):
                self._writeln("Commands:\n", colors=COLORS.HEADER)
                with self._indent():
                    self.render_multiple_command_syntaxes(result.specs)

    def render_single_command(self, spec):
        self._writeln(spec.command_description or spec.command_name, colors=COLORS.DESCRIPTION)
        self._writeln()
        with self._indent():
            self._writeln("Syntax:\n", colors=COLORS.HEADER)
            with self._indent():
                self.render_command_syntax(spec)
            self._writeln()
            if spec.argspec:
                self.render_argument_details("Arguments", list(spec.iter_argspec()), hidden_args=spec.metadata.auto_complete_hidden_args)
            optionspec_as_argspec = list(self._optionspec_to_argspec(spec, all_option_names=True))
            if optionspec_as_argspec:
                self.render_argument_details("Options", optionspec_as_argspec)

    def render_argument_details(self, header, argspec, hidden_args=[]):
        self._writeln("{}:\n".format(header), colors=COLORS.HEADER)
        with self._indent():
            arg_name_and_whitespace_length = self._current_indent_length + max(len(arg.name) for arg in argspec) + 2
            remaining_length = max(MAX_HELP_LINE_LENGTH - arg_name_and_whitespace_length, 20)
            name_template = "{{:{}}}".format(arg_name_and_whitespace_length)
            for arg in argspec:
                if arg.name in hidden_args:
                    continue
                arg_name = itertools.chain([arg.name], itertools.cycle([""]))
                description = arg.type.get_description()
                if description[-1] not in ".;!?,":
                    description += "."
                for description_part in textwrap.wrap(description, remaining_length):
                    self._write(name_template.format(next(arg_name)), colors=COLORS.COMMAND_ARG)
                    self._writeln(description_part)
                self._writeln()

    def render_multiple_command_syntaxes(self, specs):
        name_length = max(len(spec.command_name) for spec in specs)
        for spec in specs:
            self.render_command_syntax(spec, name_length)
            self._writeln()

    def render_command_syntax(self, spec, common_name_length=None):
        name_length = common_name_length or len(spec.command_name)
        if spec.command_prefix:
            self._write(spec.command_prefix + COMMAND_PREFIX_SEPARATOR, colors=COLORS.COMMAND_PREFIX)
        self._write(spec.command_basename + " ", colors=COLORS.COMMAND_NAME)
        if common_name_length and common_name_length > len(spec.command_name):
            self._write(" " * (common_name_length - len(spec.command_name)))
        cmd_name_and_whitespace_length = self._current_indent_length + name_length + 1
        is_first_in_line = True
        total_length = 0
        max_length = MAX_HELP_LINE_LENGTH - cmd_name_and_whitespace_length
        for arg in itertools.chain(
            spec.iter_argspec(),
            self._optionspec_to_argspec(spec, all_option_names=True, option_name_separator="|"),
            ):
            if arg.name in spec.metadata.auto_complete_hidden_args:
                continue
            is_flag = isinstance(arg.type, FlagArgTypeEmulator)
            length = len(arg.name)
            if not is_flag:
                length += self._get_type_name_length(arg.type) + 1  # +1 for equals sign
            if not arg.is_mandatory:
                length += 2  # Brackets
            if not is_first_in_line:
                length += 1  # Space
            total_length += length
            if total_length > max_length:
                is_first_in_line = True
                total_length = length - 1  # -1 for spece, which isn't needed for the first arg
                self._write("\n" + " " * (name_length + 1))
            elif not is_first_in_line:
                self._write(" ")
            is_first_in_line = False
            if not arg.is_mandatory:
                self._write("[", colors=COLORS.ARG_BRACKETS)
            self._write(arg.name, colors=COLORS.ARG_NAME)
            if not is_flag:
                self._write("=", colors=COLORS.ARG_EQUALS)
                self.render_type_syntax(arg)
            if not arg.is_mandatory:
                self._write("]", colors=COLORS.ARG_BRACKETS)
        self._writeln()

    def _optionspec_to_argspec(self, spec, all_option_names=False, option_name_separator=", ", with_global=True):
        for option in spec.options:
            if option.is_hidden_from_help() or (not with_global and option._is_global_option()):
                continue
            name = format_option(option.get_identifying_name()) if not all_option_names \
                   else option_name_separator.join(format_option(n) for n in option.get_ordered_option_names())
            yield Munch(
                name=name,
                is_mandatory=False,
                type=option.get_argument_type() if not option.is_flag()
                     else FlagArgTypeEmulator(name, option.get_description()),
                )

    def _get_type_name_length(self, type):
        is_list = type._IS_LIST
        type = type._element_type_instance if is_list else type
        length = len(type.get_name())
        if isinstance(type, TypeChoicesHelpMixin):
            parts = tuple(self._iter_type_name_parts(type))
            length += sum(len(part) for part in parts)  # Length of each part
            length += (len(parts) - 1) * len(type.TYPE_NAME_SEPARATOR)  # Total length of separators
            length += sum(len(value) for value in (type.TYPE_NAME_PREFIX, type.TYPE_NAME_SUFFIX))
        if is_list:
            length += length + len(self.LIST_REPEAT_PREFIX) + len(self.LIST_REPEAT_SUFFIX)
        return length

    def _iter_type_name_parts(self, choice_type):
        length = 0
        for part in choice_type.get_type_name_parts():
            length += len(part)
            if length > self.MAX_CHOICES_LENGTH:
                yield "..."
                break
            yield part
            length += len(choice_type.TYPE_NAME_SEPARATOR)

    def render_type_syntax(self, arg):
        if arg.type._IS_LIST:
            self._render_list_type_syntax(arg)
        else:
            self._render_singular_type_syntax(arg)

    def _render_list_type_syntax(self, arg):
        self._render_singular_type_syntax(arg)
        self._write(self.LIST_REPEAT_PREFIX, colors=COLORS.ARG_BRACKETS)
        self._render_singular_type_syntax(arg)
        self._write(self.LIST_REPEAT_SUFFIX, colors=COLORS.ARG_BRACKETS)

    def _render_singular_type_syntax(self, arg):
        arg_type = arg.type._element_type_instance if arg.type._IS_LIST else arg.type
        if isinstance(arg_type, TypeChoicesHelpMixin) and arg_type.should_render_choices():
            parts = arg_type.get_type_name_parts()
            if len(parts) == 1:
                self._write(parts[-1], colors=COLORS.ARG_TYPE)
            else:
                self._write(arg_type.TYPE_NAME_PREFIX, colors=COLORS.ARG_BRACKETS)
                parts = tuple(self._iter_type_name_parts(arg_type))
                for i, part in enumerate(parts, 1):
                    self._write(part, colors=COLORS.ARG_TYPE)
                    if i < len(parts):
                        self._write(arg_type.TYPE_NAME_SEPARATOR, colors=COLORS.ARG_BRACKETS)
                self._write(arg_type.TYPE_NAME_SUFFIX, colors=COLORS.ARG_BRACKETS)
        else:
            self._write(arg_type.get_name(), colors=COLORS.ARG_TYPE)

    def _write_description(self, text):
        self._writeln(text, colors=COLORS.DESCRIPTION)

    def _write_header(self, text):
        self._writeln("{}:".format(text), colors=COLORS.HEADER)

HELP_RENDERER = HelpRenderer()

class CommandSyntaxRenderer(HelpRenderer):
    def _render(self, spec):
        self.render_command_syntax(spec, **self._properties)

COMMAND_SYNTAX_RENDERER = CommandSyntaxRenderer()

@commandgroup(provides=(HelpSubject,), manifest="__internal__")
class HelpCommands(object):

    def __init__(self, manager):
        super(HelpCommands, self).__init__()
        self._manager = manager

    def _iter_command_specs(self):
        return (s for s in self._manager.get_specs() if not s.is_hidden)

    def _get_command_prefixes(self):
        return set(s.command_prefix for s in self._iter_command_specs() if s.command_prefix)

    def get_possible_values(self, _type_cls, _api, starts_with=None):
        commands = [s.command_name for s in self._iter_command_specs()]
        return itertools.chain(self._get_command_prefixes(), commands)

    def _search(self, text, category=None, result=None):
        if result is None:
            text = text.lower()
            result = SearchResults(
                description="Search results: {}".format(text),
                categories=[],
                specs=[],
                )
        if category is None:
            category = self._manager.spec_tree
        result.specs.extend(spec for spec in category.iter_specs()
                            if not spec.is_hidden
                            and text in spec.command_name.lower()
                            or any(((text in arg.name.lower() and arg.name.lower() not in spec.metadata.auto_complete_hidden_args) or text in arg.type.get_name().lower()) for arg in spec.argspec.itervalues()))
        for nested_category in category.iter_available_categories():
            if text in nested_category.category and nested_category.should_be_visible():
                result.categories.append(nested_category)
            if nested_category.should_be_visible():
                self._search(text, nested_category, result)
        return result

    @internal
    @command(
        description="Display commands' descriptions and syntax",
        types=(HelpSubject, Str),
        substitute=dict(subject="on"),
        )
    @render(
        renderer=HELP_RENDERER,
        disable_renderer_override=True,
        )
    def help(self, _api, subject=None, search=None):
        if search:
            if subject:
                raise BadArguments('Please specify either "on" or "search"')
            result = self._search(search)
            if result.is_empty():
                result.description = 'No matching commands found: {}'.format(search)
            return result
        category = self._manager.spec_tree
        description = "General help"
        if subject:
            subject = subject.lower()
            spec = self._manager.get_spec(subject, safe=True)
            # Help on command
            if spec is not None:
                return SearchResults(specs=[spec])
            # Help on category
            category = category.get_nested_category(subject)
            if not (category and category.has_available_nested_specs()):
                raise BadValue("No such command or category: {}".format(subject))
            description = "Category: {}".format(category.command_prefix)
        return SearchResults(
            description=description,
            categories=tuple(category.iter_available_categories()),
            specs=list(spec for spec in category.iter_specs() if not spec.is_hidden),
            )

class SearchResults(Munch):
    def __init__(self, description='', categories=(), specs=()):
        _locals = locals()
        for attr in ('categories', 'specs'):
            if not isinstance(_locals[attr], (tuple, list, set)):
                _locals[attr] = tuple(_locals[attr])
        super(SearchResults, self).__init__(
            description=description,
            categories=categories,
            specs=specs,
            )
    def is_empty(self):
        return not self.categories and not self.specs
    def is_single_command(self):
        return len(self.specs) == 1 and not self.categories
