from __future__ import print_function

import os
import json
import sys
from ._compat import iteritems
from munch import munchify, unmunchify, Munch
from copy import deepcopy
from contextlib import contextmanager

default_config = munchify(
    dict(
        paging=True,
        log_file=None,
        quiet=False,
        keep_going=False,
        color="auto",
        username=None,
        password=None,
        history=True,
        history_file=None,
    )
)
''' Default configuration '''

def _warning(msg):
    print("Warning: {}".format(msg), file=sys.stderr)

def _read_config_file(filename):
    if os.access(filename, os.F_OK):
        try:
            with open(filename) as f:
                c = json.load(f)
            if isinstance(c, dict):
                return c
            _warning('"{}" has an unexpected structure - ignoring. It is recommended to recreate the'
                     ' default configuration file. See help for more information.'.format(filename))
        except ValueError:
            _warning('"{}" is not a valid JSON document - ignoring.'.format(filename))
        except OSError as e:
            _warning('Failed to read configuration from "{}" - ignoring ({}).'.format(filename, e))
    return {}

def _write_config_file(config, filename):
    config_dir = os.path.dirname(filename)
    if not os.access(config_dir, os.F_OK):
        os.makedirs(config_dir)
    with open(filename, "w") as f:
        json.dump(config, f, indent=2)
        f.write('\n')

def write_default_config(filename):
    """ Write the default configuration to the configuration file. This may
    raise an OSError. """
    return _write_config_file(default_config, filename)

def write_current_config(filename):
    """ Write the current configuration to the configuration file. This may
    raise an OSError. """
    return _write_config_file(config._get_config_dict(), filename)

class LazyInitConfigWrapper(object):
    def __init__(self, filename=None):
        super(LazyInitConfigWrapper, self).__init__()
        self._filename = filename
        self._config_dict = None
        self._saved_configs_stack = []
    def _reset(self, filename=None):
        self._filename = filename
        self._config_dict = None
    def _init(self):
        if self._config_dict is None:
            c = deepcopy(default_config)
            if self._filename:
                c.update(_read_config_file(self._filename))
            self._config_dict = c
    def _get_config_dict(self):
        self._init()
        return self._config_dict
    def _save(self):
        self._init()
        self._saved_configs_stack.append(self._config_dict)
        self._config_dict = self._clone_config(self._config_dict)
    def _restore(self):
        self._init()
        assert self._saved_configs_stack, "Configuration has not been saved - nothing to restore from"
        self._config_dict = self._saved_configs_stack.pop()
    @contextmanager
    def _altered_config_context(self, **kwargs):
        self._save()
        try:
            for key, value in iteritems(kwargs):
                assert key in self._config_dict, "No such configuration key: {!r}".format(key)
                self._config_dict[key] = value
            yield
        finally:
            self._restore()
    def _clone_config(self, value):
        if isinstance(value, dict):
            return Munch((k, self._clone_config(v)) for k, v in iteritems(value))
        if isinstance(value, (set, list)):
            return type(value)(value)
        return value
    def __getattr__(self, attr):
        self._init()
        return getattr(self._config_dict, attr)
    def __setattr__(self, attr, value):
        if attr in ("_config_dict", "_filename", "_saved_configs_stack"):
            super(LazyInitConfigWrapper, self).__setattr__(attr, value)
        else:
            self._init()
            setattr(self._config_dict, attr, value)
    def __getitem__(self, key):
        self._init()
        return self._config_dict[key]
    def __setitem__(self, key, value):
        self._init()
        self._config_dict[key] = value
    def __contains__(self, key):
        self._init()
        return key in self._config_dict
    def __iter__(self):
        self._init()
        return iter(self._config_dict)
    def __repr__(self):
        self._init()
        return str(unmunchify(self._config_dict))

config = LazyInitConfigWrapper()

def initialize_config(filename):
    """ Initialize configuration using the specified configuration file (which
    may not exist). This method must be called in the context of a "main"
    logic, preferably before initializing the argument parser, which may use
    configuration values as defaults for command line arguments. """
    global config
    config._reset(filename)

def altered_config(**kwargs):
    """ Altered configuration context. This context saves the current
    configuration, and then alters the configuration according to the specified
    keyword arguments. Upon exiting the context, the saved configuration is
    restored. """
    return config._altered_config_context(**kwargs)
