import platform
import struct
import os
import re

DEFAULT_WIDTH, DEFAULT_HEIGHT = 80, 24

def get_console_size():
    """ Get the current console size as a tuple: ``(width, height)`` """
    return _get_console_size() or (DEFAULT_WIDTH, DEFAULT_HEIGHT)

def _get_console_size():
    os_name = platform.system().lower()
    global _get_console_size
    if re.search("(linux|aix|darwin|cygwin|sunos)", os_name):
        _get_console_size = _get_console_size_unix
    elif "windows" in os_name:
        _get_console_size = _get_console_size_windows
    else:
        raise AssertionError("Unsupported operating system: {}".format(os_name))
    return _get_console_size()

def _get_console_size_unix():
    cr = _get_size_from_fd_unix(0) or _get_size_from_fd_unix(1) or _get_size_from_fd_unix(2)
    if not cr:
        try:
            fd = os.open(os.ctermid(), os.O_RDONLY)
            cr = _get_size_from_fd_unix(fd)
            os.close(fd)
        except:
            pass
    if not cr:
        try:
            cr = (os.getenv('LINES', DEFAULT_HEIGHT), os.getenv('COLUMNS', DEFAULT_WIDTH))
        except:
            return None
    return int(cr[1]), int(cr[0])

def _get_size_from_fd_unix(fd):
    try:
        import fcntl, termios
        cr = struct.unpack('hh', fcntl.ioctl(fd, termios.TIOCGWINSZ,'1234'))
    except:
        return None
    return cr[0] or DEFAULT_HEIGHT, cr[1] or DEFAULT_WIDTH

def _get_console_size_windows():
    try:
        from ctypes import windll, create_string_buffer
        # stdin handle is -10
        # stdout handle is -11
        # stderr handle is -12
        h = windll.kernel32.GetStdHandle(-12)
        csbi = create_string_buffer(22)
        res = windll.kernel32.GetConsoleScreenBufferInfo(h, csbi)
        if res:
            bufx, bufy, curx, cury, wattr, left, top, right, bottom, maxx, maxy = \
                    struct.unpack("hhhhHhhhhhh", csbi.raw)
            # On windows do not print on the right-most character,
            # as it causes a new line to be printed
            sizex = right - left
            sizey = bottom - top + 1
            return sizex, sizey
    except:
        pass
    size = _get_console_size_cygwin()
    global _get_console_size
    _get_console_size = lambda: size
    return size

def _get_console_size_cygwin():
    try:
        import subprocess
        proc = subprocess.Popen(["tput", "cols"], stdin=subprocess.PIPE, stdout=subprocess.PIPE)
        output = proc.communicate(input=None)
        cols = int(output[0])
        proc = subprocess.Popen(["tput", "lines"], stdin=subprocess.PIPE, stdout=subprocess.PIPE)
        output = proc.communicate(input=None)
        rows = int(output[0])
        return (cols,rows)
    except:
        return None

if __name__ == "__main__":
    def test():
        x, y = get_console_size()
        text = "{} x {}".format(x, y)
        print("{{:^{}}}".format(x).format(text))
    test() or test()
