from os import getenv
from .utils.nothing import NOTHING

class Variables(object):
    class Variable(object):
        POSITIVES = ("yes", "on", "1", "true")
        NEGATIVES = ("no", "off", "0", "false")
        def __init__(self, variables, name, default=None, flag=False, internal=False, help=""):
            self._variables = variables
            self.name = name
            self.default = default
            self.is_flag = flag
            self.is_internal = internal
            self.help = help
        def is_set(self):
            return self._variables._getenv(self.name, NOTHING) is not NOTHING
        def get(self, default=NOTHING):
            value = self._variables._getenv(self.name, default)
            if value is NOTHING:
                return self.default
            return value
        def is_flagged(self):
            assert self.is_flag, self.name + " is not a flag"
            value = self.get()
            if self.default:
                # If default is positive, return negative only if explicitly negative
                return value is None or not value.lower() in self.NEGATIVES
            else:
                # If default is negative, return positive only if explicitly positive
                return value is not None and value.lower() in self.POSITIVES
        def __repr__(self):
            return "<Variable {name} [default {default!r}, flag={is_flag}, internal={is_internal}]>".format(self.__dict__)
    def __init__(self):
        self._variables = {}
        self._prefix = None
    def _getenv(self, name, default=None):
        assert self._prefix, ("Environment variables prefix not set.\n"
                              "Here's how to do it, in your main module:\n"
                              "from infinishell.infishell.environment import variables\n"
                              "variables.set_prefix('MYSHELL')")
        return getenv("{}_{}".format(self._prefix, name), default)
    def add(self, name, default=None, flag=False, internal=False, help=""):
        """
        Define a variable.

        :param name: Variable name (excluding the prefix)
        :param default: Default value when variable is not set
        :param flag: Whether or not this variable is a flag
        :param internal: Whether or not this variable should be displayed to
                         the user when listing all available variables
        :param help: Help string
        """
        var = self.Variable(self, name=name, default=default, flag=flag, internal=internal, help=help)
        self._variables[var.name] = var
    def __getitem__(self, item):
        var = self._variables.get(item)
        assert var, "No such environment variable: {}\nAvailable vars: {}".format(item, self._variables.keys())
        return var
    def __getattr__(self, attr):
        return self[attr]
    def is_set(self, name):
        """
        Returns whether or not this variable is set.

        :param name: Variable name
        :returns: ``True`` if set, otherwise ``False``.
        """
        return self[name].is_set()
    def get(self, name, default=NOTHING):
        """
        Get variable value.

        :param name: Variable name
        :param default: Override defined default with this one, in case
                        variable is not set
        :returns: Variable value, the specified default, or the predefined
                  default.
        """
        return self[name].get(default)
    def is_flagged(self, name):
        """
        Returns whether or not this flag variable is "true".

        :param name: Variable name
        :returns: ``True`` if this flag is flagged, otherwise ``False``.
        """
        return self[name].is_flagged()
    def get_all(self):
        """
        Returns all defined variables (both set and unset).

        :returns: ``list`` of all variable instances.
        """
        return self._variables.values()
    def set_prefix(self, prefix):
        """
        Set variables prefix. This method must be called once, upon
        initialization of the shell, preferrably before initializing the
        argument parser.

        :param prefix: The chosen prefix
        """
        self._prefix = prefix
    def get_prefix(self):
        """ Returns the currently set prefix. """
        return self._prefix


variables = Variables()

variables.add("DEBUG", flag=True, internal=True, help="General debugging features")
variables.add("DEBUG_PM", flag=True, internal=True, help="Enable post-mortem debugging for command exceptions")
variables.add("RLCONTEXT_DEBUG", flag=True, internal=True, help="RLContext-completion debug prints")
variables.add("RLCONTEXT_DEBUG_DISPLAY", flag=True, internal=True, help="RLContext-display debug prints")
variables.add("COMMAND_DEBUG", flag=True, internal=True, help="Start debugger upon entering a command")
