import sys
from .utils import decamelize
from ._compat import string_types, unicode

class ExceptionFormatter(object):
    ''' Abstract base class for exception formatters. '''
    def format(self, exc_info):
        '''
        Format the specified exception.

        :param exc_info: Exception info (see :py:func:`sys.exc_info`)
        :returns: Formatted exception as a string.
        '''
        raise NotImplementedError

class DefaultExceptionFormatter(ExceptionFormatter):
    '''
    Default exception formatter implementation. For ``MyException('Some
    message')``, this formatter outputs ``MyException: Some message``.

    Subclasses should override the following methods:

    ==============================  ==========================================
    Method                          Description
    ==============================  ==========================================
    _format_exception_name          Accepts exception type name as string,
                                    returns string to display. Default
                                    implementation returns the name as is.
    _format_exception_details       Accepts exception instance, returns
                                    message to display. Default implementation
                                    returns ``unicode(e)``.
    ==============================  ==========================================
    '''
    def format(self, exc_info):
        parts = []
        parts.append(self._format_exception_name(type(exc_info[1]).__name__))
        details = self._format_exception_details(exc_info[1])
        if details is not None and (not isinstance(details, string_types) or details):
            parts.append(": ")
            parts.append(details)
        return u"".join(parts)
    def _format_exception_name(self, name):
        return name
    def _format_exception_details(self, e):
        return unicode(e)

class DecamelizeExceptionFormatter(DefaultExceptionFormatter):
    '''
    Exception formatter that formats exception type names using
    :py:func:`~infi.shell.utils.text.decamelize`.
    '''
    def _format_exception_name(self, name):
        return decamelize(name).capitalize()

def format_exception(exc_info=None):
    '''
    Format exception using :py:class:`DecamelizeExceptionFormatter`. If the
    specified ``exc_info`` is ``None``, uses :py:func:`sys.exc_info`.
    '''
    if exc_info is None:
        exc_info = sys.exc_info()
    return DecamelizeExceptionFormatter().format(exc_info)
