import datetime
import time
import os
from munch import Munch

from .exceptions import ShellException
from .utils.filesystem import pathify

DATE_FORMAT = "%Y-%m-%d %H:%M:%S"
def format_timestamp(seconds):
    return datetime.datetime.fromtimestamp(seconds).strftime(DATE_FORMAT)

class CommandsLogger(object):
    def __init__(self, filename, mode=None, with_output=False, with_timestamps=False):
        assert filename
        filename = pathify(filename)
        mode = mode or "a"
        assert mode in ("a", "w"), "Invalid mode: {}".format(mode)
        self.filename, self.mode, self.with_output, self.with_timestamps = filename, mode, with_output, with_timestamps
        self._file = None
        self._is_started = False
    def start(self):
        if not self._is_started:
            if not os.access(os.path.dirname(self.filename), os.W_OK):
                raise ShellException('No write access to "{}"'.format(self.filename))
            self._file = open(self.filename, self.mode)
            self._is_started = True
    def stop(self):
        if self._is_started:
            self._file.close()
            self._is_started = False
    def is_started(self):
        return self._is_started
    def get_info(self):
        return Munch(
            filename=self.filename,
            mode=self.mode,
            with_output=self.with_output,
            with_timestamps=self.with_timestamps,
            is_started=self.is_started(),
            )
    def log(self, command):
        assert self._is_started
        if self.with_timestamps:
            self._file.write("# {}\n".format(format_timestamp(time.time())))
        self._file.write(command.text)
        self._file.write("\n")
        self.flush()
    def log_command_output(self, line):
        if self.with_output:
            self._file.write("# ")
            self._file.write(line)
            self._file.write("\n")
    def log_note(self, msg, *args, **kwargs):
        if not msg: return
        if args or kwargs:
            msg = msg.format(*args, **kwargs)
        for line in msg.split('\n'):
            self._file.write('# {}\n'.format(line.rstrip()))
        self.flush()
    def flush(self):
        self._file.flush()
