from ._compat import reduce

class OptionTypeBase(object):
    ''' Base class for all option types. '''

    NAMES = None
    ''' A sequence of option names that activate this option. The default
    implementation of :py:meth:`.get_option_names` returns the value of this
    attribute. See also :py:meth:`.get_identifying_name`.
    '''

    def _is_global_option(self):
        return getattr(self, '__is_global_option__', False)

    def _mark_as_global_option(self):
        self.__is_global_option__ = True

    def get_option_names(self):
        ''' Get option names. This method may be overridden, to enable altering
        the option names. The default implementation returns the value of the
        ``NAMES`` class attribute.

        Note that by default, the first longest option name will be used to
        identify this option's value during runtime. See also
        :py:meth:`.get_identifying_name`,
        :py:attr:`~infi.shell.commands.commands.runtime` and
        :py:class:`~infi.shell.commands.commands.CommandOptions`.
        '''
        assert self.NAMES is not None, 'NAMES attribute not set in ' + type(self).__name__
        assert self.NAMES, 'NAMES attribute has an empty value in ' + type(self).__name__
        return self.NAMES

    def get_ordered_option_names(self):
        ''' Get option names, by ascending order of length. '''
        return sorted(self.get_option_names(), key=lambda name: len(name))

    def get_identifying_name(self):
        ''' Get the identifying name of this option. This is one of the option's
        names (returned by ``get_option_names()``), that is used to identify
        this option during runtime. The default implementation picks the first
        longest option name. '''
        if not hasattr(self, '__identifying_name__'):
            # Compute longest option name, and use it as the identifying name
            self.__identifying_name__ = reduce(
                lambda n1, n2: n1 if len(n1) > len(n2) else n2,
                reversed(self.get_option_names()))
        return self.__identifying_name__

    def get_shortest_name(self):
        ''' Get the shortest of this option's names. '''
        if not hasattr(self, '__shortest_name__'):
            self.__shortest_name__ = reduce(
                lambda n1, n2: n1 if len(n1) < len(n2) else n2,
                self.get_option_names())
        return self.__shortest_name__

    def is_hidden_from_help(self):
        ''' Returns ``False`` by default. Hidden options are still usable, and
        auto-completion still works for them. '''
        return False

    def before_command(self, value):
        ''' A hook that is called before command execution, when this option is
        passed to a command.

        Important note: Option hooks must not raise any exceptions. Exceptions
        raised from hooks are considered as assertion errors.

        :param value: The option's value. For flags, this is either ``True`` or
                      ``False``, and for options, it's the actual (computed)
                      value of the option.
        '''
        pass

    def after_command(self, value, result, exc_info):
        ''' A hook that is called after command execution, when this option is
        passed to a command.

        Important note: Option hooks must not raise any exceptions. Exceptions
        raised from hooks are considered as assertion errors.

        :param value: The option's value. For flags, this is either ``True`` or
                      ``False``, and for options, it's the actual (computed)
                      value of the option.
        :param result: The command's return value. If an exception was raised
                       during the command's execution, the value is ``None``.
        :param exc_info: The exception information, as returned by
                         :py:meth:`sys.exc_info`, in case an exception was
                         raised during the command's execution. Otherwise,
                         ``None``.
        '''
        pass

    def is_flag(self):
        ''' Denotes whether or not this option is a flag. '''
        return False

    def __repr__(self):
        return '<{}: {}{}>'.format(
            type(self).__name__,
            ', '.join(self.get_option_names()),
            ' (global option)' if self._is_global_option() else '',
            )

class OptionType(OptionTypeBase):
    ''' A normal option, which has an arugment type and a value. '''

    ARGUMENT_TYPE = None
    ''' The argument type of this option's value. The default implementation of
    :py:meth:`.get_argument_type` returns the value of this attribute. The
    argument type must an instance, deriving from
    :py:class:`~infi.shell.types.ArgType`.'''

    def get_argument_type(self):
        ''' Get the argument type of this option's value. The argument type must
        an instance, deriving from :py:class:`~infi.shell.types.ArgType`. This
        method is invoked many times, so it should always return the same
        instance. '''
        assert self.ARGUMENT_TYPE is not None, 'ARGUMENT_TYPE not set in ' + type(self).__name__
        return self.ARGUMENT_TYPE

    def get_default_value(self):
        ''' Get the default value of this option. When this option is not
        specified by the user, the return value of this method will be the
        option's value during the command's runtime. This method is invoked many
        times, so it should always return the same instance. The default
        implementation returns ``None``. '''
        return None

    def get_description(self):
        ''' Returns the description of this option's argument type. '''
        return self.get_argument_type().get_description()

class FlagType(OptionTypeBase):
    ''' A flag option, which value is ``True`` when the option is passed to a
    command, and ``False`` otherwise. '''

    def is_flag(self):
        ''' Returns ``True``. '''
        return True

    def get_default_value(self):
        ''' Returns ``False``. This method should not be overridden by
        subclasses. '''
        return False

    def get_description(self):
        ''' Flag description, for the interactive help and documentation
        purposes. Subclasses must implement this method. '''
        return "N/A"

class OptionSet(object):
    ''' A grouping of options. An option set comprises multiple options, and is
    responsible for instantiating each. It comes to ease the repeated use of
    options which are related to, or complement one another. '''

    def get_options(self):
        raise NotImplementedError

class SimpleOptionSet(OptionSet):
    ''' Passes-through all constructor arguments to the constructor of each
    option class. Option classes are specified in the ``OPTION_CLASSES`` class
    attribute, as a sequence. '''

    OPTION_CLASSES = None

    def __init__(self, **kwargs):
        assert self.OPTION_CLASSES, \
            'No option classes were specified. Use the OPTION_CLASSES class attribute.'
        self._option_kwargs = kwargs

    def get_options(self):
        return [cls(**self._option_kwargs) for cls in self.OPTION_CLASSES]
