from munch import Munch
from pyparsing import (
    alphanums,

    lineEnd,
    restOfLine,

    ParseException,
    ParserElement,

    Literal,
    Word,
    CharsNotIn,
    QuotedString,

    Combine,
    Group,
    Suppress,

    Optional,
    ZeroOrMore,
    OneOrMore,
    )

from .exceptions import BadSyntax, BadArguments

COMMAND_PREFIX_SEPARATOR = "."
MAGIC_COMMAND_PREFIX = "%"

lineEnd = Suppress(lineEnd)

CHARS = Munch(
    cmd_separator=";",
    line_breaker="\\",
    argval_separator="=",
    comment="#",
    quotes="'\"",
    option_separator="=",
    short_option_prefix="-",
    long_option_prefix="--",
    )

EXCLUDED_CHARS = "".join((
    CHARS.cmd_separator,
    CHARS.line_breaker,
    CHARS.argval_separator,
    CHARS.comment,
    CHARS.quotes,
    ))

def format_option(name):
    return "{}{}".format(CHARS.short_option_prefix if len(name) == 1 else CHARS.long_option_prefix, name)

def parse_option(name):
    assert is_option(name), 'Not an option name: {}'.format(name)
    return name.lstrip(CHARS.short_option_prefix)

def is_option(name):
    return name.startswith(CHARS.short_option_prefix)

class ParserBase(object):
    ''' Parser base class. This class implements the shell's language, but does not implement anything further than that. '''
    def __init__(self):
        self._parser = self._build_parser()
    def _get_command_word_chars(self):
        return alphanums + "_"
    def _build_parser(self):
        ParserElement.setDefaultWhitespaceChars(" \t")

        quotedString = QuotedString('"', escChar="\\") ^ QuotedString("'", escChar="\\")

        cmd_separator = Suppress(CHARS.cmd_separator)
        line_breaker = Suppress(CHARS.line_breaker + lineEnd)

        command_expression = Word(self._get_command_word_chars(), excludeChars=EXCLUDED_CHARS)
        arg_expression = Word(alphanums, alphanums + "_-")

        short_option_prefix_exp = Literal(CHARS.short_option_prefix).suppress()
        long_option_prefix_exp = Literal(CHARS.long_option_prefix).suppress()

        command = Combine(
            Optional(MAGIC_COMMAND_PREFIX) +
            command_expression +
            ZeroOrMore(COMMAND_PREFIX_SEPARATOR + command_expression)
            ).setResultsName("command")

        equals = Literal(CHARS.argval_separator).leaveWhitespace().suppress()
        value = (CharsNotIn(EXCLUDED_CHARS + " \t\n") ^ quotedString).leaveWhitespace()

        arg = ZeroOrMore(line_breaker) + Group(arg_expression.setResultsName("arg") + equals + value.setResultsName("value"))

        short_option_body = Word(alphanums, max=1).leaveWhitespace()
        long_option_body = Word(alphanums, alphanums + "_-", min=2).leaveWhitespace()
        option = ZeroOrMore(line_breaker) + Group(
            (
                # (short_option_prefix_exp + short_option_body.setResultsName("option") +
                #     (equals + value.setResultsName("value")) | (White() + value.setResultsName("value"))) |
                (short_option_prefix_exp + short_option_body.setResultsName("flag"))
                ) ^
            (
                (long_option_prefix_exp + long_option_body.setResultsName("option") + equals + value.setResultsName("value")) |
                (long_option_prefix_exp + long_option_body.setResultsName("flag"))
                )
            )

        args = Group(ZeroOrMore(arg ^ option)).setResultsName("args")

        comment = Suppress(CHARS.comment + restOfLine)

        parser = ZeroOrMore(cmd_separator | lineEnd | line_breaker) + \
            Group(ZeroOrMore(
                ZeroOrMore(line_breaker | cmd_separator) +
                Group(command + args) +
                OneOrMore(cmd_separator | lineEnd | (line_breaker + cmd_separator))
            )).setResultsName("commands")
        parser.ignore(comment)
        return parser
    def _raw_parse(self, text):
        try:
            result = self._parser.parseString(text, parseAll=True).commands
            return result
        except ParseException as e:
            if e.msg == "Expected end of text" and e.lineno == 1 and e.loc == 0:
                # Workaround for bad parser state machine. Sadly this is
                # the most common error due to how the parser is built.
                raise BadSyntax("")
            raise BadSyntax(e.msg, line=e.line, lineno=e.lineno, loc=e.loc)
    def _validate_parser_result(self, result):
        for command, args in result:
            processed = set()
            duplicates = set()
            for arg in args:
                is_arg = "arg" in dir(arg)
                if is_arg:
                    argname = arg.arg
                    argname = arg.arg = self._process_arg_name(argname)
                else:
                    argname = arg.flag or arg.option
                    assert argname, "Could not determine argument name"
                    argname = format_option(argname)
                if argname in processed:
                    duplicates.add(argname)
                    continue
                processed.add(argname)
            if duplicates:
                raise BadArguments("The same argument was specified multiple times: {}".format(next(iter(duplicates))) if len(duplicates) == 1
                                   else "Some arguments were specified multiple times: " + ", ".join(sorted(duplicates)))
    def _process_arg_name(self, name):
        '''
        Subclasses may override this method, which optionally transforms each
        argument name. A common use case may be to convert argument names to
        lower case.
        '''
        return name
    def _parse(self, text):
        results = []
        parser_result = self._raw_parse(text)
        self._validate_parser_result(parser_result)
        for command, args_and_options in parser_result:
            args, options, flags = {}, {}, []
            for thing in args_and_options:
                attrs = set(dir(thing))
                if "arg" in attrs:
                    thing.value = thing.value.rstrip()
                    args[thing.arg] = thing.value
                elif "flag" in attrs:
                    flags.append(thing.flag)
                elif "option" in attrs:
                    options[thing.option] = thing.value
            results.append(Munch(command=command, args=args, options=options, flags=flags))
        return results
    def parse(self, text):
        '''
        Subclasses must implement this method, which is the entry point for
        parsing.

        :param text: Text to parse
        :returns: An iterable of command instances, which may be created using
                  :py:meth:`~.infi.shell.commands.commands.Commands.create_command`,
                  or otherwise.
        '''
        raise NotImplementedError

class Parser(ParserBase):
    '''
    Default parser implementation, which uses a
    :py:class:`~infi.shell.commands.commands.Commands` instance to create
    commands according to the parsed text.

    :param commands: :py:class:`~infi.shell.commands.commands.Commands`
                     instance
    :param encoding: Encoding to use
    '''
    def __init__(self, commands, encoding=None):
        super(Parser, self).__init__()
        self._commands = commands
        self._encoding = encoding
    def get_commands(self):
        return self._commands
    def parse(self, text):
        commands = []
        for result in self._parse(text):
            commands.append(self._create_command(
                command=result.command,
                args=result.args,
                options=result.options,
                flags=result.flags,
                ))
        return commands
    def _process_arg_name(self, name):
        return name.lower()
    def _create_command(self, command, args, options, flags):
        return self._commands.create_command(command, args=args, options=options, flags=flags)

class ShellParser(Parser):
    ''' Derived from :py:class:`~Parser`, this parser supports a quick-help
    "``command?``" syntax that delegates to the ``help`` command. '''
    def _get_command_word_chars(self):
        return super(ShellParser, self)._get_command_word_chars() + "?"
    def _create_command(self, command, args, options, flags):
        if command.endswith("?"):
            if args or options or flags:
                raise BadSyntax("Quick-help commands accept no arguments, options or flags")
            return self._commands.help(command[:-1])
        return super(ShellParser, self)._create_command(command, args, options, flags)

