from contextlib import contextmanager
from colorama import Style
from functools import partial

from ._compat import xrange
from .exceptions import OutputError
from .constants import INDENT_LENGTH
from .shell import get_active_shell

class Renderer(object):
    """ Default renderer implementation. """
    _indent_str = " " * INDENT_LENGTH
    def render(self, obj, stream, color=True, **kwargs):
        """
        Render an object. This method is not intended to be overridden.

        :param obj: Object to render
        :param stream: Stream to write to
        :param color: Whether or not to use colors
        :param kwargs: Extra arguments, saved in ``self._properties`` (these
                       are the same arguments that are passed to the
                       :py:meth:`~infi.shell.commands.commands.render`
                       decorator)
        """
        self._stream = stream
        self._indentation = 0
        self._is_newline = True
        self._color = color
        self._color_override = None
        self._properties = kwargs
        try:
            self._render(obj)
        except IOError as e:
            if e.errno == 32: # Broken pipe
                raise OutputError(str(e))
            raise
        finally:
            self._properties = None
            try:
                self._stream.flush()
            except:
                pass
            self._stream = None
    def _render(self, obj):
        """
        Actual rendering implementation. Override this method when implementing
        your own renderer. During invocation of this method,
        ``self._properties`` is a dictionary of rendering properties, as passed
        to the :py:meth:`~infi.shell.commands.commands.render` decorator.

        :param obj: Object to render
        """
        self._stream.write(str(obj))
    def _ensure_stream_open(self):
        if self._stream.closed:
            raise OutputError("Output stream is closed")
    @contextmanager
    def _indent(self, amount=1):
        """
        Indentation context.

        >>> with self._indent(4):
        ...     for field in obj:
        ...         self._writeln('{}: {!r}'.format(field, obj[field]))
        """
        self._indentation += amount
        yield
        self._indentation -= amount
    @property
    def _current_indent_length(self):
        return self._indentation * len(self._indent_str)
    def _render_indentation(self):
        self._ensure_stream_open()
        for _ in xrange(self._indentation):
            self._stream.write(self._indent_str)
        self._is_newline = False
    def _write(self, s, colors=[]):
        """
        Write the specified string to the stream.

        :param s: String to write
        :param colors: Iterable of Colorama colors to use for coloring the
                       string
        """
        assert isinstance(colors, (list, tuple))
        self._ensure_stream_open()
        if self._stream.closed:
            raise OutputError("Output stream is closed")
        if self._is_newline:
            self._render_indentation()
        write_part = partial(self._write_part, colors=colors)
        if "\n" not in s:
            write_part(s)
        else:
            parts = s.split("\n")
            for i, part in enumerate(parts):
                is_last_part = i + 1 == len(parts)
                if not is_last_part:
                    if self._is_newline:
                        self._render_indentation()
                    write_part(part)
                    self._stream.write("\n")
                    self._is_newline = True
                elif part:
                    write_part(part)
    def _write_part(self, s, colors=[]):
        self._ensure_stream_open()
        if self._color_override:
            colors = self._color_override
        if colors and self._color:
            for color in colors:
                self._stream.write(color)
        self._stream.write(s)
        if colors and self._color:
            self._stream.write(Style.RESET_ALL)
    def _writeln(self, s="", colors=[]):
        """ The same as :py:meth:`~_write`, but appends a line-break character. """
        self._ensure_stream_open()
        self._write(s, colors)
        self._stream.write("\n")
        self._is_newline = True
    def _set_color_override(self, colors):
        self._color_override = colors


def set_renderer_properties(**properties):
    """ Set (and optionally override) renderer properties for the current shell
    command. This method is intended for use during command execution, so that
    arguments and other factors can be used to alter how the result is rendered.
    """
    shell = get_active_shell()
    assert shell.state.is_during_command(), "This method can only be called during command execution"
    shell.set_next_renderer_properties(**properties)

def get_default_renderer_properties():
    """ Get the default renderer properties of the current shell command. This
    method is intended for use during command execution. It should be well noted
    that the default renderer properties (i.e this method's return value)
    mustn't be changed. For setting custom, one-time rendering properties for
    the current shell command, use :py:func:`~set_renderer_properties`. """
    shell = get_active_shell()
    assert shell.state.is_during_command(), "This method can only be called during command execution"
    return shell.current_command.spec.render_kwargs
