from __future__ import print_function
import re
import os
import sys
if not sys.stdout.isatty():
	# remember the original setting
	oldTerm = os.environ.get('TERM','')
	os.environ['TERM'] = ''
	import readline
	# restore the orignal TERM setting
	os.environ['TERM'] = oldTerm
	del oldTerm
else:
	import readline
import time
import traceback
import colorama
import inspect
import tempfile
import textwrap
from contextlib import contextmanager
from collections import defaultdict, namedtuple
from getpass import getpass
from itertools import chain
from functools import partial

from .parser import CHARS, format_option, parse_option, is_option
from .exceptions import NoActiveReadlineContext, TooManyPossibleValues
from .environment import variables as env
from .console import get_console_size
from .constants import MAX_HELP_LINE_LENGTH, COLORS, INDENT_LENGTH
from .utils import NOTHING, Outputter
from ._compat import raw_input, xrange, str_decode, with_metaclass, PY2, unicode

try:
	import pyreadline # noqa
	_using_pyreadline = True
except ImportError:
	_using_pyreadline = False

##### DEBUG LOGGING #####

def __debug_method(msg, *args, **kwargs):
	args = [(arg() if inspect.isroutine(arg) or isinstance(arg, partial)
			# else tuple(arg) if inspect.isgenerator(arg)  # This is dangerous as it depletes the
														   # generator- uncomment only when debugging
			else arg)
			for arg in args]
	# kwargs = {key : value() if inspect.isroutine(value) else value for key, value in kwargs}
	msg = str(msg).format(*args, **kwargs)
	print(msg, file=_debug_fd)
	_debug_fd.flush()

def __dummy_debug_method(*args, **kwargs):
	pass

_init = False
_debug = None
_debug_display = None
_debug_fd = None

def _init_env():
	global _init, _debug, _debug_display, _debug_fd
	if _init:
		return
	__debug = env.is_flagged("RLCONTEXT_DEBUG")
	__debug_display = env.is_flagged("RLCONTEXT_DEBUG_DISPLAY")
	if __debug or __debug_display:
		debug_filename = "/tmp/_shell_debug"
		print("Debugging readline to {}".format(debug_filename))
		_debug_fd = open(debug_filename, "a")
	_debug = __debug_method if __debug else __dummy_debug_method
	_debug_display = __debug_method if __debug_display else __dummy_debug_method
	_init = True

##### DEBUG LOGGING #####

_contexts = []

def _push_rlcontext(c):
	_contexts.append(c)
def _pop_rlcontext():
	return _contexts.pop()
def get_active_rlcontext(safe=False):
	""" Return current :py:class:`ReadlineContext`. This function is used as a
	singleton while inside a :py:class:`ReadlineContext`'s context, i.e during
	an interactive shell session. """
	if not _contexts:
		if safe:
			return None
		raise NoActiveReadlineContext
	return _contexts[-1]

@contextmanager
def rlcontext_disabled():
	""" Context manager for running code while the currently active readline
	context is disabled. This context manager should be used to wrap code that
	uses ``readline``, such that the shell's commands history will be preserved
	before another utility attempts to use ``readline``, and then restored
	after. """
	rlc = get_active_rlcontext(safe=True)
	if not rlc:
		yield
	else:
		try:
			rlc.__exit__(None, None, None)
			yield
		finally:
			rlc.__enter__()

def get_raw_input_functions():
	""" Get functions for raw and password input. When a readline context is
	active, that context's raw and password input functions are returned. In
	any other case, Python builtin input functions are returned. """
	rlc = get_active_rlcontext(safe=True)
	if rlc:
		return rlc.raw_input, rlc.input_password
	return raw_input, getpass

@contextmanager
def colorama_disabled():
	""" Context for disabling colorama, which interferes with ``readline``
	during input. """
	colorama.deinit()
	try:
		yield
	finally:
		colorama.reinit()

class ReadlineContext(object):
	""" Context during which enhanced user input is enabled. The major
	enhancements are auto-completion and history. """

	_command_space_pattern = re.compile(r".+\s$")
	_command_prefix_pattern = re.compile(r"([^.]+\.?).*$")
	_readline_word_delimiters = " \t\n=;"
	_strip_readline_escapes_pattern = re.compile("\x01|\x02")

	def __init__(self, commands,
			history_length=500,
			history_filename=None,
			cache_timeout_sec=5,
			max_items_without_confirmation=100,
			encoding=None,
			outputter=Outputter(sys.stdin),
			):
		_init_env()
		self._api = commands.get_api()
		self._commands = commands
		self._prompt = ""
		self._history_length = history_length
		self._history_filename = os.path.expanduser(history_filename) if history_filename \
								 else self._create_temp_history_file()
		self._matches = None  # First-level cache
		self._completion_cache = CompletionCache(cache_timeout_sec)  # Second-level cache
		self._max_items_without_confirmation = max_items_without_confirmation
		self._buffer = ""
		self._auto_completion_enabled = False
		self._encoding = encoding
		self._outputter = outputter
		self._completion_context = None
		self._contextual_info = None

	def input(self, prompt="", buffer=None, auto_completion=True, history=True, input_method=raw_input):
		""" Read user input.

		:param prompt: Prompt to display at beginning of line.
		:param buffer: When calling this method consecutively, this parameter
					   must include a concatenation of all previous inputs.
		:param auto_completion: Whether or not auto completion is enabled.
		:param history: Whether or not input history is enabled.
		:param input_method: Python function to use for actually getting user
							 input. Normally this parameter shouldn't be
							 specified.
		"""
		self._prompt = self._strip_readline_escapes_pattern.subn("", prompt)[0]
		self._buffer = buffer
		self._auto_completion_enabled = auto_completion
		self._history_enabled = history
		if not history:
			history_items = self._get_history()
			readline.clear_history()
		try:
			with colorama_disabled():
				text = input_method(prompt)
			if self._encoding:
				text = str_decode(text, self._encoding)
			return text
		finally:
			self._auto_completion_enabled = False
			if not history:
				readline.clear_history()
				self._add_history(history_items)
			self._buffer = None

	def raw_input(self, prompt=""):
		""" Shorthand method for :py:meth:`input`, which reads user input
		without auto-completion and history. """
		return self.input(prompt, auto_completion=False, history=False)

	def input_password(self, prompt="Password: "):
		""" Read user input, without echoing input characters. Auto-completion
		and history are off.

		:param prompt: Question or message to display.
		"""
		return self.input(prompt, auto_completion=False, history=False, input_method=getpass)

	def yesno(self, prompt="Are you sure?"):
		""" Prompt user with a yes/no question.

		:param prompt: Question or message to display.
		"""
		while True:
			answer = self.raw_input("{} [y/n] ".format((prompt or "").strip())).lower()
			if answer in ("y", "yes"): return True
			if answer in ("n", "no"): return False

	def _get_line_buffer(self):
		""" Get current line buffer, which is a concatenation of all previous
		consecutive :py:meth:`input` results, and the current ``rteadline`` line
		buffer. """
		text = readline.get_line_buffer()
		if self._encoding:
			text = str_decode(text, self._encoding)
		if self._buffer:
			text = " ".join((self._buffer, text))
		last_separator_index = text.rfind(CHARS.cmd_separator)
		if last_separator_index >= 0:
			text = text[last_separator_index+1:]
		return text.lstrip()

	@classmethod
	def _create_temp_history_file(cls):
		handle, filename = tempfile.mkstemp()
		return filename

	@classmethod
	def _get_history(cls):
		""" Get current ``readline`` history.

		:returns: List of strings.
		"""
		return [readline.get_history_item(i) for i in xrange(1, readline.get_current_history_length()+1)]

	@classmethod
	def _add_history(cls, history):
		""" Add multiple lines to ``readline`` history.

		:param history: List of strings.
		"""
		for line in history:
			readline.add_history(line)

	def _set_history_for_user(self):
		if os.access(self._history_filename, os.F_OK):
			_debug('Reading history file: {}', self._history_filename)
			readline.read_history_file(self._history_filename)

	@classmethod
	def _configure_readline(cls, completer, delims, history_length, history=()):
		readline.set_completer(completer)
		readline.set_completer_delims(delims)
		readline.clear_history()
		readline.set_history_length(history_length)
		cls._add_history(history)

	def __enter__(self):
		self._previous_conf = dict(
				completer=readline.get_completer(),
				delims=readline.get_completer_delims(),
				history_length=readline.get_history_length(),
				history=self._get_history()
				)
		self._configure_readline(
				completer=self.complete,
				delims=self._readline_word_delimiters,
				history_length=self._history_length
				)
		if not _using_pyreadline:
			readline.set_completion_display_matches_hook(self.display)
		self._set_history_for_user()
		for config_line in self._get_readline_inputrc():
			readline.parse_and_bind(config_line)
		_push_rlcontext(self)
		return self

	def _get_readline_inputrc(self):
		""" Build and return base ``readline`` configuration. """
		lines = [
			"tab: complete",
			"\C-r: reverse-search-history",
			"set completion-query-items {}".format(self._max_items_without_confirmation),
			"set show-all-if-unmodified on",
			"set horizontal-scroll-mode off",
			"set skip-completed-text on",
			]
		if _using_pyreadline:
			lines.append("set show-all-if-ambiguous on")
		return lines

	def _write_history_to_file(self):
		try:
			history_file_parent = os.path.dirname(self._history_filename)
			if not os.path.isdir(history_file_parent):
				os.makedirs(history_file_parent)
			_debug('Writing history file: {}', self._history_filename)
			readline.write_history_file(self._history_filename)
		except:
			exc_type, exc, _ = sys.exc_info()
			msg = "Failed to write history file {}: {}".format(self._history_filename, exc_type.__name__)
			exc_str = str(exc)
			if exc_str:
				msg = "{}: {}".format(msg, exc_str)
			print(msg, file=sys.stderr)

	def __exit__(self, exc_type, exc_instance, traceback):
		_pop_rlcontext()
		self._write_history_to_file()
		self._configure_readline(**self._previous_conf)
		del self._previous_conf
		self.clear_completion_cache()
		if not _using_pyreadline:
			readline.set_completion_display_matches_hook(None)

	def display(self, substitution, matches, column_width):
		""" ``readline`` display hook. This is the entry point for displaying
		completion matches. """
		cc = self._completion_context
		# Extra padding
		column_width += 1
		try:
			if cc.is_completing_command and len(self._get_line_buffer().split()) in (0, 1):
				_debug_display("Displaying command name completion results")
				# Note that this clause does not use the actual matches, but instead uses completion context data
				self._outputter.output()
				if cc.categories:
					indent_str = " " * INDENT_LENGTH
					self._outputter.output("Categories:", colors=COLORS.HEADER)
					for category in cc.categories:
						if category.should_be_visible():
							_debug_display("Displaying category: {}".format(category.category))
							self._outputter.write(indent_str)
							self._outputter.output(category.category, colors=COLORS.CATEGORY_NAME)
				if cc.specs:
					for i, (specs, header) in enumerate((
						((spec for spec in cc.specs if (not spec.is_magic and not spec.is_deprecated and not spec.is_hidden)), "Commands:"),
						((spec for spec in cc.specs if spec.is_magic), "Special commands:"),
						)):
						matches = []
						# Recalculate column width, since displayed matches are entirely different than actual matches
						column_width = 0
						for spec in specs:
							column_width = max(column_width, len(spec.command_basename))
							matches.append(self._outputter.colorize(spec.command_basename, COLORS.COMMAND_NAME))
						column_width += 4  # Extra padding, same deal as in the beginning of this method
						if not matches:
							continue
						self._outputter.write(header, colors=COLORS.HEADER)  # No newline - _display() provides newline
						self._display(matches, column_width, redisplay_prompt=False, colored=True, indent=1)
				self._redisplay_prompt()
				return
			elif cc.is_completing_argument:
				_debug_display("Displaying argument or option name completion results")
				is_matching_name = lambda name: name.startswith(cc.completing_arg) and name not in cc.complete_args
				matches = [arg.name if arg.is_mandatory else "[{}]".format(arg.name)
						   for arg in cc.spec
						   if is_matching_name(arg.name)]
				matches.extend("[{}]".format("|".join(format_option(n) for n in option.get_ordered_option_names()))
							   for option in cc.spec.options
							   if is_matching_name(format_option(option.get_identifying_name())))
				column_width += 1 # plus two brackets, minus equals sign
			elif cc.is_completing_value and cc.arg_type._IS_LIST:
				_debug_display("Displaying argument or option list-value completion results")
				common_part_length = reduce(lambda l, part: l + len(part) + 1, list(cc.arg_type._iter_unescaped_values(matches[0], unstrip=False))[:-1], 0)
				column_width = 0  # Recalculate column width, since matches are about to change
				for i, match in enumerate(matches):
					matches[i] = match = match[common_part_length:]
					column_width = max(column_width, len(match))
				column_width += 1
			else:
				_debug_display("Displaying generic completion results")
			# Actual displaying of matches
			self._display(matches, column_width)
		except Exception:
			self._outputter.output("\n{}".format(traceback.format_exc()))
		finally:
			self._outputter.flush()

	def _display(self, matches, column_width, redisplay_prompt=True, colored=False, indent=0):
		""" Output rendered matches, considering terminal dimensions. """
		console_width = get_console_size()[0]
		if indent:
			indent_str = " " * indent * INDENT_LENGTH
			console_width -= len(indent_str)
		else:
			indent_str = None
		if not self._display_contextual_info(console_width):
			self._outputter.output()
		_debug_display("CONSOLE WIDTH: {}   COLUMN_WIDTH: {}", console_width, column_width)
		if console_width < column_width:
			column_width = console_width
		columns = console_width // column_width
		_debug_display("COLUMNS: {}", columns)
		match_count = len(matches)
		match_formatter = "{{:<{}}}".format(column_width if not colored else (column_width + 12))
		max_items_per_column = match_count // columns
		if match_count % columns:
			max_items_per_column += 1
		_debug_display("MATCHES: {}   MAX ITEMS PER COLUMN: {}   MATCH_COUNT % COLUMNS: {}"
					   .format(match_count, max_items_per_column, match_count % columns))
		for index in range(max_items_per_column):
			if indent_str:
				self._outputter.write(indent_str)
			for offset in range(0, match_count, max_items_per_column):
				if index + offset < match_count:
					self._outputter.write(match_formatter.format(matches[index+offset]))
			self._outputter.output()
		if redisplay_prompt:
			self._redisplay_prompt()

	def _redisplay_prompt(self):
		ansi_escape = re.compile(r'\x1b[^m]*m')	

		line = self._get_line_buffer()
		full_line = "{}{}".format(self._prompt, line)
		len_prompt = len(ansi_escape.sub('', self._prompt))
		len_full_line = len(ansi_escape.sub('', full_line))
		start_ind = int(readline.get_begidx())
		end_ind = int(readline.get_endidx())

		_debug("FULL_LINE: {} , LINE: {}, PROMPT: {} , START_IND: {} , END_IND: {}".format(len_full_line, len(line), len_prompt, start_ind, end_ind))

		self._outputter.write("{}".format(full_line))
		if end_ind != len(line):
			self._outputter.write("\033[{}D".format(len_full_line - len_prompt - end_ind))


	def _display_contextual_info(self, console_width=None):
		""" Render and output contextual information, if such is set. This
		method is called before matches are displayed, or when there are no
		matches to display. """
		if console_width is None:
			console_width = get_console_size()[0]
		if self._contextual_info:
			self._outputter.output()
			self._contextual_info.render(self._outputter, console_width)
			return True
		return False

	def complete(self, text, state):
		""" ``readline`` completion hook. This is the entry point for
		auto-completion logic. """
		_debug("\n\n")
		try:
			_debug("STATE: {} TEXT: '{}'", state, text)
			# If state is zero, a new completion is starting
			if state == 0:
				self._matches = self._complete(text)
			if self._matches and state < len(self._matches):
				result = self._matches[state]
				if PY2 and result is not None and self._encoding:
					result = result.encode(self._encoding)
				# If there is only one result and it contains spaces, surround it with quote marks
				if len(self._matches) == 1 and " " in result[:len(result)-1]:
					result = '"{}" '.format(result[:len(result)-1])
			else:
				self._cached_matches = None
				result = None
			_debug("MATCH: {}".format(result))
			return result
		except BaseException:
			import traceback
			_debug(traceback.format_exc())

	def _complete(self, text):
		self._completion_context = CompletionContext()
		self._contextual_info = None
		if not self._auto_completion_enabled:
			return None
		text = self._get_line_buffer()
		parts = text.split()

		start_ind = readline.get_begidx()
		if start_ind > 0 and text[start_ind-1] == '=':
			start_ind = text.rfind(' ', 0, start_ind) + 1 
		ends_with_space = self._command_space_pattern.match(text)
		if not parts or (len(parts) == 1 and not ends_with_space):
			return self._complete_command(text)
		# Command is complete, carry on to argument or value
		command_name = parts[0]
		spec = self._commands.get_spec(command_name)
		if not spec:
			return None
		
		end_ind = readline.get_endidx()
		_debug("START_IND {} , END_IND {} , TEXT_LEN {}, TEXT '{}' ", start_ind, end_ind, len(text), text)
		next_space = text[start_ind:].find(" ")
		next_space = next_space + start_ind if next_space != -1 else next_space
		# If there is a space and before the space there is another '=' sign, keep end_ind as is
		if next_space != -1 and "=" in text[end_ind:next_space]:
			end_ind = end_ind
		# If there is a space and there is no '=' sign before it , take everything until the space
		elif next_space != -1 and "=" not in text[end_ind:next_space]:
			end_ind = next_space
		# If there is no space and there is a '=' sign, keep end_ind as is
		elif next_space == -1 and "=" in text[end_ind:]:
			end_ind = end_ind
		# If there is no space and there is not '=' sign, take everything until end of line
		elif next_space == -1 and "=" not in text[end_ind:]:
			end_ind = len(text)
		_debug("NEXT_SPACE_IND: {}, NEW_END_IND: {}",next_space, end_ind)
			

		if end_ind != len(text):
			_debug("COMPLETING IN MIDDLE OF LINE: True")
			arg_name = text[start_ind:end_ind]
			argument_fractions = parts[1:]
		elif ends_with_space:
			_debug("COMPLETING IN MIDDLE OF LINE: False")
			arg_name = ""
			argument_fractions = parts[1:]
		else:
			_debug("COMPLETING IN MIDDLE OF LINE: False")
			arg_name, argument_fractions = parts[-1], parts[1:-1]
		return self._complete_arg(spec, arg_name, argument_fractions)

	def _complete_command(self, text):
		if not text or not text.strip():
			specs = list(self._commands.spec_tree.iter_specs())
			categories = list(self._commands.spec_tree.iter_available_categories())
		else:
			specs, categories = self._commands.search(text)
		matches = list(chain(
			(spec.command_name + " " for spec in specs if ((not spec.is_deprecated and not spec.is_hidden) or spec.is_magic)),
			(category.command_prefix + "." for category in categories if category.should_be_visible()),
			))
		# Add command-related contextual data
		self._completion_context.completing_command(
			specs=specs,
			categories=categories,
			)
		return matches

	def _complete_arg(self, spec, arg_name, argument_fractions):
		_debug("COMPLETING ARG: {} , ARG_FRACTIONS: {}".format(arg_name, argument_fractions))
		existing = set(part.split("=")[0] for part in argument_fractions if "=" in part)
		# Add argument-related contextual data
		self._completion_context.completing_argument(
			spec=spec,
			complete_args=existing,
			completing_arg=arg_name,
			)
		argvalue = arg_name.split("=", 1)
		if len(argvalue) == 1:
			_debug("COMPLETING ARG, OPTION OR FLAG: '{}'", arg_name)
			# Complete argument
			arg_name = arg_name.lower()
			_debug("ALL AVAILABLE ARGS: {}", spec.argspec.keys)
			option_names, flag_names = [], []
			for option in spec.options:
				current_option_names = tuple(format_option(n) for n in option.get_ordered_option_names())
				(flag_names if option.is_flag() else option_names).extend(current_option_names)
				if any(n in existing for n in current_option_names):
					existing.update(current_option_names)
			_debug("ALL AVAILABLE OPTIONS: {}", option_names)
			_debug("ALL AVAILABLE FLAGS: {}", flag_names)
			_debug("EXISTING ARGS, OPTIONS AND FLAGS: {}", existing)
			result = None
			matching_argspecs = [arg for arg in spec.iter_argspec() if arg.name not in existing and arg.name.startswith(arg_name) and arg.name not in spec.metadata.auto_complete_hidden_args]
			if len(matching_argspecs) == 1 and matching_argspecs[0].is_mandatory:
				# There's only one argument and it's mandatory, so options should not be suggested
				result = ["{}=".format(matching_argspecs[0].name)]
			else:
				# Suggest options as well
				result = ["{}=".format(name) for name in chain(
					(arg.name for arg in matching_argspecs),
					(option_name for option_name in option_names if option_name not in existing and option_name.startswith(arg_name)),
					)]
				result.extend("{} ".format(flag_name) for flag_name in flag_names
							  if flag_name not in existing and flag_name.startswith(arg_name))
			_debug("MATCHES: {}", result)
			return result
		# Complete value
		return self._complete_value(spec, argvalue[0], argvalue[1])

	def _complete_value(self, spec, arg_name, partial_value):
		_debug("ARG OR OPTION NAME: {}", arg_name)
		_debug("PARTIAL VALUE: '{}'", partial_value)
		# Get argument type
		arg_type = None
		if is_option(arg_name):
			option = spec.options[parse_option(arg_name)]
			if not option.is_flag():
				_debug("Completing option value")
				arg_type = option.get_argument_type()
		else:
			argspec = spec.argspec.get(arg_name)
			if argspec:
				_debug("Completing argument value")
				arg_type = argspec.type
		if arg_type is None:
			_debug("No info on specified arg or option")
			return None
		# Add value-related contextual data
		self._completion_context.completing_value(
			arg=arg_name,
			arg_type=arg_type,
			)
		# Prepare contextual info for display
		self._contextual_info = ArgTypeContextualInfo(arg_type)
		# Get values from cache, or compute them
		if arg_type._IS_LIST:
			_debug("11111")    
			all_values = self._get_list_possible_values(arg_type, partial_value)
		elif not arg_type.COMPLETION_CACHE:
			_debug("22222")    
			all_values = self._compute_possible_values(arg_type, partial_value)
		else:
			_debug("33333")    
			all_values = self._get_cached_possible_values(arg_type, partial_value)
		# Filter matches
		_debug("All values: {}", all_values)
		if arg_type.AUTO_COMPLETION_CASE_INSENSITIVE:
			partial_value_lower = partial_value.lower()
			matches = [v for v in all_values if v.lower().startswith(partial_value_lower)] if all_values else None
		else:
			matches = [v for v in all_values if v.startswith(partial_value)] if all_values else None
		_debug("Matches: {}", matches)
		if not matches:
			_debug("No matches")
			self._display_contextual_info()
			self._outputter.output("<no matches>")
			self._redisplay_prompt()
		return matches

	def _compute_possible_values(self, type, partial_value):
		""" Compute possible values from scratch, for specified type and partial
		value. Values are not read from, nor written to the completion cache.
		"""
		# Save values with a trailing space character, for spacing when the completion is done
		new_raw_values = type.get_all_possible_values(self._api, partial_value)
		_debug("New raw values for {}: {}", type, new_raw_values)
		values = list(unicode(value)+" " for value in new_raw_values)
		_debug("New values for {}: {}", type, values)
		return values

	def _get_cached_possible_values(self, type, partial_value):
		""" Get possible values for specified type and partial value. If values
		were not cached yet, or have expired, then they are computed again. """
		if type.is_cacheable():
			# Cacheable types are computed once. Therefore all possible values
			# should be retrieved (and stored). Values are filtered later on
			# anyway, during _complete_value().
			partial_value = ""
		values = self._completion_cache.get(type, partial_value)
		_debug('Cache hit for "{}:{}*"? {}', type, partial_value, values is not None)
		if values is None:
			_debug("Retrieving new values for {}", type)
			try:
				values = self._compute_possible_values(type, partial_value)
			except TooManyPossibleValues:
				_debug("Too many possible values for {}", type)
				return ()
			except:
				import traceback
				_debug("Failed to get new values for {}: {}", type, traceback.format_exc)
				return ()
			self._completion_cache.put(type, partial_value, values)
		return values

	def _get_colliding_values(self, type,existing_values, possible_values, current_val):
		return type.get_colliding_values(self._api,existing_values, possible_values, current_val)

	def _get_list_possible_values(self, listtype, partial_value):
		""" Special handling for computation of list subtypes possible values.
		"""
		return listtype.get_possible_values(self._api, partial_value, self._get_cached_possible_values, self._get_colliding_values)

	def clear_completion_cache(self):
		""" Remove all computed possible values from completion cache. """
		self._completion_cache.discard()

	def load_history_for_user(self, history_filename):
		self._write_history_to_file()
		readline.clear_history()
		self._history_filename = os.path.expanduser(history_filename) if history_filename \
								 else self._create_temp_history_file()
		self._set_history_for_user()


	def pop_from_history(self):
		""" Remove and return a line from ``readline`` history. """
		# *sigh*...
		pos = readline.get_current_history_length() - 1
		text = readline.get_history_item(pos)
		readline.remove_history_item(pos)
		return text

	def push_to_history(self, text):
		""" Add line to ``readline`` history. """
		readline.add_history(text)

CachedValues = namedtuple("CachedValues", ("starts_with", "timestamp", "values"))

class CompletionCache(object):
	def __init__(self, timeout):
		super(CompletionCache, self).__init__()
		self._cache = defaultdict(list)
		self._timeout = timeout
	def get(self, arg_type, partial_value):
		oldest_timestamp = time.time() - self._timeout
		for values in self._cache[arg_type]:
			if values.timestamp >= oldest_timestamp and (not values.starts_with or partial_value.startswith(values.starts_with)):
				if not values.starts_with:
					return list(values.values)
				return values.values
		return None
	def put(self, arg_type, partial_value, new_values):
		values = self._cache[arg_type]
		self._clean_expired(values)
		values.append(CachedValues(partial_value, time.time(), new_values))
	def _clean_expired(self, values):
		now = time.time()
		values[:] = (value for value in values if now - value.timestamp < self._timeout)
	def discard(self):
		self._cache.clear()

class ContextualInfo(object):
	""" Contextual information to render when displaying completion matches. """
	def __init__(self):
		self.outputter = None
		self.console_width = None
	def render(self, outputter, console_width):
		self.outputter = outputter
		self.console_width = console_width
		self._render()
	def _render(self):
		raise NotImplementedError
	def print_wrapped(self, text, colors=()):
		for line in textwrap.wrap(text, min(MAX_HELP_LINE_LENGTH, self.console_width)):
			self.outputter.output(line, colors=colors)

class ArgTypeContextualInfo(ContextualInfo):
	""" Contextual information of argument value. When this class is used as the
	contextual info, the argument type's description is rendered along with any
	(or no) matches. """
	def __init__(self, type):
		self.type = type
	def _render(self):
		self.print_wrapped(self.type.get_description(), colors=COLORS.COMPLETION_DESCRIPTION)

class CompletionContextMeta(type):
	KINDS = "command argument value"
	def __new__(mcls, name, bases, attrs):
		if name != "_WithMetaclassBase":
			for kind in mcls.KINDS.split():
				attrs.update(mcls.create_getter_setter(kind))
		return type.__new__(mcls, name, bases, attrs)
	@classmethod
	def create_getter_setter(mcls, kind):
		def getter(self):
			_debug("Completion context: is completing {}? {}, now completing {}", kind, self.kind == kind, self.kind)
			return self.kind == kind
		getter.__name__ = "is_completing_" + kind
		def setter(self, **properties):
			_debug("Completion context: now completing {}", kind)
			self.kind = kind
			self.set(**properties)
		setter.__name__ = "completing_" + kind
		return (getter.__name__, property(getter)), (setter.__name__, setter)


class CompletionContext(with_metaclass(CompletionContextMeta)):
	""" Contextual data for displaying completion matches. """
	def __init__(self):
		self.kind = None
		self.properties = {}
		def __setattr__(self, attr, value):
			self.properties[attr] = value
		self.__setattr__ = __setattr__
	def __getattr__(self, attr):
		return self.properties.get(attr, NOTHING)
	def set(self, **properties):
		self.properties.update(properties)
	def __repr__(self):
		return "<{} kind={} properties={}>".format(type(self).__name__, self.kind, self.properties)

def iter_get_last_element(it):
	element = NOTHING
	while True:
		try:
			element = next(it)
		except StopIteration:
			return element
