import logging
import sys
import traceback
import re
from colorama import Fore, Style, init as init_colorama

from .exceptions import (
    ShellException,
    FatalShellError,
    BadSyntax,
    CommandFailed,
    OutputError,
    SpecialCaseException,
    )
from .parser import CHARS
from .rlcontext import ReadlineContext
from .formatters import DecamelizeExceptionFormatter
from .utils import (
    UnicodeSafeStreamWrapper,
    PaginatedStreamWrapper,
    PaginationAborted,
    LineHookStreamWrapper,
    Outputter,
    )
from .logger import CommandsLogger
from .environment import variables as env
from ._compat import unicode, iteritems, with_metaclass

_ERROR_COLORS = [Fore.RED, Style.BRIGHT]
_WARNING_COLORS = [Fore.YELLOW, Style.NORMAL]

log = logging.getLogger("shell")

init_colorama(autoreset=False)
# Note: keep autoreset=False, because it messes with readline

_active_shells = []
def get_active_shell(safe=False):
    if not _active_shells:
        if safe:
            return None
        raise ShellException("No shell is active")
    return _active_shells[0]
def _push_active_shell(shell):
    _active_shells.insert(0, shell)
def _pop_active_shell(shell):
    popped = _active_shells.pop(0)
    assert shell is popped

class Executor(object):
    """
    Non-interactive commands executor and base class for shells.
    An executor is the glue between the parser, the commands, and the renderers.
    A shell, for that matter, is a user-interactive executor.
    This class provides general shell-related logic, and has a state (:py:class:`.ShellState`).
    """

    def __init__(self, parser=None, stream=None, error_stream=None, exc_formatter=None, paging=True,
                 color=None, errors_color=None, log_filename=None, quiet=False, keep_going=False):
        """
        :param parser: :py:class:`.Parser` to use (mandatory).
        :param stream: Standard output stream (default: ``sys.stdout``).
        :param error_stream: Standard error stream (default: ``sys.stderr``)
        :param exc_formatter: :py:class:`.ExceptionFormatter` to use (default:
                              :py:class:`.DecamelizeExceptionFormatter`)
        :param paging: Whether or not to enable paginated output.
        :param color: Whether or not to enable color for standard output
                      (default: enable if stream is a TTY)
        :param errors_color: Whether or not to enable color for standard error
                             (default: enable if stream is a TTY)
        :param log_filename: Path of commands log file (default: no logging).
        :param quiet: Whether or not to suppress normal output (except
                      failures).
        :param keep_going: Whether or not to stop immediately upon command
                           failure.
        """
        if parser is None:
            raise TypeError("'parser' argument is mandatory")
        self._parser = parser
        stream = stream or sys.stdout
        error_stream = error_stream or sys.stderr
        self._outputter = Outputter(
            UnicodeSafeStreamWrapper(stream),
            color=stream.isatty() if color is None else color,
            quiet=quiet,
            )
        self._errorrer = Outputter(
            UnicodeSafeStreamWrapper(error_stream),
            color=error_stream.isatty() if errors_color is None else errors_color,
            quiet=False,
            )
        self._paging = stream.isatty() and paging and PaginatedStreamWrapper.is_pagination_available()
        self._exc_formatter = exc_formatter or DecamelizeExceptionFormatter()
        self._logger = None
        self._quiet = quiet
        self._keep_going = keep_going
        self._warnings = []
        self.set_renderer_override(None)
        if log_filename:
            self.start_logging(log_filename)
        self.state = ShellState()
        self.current_command = None
        self._reset_custom_renderer_properties()
    def is_interactive(self):
        """ Return whether or not this is an interactive shell. """
        return False
    def is_interactive_terminal(self):
        """ Returns whether or not this shell is running in an interactive
        terminal. Returns ``True`` when ``stdin`` and ``stdout`` are both TTYs,
        otherwise ``False``. """
        return self._outputter.get_stream().isatty() and sys.stdin.isatty()
    def has_colors(self):
        """ Return whether or not the standard outputter is colored. """
        return self._outputter.has_colors()
    def get_renderer_override(self):
        """ Get current overriding-renderer (see
        :py:meth:`.set_renderer_override`). """
        return self._renderer_override
    def set_renderer_override(self, renderer):
        """ Set an overriding-renderer. Set renderer will be used to render all
        command results, and will override both the shell's default renderer,
        and commands' specific renderers. Pass ``None`` as the renderer to
        clear. """
        self._renderer_override = renderer
    def set_next_renderer_properties(self, **properties):
        """ Set custom renderer properties for the next command only. This
        method can be used multiple times during a command's execution. """
        self._custom_renderer_properties.update(properties)
    def _reset_custom_renderer_properties(self):
        self._custom_renderer_properties = {}
    def _reset_command_warnings(self):
        del self._warnings[:]
    def __enter__(self):
        _push_active_shell(self)
    def __exit__(self, _, __, ___):
        _pop_active_shell(self)
    def _execute(self, command):
        """ Execute :py:class:`.Command` and render result. This method handles
        command execution and rendering, command and output logging, streams
        (including pagination), and alters the shell's state. Command exceptions
        are not handled. """
        try:
            if command.spec.is_deprecated and self.is_interactive():
                self.warning("This command is deprecated and will be removed in the next feature release, please use with caution.")
            self.state.set_during_command()
            self.current_command = command
            log.debug("Executing command: %s", command)
            is_logging = self.is_logging() and not command.spec.is_magic
            is_logging_output = is_logging and self._logger.with_output
            if is_logging:
                self._logger.log(command)
            api = self._parser.get_commands().get_api()  # API should not be part of the infra at all
            with command.context(api):
                result = command.execute(api)
                self.state.set_after_command()
                if self._quiet:
                    return
                stream = self._outputter.get_stream()
                if is_logging_output:
                    stream = LineHookStreamWrapper(stream, self._logger.log_command_output)
                if self._paging:
                    stream = PaginatedStreamWrapper(stream, color=self._outputter.has_colors())
                    stream.capture_console_dimensions()
                stream = self._prepare_stream(stream)
                self._output_command_warnings()
                # Flush any error output before emitting normal output
                self._errorrer.flush()
                try:
                    command.render(result, stream,
                        renderer=self._renderer_override,
                        color=self._outputter.has_colors(),
                        custom_properties=self._custom_renderer_properties,
                        )
                except PaginationAborted:
                    if is_logging_output:
                        self._logger.log_command_output('--- USER ABORTED, OUTPUT TRUNCATED ---')
                finally:
                    if is_logging_output:
                        self._logger.flush()
        finally:
            self.current_command = None
            self._reset_custom_renderer_properties()
            self._reset_command_warnings()
            self.state.set_before_command()
    def _prepare_stream(self, stream):
        """ After a command had finished its execution without an error, prepare
        the output stream for rendering. This method is intended for subclasses
        to override. The return value may either be the same stream, or another.

        The argument stream should be considered as the standard output, so when
        returning another stream, it is encoutaged to wrap the argument stream,
        and discouraged to use ``sys.stdout`` directly.

        :returns: The stream to use for rendering the command's output.
         """
        return stream
    def output(self, *args, **kwargs):
        """ Write to output stream.

        :param msg: text to output
        :param colors: colors to prepend to text, if colors are enabled
        """
        # Flush any error output before emitting normal output
        self._errorrer.flush()
        self._outputter.output(*args, **kwargs)
    def error(self, *args, **kwargs):
        """ Write to error stream.

        :param msg: text to output
        :param colors: colors to prepend to text, if colors are enabled
        """
        self._error(_ERROR_COLORS, *args, **kwargs)
    def warning(self, *args, **kwargs):
        """ Write to error stream.

        :param msg: text to output
        :param colors: colors to prepend to text, if colors are enabled
        """
        self._error(_WARNING_COLORS, *args, **kwargs)
    def _error(self, error_colors, *args, **kwargs):
        # Flush any normal output before emitting error output
        self._outputter.flush()
        for key, value in (
            ("colors", error_colors),
            ("consider_quiet", False),
            ):
            kwargs.setdefault(key, value)
        self._errorrer.output(*args, **kwargs)
    @property
    def output_stream(self):
        """ Get output stream, for direct use. Note that direct use of the
        output and error streams is discouraged. See the :py:meth:`.output` and
        :py:meth:`.error` methods. """
        return self._outputter.get_stream()
    @property
    def error_stream(self):
        """ Get error stream, for direct use. Note that direct use of the output
        and error streams is discouraged. See the :py:meth:`.output` and
        :py:meth:`.error` methods. """
        return self._errorrer.get_stream()
    def _exception(self, tb=False):
        exc_info = sys.exc_info()
        if env.is_flagged("DEBUG_PM"):
            self._exception_post_mortem(exc_info)
        if tb or env.is_flagged("DEBUG"):
            msg = "".join(traceback.format_exception(*exc_info))
        else:
            msg = self._exc_formatter.format(exc_info)
        self.error(msg)
        if log.isEnabledFor(logging.ERROR):
            log.error(msg)
    def _exception_post_mortem(self, exc_info):
        try:
            import ipdb as pdb  # noqa
        except ImportError:
            import pdb  # noqa
        self.error("\n### POST MORTEM ###")
        pdb.post_mortem(exc_info[2])
    def _unhandled_exception(self):
        self._exception(tb=True)
    def _special_case_exception(self, x):
        print('NOT IMPLEMENTED')
        raise NotImplementedError()
    def start_logging(self, filename, mode="a", with_output=False, with_timestamps=False):
        """ Start command logger.

        :param filename: Log file name.
        :param mode: File mode (``"a"``, ``"w"``)
        :param with_output: Whether or not to log command outputs (as comments).
        :param with_timestamps: Whether or not to log timestamps (as comments).
        :raises: ShellException
        """
        if self.is_logging():
            raise ShellException("Already logging, cannot start logging again")
        logger = CommandsLogger(filename, mode=mode, with_output=with_output, with_timestamps=with_timestamps)
        try:
            logger.start()
        except (IOError, OSError) as e:
            raise ShellException("Cannot log to {}: {}".format(logger.filename, e))
        self._logger = logger
    def stop_logging(self):
        """ Stop command logger. """
        if not self.is_logging():
            raise ShellException("Logging is already inactive, there's nothing to stop")
        self._logger.stop()
        self._logger = None
    def is_logging(self):
        """ Return whether or not command logging is enabled. """
        return self._logger is not None
    def get_logging_properties(self):
        """ Get properties dict of active logger, or ``None`` if logger is
        inactive. """
        return self._logger.get_info() if self.is_logging() else None
    def log_note(self, msg, *args, **kwargs):
        """ Write a note to the command log (as a comment). """
        if self.is_logging():
            self._logger.log_note(msg, *args, **kwargs)
    def enable_stream_encoding(self):
        self._outputter.get_stream().enable_encoding()
        self._errorrer.get_stream().enable_encoding()
    def disable_stream_encoding(self):
        self._outputter.get_stream().disable_encoding()
        self._errorrer.get_stream().disable_encoding()
    def add_warning(self, msg):
        """ Add a warning to the list of warnings for the current command.
        Warnings are displayed before rendering the result of the current
        command, and only if the command was successful. """
        assert self.state.is_during_command(), "Warnings can be added only during command execution"
        self._warnings.append(msg)
    def _output_command_warnings(self):
        while self._warnings:
            self.warning(self._warnings.pop(0))
    def execute(self, text):
        """ Execute specified text as a non-interactive shell. """
        with self:
            return self._execute_text(text)
    def _execute_text(self, text):
        try:
            commands = self._parser.parse(text)
        except BadSyntax:
            self._exception()
            return False
        success = True
        for command in commands:
            try:
                self._execute(command)
            except (CommandFailed, BadSyntax, NotImplementedError):
                success = False
                self._exception()
                if not self._keep_going:
                    break
            except EOFError:
                # Raised upon "exit", "quit", or keyboard abort (CTRL+D POSIX / CTRL+Z Windows)
                # This is a valid scenario in a script, especially when debugging, one might
                # want to add an "exit" command in the middle of a script, but it should not
                # fail the script
                break
            except UnicodeError:
                # May be thrown by a renderer, but regardless of whether or not the command
                # itself succeeded, so should not fail the script
                self._exception()
            except OutputError:
                # Thrown when the output stream is closed prematurely (broken pipe, etc'),
                # should definitely fail the script
                return False
            except KeyboardInterrupt:
                # User aborted, fail the script
                self.output("\nAborted")
                return False
            except FatalShellError:
                self._exception()
                return False
            except:
                success = False
                self._unhandled_exception()
                if not self._keep_going:
                    break
        return success


ExecutorBase = Executor  # For backwards compatibility (Executor was fully merged to ExecutorBase)

_input_line_pre_logging_replacers = {
    re.compile("\\\\\s*\\n") : " ",
    re.compile("\\n") : "; ",
    }
def _process_input_line_for_logging(text):
    for pattern, replacement in iteritems(_input_line_pre_logging_replacers):
        text = pattern.sub(replacement, text)
    return text

class Shell(Executor):
    """ Interactive shell implementation, based on :py:class:`.Executor`. """

    _strip_prompt_control_chars_pattern = re.compile(r"\x1b[^m]*m|\x01|\x02")

    def __init__(self, history_filename=None, encoding=None, prompt="", prompt_sep="> ", welcome_msg=None, exit_msg=None, **kwargs):
        """
        Create a new ``Shell``. This constructor accepts all
        :py:class:`.Executor` constructor arguments, with some additions.

        :param history_filename: Path for loading and storing command history
                                 (default: no history).
        :param encoding: Encoding of user input (default: Python default).
        :param prompt: Prompt string, or a function ``prompt(shell)``.
        :param prompt_sep: Prompt seperator string, or a function ``sep(shell)``.
        :param welcome_msg: Message to display on startup.
        :param exit_msg: Message to display on exit.
        :param kwargs: :py:class:`.Executor` arguments.
        """
        super(Shell, self).__init__(**kwargs)
        self._encoding = encoding
        self._prompt = prompt
        self._prompt_sep = prompt_sep
        self._welcome_msg = unicode(welcome_msg) if welcome_msg else None
        self._exit_msg = unicode(exit_msg) if exit_msg else None
        self._history_filename = history_filename
    def is_interactive(self):
        return True
    def start(self, _hook=None):
        """ Start interactive shell and enter main loop. 
        :param _hook: hook to run before loop start, returns True / False 
        """
        with self:
            log.info("Start interactive session")
            if self._welcome_msg:
                self.output(self._welcome_msg, colors=[Style.BRIGHT])
            self._on_start()
            if _hook and _hook() == False:
                log.info("EULA was not signed and cannot be snoozed")
                return
            self._loop()
            if self._exit_msg:
                self.output(self._exit_msg)
            log.info("End interactive session")
    def _on_start(self):
        pass
    def _input(self, rlcontext):
        while True:
            try:
                prompt = self._get_prompt()
                prompt_len = len(self._strip_prompt_control_chars_pattern.sub("", prompt))
                prompt_sep = self._get_prompt_sep()
                buffer = readline_buffer = rlcontext.input(prompt + prompt_sep).strip()
                while buffer.endswith(CHARS.line_breaker):
                    readline_buffer = readline_buffer.replace(CHARS.line_breaker, " ")
                    prompt_replacement = " " * prompt_len if prompt_len < 3 else (" " * (prompt_len - 3) + "...")
                    continuation = rlcontext.input(prompt_replacement + prompt_sep, readline_buffer).strip()
                    if not continuation:
                        buffer = buffer[:-1]
                        break
                    readline_buffer += continuation
                    buffer = u"{}\n{}".format(buffer, continuation)
                if buffer != readline_buffer:
                    for _ in xrange(buffer.count("\n")+1):
                        rlcontext.pop_from_history()
                    rlcontext.push_to_history(buffer.replace(CHARS.line_breaker, ""))
                if buffer:
                    if log.isEnabledFor(logging.INFO):
                        log.info("Input: %s", _process_input_line_for_logging(buffer))
                    return buffer
            except KeyboardInterrupt:
                self.output("\nIgnored")
    def _get_prompt_part(self, part):
        if hasattr(part, "__call__"):
            try:
                part = unicode(part(self))
            except Exception:
                self.error("Failed to compute prompt:\n" + traceback.format_exc())
                return ""
        return part
    def _get_prompt(self):
        return self._get_prompt_part(self._prompt)
    def _get_prompt_sep(self):
        return self._get_prompt_part(self._prompt_sep)
    def _loop(self):
        with ReadlineContext(self._parser.get_commands(),
                             encoding=self._encoding,
                             history_filename=self._history_filename,
                             outputter=self._outputter,
                             ) as rlcontext:
            while True:
                try:
                    for command in self._parser.parse(self._input(rlcontext)):
                        self._execute(command)
                except KeyboardInterrupt:
                    self.output("\nAborted")
                except SpecialCaseException as x:
                    # import IPython; IPython.embed()
                    self._special_case_exception(x)
                except (CommandFailed, BadSyntax, NotImplementedError):
                    self._exception()
                except EOFError:
                    # Raised upon "exit", "quit", or keyboard abort (CTRL+D POSIX / CTRL+Z Windows)
                    break
                except UnicodeError:
                    self._exception()
                except FatalShellError:
                    self._exception()
                    break
                except:
                    self._unhandled_exception()


class ShellStateMeta(type):
    def __new__(mcls, name, bases, attrs):
        if name != "_WithMetaclassBase":
            def _create_methods(state_name):
                def is_(self):
                    return self._state == state_name
                def set_(self):
                    self.set(state_name)
                return is_, set_
            for state_name in attrs.get("STATES", ()):
                lowercase_state = state_name.lower()
                for method in _create_methods(state_name):
                    method.__name__ += lowercase_state
                    attrs[method.__name__] = method
        return type.__new__(mcls, name, bases, attrs)


class ShellState(with_metaclass(ShellStateMeta)):
    '''
    A shell state object has 3 attributes for each state name (state names are
    listed in the STATES attribute). For example, for BEFORE_COMMAND, the
    following attributes are available:
     * state.BEFORE_COMMAND       :  state name
     * state.is_before_command()  :  returns True if this is the current state
     * state.set_before_command() :  sets this state as the current state
    '''
    STATES = [
        "BEFORE_COMMAND",
        "DURING_COMMAND",
        "AFTER_COMMAND",
        ]
    DEFAULT_STATE = STATES[0]
    def __init__(self):
        super(ShellState, self).__init__()
        self.reset()
    def reset(self):
        self.set(self.DEFAULT_STATE)
    def set(self, state):
        state = state.upper()
        assert state in self.STATES
        self._state = state
    def get(self):
        return self._state
    def is_in(self, *states):
        return any(state.upper() == self._state for state in states)
