import re
import time
import os
from itertools import chain
from decimal import Decimal
from numbers import Number
from collections import OrderedDict
from munch import Munch
from datetime import timedelta

from .exceptions import BadValue
from .utils import NOTHING, networking
from .commands import DynamicValue
from ._compat import izip, iterkeys, string_types, with_metaclass

class ListCreator(property):
    def __init__(self, **properties):
        self.__list_properties__ = properties
        super(ListCreator, self).__init__()
    def __get__(self, cls, owner):
        return List.of(owner, **self.__list_properties__)

class TypeChoicesHelpMixin(object):
    """ Mixin for types which name should be displayed as options. For example,
    a simple boolean type should be displayed as ``{YES|NO}``, instead of e.g
    ``YesNoType``. """
    TYPE_NAME_SEPARATOR = "|"
    """ String to be used for separating name parts. Subclasses may override
    this. """
    TYPE_NAME_PREFIX = "{"
    """ String to be used as a prefix when displaying type name. """
    TYPE_NAME_SUFFIX = "}"
    """ String to be used as a suffix when displaying type name. """
    def get_type_name_parts(self):
        """
        Get type name parts to display.

        :returns: A tuple or list of strings.
        """
        raise NotImplementedError
    def should_render_choices(self):
        """ Denote whether or not the mixin's logic is active. If this method
        returns ``True``, the type's values (choices) will be rendered instead
        of the type's name. The default implementation of this method returns
        ``True``. """
        return True

class ArgType(object):
    AUTO_COMPLETION_CASE_INSENSITIVE = False
    """Denotes whether or not auto-completion results will accept lower as well
    as upper case values and complete them"""
    CACHEABLE = False
    """ Denotes whether or not the interpreted values of this argument type
    should be cached, so that they are computed no more than once. """
    COMPLETION_CACHE = True
    """ Denotes whether or not the values returned from
    :py:func:`~get_possible_values` may be cached for a short period of time,
    to prevent CPU or IO overload when the user attempts to complete the same
    value in quick succession. """
    CONSTANTS = None
    """ Subtypes may define this attribute as a dict between constant names and
    their corresponding constant values. When more than one type in the
    hierarchy defines this attribute, constants are accumulated, such that
    subtype constants take precedence in case of duplicate names. """
    _IS_LIST = False
    def __init__(self):
        if type(self) is ArgType:
            raise TypeError("Only subclasses of ArgType may be instantiated")
        self._value_providers = []
        self._constants = self._get_constants()
        self._constant_names = set(self._constants)
    def _iter_base_hierarchies(self, reverse=False, t=None):
        t = t or type(self)
        if not reverse:
            yield t
        for base in t.__bases__:
            for supertype in self._iter_base_hierarchies(reverse, base):
                yield supertype
        if reverse:
            yield t
    def iter_type_resolution_order(self, reverse=False):
        """ Iterate over base classes by normal or reverse resolution order. """
        types = set()
        for type in self._iter_base_hierarchies(reverse=reverse):
            if type in types:
                continue
            types.add(type)
            yield type
    def _get_constants(self):
        # Get constants dict for this type and all of its supertypes. Keys are
        # in upper-case.
        constants = {}
        for cls in self.iter_type_resolution_order(reverse=True):
            class_constants = getattr(cls, "CONSTANTS", None)
            if class_constants:
                for key, value in class_constants.iteritems():
                    constants[key.upper()] = value
        return constants
    def get_constant(self, name):
        """ Get constant value by constant name. If no such constant is defined
        for this type or any of its supertypes, :py:const:`NOTHING` is returned. """
        return self._constants.get(name.upper(), NOTHING)
    def get_constant_name(self, value):
        """ Get constant name by constant value. If no such constant is defined
        for this type or any of its superclasses, `None` is returned. If more
        than one constant matches the specified value, there's no guarantee
        which of the matching names is returned. """
        for k, v in self._constants.iteritems():
            if v == value:
                return k
        return None
    def verify(self, arg, value):
        """ Verify actual value. This method should be invoked after the actual
        value had been computed, so the value may not be a string. Argument name
        is only used for raising errors. When verification fails, an exception is
        raised. This method should not be passed constant values, for they
        should never be verified. """
        error_message = self._verify(value)
        if error_message:
            self._raise_value_error(arg, value, error_message)
    def verify_raw(self, arg, value):
        """ Verify raw value. This method should be invoked before the actual
        value had been computed, so the value is expected to be a string.
        Argument name is only used for raising errors. When verification fails,
        an exception is raised. This method should not be passed constant
        values, for they should never be verified. """
        error_message = self._verify_raw(value)
        if error_message:
            self._raise_value_error(arg, value, error_message)
    def _raise_value_error(self, arg, value, error):
        msg = u"'{}' {}".format(arg, error)
        if value is not None and not (isinstance(value, string_types) and not value.strip()) and \
                not error.startswith('is set to '):
            msg = u"{}: {}".format(msg, value)
        raise BadValue(msg)
    def _verify(self, value):
        """ Implement this method for custom verification of actual value. When
        verification is successful, the return value must be None. Otherwise,
        the return value must be an error string. """
        pass
    def _verify_raw(self, value):
        """ Implement this method for custom verification of raw value.  When
        verification is successful, the return value must be None. Otherwise,
        the return value must be an error string. """
        pass
    def get_actual(self, arg, value):
        """ Compute actual value from raw value. Argument name is only used for
        raising errors. When computation of actual value fails, an exception is
        raised. This method should not be passed constant values, for they
        never need any conversion. """
        success, value_or_error = self._get_actual(value)
        if not success:
            self._raise_value_error(arg, value, value_or_error)
        return value_or_error
    def _get_actual(self, value):
        """ Implement this method for custom computation of actual value. The
        return value is a 2-item tuple of which first element is a boolean,
        denoting whether or not the computation succeeded. If the computation
        succeeded, then the second item holds the result. Otherwise, the second
        item holds an error string.

        If the actual value should be computed within the context of the
        command, i.e just before invoking the command method, then the second
        item should be an instance of
        :py:class:`~infi.shell.commands.commands.DynamicValue`. This is useful
        when value computation failures should be considered as command
        failures, rather than argument parsing failures.
        :py:class:`~infi.shell.types.Dynamic` is a base type providing just this
        logic.
        """
        return True, value
    def _register_value_provider(self, provider):
        self._value_providers.append(provider)
    def get_possible_values(self, _api, starts_with=None):
        """
        Get possible values for auto-completion. In most cases, subtypes should
        override this method. The returned values must be valid textual argument
        values, i.e values which the user could've typed. Order is not
        important.

        :param _api: API object
        :param starts_with: When specified, only values starting with this
                            string should be returned
        :returns: An iterable of strings.
        """
        return chain(*[p.get_possible_values(type(self), _api, starts_with) for p in self._value_providers])
    def get_all_possible_values(self, _api, starts_with=None):
        """
        Get all possible values, including constants. Subtypes should not
        override this method.

        :returns: An iterable of strings.
        """
        return chain(
            self._constant_names,  # Constant names
            self.get_possible_values(_api, starts_with=starts_with),
            )
    def get_colliding_values(self, _api, existing_values, possible_values, current_value):
        """ Overide this method in order to get all forbidden values based on entered values and current partial value.
        """
        return ()
    def is_cacheable(self):
        return self.CACHEABLE
    @classmethod
    def get_name(cls):
        """ Argument type name. Subclasses generally have no reason to override
        this, but they may. """
        return cls.__name__.upper()
    def get_description(self):
        """ Argument type description, for the interactive help and
        documentation purposes. Subclasses must implement this method. """
        return "N/A"
    def __repr__(self):
        return "ArgType(name={!r})".format(self.get_name())
    def __eq__(self, other):
        return type(self).__name__ == type(other).__name__
    def __hash__(self):
        return hash(type(self).__name__) * hash(id(self))
    List = ListCreator()
    """ This attributes holds an auto-generated List subtype, of which element
    type is this type. Use this attribute as a type, when multiple values are
    needed. """

class Int(ArgType):
    """ Simple integer type. If value begins with "0x", then it's treated as a
    hexadecimal integer. """
    _pattern = re.compile(r"^\s*-?((0x[0-9a-f]+)|(\d+))\s*$", re.I)
    _hex_pattern = re.compile(r"^\s*-?0x[0-9a-f]+\s*$", re.I)
    def _verify_raw(self, value):
        if not self._pattern.match(value):
            return "is not an integer"
    def _get_actual(self, value):
        try:
            base = 16 if self._hex_pattern.match(value) is not None else 10
            return True, int(value, base)
        except (ValueError, TypeError):
            return False, "is not an integer"
    def get_description(self):
        return "An integer number"

class Range(Int):
    """ Base class for integers within a range of integers (inclusive, and
    supports hex). Subclasses must override cls.MIN and cls.MAX, or implement
    _get_min() and _get_max(). """
    MIN = None
    """ Override this attribute to set minimum value. """
    MAX = None
    """ Override this attribute to set maximum value. """
    def _get_min(self):
        """ Override this method to set minimum value. """
        return self.MIN
    def _get_max(self):
        """ Override this method to set maximum value. """
        return self.MAX
    def _verify(self, value):
        result = super(Range, self)._verify(value)
        if result is not None:
            return result
        min, max = self._get_min(), self._get_max()
        if min is None:
            if max is None:
                # No range, any value is valid
                return None
            # Only max
            return None if value <= max else "is set to {}, but cannot be greater than {}".format(value, max)
        elif max is None:
            # Only min
            return None if value >= min else "is set to {}, but cannot be smaller than {}".format(value, min)
        else:
            # Min and max
            return None if (min <= value <= max) else \
                   "is set to {}, but cannot be smaller than {} or greater than {}".format(value, min, max)
    def get_description(self):
        min, max = self._get_min(), self._get_max()
        if min is None:
            if max is None:
                # No range, any value is valid
                return super(Range, self).get_description()
            # Only max
            return "An integer number equal to or greater than {}".format(max)
        elif max is None:
            # Only min
            return "An integer number equal to or lesser than {}".format(min)
        else:
            # Min and max
            return "An integer number between {} and {} (inclusive)".format(min, max)

class Natural(Range):
    """ Predefined range: non-negative numbers. """
    MIN = 0
    def get_description(self):
        return "A positive integer number, or zero"

class Positive(Range):
    """ Predefined range: positive numbers. """
    MIN = 1
    def get_description(self):
        return "A positive integer number, greater than zero"

class Str(ArgType):
    """ Simple string type """
    def get_description(self):
        return "Any sequence of characters"
    def _verify(self, value):
        if not isinstance(value, string_types):
            return "is not a string"
    def _get_actual(self, value):
        return True, value.strip()

class RegexType(ArgType):
    """ Base type for values which must conform to certain patterns. Subclasses
    must implement either ``_get_pattern()`` or ``_get_patterns()``, and
    ``_get_error()``, and may implement ``_get_actual()`` in order to transform
    the value (transformation will happen after regex verification). Each
    pattern may either be a compiled pattern, or a string, in which case
    ``_get_flags()`` is used for specifying regex flags. ``_should_match()`` may
    also be overridden to return ``False``, denoting the input must not match
    any of the patterns. """
    def __init__(self):
        super(RegexType, self).__init__()
        self._patterns = [p if not isinstance(p, string_types) else re.compile(p.rstrip("$") + "$", self._get_flags())
                          for p in self._get_patterns()]
        self._error = self._get_error()
    def _get_pattern(self):
        """
        Subclasses must either implement this method, or ``_get_patterns()``.

        :returns: String regex pattern.
        """
        raise NotImplementedError
    def _get_patterns(self):
        return [self._get_pattern()]
    def _get_flags(self):
        """
        Subclasses may implement this method, which only takes effect on
        precompiled patterns.

        :returns: Logical AND of regex flags.
        """
        return 0
    def _get_error(self):
        """
        Subclasses must implement this method.

        :returns: Error string, excluding the subject, e.g "is not valid".
        """
        raise NotImplementedError
    def _verify_raw(self, value):
        error = super(RegexType, self)._verify_raw(value)
        if error:
            return error
        matched = self._get_match(value) is not None
        if matched != self._should_match():
            return self._error
    def _get_match(self, value):
        for p in self._patterns:
            m = p.match(value)
            if m:
                return m
        return None
    def _should_match(self):
        """
        Subclasses may implement this method, denoting whether the input should
        or shouldn't match the patterns.

        :returns: Boolean.
        """
        return True

class EnumType(ArgType):
    """ Base type for a predefined list of values. Subclasses must implement
    ``_get_values()``, and may implement ``_get_actual()`` in order to
    transform the value (transformation will happen after verification). """
    def __init__(self):
        super(EnumType, self).__init__()
        self._values = tuple(str(v) for v in self._get_values())
        self._value_mapping = {v.upper() : v for v in self._values}
        self._error = "is not one of [{}]".format(", ".join(self._values))
    def _get_values(self):
        """
        Subclasses must implement this method. This method must not return an
        actual iterator.

        :returns: An iterable of all valid values.
        """
        raise NotImplementedError
    def _verify_raw(self, value):
        if not value or value.upper() not in self._value_mapping:
            return self._error
    def _get_actual(self, value):
        is_successful, value = super(EnumType, self)._get_actual(value)
        if not is_successful:
            return is_successful, value
        return True, self._value_mapping[value.upper()]
    def get_description(self):
        if len(self._values) == 1:
            return "Only one value: {}".format(self._values[0])
        all_but_last = ", ".join(self._values[:-1])
        return "{} or {}".format(all_but_last, self._values[-1])
    def _register_value_provider(self, provider):
        raise TypeError("{} cannot register value providers".format(type(self).__name__))
    def get_possible_values(self, _api, starts_with=None):
        if starts_with:
            if self.AUTO_COMPLETION_CASE_INSENSITIVE:
                return [const for const in self._values if str(const).lower().startswith(starts_with.lower())]
            else:   
                return [const for const in self._values if str(const).startswith(starts_with)]
        return self._values

class ChoiceType(TypeChoicesHelpMixin, EnumType):
    """ Like :py:class:`~EnumType`, but interactive help displays the actual
    choices instead of the type name. """
    def get_type_name_parts(self):
        return self._values

class MappedEnumType(EnumType):
    """ Base type for a predefined list of raw values, mapped to actual values.
    Subclasses must implement ``_get_mapping()``, which is invoked once when
    the type is instantiated, and must return a dictionary of which keys and
    values are the raw and actual values, respectively. """
    def __init__(self):
        self._mapping = OrderedDict(self._get_mapping())
        super(MappedEnumType, self).__init__()
    def _get_values(self):
        return tuple(iterkeys(self._mapping))
    def _get_mapping(self):
        """
        Subclasses must implement this method.

        :returns: A dictionary (or an iterable of 2-tuples) of which keys and
                  values are the raw and actual values, respectively.
        """
        raise NotImplementedError
    def _get_actual(self, value):
        is_successful, value = super(MappedEnumType, self)._get_actual(value)
        if not is_successful:
            return is_successful, value
        # At this stage the value had been validated, so no need to worry about KeyError
        return True, self._mapping[value]

class MappedChoiceType(TypeChoicesHelpMixin, MappedEnumType):
    """ Like :py:class:`~MappedEnumType`, but interactive help displays the
    actual choices instead of the type name. """
    def get_type_name_parts(self):
        return self._values

class Bool(ArgType):
    """ Boolean value, accepting the following flavors: true, false, yes, no,
    on, off. Any case is accepted. """
    _true = "true", "yes", "on"
    _false = "false", "no", "off"
    _all_values = tuple(chain(*izip(_true, _false)))
    def _get_actual(self, value):
        value = value.lower()
        if value in self._true:
            return True, True
        elif value in self._false:
            return True, False
        return False, "is not a valid boolean value"
    def get_description(self):
        return "A boolean: {}".format(", ".join(self._all_values))
    def _register_value_provider(self, provider):
        raise TypeError("{} cannot register value providers".format(type(self).__name__))
    def get_possible_values(self, _api, starts_with=None):
        if starts_with:
            return tuple(v for v in self._all_values if v.startswith(starts_with))
        return self._all_values

class YesNo(TypeChoicesHelpMixin, MappedEnumType):
    @classmethod
    def get_type_name_parts(cls):
        return ("YES", "NO", "Y", "N")
    def get_description(self):
        return "Either yes, or no"
    def _get_mapping(self):
        return dict(yes=True, no=False)
    def _verify_raw(self, value):
        if not value.upper() in self.get_type_name_parts():
            return "is not one of [{}]".format(", ".join((v.lower() for v in self.get_type_name_parts())))
    def _get_real_values(self):
        return dict(yes=True, no=False, y=True, n=False)
    def _get_actual(self, value):
        return True, self._get_real_values()[value.lower()]

class ListMeta(type):
    def __new__(mcls, name, bases, attrs):
        # Create class as usual
        newcls = type.__new__(mcls, name, bases, attrs)
        if name == "_WithMetaclassBase":
            return newcls

        # Type validations
        if not ArgType in bases:
            if newcls.ELEMENT_TYPE is None:
                raise TypeError("List types must declare an ELEMENT_TYPE attribute, but {} doesn't "
                                "declare it.".format(name))
            for attr in (
                "CONSTANTS",
                "CACHEABLE",
                "get_possible_values()",
                ):
                is_method = attr.endswith("()")
                if is_method and attr in attrs or not is_method and getattr(newcls, attr, None) is not None:
                    raise TypeError('List types mustn\'t declare {subject} "{attr}", but {name} declares it. '
                                    'Declare "{attr}" in element type instead.'.format(
                                        subject="method" if is_method else "attribute",
                                        attr=attr,
                                        name=name,
                                        ))
            own_properties = None
            if "PROPERTIES" in attrs:
                if not (newcls.PROPERTIES is None or isinstance(newcls.PROPERTIES, dict)):
                    raise TypeError('{} type "PROPERTIES" attribute: expected dict or None, but got: {!r}'.format(name, type(newcls.PROPERTIES)))
                if newcls.PROPERTIES:
                    for key in newcls.PROPERTIES:
                        if key not in newcls.DEFAULT_PROPERTIES:
                            raise TypeError('{} type "PROPERTIES" attribute: invalid property: {}. Valid properties: {}'
                                            .format(name, key, ", ".join(sorted(newcls.DEFAULT_PROPERTIES))))
                # Save own properties aside, and remove attribute such that newcls.PROPERTIES will refer to
                # superclass properties, because we want to use Python's MRO rather than search the bases manually
                own_properties = newcls.PROPERTIES
                del newcls.PROPERTIES
            # Compute list's properties. The reason behind this is to enable the developer to override only part of
            # the properties, e.g if a base class has {'foo': 0, 'bar': 1}, a subclass may define {'foo': 2} in
            # order to override 'foo', and 'bar' will be kept as 1.
            final_properties = dict(newcls.DEFAULT_PROPERTIES)    # Defaults
            final_properties.update(newcls.PROPERTIES)            # Superclass properties
            if own_properties:
                final_properties.update(own_properties)           # Own properties
            newcls.PROPERTIES = final_properties
        # Create and add relevant regex patterns
        re_esc_separator = re.escape(newcls.SEPARATOR)
        re_esc_escape = re.escape(newcls.ESCAPE)
        re_repl_escape = mcls._escape_replacement(newcls.ESCAPE)
        re_repl_separator = mcls._escape_replacement(newcls.SEPARATOR)
        # Splitter patterns
        unstripped_list_split_pattern = r"(?<!(?<!{esc}){esc}){sep}".format(esc=re_esc_escape, sep=re_esc_separator)
        newcls._unstripped_list_split_pattern = re.compile(unstripped_list_split_pattern)
        list_split_pattern = r"\s*{}\s*".format(unstripped_list_split_pattern)
        newcls._list_split_pattern = re.compile(list_split_pattern)
        # Escaping and unescaping patterns
        newcls._unescape_patterns = (
            # Double escape
            (re.compile(re_esc_escape * 2), re_repl_escape),
            # Escaped list separator
            (re.compile(re_esc_escape + re_esc_separator), re_repl_separator),
            )
        newcls._escape_patterns = (
            # Double escape
            (re.compile(re_esc_escape), re_repl_escape * 2),
            # Escaped list separator
            (re.compile(re_esc_separator), re_repl_escape + re_repl_separator),
            )
        return newcls
    @classmethod
    def _escape_replacement(mcls, s):
        return "\\\\" if s == "\\" else s


class List(ArgType, with_metaclass(ListMeta)):
    """
    Base type for list types. Subclasses must define ``ELEMENT_TYPE``, which
    must be a subclass of :py:class:`ArgType`.

    Alternatively, list types can be created dynamically, e.g: ``Int.List``, or
    ``List.of(Str, allow_duplicates=False)``. See ``DEFAULT_PROPERTIES``
    attribute for available keyword arguments.
    """
    _IS_LIST = True
    __dynamic_subtypes__ = dict()

    ELEMENT_TYPE = None
    """ List element type. This attribute must be set to an :py:class:`ArgType`
    subclass. """

    SEPARATOR = ","
    """ Literal to be used for separating list elements. """

    ESCAPE = "\\"
    """ Literal to be used for escaping list separators. When this literal is
    used twice in succession, it is interpreted as a double escape. """

    CACHEABLE = None
    """ This attribute mustn't be overridden. Its value is inherited from the
    list's element type. """

    CONSTANTS = None
    """ This attribute mustn't be overridden. Its value is inherited from the
    list's element type. """

    DEFAULT_PROPERTIES = dict(
        allow_duplicates=True,  # Whether or not to allow duplicate list items
        )

    PROPERTIES = dict(DEFAULT_PROPERTIES)
    """ List's caracteristic properties. Override this attribute in order to
    override fields in ``DEFAULT_PROPERTIES``. Note that the value must be a
    dictionary, but doesn't have to contain all of the keys that are defined in
    ``DEFAULT_PROPERTIES``.
    """


    def __init__(self, *args, **kwargs):
        super(List, self).__init__(*args, **kwargs)
        self._element_type_instance = self.ELEMENT_TYPE()
    def _iter_unescaped_values(self, values, unstrip=True):
        split = (self._list_split_pattern if unstrip else self._unstripped_list_split_pattern).split(values)
        return (self._unescape(v) for v in split)
    def _unescape(self, value):
        for pattern, replacement in self._unescape_patterns:
            value = pattern.subn(replacement, value)[0]
        return value
    def _escape(self, value):
        for pattern, replacement in self._escape_patterns:
            value = pattern.subn(replacement, value)[0]
        return value
    def _verify(self, value):
        # Do nothing here - all verifications are done during get_actual() in this type
        pass
    def _verify_raw(self, value):
        # Do nothing here - all verifications are done during get_actual() in this type
        pass
    def get_actual(self, arg, value):
        """ In this type, this method performs raw value verification, value
        conversion, and actual value verification. This method has to be
        overridden instead of:py:meth:`_get_actual()`, since ``arg`` is needed
        for verification error messages. """
        values = []
        unescaped_values = list(self._iter_unescaped_values(value))
        existing_values = set()
        total_values = len(unescaped_values)
        for i, raw_element_value in enumerate(unescaped_values, 1):
            # If this is the last value and it's blank, that's the end of the list,
            # unless there is only one element in the list, in which case this is
            # an empty value which should be processed anyway
            if i == total_values and i != 1 and not raw_element_value.strip():
                break
            if not self.PROPERTIES["allow_duplicates"]:
                if raw_element_value in existing_values:
                    error = 'list contains a duplicate item'
                    self._raise_value_error(arg, raw_element_value, error)
                existing_values.add(raw_element_value)
            # If raw value represents a constant declared in this type, add the constant's value
            actual_element_value = self._element_type_instance.get_constant(raw_element_value)
            if actual_element_value is not NOTHING:
                if issubclass(self.ELEMENT_TYPE, Dynamic):
                    actual_element_value = DynamicValue(actual_element_value, self._dummy_dynamic_func)
                values.append(actual_element_value)
                continue
            # Verify raw value
            error = self._element_type_instance._verify_raw(raw_element_value)
            if error:
                if total_values > 1:
                    error = 'list item {} {}{}'.format(i, '("{}") '.format(raw_element_value) if raw_element_value else '', error)
                self._raise_value_error(arg, value, error)
            # Get actual value
            success, actual_element_value = self._element_type_instance._get_actual(raw_element_value)
            if not success:
                # actual_element_value is an error message, since success == False
                if total_values > 1:
                    error = 'list item {} {}{}'.format(i, '("{}") '.format(raw_element_value) if raw_element_value else '', actual_element_value)
                self._raise_value_error(arg, value, error)
            # Verify actual value
            error = self._element_type_instance._verify(actual_element_value)
            if error:
                if total_values > 1:
                    error = 'list item {} {}{}'.format(i, '("{}") '.format(raw_element_value) if raw_element_value else '', error)
                self._raise_value_error(arg, value, error)
            values.append(actual_element_value)
        if issubclass(self.ELEMENT_TYPE, Dynamic):
            values = DynamicValue(values, self._get_dynamics)
        return values
    def _dummy_dynamic_func(self, api, arg, value):
        return arg, value
    def _get_dynamics(self, api, arg, dynamics):
        return arg, [dynamic.get(api, arg)[1] for dynamic in dynamics]
    def get_possible_values(self, _api, starts_with, get_element_type_possible_values, get_element_type_colliding_values):
        if starts_with:
            # Split input
            stripped_elements = list(self._iter_unescaped_values(starts_with))
            complete_elements = set(stripped_elements[:-1])
            unstripped_elements = list(self._iter_unescaped_values(starts_with, unstrip=False))
            complete_unstripped_elements = unstripped_elements[:-1]
            padding = unstripped_elements[-1]
            padding = padding if not padding.strip() else ""
            # Get possible values (for last element only) from element type (consider caching)
            values = get_element_type_possible_values(self._element_type_instance, stripped_elements[-1])
            if not values:
                return ()
            # rlcontext computes and caches possible values with padding, so remove padding
            values = (v[:-1] for v in values)
            # Filter out values which already exist in the list
            values = (v for v in values if v not in complete_elements)
            # Filter out values that are not acceptable
            possible_values = list(values)
            values = (i for i in possible_values)
            values = (v for v in values if v not in get_element_type_colliding_values(self._element_type_instance, complete_elements, possible_values, stripped_elements[-1]))
            # Return possible values, each an escaped list
            return (self.SEPARATOR.join(self._escape(element) for element in chain(complete_unstripped_elements, [padding + value + " "])) for value in values)
        # Get possible values from element type as usual
        values = get_element_type_possible_values(self._element_type_instance, None)
        return (self._escape(v) for v in values) if values else ()

    @classmethod
    def of(cls, element_type, **properties):
        """ Shorthand method for creating a list type of the specified element
        type. Created types are cached, so consecutive calls return the same
        class object. """
        key = (element_type, None if not properties else tuple(sorted(properties.iteritems())))
        cached = cls.__dynamic_subtypes__.get(key)
        if not cached:
            cached = cls.__dynamic_subtypes__[key] = cls.__create_dynamic_subtype__(element_type, properties)
        return cached
    @classmethod
    def __create_dynamic_subtype__(cls, element_type, properties):
        name = element_type.get_name()
        description = element_type().get_description() + " (multiple values, separated by commas)"
        case_insensitive = element_type().AUTO_COMPLETION_CASE_INSENSITIVE
        class subtype(cls):
            AUTO_COMPLETION_CASE_INSENSITIVE = case_insensitive
            ELEMENT_TYPE = element_type
            if properties:
                PROPERTIES = properties
            @classmethod
            def get_name(self):
                return name
            def get_description(self):
                return description
        subtype.__name__ = element_type.__name__ + 'List'
        return subtype

class Measure(ArgType):
    """ Base type for measures (e.g distance, volume, storage capacity, etc').

    Subclasses must implement:

    ==================          ===================================================
    Method                      Description
    ==================          ===================================================
    _get_error()                returns generic error string
    _get_units()                returns an iterable of unit symbols
    _process_actual()           gets amount and unit, and returns actual value
    ==================          ===================================================

    Subclasses may implement:

    ======================      ===================================================
    Method                      Description
    ======================      ===================================================
    _is_negative_enabled()      returns boolean
    _get_fraction_digits()      returns None or 0 for integer, -1 for unlimited,
                                positive number for limited
    ======================      ===================================================
    """
    def __init__(self):
        super(Measure, self).__init__()
        p = ["("]
        if self._is_negative_enabled():
            p.append("-?")
        p.append("[0-9]+")
        fraction_digits = self._get_fraction_digits()
        if fraction_digits:
            p.append("(?:\\.[0-9]{})?".format("+" if fraction_digits < 0 else "{{1,{}}}".format(fraction_digits)))
        p.append(")({})?$".format("|".join(self._get_units())))
        self._pattern = re.compile("".join(p), re.IGNORECASE)
    def _verify(self, value):
        if not isinstance(value, Number) or (not self._is_negative_enabled() and value < 0):
            return self._get_error()
    def _get_actual(self, value):
        match = self._pattern.match(value)
        if not match:
            return False, self._get_error()
        size = Decimal(match.group(1))
        if size == int(size):
            size = int(size)
        unit = match.group(2)
        actual = self._process_actual(size, unit)
        if actual is None:
            return False, self._get_error()
        return True, actual
    def _process_actual(self, size, unit):
        """
        Subclasses must implement this method.

        :param size: measured amount
        :param unit: measured unit string
        :returns: Actual value.
        """
        raise NotImplementedError
    def _get_error(self):
        """
        Subclasses must implement this method.

        :returns: Error string, excluding the subject, e.g "is not valid".
        """
        raise NotImplementedError
    def _is_negative_enabled(self):
        """
        Subclasses may implement this method.

        :returns: Boolean, defaults to ``True``.
        """
        return True
    def _get_fraction_digits(self):
        """
        Subclasses may implement this method.

        :returns: ``None`` or 0 for integer, -1 for unlimited, positive number
                  for limited.
        """
        return -1
    def _get_units(self):
        """
        Subclasses must implement this method.

        :returns: An iterable of unit strings.
        """
        raise NotImplementedError
    NUMBER_PATTERN = re.compile(r"^((?:\.\d+)|(?:\d+(?:\.\d+)?))")
    def _iter_possible_values(self, starts_with):
        matches = self.NUMBER_PATTERN.findall(starts_with)
        if not matches:
            return
        number = matches[0]
        for raw_unit in self._get_units():
            for unit in (raw_unit.lower(), raw_unit.upper()):
                complete_expression = "{}{}".format(number, unit)
                if complete_expression.startswith(starts_with):
                    yield complete_expression
    def get_possible_values(self, _api, starts_with=None):
        if not starts_with:
            return []
        return list(self._iter_possible_values(starts_with))

class Timestamp(ArgType):
    """ Timestamp type. Subclasses may override DATE_PATTERN and/or
    DATE_TIME_PATTERN. """
    DATE_PATTERN = "%Y-%m-%d"
    DATE_TIME_PATTERN = "%Y-%m-%d %H:%M:%S"
    def _get_actual(self, value):
        value = value.strip().lower()
        timestamp = self._parse_date(value, self.DATE_PATTERN, self.DATE_TIME_PATTERN)
        if not timestamp:
            return False, "is not a valid timestamp"
        return True, int(timestamp*1000)
    def _parse_date(self, date, *formats):
        for format in formats:
            try:
                return time.mktime(time.strptime(date, format))
            except ValueError:
                pass
        return None
    def get_description(self):
        return "A timestamp, e.g: '2011-04-13', '2011-04-13 18:30:12'"

class TimeDelta(ArgType):
    """ TimeDelta type. Input format is hh:mm:ss and the actual value is the number of seconds. """
    TIME_PATTERN = re.compile(r'^(\d+)(?::([0-5]?\d)(?::([0-5]?\d))?)?$')
    LIMIT_SECONDS = None
    def _get_actual(self, value):
        match = self.TIME_PATTERN.match(value)
        if not match:
            return False, "is not a valid time delta"
        else:
            result = 0
            for i in match.groups():
                if i is not None:
                    result *= 60
                    result += int(i)
        return True, result
    def _verify(self, value):
        if self.LIMIT_SECONDS:
            if value > self.LIMIT_SECONDS:
                return "interval limit is {}".format(timedelta(seconds=self.LIMIT_SECONDS))
    def get_description(self):
        return "A time delta, e.g: '02:02', '05:11:06'"

class OctalFilePermissions(RegexType):
    """
    UNIX-style numeric file permissions, e.g: 505, 777, etc'. The resulting
    object is an integer, but the value is octal, so e.g "755" is translated
    to an integer of value "755".
    """
    def _get_pattern(self):
        return "[0-7]{3}"
    def _get_actual(self, value):
        try:
            return True, self._to_int(value)
        except ValueError as e:
            # This is never supposed to happen
            return False, "Conversion error: " + str(e)
    def _to_int(self, string_value):
        return int(string_value)
    def _get_error(self):
        return "is not a valid UNIX file permission"
    def get_description(self):
        return 'UNIX-style file permissions, e.g "505", "777", etc\''

class DecimalFilePermissions(OctalFilePermissions):
    """
    UNIX-style numeric file permissions, e.g: 505, 777, etc'. The resulting
    object is an integer, and the value is decimal, so e.g "755" is translated
    to an integer of value "493".
    """
    def _to_int(self, string_value):
        # 8 is the base of the source string so the result is translated to decimal
        return int(string_value, 8)

class IPAddress(RegexType):
    """
    IPv4 address. This type processes an IP address into a dictionary with the
    following keys: ip, ip_parts, ip_int. The resulting dictionary also supports
    attribute access.
    """
    PATTERN = r"(?P<ip>((25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?))"
    def _get_pattern(self):
        return self.PATTERN
    def _get_actual(self, value):
        return True, self._build_result(self._get_match(value))
    def _build_result(self, match):
        ip = match.group("ip")
        return Munch(
            ip=ip,
            ip_parts=ip.split("."),
            ip_int=networking.ip_to_int(ip),
            )
    def _get_error(self):
        return "is not a valid IPv4 address"
    def get_description(self):
        return "A valid IPv4 address"

class IPAddressCIDR(IPAddress):
    """
    IPv4 address with CIDR notation, e.g: 192.168.0.1/24. This type processes an
    IP address with CIDR notation into a dictionary with the following keys: ip,
    ip_parts, ip_int, cidr, ip_and_cidr, netmask. The resulting dictionary also
    supports attribute access.
    """
    TYPE_NAME = "IPADDRESS/CIDR"
    def _get_pattern(self):
        return super(IPAddressCIDR, self)._get_pattern() + r"/(?P<cidr>[0-9]|[1-2][0-9]|3[0-2])"
    def _build_result(self, match):
        result = super(IPAddressCIDR, self)._build_result(match)
        cidr = int(match.group("cidr"))
        result.update(
            cidr=cidr,
            ip_and_cidr=match.string,
            netmask=networking.cidr_to_netmask(cidr),
            )
        return result
    def _get_error(self):
        return "is not a valid IPv4-CIDR pair"
    @classmethod
    def get_name(cls):
        return cls.TYPE_NAME
    def get_description(self):
        return "A valid IPv4 network address in CIDR notation, e.g. 192.168.1.0/24"

class Netmask(EnumType):
    """
    IPv4 netmask. This type processes an IP netmask into a dictionary with the
    following keys: netmask, netmask_parts, cidr. The resulting dictionary also
    supports attribute access.
    """
    def _get_values(self):
        return [
            "255.255.255.255", "255.255.255.254", "255.255.255.252", "255.255.255.248",
            "255.255.255.240", "255.255.255.224", "255.255.255.192", "255.255.255.128",
            "255.255.255.0", "255.255.254.0", "255.255.252.0", "255.255.248.0",
            "255.255.240.0", "255.255.224.0", "255.255.192.0", "255.255.128.0",
            "255.255.0.0", "255.254.0.0", "255.252.0.0", "255.248.0.0",
            "255.240.0.0", "255.224.0.0", "255.192.0.0", "255.128.0.0",
            "255.0.0.0", "254.0.0.0", "252.0.0.0", "248.0.0.0",
            "240.0.0.0", "224.0.0.0", "192.0.0.0", "128.0.0.0",
            "0.0.0.0",
            ]
    def _get_actual(self, value):
        success, netmask = super(Netmask, self)._get_actual(value)
        if not success:
            return False, netmask
        return True, Munch(
            netmask=netmask,
            netmask_parts=netmask.split("."),
            cidr=networking.netmask_to_cidr(netmask),
            )
    def get_description(self):
        return "An IPv4 netmask"

class Hostname(RegexType):
    """ IPv4 address or RFC-1123 hostname. """
    def _get_patterns(self):
        return (
            IPAddress.PATTERN,
            r"^(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]*[a-zA-Z0-9])\.)*([A-Za-z0-9]|[A-Za-z0-9][A-Za-z0-9\-]*[A-Za-z0-9])$",
            )
    def _get_error(self):
        return "is not a valid hostname or IP address"
    def get_description(self):
        return "A valid hostname or IPv4 address"

class Port(Range):
    """ IPv4 port. """
    MIN = 1
    MAX = 65535

class EmailAddress(RegexType):
    """ Email address. """
    def _get_pattern(self):
        return (
            # dot-atom
            r"(^[-!#$%&'*+/=?^_`{}|~0-9A-Z]+(\.[-!#$%&'*+/=?^_`{}|~0-9A-Z]+)*"
            # quoted-string
            r'|^"([\001-\010\013\014\016-\037!#-\[\]-\177]|\\[\001-011\013\014\016'
            r'-\177])*"'
            # domain
            r')@(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+[A-Z]{2,6}\.?$'
            )
    def _get_flags(self):
        return re.IGNORECASE
    def _get_error(self):
        return "is not a valid email address"
    def get_description(self):
        return "A valid email address"

class Percent(RegexType):
    """ Integer between 0 and 100, optionally followed by a percent sign.
    Subclasses may override ``MIN`` and ``MAX``, or set them to ``None``, in
    order to enable percent values other than 0-100%. """
    MIN = 0
    MAX = 100
    def _get_pattern(self):
        return r"^\s*(?P<value>-?\d+) ?%?\s*"
    def _get_error(self):
        return "is not a valid percent value"
    def _get_actual(self, value):
        match = self._get_match(value)
        assert match
        number_str = match.group("value")
        try:
            return True, int(number_str, 10)
        except (ValueError, TypeError):
            return False, self._get_error()
    def _verify(self, value):
        if self.MIN is not None:
            if self.MAX is not None:
                if value < self.MIN or value > self.MAX:
                    return "is not between {} and {}".format(self.MIN, self.MAX)
            else:
                if value < self.MIN:
                    return "is not greater than {}".format(self.MIN)
        elif self.MAX is not None:
            if value > self.MAX:
                return "is not lesser than {}".format(self.MAX)
    def get_description(self):
        if self.MIN is not None:
            if self.MAX is not None:
                minmax = " between {} and {}".format(self.MIN, self.MAX)
            else:
                minmax = ", greater than {}".format(self.MIN)
        elif self.MAX is not None:
            minmax = ", lesser than {}".format(self.MAX)
        else:
            minmax = ""
        return "An integer{}, optionally followed by a percent symbol".format(minmax)

class Path(ArgType):
    """ A filesystem path, with auto-completion. Path may or may not exist. """
    COMPLETION_CACHE = False
    _VERIFY_EXISTENCE = False
    _ENSURE_IS_DIR = False
    _ENSURE_IS_FILE = False
    def _verify(self, path):
        if self._VERIFY_EXISTENCE:
            if not os.path.exists(path):
                return "path does not exist"
            if not os.access(path, os.R_OK):
                return "path is not readable"
        if self._ENSURE_IS_DIR:
            if not os.path.isdir(path):
                if os.path.isfile(path):
                    return "oath is a file, not a directory"
                return "path is not a directory"
        if self._ENSURE_IS_FILE:
            if not os.path.isfile(path):
                if os.path.isdir(path):
                    return "path is a directory, not a file"
                return "path is not a file"
        return None
    def _get_actual(self, path):
        return True, self._parse_path(path)
    def _parse_path(self, path):
        return os.path.realpath(os.path.expanduser(os.path.expandvars(path)))
    def get_possible_values(self, _api, starts_with=None):
        if not starts_with:
            try:
                return os.listdir(".")
            except Exception:
                return None
        parsed_path = self._parse_path(starts_with)
        dirname = os.path.dirname(parsed_path)  # Returns empty string if dirname is cwd
        basename = os.path.basename(parsed_path)
        if os.path.isdir(dirname):
            results = None
            while results is None:
                results = list(self._iterate_matching_entries(dirname, basename))
                if len(results) == 1 and os.path.isdir(results[0]):
                    results, dirname, basename = None, results[0], ""
            return results
        return None
    def _iterate_matching_entries(self, dirname, basename_prefix):
        for entry in os.listdir(dirname or "."):
            if entry.startswith(basename_prefix):
                fullpath = os.path.join(dirname, entry)
                if os.path.isdir(fullpath):
                    fullpath += os.sep
                yield fullpath
    def get_description(self):
        return "A path name"

class ExistingPath(Path):
    """ An existing filesystem path, with auto-completion. """
    _VERIFY_EXISTENCE = True
    @classmethod
    def get_name(cls):
        return "PATH"
    def get_description(self):
        return "Path of an existing file or directory"

class File(Path):
    """ A filesystem path of an existing file. """
    _VERIFY_EXISTENCE = True
    _ENSURE_IS_FILE = True
    def get_description(self):
        return "Path of an existing file"

class Directory(Path):
    """ A filesystem path of an existing directory. """
    _VERIFY_EXISTENCE = True
    _ENSURE_IS_DIR = True
    def get_description(self):
        return "Path of an existing directory"

class Dynamic(ArgType):
    """ Base type for dynamic types. Dynamic types compute the actual value
    within the context of the command itself, so failures are treated as
    command failures, rather than argument parsing failures. Subclasses must
    implement :py:meth:`._get_dynamic`.
    """
    def _get_actual(self, value):
        return True, DynamicValue(value, self._get_dynamic)
    def _get_dynamic(self, api, arg, value):
        """ Subclasses must implement this method, which should return the
        actual value, or raise an exception. """
        raise NotImplementedError
