from os.path import isfile, isdir, exists as path_exists, dirname, basename, abspath, expanduser
import os
from contextlib import contextmanager
from ..exceptions import FilesystemException

READ = os.R_OK
WRITE = os.W_OK
_mode_names = {
    READ : "read",
    WRITE : "write",
    }

def makedirs(dir, safe=False):
    '''
    Create directories, recursively.

    :param dir: Directory name (or canonical path) to create
    :param safe: Whether or not to suppress errors
    :returns: Insignificant if safe is ``False``. Otherwise, ``True`` if
              operation succeeded, else ``False``.
    '''
    if not os.path.isdir(dir):
        try:
            _check_permissions_recursive(dir, WRITE)
        except:
            if not safe:
                raise
            return False
        try:
            os.makedirs(dir)
        except Exception as e:
            if not safe:
                raise FilesystemException("Failed to create directory \"{}\": {}".format(dir, e))
            return False
        return True

def _check_permissions_recursive(dir, *modes):
    '''
    Check read/write/execute permissions on the specified path. If the path does
    not exist, recurse on the parent directory.
    '''
    if not path_exists(dir):
        _check_permissions_recursive(dirname(dir), *modes)
    for mode in modes:
        if not os.access(dir, mode):
            raise FilesystemException("No {} access for {}".format(_mode_names[mode], dir))

def pathify(*path):
    ''' Represent the specified path parts as one absolute path. '''
    return abspath(os.sep.join(
        expanduser(p.strip().rstrip(os.sep)) for p in (_p for _p in path if _p and _p.strip())
        ))

class DataDir(object):
    ''' Data directory manager. '''
    def __init__(self):
        self.__initialized__ = False
    def __ensure_initialized__(self):
        assert self.__initialized__, "Data directory not initialized"
    def init(self, basepath):
        '''
        Initialize data directory in the specified base path. This method must
        be called in the context of a main function, before having used any
        code that relies on the data directory. Initialization never deletes
        any files, only ensures the data directory exists. If the specified
        base path is not accessible or not writable, data directory resources
        will be unavailable.

        :param basepath: Data directory base path
        '''
        self.basepath = abspath(expanduser(basepath)).rstrip(os.sep)
        makedirs(self.basepath, safe=True)
        self.__initialized__ = True
        self.__paths__ = dict()
    def get_path(self, *path):
        ''' Get absolute path of a resource within the data directory. '''
        self.__ensure_initialized__()
        real_path = self.__paths__.get(path)
        if real_path is None:
            real_path = self.__paths__[path] = pathify(self.basepath, *path)
        return real_path
    @contextmanager
    def open(self, mode, *path):
        '''
        Open the specified file path for reading or writing, relative to the
        data directory.

        :param mode: File mode, the same as for Python's built-in
                     :py:func:`.open`
        :param path: Path parts, the same as for :py:func:`pathify`
        '''
        real_path = self.get_path(*path)
        if isdir(path):
            raise FilesystemException("Path is a directory, not a file: {}".format(real_path))
        with open(real_path, mode) as f:
            yield f
    def read(self, *path):
        '''
        Read and return contents of the specified file path, relative to the
        data directory.

        :param path: Path parts, the same as for :py:func:`pathify`
        '''
        self.__ensure_initialized__()
        if not self.exists(*path):
            raise FilesystemException("File does not exist: {}".format(self.get_path(*path)))
        with self.open('r', *path) as f:
            return f.read()
    def write(self, data, *path):
        '''
        Write the specified data into the specified file path, relative to the
        data directory.

        :param data: Data to write
        :param path: Path parts, the same as for :py:func:`pathify`
        '''
        self.__ensure_initialized__()
        if len(path) > 1:
            makedirs(dirname(self.get_path(*path)))
        with self.open('w', *path) as f:
            f.write(data)
    def exists(self, *path):
        '''
        Test whether the specified path exists, relative to the data directory.

        :param path: Path parts, the same as for :py:func:`pathify`
        :returns: ``True`` if exists, otherwise ``False``.
        '''
        self.__ensure_initialized__()
        return path_exists(self.get_path(*path))
    def is_file(self, *path):
        '''
        Test whether the specified path exists and is a file, relative to the
        data directory.

        :param path: Path parts, the same as for :py:func:`pathify`
        :returns: ``True`` if path exists and is a file, otherwise ``False``.
        '''
        self.__ensure_initialized__()
        return isfile(self.get_path(*path))
    def is_dir(self, *path):
        '''
        Test whether the specified path exists and is a directory, relative to
        the data directory.

        :param path: Path parts, the same as for :py:func:`pathify`
        :returns: ``True`` if path exists and is a directory, otherwise
                  ``False``.
        '''
        self.__ensure_initialized__()
        return isdir(self.get_path(*path))
    def __repr__(self):
        return '<DataDir Singleton>'

datadir = DataDir()
''' Data directory singleton. '''
