from ..rlcontext import get_raw_input_functions

def input_text(prompt=""):
    """ Read and return user input. Auto-completion and history are off.
    
    :param prompt: Message to display.
    :returns: User's input.
    """
    return get_raw_input_functions()[0](prompt)

def input_password(prompt="Password: "):
    """ Read and return user input, without echoing input characters.
    Auto-completion and history are off.
    
    :param prompt: Question or message to display.
    :returns: User's input.
    """
    return get_raw_input_functions()[1](prompt)

def yesno(prompt="Are you sure?"):
    """ Prompt user with a yes/no question.

    :param prompt: Question or message to display.
    :returns: ``True`` if user chose "yes", otherwise ``False``.
    """
    actual_prompt = "{} [y/n] ".format((prompt or "").strip())
    while True:
        answer = input_text(actual_prompt).lower().strip()
        if answer in ("y", "yes"): return True
        if answer in ("n", "no"): return False

def input_choice(prompt, *choices):
    """ Prompt user with a choice, and returns the value associated with that
    choice. The user must type the name of the desired choice, or only the first
    character(s) of the desired choice, if these characters match exactly one of
    the choices unambiguously.

    :param prompt: Question or message to display.
    :param choices: An iterable of 2-tuples, each of which includes a choice
                    word and a value.
    :returns: The value that matches the user's choice.
    """
    actual_prompt = "{} [{}] ".format((prompt or "").strip(), "/".join(choice[0] for choice in choices).strip())
    while True:
        user_choice = input_text(actual_prompt).lower().strip()
        if not user_choice: continue
        matches = []
        matched_value = None
        for choice, value in choices:
            if choice.startswith(user_choice):
                matches.append(choice)
                matched_value = value
        if matches:
            if len(matches) == 1:
                return matched_value
            print("Ambiguous choice, may mean either {} or {}".format(", ".join(matches[:-1]), matches[-1]))
        else:
            print("Invalid choice")

def input_simple_choice(prompt, *choices):
    """ Prompt user with a choice, and returns it. The user must type the name
    of the desired choice, or only the first character(s) of the desired choice,
    if these characters match exactly one of the choices unambiguously.

    :param prompt: Question or message to display.
    :param choices: An iterable of choice words.
    :returns: The user's choice.
    """
    return input_choice(prompt, *((c, c) for c in choices))
