import itertools

from .nothing import NOTHING
from .._compat import string_types, im_func, iteritems, iterkeys, PY2

def getattr_deep(obj, attrs, default=NOTHING):
    '''
    Get canonical attribute from a reference tree.

    :param attrs: Canonical attribute path, separated by dots
    :param default: Default value to return if attribute is not found. If the
                    default is missing and the attribute is not found, then an
                    :py:class:`AttributeError` is raised.
    :returns: Attribute value, or the specified default, or raises an
              :py:class:`AttributeError`
    '''
    curr = obj
    for attr in attrs.split('.'):
        curr = getattr(curr, attr, NOTHING)
        if curr is NOTHING:
            if default is NOTHING:
                raise AttributeError('Instance of type {} has no such nested attribute: {}'.format(type(obj), attrs))
            return default
    return curr

def getitem_deep(obj, attrs, default=NOTHING):
    '''
    Get canonical attribute from a dictionaries tree.

    :param attrs: Canonical attribute path, separated by dots
    :param default: Default value to return if attribute is not found. If the
                    default is missing and the attribute is not found, then an
                    :py:class:`KeyError` is raised.
    :returns: Attribute value, or the specified default, or raises an
              :py:class:`KeyError`
    '''
    curr = obj
    for attr in attrs.split('.'):
        curr = curr.get(attr, NOTHING)
        if curr is NOTHING:
            if default is NOTHING:
                raise KeyError('Instance of type {} has no such nested attribute: {}'.format(type(obj), attrs))
            return default
    return curr

class LazyIterator(object):
    ''' Lazy iterator, which accepts a callback method that returns the actual
    iterator. The callback method will only be called when iteration is
    started, i.e when ``next()`` is called the first time. '''
    def __init__(self, callback):
        self._callback = callback
        self._actual_iterator = None
    def __iter__(self):
        return self
    def next(self):
        if self._actual_iterator is None:
            self._actual_iterator = self._callback()
            assert self._actual_iterator is not None, 'Callback returned None instead of iterator'
        return self._actual_iterator.next()

class Lazy(object):
    ''' Lazy value container, which accepts a callback method that returns the
    actual value. '''
    NOT_INITIALIZED = object()
    def __init__(self, callback):
        self._callback = callback
        self._value = self.NOT_INITIALIZED
    @property
    def value(self):
        ''' Holds the actual value. The first time this attribute is retrieved,
        the callback is invoked. '''
        if self._value is self.NOT_INITIALIZED:
            self._value = self._callback()
        return self._value

def represent(owner, *attrs):
    '''
    Helper function for generating :py:meth:`__repr__` output.

    :param owner: Object owning the attribute, i.e "self"
    :param attrs: Attributes to output
    :returns: A string representation of the owner and its attributes
    '''
    return '<{}: {}>'.format(type(owner).__name__, ", ".join("{}={!r}".format(attr, getattr(owner, attr)) for attr in (attrs)))

class strlist(list):
    ''' An extension of ``list``, for creating lists out of space-separated
    strings. This type of list also supports the subtraction operator, so
    ``list`` may be subtracted from or added to ``strlist``, and vice versa, and
     ``string`` may be subtracted from or added to ``strlist`` as well. '''
    def __init__(self, value=None):
        if value is None:
            list.__init__(self)
            return
        elif isinstance(value, string_types):
            value = value.split()
        elif not hasattr(value, '__iter__'):
            raise TypeError('Expected iterable or string, but got {!r}'.format(type(value)))
        list.__init__(self, value)
    def __add__(self, value):
        if isinstance(value, string_types):
            return strlist(itertools.chain(self, (value,)))
        if hasattr(value, '__iter__'):
            pending_addition = None if isinstance(value, (list, set, tuple, dict)) else []
            for i, item in enumerate(value):
                if not isinstance(item, string_types):
                    raise TypeError('Iterable item at index {}: expected string, got {!r}'.format(i, type(item)))
                if pending_addition is not None:
                    pending_addition.add(item)
            return strlist(itertools.chain(self, pending_addition if pending_addition is not None else value))
        return NotImplemented
    def __radd__(self, other):
        if isinstance(other, string_types):
            other = (other,)
        elif not hasattr(other, '__iter__'):
            return NotImplemented
        return im_func(strlist.__add__)(other, self)
    def __sub__(self, value):
        if isinstance(value, string_types):
            return strlist(item for item in self if item != value)
        if hasattr(value, '__iter__'):
            pending_removal = set()
            for i, item in enumerate(value):
                if not isinstance(item, string_types):
                    raise TypeError('Iterable item at index {}: expected string, got {!r}'.format(i, type(item)))
                pending_removal.add(item)
            return strlist(item for item in self if item not in pending_removal)
        return NotImplemented
    def __rsub__(self, other):
        if not isinstance(other, (list, tuple)):
            return NotImplemented
        return im_func(strlist.__sub__)(other, self)

class DictProxy(object):
    '''
    Proxy for multiple dictionaries, emulating a single dictionary. The
    constructor accepts multiple dictionaries, such that the precedence order
    is reversed. Also, the proxy allows setting and deleting keys without
    changing the proxied dictionaries. For convenience, the proxy also features
    attribute access.

    Example:
    >>> d1 = {'a': 1, 'b': 2}
    >>> d2 = {'c': 3, 'b': 4}
    >>> p = DictProxy(d1, d2)
    >>> p['a']
    1
    >>> p['b']
    4
    >>> p['c']
    3
    >>> p.b
    4
    >>> sorted(p.iterkeys())
    ['a', 'b', 'c']
    >>> sorted(p.itervalues())
    [1, 3, 4]
    >>> p['b'] = 8
    >>> p['b']
    8
    >>> d1['b']
    2
    >>> d2['b']
    4
    >>> del p['b']
    >>> p['b']
    Traceback (most recent call last):
        ...
    KeyError: 'b'
    >>> d1['b']
    2
    >>> d2['b']
    4
    '''
    class __placeholder__(object):
        def __init__(self, name):
            self._name = '<' + name + '>'
        def __repr__(self):
            return self._name
    __deleted_key__ = __placeholder__('DeletedKey')
    __nothing__ = __placeholder__('Nothing')
    def __init__(self, *dictionaries):
        self.__missing_key__ = self.__placeholder__('MissingKey')
        self.__dictionaries__ = [{}]
        for d in reversed(dictionaries):
            if d is None:
                continue
            if not isinstance(d, (dict, DictProxy)):
                raise TypeError('Expected dict, but got ' + repr(type(d)) + ' instead')
            self.__dictionaries__.append(d)
    def __getitem__(self, item):
        return self.__getitem_impl__(item)
    def __missing__(self, key):
        raise KeyError(key)
    def __getitem_impl__(self, *args):
        item = args[0]
        default = args[1] if len(args) == 2 else self.__missing_key__
        for d in self.__dictionaries__:
            value = d.get(item, self.__missing_key__)
            if value is self.__deleted_key__:
                break
            if value is not self.__missing_key__:
                return value
        if default is not self.__missing_key__:
            return default
        raise KeyError(item)
    def __getattr__(self, attr):
        return self[attr]
    def __setitem__(self, item, value):
        self.__dictionaries__[0][item] = value
    def __setattr__(self, attr, value):
        if attr in ("__dictionaries__", "__missing_key__"):
            object.__setattr__(self, attr, value)
        else:
            self[attr] = value
    def __delitem__(self, item):
        self[item]
        self[item] = self.__deleted_key__
    def __delattr__(self, attr):
        del self[attr]
    def __contains__(self, item):
        value = self.__getitem_impl__(item, self.__nothing__)
        return value is not self.__nothing__
    def __iter__(self):
        return self.iterkeys()
    def __len__(self):
        return len(self.__keyset__())
    def __keyset__(self):
        deleted = set(key for key, value in iteritems(self.__dictionaries__[0]) if value is self.__deleted_key__)
        return set(key for key in itertools.chain(*(iterkeys(d) for d in self.__dictionaries__)) if key not in deleted)
    def iterkeys(self):
        return (key for key in self.__keyset__())
    def keys(self):
        return list(self.iterkeys())
    def viewkeys(self):
        raise NotImplementedError
    def itervalues(self):
        return (self[key] for key in self)
    def values(self):
        return list(self.itervalues())
    def viewvalues(self):
        raise NotImplementedError
    def iteritems(self):
        return ((key, self[key]) for key in self)
    def items(self):
        return list(self.iteritems())
    def viewitems(self):
        raise NotImplementedError
    def clear(self):
        self.__dictionaries__ = [{}]
    def copy(self):
        return {key : value for key, value in self.iteritems()}
    def get(self, key, default=None):
        return self.__getitem_impl__(key, default)
    def has_key(self, key):
        return key in self.iterkeys()
    def pop(self, *args):
        if not args:
            raise TypeError('pop expected at least 1 arggument, got 0')
        if len(args) > 2:
            raise TypeError('pop expected at most 2 arguments, got ' + str(len(args)))
        value = self.__getitem_impl__(*args)
        del self[args[0]]
        return value
    def popitem(self):
        raise NotImplementedError
    def setdefault(self, key, value):
        final_value = self.__getitem_impl__(key, self.__nothing__)
        if final_value is self.__nothing__:
            self[key] = final_value = value
        return final_value
    def update(self, d=None, **fields):
        if d is not None:
            assert not fields
            if PY2:
                tuples = d.iteritems() if hasattr(d, 'iteritems') else d
            else:
                tuples = d.items() if hasattr(d, 'items') else d
        else:
            tuples = iteritems(fields)
        for k, v in tuples:
            self[k] = v
    def __repr__(self):
        return '{{{}}}'.format(', '.join('{!r}: {!r}'.format(key, value) for key, value in self.iteritems()))

class DictEmulator(object):
    def __init__(self, getter_func, contains_func=None):
        self.__getter_func__ = getter_func
        self.__contains_func__ = contains_func
    def __getitem__(self, item):
        return self.__getter_func__(item)
    def __getattr__(self, attr):
        return self[attr]
    def get(self, item, default=None):
        assert self.__contains_func__, 'In order to use ' + type(self).__name__ \
               + '.get(), you must specify contains_func in the constructor'
        if self.__contains_func__(item):
            return self.__getter_func__(item)
        return default

if __name__ == '__main__':
    import doctest
    print(doctest.testmod())
