# The code in this module has been taken from Peter Norvig's article on how to
# write a spelling corrector: http://norvig.com/spell-correct.html
import itertools
from .._compat import xrange

DEFAULT_ALPHABET = 'abcdefghijklmnopqrstuvwxyz'

class SingleSuggestionSpellingCorrector(object):
    """ Spelling corrector that is faster, but returns the first suggestion only. """
    def __init__(self, known_words, alphabet=DEFAULT_ALPHABET):
        """
        :param known_words: Iterable of all words the speller should know
        """
        self._known_words = set(known_words)
        self._alphabet = alphabet
    def edits1(self, word):
        # Note: do not convert these list comprehentions to generators - they will become highly inefficient
        splits     = [(word[:i], word[i:]) for i in xrange(len(word) + 1)]
        deletes    = [a + b[1:] for a, b in splits if b]
        transposes = [a + b[1] + b[0] + b[2:] for a, b in splits if len(b)>1]
        replaces   = [a + c + b[1:] for a, b in splits for c in self._alphabet if b]
        inserts    = [a + c + b     for a, b in splits for c in self._alphabet]
        return set(itertools.chain(deletes, transposes, replaces, inserts))
    def known_edits2(self, word):
        return set(e2 for e1 in self.edits1(word) for e2 in self.edits1(e1) if e2 in self._known_words)
    def known(self, words):
        return set(w for w in words if w in self._known_words)
    def correct(self, word):
        """
        :param word: Word to correct
        :returns: A list of suggestions.
        """
        if word in self._known_words:
            return [word]
        candidates = self.edits1(word)
        for candidate in candidates:
            if candidate in self._known_words:
                return [candidate]
        if len(candidates) < 1000:
            for e1 in candidates:
                for e2 in self.edits1(e1):
                    if e2 in self._known_words:
                        return [e2]
        return []

class MultiSuggestionsSpellingCorrector(object):
    """ Spelling corrector that is slower, but returns multiple suggestions. """
    def __init__(self, known_words, alphabet=DEFAULT_ALPHABET):
        """
        :param known_words: Iterable of all words the speller should know
        """
        self._known_words = set(known_words)
        self._alphabet = alphabet
    def edits1(self, word):
        # Note: do not convert these list comprehentions to generators - they will become highly inefficient
        splits     = [(word[:i], word[i:]) for i in xrange(len(word) + 1)]
        deletes    = [a + b[1:] for a, b in splits if b]
        transposes = [a + b[1] + b[0] + b[2:] for a, b in splits if len(b)>1]
        replaces   = [a + c + b[1:] for a, b in splits for c in self._alphabet if b]
        inserts    = [a + c + b     for a, b in splits for c in self._alphabet]
        return set(deletes + transposes + replaces + inserts)
    def known_edits2(self, word):
        return set(e2 for e1 in self.edits1(word) for e2 in self.edits1(e1) if e2 in self._known_words)
    def known(self, words):
        return set(w for w in words if w in self._known_words)
    def correct(self, word):
        """
        :param word: Word to correct
        :returns: A list of suggestions.
        """
        candidates = self.known([word]) or self.known(self.edits1(word)) or self.known_edits2(word)
        return candidates

SpellingCorrector = SingleSuggestionSpellingCorrector
""" Default spelling corrector. """
