from itertools import islice
import re

__all__ = ['LineHookStreamWrapper', 'ConditionalLineStreamWrapper']

LINE_SPLITTER = re.compile(r"\r?\n\r?")

class HookStreamWrapperBase(object):
    def __init__(self, stream, *hooks):
        self._stream = stream
        self._hooks = list(hooks)
        self._buffer_reset()
    def add_hook(self, hook):
        '''
        :param hook: Hook function to add
        '''
        self._hooks.append(hook)
    def _invoke_hooks(self, line):
        for hook in self._hooks:
            hook(line)
    def _buffer_reset(self):
        self._buffer = []
    def _buffer_append(self, s):
        self._buffer.append(s)
    def _buffer_get_and_reset(self):
        buffer = "".join(self._buffer)
        self._buffer_reset()
        return buffer
    @property
    def closed(self):
        return self._stream.closed
    def isatty(self):
        return self._stream.isatty()
    def writelines(self, iterable):
        for s in iterable:
            self.write(s)
    def write(self, s):
        raise NotImplementedError
    def flush(self):
        self._stream.flush()
    def close(self):
        self._stream.close()
    def __repr__(self):
        return "<{}: {!r}>".format(type(self).__name__, self._stream)

class LineHookStreamWrapper(HookStreamWrapperBase):
    '''
    Stream wrapper that invokes hook functions whenever a line break is
    encountered.

    This stream wrapper implements the same functions as ``sys.stdout``.

    :param stream: Stream to wrap
    :param hooks: Hooks to invoke each time a line break is encountered
    '''
    def write(self, s):
        lines = LINE_SPLITTER.split(s)
        self._buffer_append(lines[0])
        self._stream.write(lines[0])
        # If parameter string holds more than one line
        if len(lines) > 1:
            self._stream.write('\n')
            self._invoke_hooks(self._buffer_get_and_reset())
            # Process remaining lines, except last one which doesn't end with a newline char
            for line in islice(lines, 1, len(lines)-1):
                self._stream.write(line)
                self._stream.write('\n')
                self._invoke_hooks(line)
            # Buffer last line (which is only partial, doesn't end with a newline char)
            self._buffer_append(lines[-1])
            self._stream.write(lines[-1])

class ConditionalLineStreamWrapper(HookStreamWrapperBase):
    '''
    Stream wrapper that uses predicates to determine whether each line should be
    printed or not.

    This stream wrapper implements the same functions as ``sys.stdout``.

    :param stream: Stream to wrap
    :param hooks: Predicate functions that accept a line string (without newline
                  characters) and return a boolean.
    '''
    def _should_write(self, line):
        return all(predicate(line) for predicate in self._hooks)
    def _conditional_write(self, line, newline=True):
        if self._should_write(line):
            self._stream.write(line)
            if newline:
                self._stream.write('\n')
    def write(self, s):
        lines = LINE_SPLITTER.split(s)
        self._buffer_append(lines[0])
        # If parameter string holds more than one line
        if len(lines) > 1:
            self._conditional_write(self._buffer_get_and_reset())
            # Process remaining lines, except last one which doesn't end with a newline char
            for line in islice(lines, 1, len(lines)-1):
                self._conditional_write(line)
            # Buffer last line (which is only partial, doesn't end with a newline char)
            self._buffer_append(lines[-1])
    def flush(self):
        buffered_text = self._buffer_get_and_reset()
        if buffered_text:
            self._conditional_write(buffered_text, newline=False)
        super(ConditionalLineStreamWrapper, self).flush()
