from contextlib import contextmanager
from colorama import Style

class Outputter(object):
    """ Wrapper for stream-like objects, supporting colors, quiet output and
    error handling. """
    def __init__(self, stream, color=None, quiet=False):
        self._stream = stream
        self._color = color
        self._quiet = quiet
        self._closed = False
    @contextmanager
    def _broken_pipe_context(self):
        try:
            yield
        except IOError as e:
            if e.errno == 32:  # Broken pipe
                self._closed = True
            else:
                raise
    def output(self, msg=None, colors=(), consider_quiet=True, nl=True, flush=False):
        """
        :param msg: text to output
        :param colors: colors to prepend to text, if colors are enabled
        :param consider_quiet: whether or not to suppress output if quiet mode is on
        :param nl: whether or not to append line-feed at text end
        :param flush: whether or not to flush explicitly after writing
        """
        if self.closed:
            return
        if consider_quiet and self._quiet:
            return
        if msg is None:
            if nl:
                self._stream.write("\n")
                if flush:
                    self.flush()
            return
        assert isinstance(colors, (list, tuple))
        with self._broken_pipe_context():
            if colors and self._color:
                for color in colors:
                    self._stream.write(color)
            self._stream.write(msg)
            if colors and self._color:
                self._stream.write(Style.RESET_ALL)
            if nl:
                self._stream.write("\n")
            if flush:
                self.flush()
    def write(self, *args, **kwargs):
        """
        Shorthand method for :py:meth:`~output`, specifying ``nl=False``. This
        method accepts the same parameters as :py:meth:`~output`, except ``nl``.
        """
        return self.output(*args, nl=False, **kwargs)
    def colorize(self, msg, colors=()):
        """
        Add colors to string, and return new string. If outputter is configured
        to disallow colors, then the input string is returned as is.
        """
        if not (colors or self._color):
            return msg
        assert isinstance(colors, (list, tuple))
        return "".join(colors) + msg + Style.RESET_ALL
    @property
    def closed(self):
        return self._closed or self._stream.closed
    def flush(self):
        if self.closed:
            return
        with self._broken_pipe_context():
            self._stream.flush()
    def close(self):
        if self.closed:
            return
        with self._broken_pipe_context():
            self._stream.close()
    def get_stream(self):
        return self._stream
    def has_colors(self):
        return self._color
