import re
from munch import Munch
from colorama import Fore, Style
from itertools import count
from ...console import get_console_size
from ..._compat import itervalues, string_types, unicode
from ..getch import getch, is_available as is_getch_available

__all__ = ['PaginatedStreamWrapper', 'PaginationAborted']

CHARS = Munch(
    print_all = "g",
    quit = "q",
    )

_KEYS = ["Space", "Enter"] + [c.upper() for c in itervalues(CHARS)]
_PROMPT = "More..."
MORE_RESULTS_PROMPT = "{} ({})".format(_PROMPT, ", ".join(_KEYS))
_COLORED_KEYS = [Fore.BLUE + key + Fore.RESET for key in _KEYS]
COLORED_MORE_RESULTS_PROMPT = Style.BRIGHT + "{} ({})".format(_PROMPT, ", ".join(_COLORED_KEYS)) + Style.RESET_ALL
_prompt_len = len(MORE_RESULTS_PROMPT)
ERASE_PROMPT_STR = "\10"*_prompt_len + " "*_prompt_len + "\10"*_prompt_len

_enum = count()
ACTIONS = Munch(
    page_down = next(_enum),
    down = next(_enum),
    quit = next(_enum),
    print_all = next(_enum),
    )

LINE_SPLITTER = re.compile(r"\r?\n\r?")

class PaginatedStreamWrapper(object):
    '''
    Stream wrapper that prompts the user each time a full page had been
    written. The user can then choose to advance one line, one page, or disable
    prompting altogether.

    Note that before using this stream, one must call
    :py:meth:`capture_console_dimensions` in order to save the current
    dimensions of the terminal console. This doesn't happen automatically to
    prevent inconsistent output in case the console dimensions (or the way
    they're reported) change during pagination.

    This stream wrapper implements the same functions as ``sys.stdout``.

    :param stream: Stream to wrap
    :param color: Whether or not to output a colored prompt
    '''
    def __init__(self, stream, color=False):
        self._stream = stream
        self._buffer_reset()
        self._print_till_end = False
        self._aborted = False
        self._closed = False
        self._console_width, self._console_height = 0, 0
        self._page_weight = 0
        self._with_color = color
    def _buffer_reset(self):
        self._buffer = []
    def _buffer_append(self, s):
        self._buffer.append(s)
    def _buffer_get_and_reset(self):
        buffer = "".join(self._buffer)
        self._buffer_reset()
        return buffer
    def capture_console_dimensions(self):
        ''' Capture console dimensions. Call this method before each "writing
        session" when using this stream. '''
        self._console_width, self._console_height = get_console_size()
        self._console_height -= 1 # Spare one line for the prompt
    def writelines(self, iterable):
        for s in iterable:
            self.write(s)
    def write(self, s):
        self._validate_write_state()
        if not isinstance(s, string_types):
            s = unicode(s)
        if not s:
            return
        if self._print_till_end:
            self._stream.write(s)
            return
        lines = LINE_SPLITTER.split(s)
        self._buffer_append(lines[0])
        # If parameter string holds more than one line
        if len(lines) > 1:
            # Print current buffer and reset it
            self._print_paginated_line(self._buffer_get_and_reset())
            # Print remaining lines, except the last one which doesn't end with a new-line char
            for line in lines[1:-1]:
                self._check_aborted()
                self._print_paginated_line(line)
            self._buffer_append(lines[-1])
    def _print_paginated_line(self, line):
        if self._print_till_end:
            self._print_line(line)
            return
        line_length = len(line)
        line_weight = (1 if not line_length
                       else line_length / self._console_width + 1 if line_length % self._console_width
                       else line_length / self._console_width)
        self._page_weight += line_weight
        if self._page_weight > self._console_height:
            action = self._get_action()
            if action is ACTIONS.page_down:
                self._page_weight = line_weight
            elif action is ACTIONS.print_all:
                self._print_till_end = True
            elif action is ACTIONS.quit:
                self._aborted = True
                return
            elif action is ACTIONS.down:
                pass
        self._print_line(line)
    def _get_action(self):
        self._stream.write(COLORED_MORE_RESULTS_PROMPT if self._with_color else MORE_RESULTS_PROMPT)
        self._stream.flush()
        while True:
            c = getch().lower()
            if c == " ":
                action = ACTIONS.page_down
                break
            if c in ("\r", "\n"):
                action = ACTIONS.down
                break
            if c == CHARS.print_all:
                action = ACTIONS.print_all
                break
            if c == CHARS.quit:
                action = ACTIONS.quit
                break
        self._stream.write(ERASE_PROMPT_STR)
        return action
    def _print_line(self, line):
        self._stream.write(line)
        self._stream.write("\n")
    def flush(self):
        self._stream.flush()
    @property
    def closed(self):
        return self._closed or self._stream.closed
    def close(self):
        self._closed = True
        # Do not close wrapped stream
    def isatty(self):
        return self._stream.isatty()
    def _validate_write_state(self):
        self._check_aborted()
        if self.closed:
            raise ValueError("Stream is closed")
        assert self._console_width and self._console_height, "Console dimensions not set"
    def _check_aborted(self):
        if self._aborted:
            raise PaginationAborted
    def __repr__(self):
        return "<{}: {!r}>".format(type(self).__name__, self._stream)
    @staticmethod
    def is_pagination_available():
        return is_getch_available

class PaginationAborted(Exception):
    pass
