import time

class Timer(object):
    '''
    Utility for measuring time deltas.
    
    Simple use:

    >>> t = Timer()
    >>> t.start()
    >>> # Do some work
    >>> t.stop()
    >>> print(t.get_seconds())

    Use as a context:

    >>> t = Timer()
    >>> with t:
    >>>     pass  # Do some work
    >>> print(t.get_seconds())
    '''
    def __init__(self):
        self.start_time, self.stop_time, self.delta = 0, 0, 0
    def __enter__(self):
        self.start()
    def __exit__(self, _, __, ___):
        self.stop()
    def start(self):
        self.start_time = time.time()
    def stop(self):
        self.stop_time = time.time()
        self.delta = self.stop_time - self.start_time
    def get_seconds(self):
        return self.delta
    def __repr__(self):
        return '<{} instance>'.format(type(self).__name__)

class _NullTimer(Timer):
    def start(self): pass
    def stop(self): pass
    def __enter__(self): pass
    def __exit__(self, _, __, ___): pass

NULL_TIMER = _NullTimer()
''' Dummy timer, does nothing. '''
